#ifdef WIN32 
#pragma warning(disable : 4786)
#endif

#include "BLI_arithb.h"
#include "aqsis_File.h"
#include "aqsis_defines.h"
#include "ri.h"
#include "parameters.h"      // Dobby shading engine DSO parameter handling

#include <math.h>

#include <sstream>

using namespace std;

// Some includes required for *nixes
#ifndef WIN32
#include <sys/stat.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <signal.h>
#include <stdlib.h>
#include <unistd.h>
#else
#include <stdio.h>
#include <sys/stat.h>
#include <sys/types.h>
#define	snprintf _snprintf
#endif

//-------------------------------------------------------------- WIN32 SPECIFIC
#ifdef WIN32 

#include<windows.h>

#ifndef FILE_MAXDIR
#define FILE_MAXDIR  160
#endif

#ifndef FILE_MAXFILE
#define FILE_MAXFILE 80
#endif

extern "C" { extern char bprogname[]; }

// add drive character if not in path string, using blender executable location as reference
static void addDrive(string &path)
{
	int sp = path.find_first_of(":");
	if (sp==-1) {
		string blpath = bprogname;
		sp = blpath.find_first_of(":");
		if (sp!=-1) path = blpath.substr(0, sp+1) + path;
	}
}

#endif // WIN32
//------------------------------------------------------- END OF WIN32 SPECIFIC

//---------------------------------------- PLATFORM-SPECIFIC DIRECTORY CREATION
// NB: Now use BLI_recurdir_fileops()
#if 0
#ifdef WIN32
// MS-Windows
static int createDir(char* name)
{
	if (BLI_exists(name))
		return 2;	//exists
	if (CreateDirectory((LPCTSTR)(name), NULL)) {
		cout << "Directory: " << name << " created\n";
		return 1;	// created
	}
	else	{
		cout << "Could not create directory: " << name << endl;
		return 0;	// fail
	}
}
#else
// POSIX unices
static int createDir(char* name)
{
        if (BLI_exists(name))
                return 2;       // exists
        // mode: directory with read, write and execute permissions for
        //       the user and group that are currently executing Blender
        mode_t mode = S_IFDIR | S_IRUSR | S_IWUSR | S_IXUSR | 
                      S_IRGRP | S_IWGRP | S_IXGRP;
        if (mkdir(name, mode) == 0)
        {
                cout << "Directory: " << name << " created." << endl;
                return 1;       // created;
        }
        else
        {
                cout << "Could not create directory: " << name << endl;
                return 0;       // failed
        }
}
#endif
#endif
//--------------------------------- END OF PLATFORM-SPECIFIC DIRECTORY CREATION

#ifndef PI
#define PI 3.14159265359
#endif

bool aqsisFileRender_t::initExport()
{
	ribpath = "";
	bool dir_failed = false;
	// try the user setting setting first, export dir must be set and exist

        // find the path in which to output the generated RIB files.  the
        //  logic here is:
        //    1. first check AQ_ribpath.  if AQ_ribpath has been specified
        //        then use that unchanged.
        //    2. check the image output path.  if that has been specified then
        //        use a sub-directory of the image output path called RIB.
        //    3. if all else fails, try to use the current directory.
        if (strlen(G.scene->r.AQ_ribpath) == 0)
        {
                cout << "No output directory for RIB files is set." << endl
                     << "Trying a RIB sub-directory of the image output "
                     << "directory."
                     << endl;
                if (strlen(G.scene->r.pic) == 0)
                {
                        cout << "No image output directory is set. "
                             << "Trying the current directory."
                             << endl;
                        ribpath = ".";
                        ribpath += PATH_DELIMITER;
                }
                else
                {
                        int ls;
                        string pic = G.scene->r.pic;
                        bool wasrelative = strncmp(&pic[0], "//", 2) == 0;
                        if (wasrelative)
                        {
                                string sce = G.sce;
                                ls = sce.find_last_of(PATH_DELIMITER);
                                sce = sce.substr(0, ls);
                                
                                pic = sce + pic.substr(1, pic.size());
                        }
                        ls = pic.find_last_of(PATH_DELIMITER);
                        ribpath = pic.substr(0, ls+1);
                        ribpath += "RIB";
                        ribpath += PATH_DELIMITER;
                }
        }
        else
        {
                ribpath = G.scene->r.AQ_ribpath;
                if (ribpath.find_last_of(PATH_DELIMITER) != 
                    (ribpath.length()-1))
                {
                        ribpath += PATH_DELIMITER;
                }
        }
        cout << "Using RIB output directory: " << ribpath << endl;
        
        // check if the RIB output directory exists
        if (!BLI_exists(&ribpath[0]))
        {
                cout << "RIB output directory does not exist!" << endl;
                cout << "Trying to create: " << ribpath << endl;
                BLI_recurdir_fileops(&ribpath[0]);
                // retrospectively check whether the directory was created,
                //  since BLI_recurdir_fileops doesn't have a return code.
                if (!BLI_exists(&ribpath[0]))
                {
                        cout << "Could not create directory: "
                             << &ribpath[0] << endl;
                        dir_failed = true;
                }
        }
#ifdef WIN32
        // add drive character for windoze
        addDrive(ribpath);
#endif
        
	// for all
	if (dir_failed) return false;

	// This goes here before the RiBegin, as it sets up specific options 
        //  that control the initialisation of the RIB processor.
	RtToken indentType[] = {"Space"};
	RtInt indentSize[] = {2};

	RiOption("RI2RIB_Indentation", "Type", indentType, "Size", indentSize, 
                 RI_NULL);

        stringstream strFileName;
        char numStr[10];
        snprintf(numStr, 10, "%.4d", G.scene->r.cfra);
        string pic = string(G.scene->r.pic);
        int rlast = pic.find_last_of(PATH_DELIMITER);
        string fName = pic.substr(rlast, pic.length()-rlast);
        strFileName << &ribpath[0] << &fName[0] << numStr << ".rib" << ends;

	ribfile = strFileName.str();
	RiBegin( &ribfile[0] );

        // print comments about Ri copyright
        RiArchiveRecord(RI_COMMENT,
                " The RenderMan (R) Interface Procedures and Protocol are:",
                RI_NULL);
        RiArchiveRecord(RI_COMMENT,
                "       Copyright 1988, 1989, Pixar", RI_NULL);
        RiArchiveRecord(RI_COMMENT,
                "        All Rights Reserved", RI_NULL);

	// set the shader path option
	if (strlen(G.scene->r.AQ_shaderpath) > 0)
	{
		char* ppath = G.scene->r.AQ_shaderpath;
		RiOption("searchpath", "shader", &ppath, RI_NULL);
	}

        // set the Dobby DSO path option
        if (strlen(G.scene->r.AQ_dobbypath) > 0)
        {
                char *dobby_path = G.scene->r.AQ_dobbypath;
                RiOption("searchpath", "dsolibs", &dobby_path, RI_NULL);
        }

	return true;
}


bool aqsisFileRender_t::finishExport()
{
	RiEnd();

	// file exported, now render
	executeAqsis(ribfile);
	return true;
}


bool aqsisFileRender_t::executeAqsis(const string &theribfile)
{
        // set the command for rendering depending upon the currently-selected
        //  renderer
	std::stringstream command;
        switch (R.r.AQ_renderer)
        {
                case AQ_RENDERER_AIR:
                        command << "air " << "\"" << &theribfile[0]
                                << "\"" << std::ends;
                        break;
                case AQ_RENDERER_3DELIGHT:
                        command << "renderdl " << "\"" << &theribfile[0]
                                << "\"" << std::ends;
                        break;
                case AQ_RENDERER_PIXIE:
                        command << "rndr " << "\"" << &theribfile[0] << "\""
                                << std::ends;
                        break;
                case AQ_RENDERER_RDC:
                        command << "renderdc " << "\"" << &theribfile[0]
                                << "\"" << std::ends;
                        break;
                case AQ_RENDERER_PRMAN:
                        command << "prman " << "\"" << &theribfile[0]
                                << "\"" << std::ends;
                case AQ_RENDERER_AQSIS:
                default:
                        command << "aqsis " 
                                << "\"" << &theribfile[0] << "\""
                                << " -progress" << std::ends; 
        }
        
#ifndef WIN32
	sigset_t aqsis,old;
	sigemptyset(&aqsis);
	sigaddset(&aqsis, SIGVTALRM);
	sigprocmask(SIG_BLOCK, &aqsis, &old);
	int ret=system(command.str().c_str());
	sigprocmask(SIG_SETMASK, &old, NULL);
	if (WIFEXITED(ret))
	{
		if (WEXITSTATUS(ret)) 
                        cout << "Executed -" << &command.str()[0] << "-" 
                             << endl;
		switch (WEXITSTATUS(ret))
		{
			case 0: 
                                cout << "Render completed successfully\n";  
                                return true;
			case 127: 
                                cout << "Renderer not found\n";  
                                return false;
			case 126: 
                                cout << "Renderer: permission denied\n";  
                                return false;
			default: 
                                cout << "Renderer exited with errors\n";  
                                return false;
		}
	}
	else if (WIFSIGNALED(ret))
		cout << "Renderer crashed\n";
	else
		cout << "Unknown error\n";
	return false;
#else
	int ret=system(command.str().c_str());
	return ret==0;
#endif
}

// write a frame
void aqsisFileRender_t::writeFrame()
{
	// find the name of the image file
	int number = G.scene->r.cfra;
	char numStr[10];
	snprintf(numStr, 10, "%.4d", number);
        string pic = G.scene->r.pic;
        bool wasrelative = strncmp(&pic[0], "//", 2) == 0;
        if (wasrelative)
        {
                string sce = G.sce;
                int ls = sce.find_last_of(PATH_DELIMITER);
                sce = sce.substr(0, ls);

                pic = sce + pic.substr(1, pic.size());
        }
        std::stringstream strFrameName;
        strFrameName << &pic[0] << numStr;
        if (G.scene->r.scemode & R_EXTENSION)
        {
                strFrameName << ".tif";
        }
        strFrameName << ends;

	RiFrameBegin(number);

                // find the output mode (RGB, RGBA, or B&W)
                RtToken mode;
                switch (R.r.planes)
                {
                        case R_PLANESBW:        // B&W
                                mode = RI_RGB;
                                // we use an imager shader to simulate B&W, 
                                //  rather than setting RiColorSamples(), 
                                //  which is a more obscure part of the 
                                //  RenderMan interface
                                RiImager("BL_imager_bw", RI_NULL);
                                break;
                        case R_PLANES24:        // RGB
                                mode = RI_RGB;  break;
                        case R_PLANES32:        // RGBA
                        default:
                                mode = RI_RGBA; break;
                }
        
                // output the correct display type
		switch( R.r.AQ_displaytype )
		{
			case AQ_DISP_FB:
				RiDisplay(&strFrameName.str()[0], 
                                          RI_FRAMEBUFFER, mode, RI_NULL );
				break;

			case AQ_DISP_FILE:
				RiDisplay(&strFrameName.str()[0], RI_FILE, 
                                          mode, RI_NULL );
				break;

			case AQ_DISP_CUST:
				if( strlen(R.r.AQ_customdisplay) > 0 )
                                {
					RiDisplay(&strFrameName.str()[0], 
                                                  R.r.AQ_customdisplay, 
                                                  mode, RI_NULL);
                                }
				else
				{
                                        cout << "Custom display type with no "
                                             << "name, defaulting to "
                                             << "framebuffer." << endl;
					RiDisplay(&strFrameName.str()[0], 
                                                  RI_FRAMEBUFFER, mode, 
                                                  RI_NULL);
				}
				break;
		}
                
		writeCamera();
		RiWorldBegin();
                        if (R.r.AQ_shadebug == AQ_SHADEBUG_OFF)
			        writeLamps();
			writeAllObjects(0);
		RiWorldEnd();

	RiFrameEnd();
}



// write main camera
void aqsisFileRender_t::writeCamera()
{
	// render resolution including the percentage buttons (aleady 
        //  calculated in initrender for R renderdata)
	int xres = R.r.xsch;
	int yres = R.r.ysch;
	RiFormat( xres, yres, 1.0 );

        // aspect ratio can be set in Blender using aspX & aspY.  These
        //  factors act to change the screen window (this can be seen using
        //  the Blender GUI), so we use ScreenWindow here.
        // the logic of this is just totally FUBAR, so don't ask me - I'm
        //  just mimicking Blender's behaviour!  this is mostly a copy of
        //  code from drawview.c
        RtFloat aspect = float(xres*R.r.xasp)/float(yres*R.r.yasp);
        RtFloat iaspect = float(xres) / float(yres);
        if (iaspect < 1)
                iaspect = float(1) / iaspect;
        RtFloat xscr, yscr;
        if (aspect > 1)
        {
                xscr = iaspect;
                yscr = iaspect / aspect;
        }
        else
        {
                xscr = iaspect * aspect;
                yscr = iaspect;
        }
        RiScreenWindow(-xscr, xscr, -yscr, yscr);

        // projection
	RtFloat factor;
	if(xres >= yres)
		factor = RtFloat(yres)/RtFloat(xres);
	else
		factor = RtFloat(xres)/RtFloat(yres);
	RtFloat focal=360.0f * atan(factor * 16.0f / mainCamLens) / RI_PI;
	RiProjection( RI_PERSPECTIVE, RI_FOV, &focal, RI_NULL );

	Camera *cam= reinterpret_cast<Camera*>(maincam_obj->data);
	RiClipping(cam->clipsta, cam->clipend);

	// depth of field
	if (R.r.AQ_dof != 0)
	{
		RiDepthOfField(cam->fstop, cam->focallength,
			cam->focaldistance);
	}

	// other options
	RiShadingRate(R.r.AQ_shadingrate);
	RiPixelSamples(R.r.AQ_samples_x, R.r.AQ_samples_y);
	RtFilterFunc pixelFilter;
	switch (R.r.AQ_pixelfilter)
	{
		case AQ_PIXFILT_GAUSSIAN:
			pixelFilter = RiGaussianFilter; break;
		case AQ_PIXFILT_BOX:
			pixelFilter = RiBoxFilter; break;
		case AQ_PIXFILT_TRIANGLE:
			pixelFilter = RiTriangleFilter; break;
		case AQ_PIXFILT_SINC:
			pixelFilter = RiSincFilter; break;
		case AQ_PIXFILT_CATMULLROM:
			pixelFilter = RiCatmullRomFilter; break;
		default:
			pixelFilter = RiGaussianFilter; break;
	}
	RiPixelFilter(pixelFilter, R.r.AQ_pixfiltwid_x, R.r.AQ_pixfiltwid_y);
	RiExposure(R.r.AQ_gain, R.r.AQ_gamma);
	if (R.r.mode & R_MBLUR)
		RiShutter(0.0, R.r.blurfac);
	else
		RiShutter(0.0, 0.0);
        RtToken interpStyle;
        switch (R.r.AQ_shadinterp)
        {
                case AQ_SHADINTERP_CONSTANT:
                        interpStyle = RI_CONSTANT; break;
                case AQ_SHADINTERP_SMOOTH:
                        interpStyle = RI_SMOOTH; break;
        }
        RiShadingInterpolation(interpStyle);


	RiArchiveRecord(RI_COMMENT, " Camera transformation", RI_NULL);
        objTransforms[maincam_obj].writeCameraMotion(R.r.mode & R_MBLUR);
}



/**
 * Write all objects.
 *
 * @param shadowpass	Indicates that the objects are being written for a
 *			  shadow pass.
 */
void aqsisFileRender_t::writeAllObjects(int shadowpass)
{
        bool mblur = ((!shadowpass) && (R.r.mode & R_MBLUR));
	map<Object*, AQObject*>::const_iterator obi;
	// First output all objects as separate RIBs
	for(obi=m_objects.begin(); obi!=m_objects.end(); ++obi)
	{
                AQObject* aqobj = obi->second;
                assert(aqobj);
		Object* obj = aqobj->getBlenderObject();
                assert(obj);
                assert(obj == obi->first);

		// Write out individual (non-duplicated) meshes to separate
		//  RIB files
		if (dup_srcob.find(obj) == dup_srcob.end())
                        aqobj->writeObject(ribpath, mblur);
	}

	// Then reference the RIB's in our main file.
	for(obi=m_objects.begin(); obi!=m_objects.end(); ++obi)
	{
		Object*   obj   = obi->first;
                AQObject* aqobj = obi->second;
                assert(obj);
                assert(aqobj);
                assert(aqobj->getBlenderObject() == obj);
	
		// skip any duplivert objects
		if (dup_srcob.find(obj) != dup_srcob.end())
			continue;

		// if this is a shadow pass and an object's material is not 
		//  traceable then skip it
		Material* fmat = aqobj->getMaterial();
		if (shadowpass && !(fmat->mode & MA_TRACEBLE))
			continue;

		// write out the transformation and setup
		char* name = obj->id.name;
		std::stringstream strComment;
		strComment << " Object " << name << std::ends;
		RiArchiveRecord(RI_COMMENT, &strComment.str()[0], RI_NULL);
		RiAttributeBegin();
                objTransforms[obj].writeObjectMotion(mblur);
		// Output object name as identifier attribute.
		RiAttribute(RI_IDENTIFIER, RI_NAME, &name, RI_NULL);

		// turn on any layer-specific lights
		switchLayerLamps(obj->lay);

		// write out the object material
                if (!shadowpass)
                        writeMaterial(fmat);

		// Reference the RIB file created earlier.
                RiReadArchive( &aqobj->getFileName(ribpath,R.r.AQ_relpath)[0], 
                               0, RI_NULL );

		RiAttributeEnd();
	}

	// Now all duplivert objects (if any) as instances of main object
	// The original object has been included in the VlakRen renderlist 
	// above (see convertBlenderScene.c)
	// All others are written here.
	map<Object*, Object*>::const_iterator dupObi;
	for(dupObi=dup_srcob.begin(); dupObi != dup_srcob.end(); ++dupObi)
	{
		// fetch the original object and its duplicated version
		Object* obj_dupl = dupObi->first;
		Object* obj_orig = dupObi->second;
                assert(obj_dupl);
                assert(obj_orig);

		// skip non-mesh objects
		if ((obj_orig == NULL) || (obj_orig->type != OB_MESH))
			continue;
	
		// if this is a shadow pass and an object's material is not
		//  traceable then skip it
		obi = m_objects.find(obj_orig);
                //assert(obi != m_objects.end());
                if (obi == m_objects.end())
                {
                        std::cout << "Cannot find original object: "
                                  << obj_orig->id.name << std::endl;
                        assert(false);
                }
                AQObject* aqobj = obi->second;
                assert(aqobj);
		Material* fmat = aqobj->getMaterial();
		if (shadowpass && ((fmat == NULL) || 
                    !(fmat->mode & MA_TRACEBLE)))
                {
			continue;
                }
	
		// write out the object name
		char* name = obj_dupl->id.name;
		std::stringstream strComment;
		strComment << " Object " << name << std::ends;
		RiArchiveRecord(RI_COMMENT, &strComment.str()[0], RI_NULL);
	
		// object transformation matrix
                RiAttributeBegin();
                objTransforms[obj_dupl].writeObjectMotion(mblur);
	
		// Output object name as identifier attribute.
		RiAttribute(RI_IDENTIFIER, RI_NAME, &name, RI_NULL);

		// turn on any layer-specific lights
		switchLayerLamps(obj_dupl->lay);

		// write out the material
		if (!shadowpass)
                        writeMaterial(fmat);

		// Reference the RIB file created earlier.
                RiReadArchive( &aqobj->getFileName(ribpath, R.r.AQ_relpath)[0],
                               0, RI_NULL );

		RiAttributeEnd();

	}
}


// write all materials & modulators
void aqsisFileRender_t::writeMaterialsAndModulators()
{
	map<string, Material*>::const_iterator blendmat;
	for (blendmat=used_materials.begin(); blendmat!=used_materials.end(); ++blendmat) 
	{
                DY_ParamManager m;
		Material* matr = blendmat->second;
		RtContextHandle OldRi = RiGetContext();

		// Start a new RIB file for the material.
		RiBegin(&getMaterialRIBName(matr, ribpath, false)[0]);
		
                RiArchiveRecord(RI_COMMENT,
                   " The RenderMan (R) Interface Procedures and Protocol are:",
                   RI_NULL);
                RiArchiveRecord(RI_COMMENT,
                   "       Copyright 1988, 1989, Pixar", RI_NULL);
                RiArchiveRecord(RI_COMMENT,
                   "        All Rights Reserved", RI_NULL);
                
#if 0
		std::vector<RtToken> mapTypes;
		std::vector<RtToken> blendTypes;
		std::vector<RtFloat> texColors;
		std::vector<RtFloat> texFilterColors;
		std::vector<RtFloat> texContrasts;
		std::vector<RtFloat> texBrightnesses;
		std::vector<RtFloat> texColValues;
		std::vector<RtFloat> texCspecValues;
		std::vector<RtFloat> texSpecValues;
		std::vector<RtFloat> texCmirValues;
		std::vector<RtFloat> texRefValues;
		std::vector<RtFloat> texHardValues;
		std::vector<RtFloat> texAlphaValues;
		std::vector<RtFloat> texEmitValues;
		std::vector<RtFloat> texTransluValues;
		std::vector<RtFloat> texNorValues;
		std::vector<RtFloat> texDispValues;
		std::vector<RtFloat> texNoiseDepths;
		std::vector<RtFloat> texNoiseTurbs;
		std::vector<RtFloat> texNoiseSizes;
		std::vector<RtFloat> texNoiseTypes;
		std::vector<RtFloat> texNoiseSTypes;
		std::vector<RtFloat> texOffsets;
		std::vector<RtFloat> texSizes;

		// blendermappers
		for (int m=0;m<8;m++) 
		{
			char* type = "";
			char* blendType = "mix";
			RtFloat texColValue = 0,
				texCspecValue = 0,
				texSpecValue = 0,
				texCmirValue = 0,
				texRefValue = 0,
				texHardValue = 0,
				texAlphaValue = 0,
				texEmitValue = 0,
				texTransluValue = 0,
				texNorValue = 0,
				texDispValue = 0;
			RtColor texColor = {0.0f,0.0f,0.0f};
			RtColor texFilterColor = {0.0f,0.0f,0.0f};
			RtFloat texContrast = 0.0f;
			RtFloat texBrightness = 0.0f;
			RtFloat texNoiseDepth = 1.0f;
			RtFloat texNoiseTurb = 0.0f;
			RtFloat texNoiseSize = 1.0f;
			RtFloat texNoiseType = 1.0f;
			RtPoint texOffset = {1.0f, 1.0f, 1.0f};
			RtPoint texSize = {1.0f, 1.0f, 1.0f};
			RtFloat texNoiseSType = 0.0f;
			if( !matr->septex & (1<<m) )
			{
				// Empty texture slots are left as emtpy string.
				MTex* mtex = matr->mtex[m];
				if (mtex!=NULL && mtex->tex!=NULL)
				{
					Tex* tex = mtex->tex;
					map<string, pair<Material*, MTex*> >::const_iterator mtexL = used_textures.find(string(tex->id.name));
					if (mtexL!=used_textures.end())
					{
						switch(tex->type)
						{
							case TEX_CLOUDS:
								type = "clouds";
								break;
							case TEX_WOOD:
								type = "wood";
								break;
							case TEX_MARBLE:
								type = "marble";
								break;
							case TEX_MAGIC:
								type = "magic";
								break;
							case TEX_BLEND:
								type = "blend";
								break;
							case TEX_STUCCI:
								type = "stucci";
								break;
							case TEX_NOISE:
								type = "noise";
								break;
							case TEX_IMAGE:
								type = "image";
								break;
							case TEX_PLUGIN:
								type = "plugin";
								break;
							case TEX_ENVMAP:
								type = "envmap";
								break;
						}
						switch(mtex->blendtype)
						{
							case MTEX_MUL:
								blendType="mul";
								break;
							case MTEX_ADD:
								blendType="add";
								break;
							case MTEX_SUB:
								blendType="sub";
								break;
						}

						if(mtex->mapto & MAP_COL)
							texColValue = 1;
						if(mtex->mapto & MAP_COLSPEC)
							texCspecValue = 1;
						if(mtex->mapto & MAP_SPEC || mtex->maptoneg & MAP_SPEC)
							texSpecValue = (mtex->mapto & MAP_SPEC)? 1 : -1;
						if(mtex->mapto & MAP_COLMIR)
							texCmirValue = 1;
						if(mtex->mapto & MAP_REF || mtex->maptoneg & MAP_REF)
							texRefValue = (mtex->mapto & MAP_REF)? 1 : -1;
						if(mtex->mapto & MAP_HAR || mtex->maptoneg & MAP_HAR)
							texHardValue = (mtex->mapto & MAP_HAR)? 1 : -1;
						if(mtex->mapto & MAP_ALPHA || mtex->maptoneg & MAP_ALPHA)
							texAlphaValue = (mtex->mapto & MAP_ALPHA)? 1 : -1;
						if(mtex->mapto & MAP_EMIT || mtex->maptoneg & MAP_EMIT)
							texEmitValue = (mtex->mapto & MAP_EMIT)? 1 : -1;
						if(mtex->mapto & MAP_TRANSLU)
							texTransluValue = 1;
						if(mtex->mapto & MAP_NORM || mtex->maptoneg & MAP_NORM)
							texNorValue = (mtex->mapto & MAP_NORM)? 1 : -1;
						if(mtex->mapto & MAP_DISPLACE)
							texDispValue = 1;

						texColor[0] = mtex->r;
						texColor[1] = mtex->g;
						texColor[2] = mtex->b;
						texFilterColor[0] = tex->rfac;
						texFilterColor[1] = tex->gfac;
						texFilterColor[2] = tex->bfac;
						texContrast = tex->contrast;
						texBrightness = tex->bright;
						texNoiseDepth = tex->noisedepth;
						texNoiseTurb = tex->turbul;
						texNoiseSize = tex->noisesize;
						texNoiseType = tex->noisetype;
						texOffset[0] = mtex->ofs[0];
						texOffset[1] = mtex->ofs[1];
						texOffset[2] = mtex->ofs[2];
						texSize[0] = mtex->size[0];
						texSize[1] = mtex->size[1];
						texSize[2] = mtex->size[2];
						texNoiseSType = tex->stype;
					}
				}
			}
			mapTypes.push_back(type);
			blendTypes.push_back(blendType);
			texColors.push_back(texColor[0]);
			texColors.push_back(texColor[1]);
			texColors.push_back(texColor[2]);
			texFilterColors.push_back(texFilterColor[0]);
			texFilterColors.push_back(texFilterColor[1]);
			texFilterColors.push_back(texFilterColor[2]);
			texContrasts.push_back(texContrast);
			texBrightnesses.push_back(texBrightness);
			texNoiseTypes.push_back(texNoiseType);
			texNoiseDepths.push_back(texNoiseDepth);
			texNoiseTurbs.push_back(texNoiseTurb);
			texNoiseSizes.push_back(texNoiseSize);
			texColValues.push_back(texColValue);
			texCspecValues.push_back(texCspecValue);
			texSpecValues.push_back(texSpecValue);
			texCmirValues.push_back(texCmirValue);
			texRefValues.push_back(texRefValue);
			texHardValues.push_back(texHardValue);
			texAlphaValues.push_back(texAlphaValue);
			texEmitValues.push_back(texEmitValue);
			texTransluValues.push_back(texTransluValue);
			texNorValues.push_back(texNorValue);
			texDispValues.push_back(texDispValue);
			texSizes.push_back(texSize[0]);
			texSizes.push_back(texSize[1]);
			texSizes.push_back(texSize[2]);
			texOffsets.push_back(texOffset[0]);
			texOffsets.push_back(texOffset[1]);
			texOffsets.push_back(texOffset[2]);
			texNoiseSTypes.push_back(texNoiseSType);
		}
#endif

                /* RAMPS */
                int ramp_col_tot, ramp_col_ipotype, ramp_spec_tot,
                    ramp_spec_ipotype;
                ramp_col_tot = ramp_col_ipotype = ramp_spec_tot 
                             = ramp_spec_ipotype = 0;
                if (matr->ramp_col)
                {
                        ramp_col_tot      = matr->ramp_col->tot;
                        ramp_col_ipotype  = matr->ramp_col->ipotype;
                }
                if (matr->ramp_spec)
                {
                        ramp_spec_tot     = matr->ramp_spec->tot;
                        ramp_spec_ipotype = matr->ramp_spec->ipotype;
                }
                std::vector<RtFloat> ramp_col_pos;
                std::vector<RtFloat> ramp_col_alpha;
                std::vector<RtFloat> ramp_col_color;
                std::vector<RtFloat> ramp_spec_pos;
                std::vector<RtFloat> ramp_spec_alpha;
                std::vector<RtFloat> ramp_spec_color;
                for (int ii=0; ii < 16; ii++)
                {
                        CBData *d;
                        
                        if (matr->ramp_col)
                        {
                                d = &matr->ramp_col->data[ii];
                                ramp_col_pos.push_back(d->pos);
                                ramp_col_alpha.push_back(d->a);
                                ramp_col_color.push_back(d->r);
                                ramp_col_color.push_back(d->g);
                                ramp_col_color.push_back(d->b);
                        }
                        else
                        {
                                ramp_col_pos.push_back(0);
                                ramp_col_alpha.push_back(0);
                                ramp_col_color.push_back(0);
                                ramp_col_color.push_back(0);
                                ramp_col_color.push_back(0);
                        }

                        if (matr->ramp_spec)
                        {
                                d = &matr->ramp_spec->data[ii];
                                ramp_spec_pos.push_back(d->pos);
                                ramp_spec_alpha.push_back(d->a);
                                ramp_spec_color.push_back(d->r);
                                ramp_spec_color.push_back(d->g);
                                ramp_spec_color.push_back(d->b);
                        }
                        else
                        {
                                ramp_spec_pos.push_back(0);
                                ramp_spec_alpha.push_back(0);
                                ramp_spec_color.push_back(0);
                                ramp_spec_color.push_back(0);
                                ramp_spec_color.push_back(0);
                        }
                }
                        
                /* Everything else */
		RtColor col, speccol, mircol, ambientcol;
		col[0] = matr->r;
		col[1] = matr->g;
		col[2] = matr->b;
		speccol[0] = matr->specr;
		speccol[1] = matr->specg;
		speccol[2] = matr->specb;
		mircol[0] = matr->mirr;
		mircol[1] = matr->mirg;
		mircol[2] = matr->mirb;
		ambientcol[0] = matr->ambr;
		ambientcol[1] = matr->ambg;
		ambientcol[2] = matr->ambb;
		int refractionmode = 0;
		if( matr->mode & MA_ZTRA )	refractionmode = 1;
		else if(matr->mode & MA_RAYTRANSP)	refractionmode = 2;

                //---- SHADERS SECTION
                RtFloat recvshadow = ((matr->mode & MA_SHADOW) != 0);
                m.addTriple("shaders", "diffusecolor", col);
                m.addTriple("shaders", "specularcolor", speccol);
                m.addTriple("shaders", "mirrorcolor", mircol);
                m.addTriple("shaders", "ambientcolor", ambientcol);
                m.addInt("shaders", "diffusemode", matr->diff_shader);
                m.addFloat("shaders", "Kd_ref", matr->ref);
                m.addFloat("shaders", "Kd_rough", matr->roughness);
                m.addFloat("shaders", "Kd_size", matr->param[0]);
                m.addFloat("shaders", "Kd_smooth", matr->param[1]);
                m.addInt("shaders", "specularmode", matr->spec_shader);
                m.addFloat("shaders", "Ks_spec", matr->spec);
                m.addFloat("shaders", "Ks_hard", matr->har);
                m.addFloat("shaders", "Ks_refr", matr->refrac);
                m.addFloat("shaders", "Ks_size", matr->param[2]);
                m.addFloat("shaders", "Ks_smooth", matr->param[3]);
                m.addFloat("shaders", "Kalpha", matr->alpha);
                m.addFloat("shaders", "Kemit", matr->emit);
                m.addFloat("shaders", "Kref", matr->ref);
                m.addFloat("shaders", "Ktranslu", matr->translucency);
                m.addInt("shaders", "traceable", matr->mode & MA_TRACEBLE);
                m.addInt("shaders", "shadeless", matr->mode & MA_SHLESS);
                m.addInt("shaders", "vertexcol", matr->mode & MA_VERTEXCOL);
                m.addInt("shaders", "vertexcolp", matr->mode & MA_VERTEXCOLP);
                m.addInt("shaders", "ztransp", matr->mode & MA_ZTRA);
                m.addInt("shaders", "onlyshadow", matr->mode & MA_ONLYSHADOW);
                m.addFloat("shaders", "raymir", matr->ray_mirror);
                m.addFloat("shaders", "refl_fresnel", matr->fresnel_mir);
                m.addFloat("shaders", "refl_fac", matr->fresnel_mir_i);
                m.addInt("shaders", "refractionmode", refractionmode);
                m.addFloat("shaders", "refr_fresnel", matr->fresnel_tra);
                m.addFloat("shaders", "refr_ior", matr->ang);
                m.addFloat("shaders", "refr_fac", matr->fresnel_tra_i);
                m.addFloat("shaders", "specTra", matr->spectra);
                m.addFloat("shaders", "add", matr->add);

                //---- DIFFUSE COLOR RAMP
                m.addInt("ramp_col", "use", ((matr->mode & MA_RAMP_COL) != 0));
                m.addInt("ramp_col", "input", matr->rampin_col);
                m.addInt("ramp_col", "blend", matr->rampblend_col);
                m.addFloat("ramp_col", "factor", matr->rampfac_col);
                m.addInt("ramp_col", "tot", ramp_col_tot);
                m.addInt("ramp_col", "ipotype", ramp_col_ipotype);
                m.addFloatV("ramp_col", "pos", 16, &ramp_col_pos[0]);
                m.addFloatV("ramp_col", "alpha", 16, &ramp_col_alpha[0]);
                m.addTripleV("ramp_col", "colors", 16, &ramp_col_color[0]);
                
                //---- SPECULAR COLOR RAMP
                m.addInt("ramp_spec", "use", ((matr->mode & MA_RAMP_SPEC) != 0));
                m.addInt("ramp_spec", "input", matr->rampin_spec);
                m.addInt("ramp_spec", "blend", matr->rampblend_spec);
                m.addFloat("ramp_spec", "factor", matr->rampfac_spec);
                m.addInt("ramp_spec", "tot", ramp_spec_tot);
                m.addInt("ramp_spec", "ipotype", ramp_spec_ipotype);
                m.addFloatV("ramp_spec", "pos", 16, &ramp_spec_pos[0]);
                m.addFloatV("ramp_spec", "alpha", 16, &ramp_spec_alpha[0]);
                m.addTripleV("ramp_spec", "colors", 16, &ramp_spec_color[0]);

                stringstream pliststr;
                pliststr << &getMaterialDobbyName(matr, ribpath, false)[0];
                string plist = pliststr.str();
                m.write(plist);

                const char* plist_str[1];
                plist_str[0] = plist.c_str();
                RiSurface("BL_surface", 
                        "uniform string plist", &plist_str,
                        "uniform float recvshadow", &recvshadow,
                        RI_NULL);
                
                RiEnd();
                RiContext( OldRi );

/*		// blendershaders + modulators
		// modulators
				// material (diffuse) color, amount controlled by colfac (see below)
				// texture flag, combination of strings
				if (mtex->texflag & (MTEX_RGBTOINT | MTEX_STENCIL | MTEX_NEGATIVE))
				{
					ts = "";
					if (mtex->texflag & MTEX_RGBTOINT) ts += "no_rgb ";
					if (mtex->texflag & MTEX_STENCIL) ts += "stencil ";
					if (mtex->texflag & MTEX_NEGATIVE) ts += "negative";
					ostr << "\t\t<texflag value=\"" << ts << "\" />\n";
				}

				// colfac, controls amount of color modulation
				ostr << "\t\t<colfac value=\"" << mtex->colfac << "\" />\n";

				// def_var
				ostr << "\t\t<def_var value=\"" << mtex->def_var << "\" />\n";

				//varfac
				ostr << "\t\t<varfac value=\"" << mtex->varfac << "\" />\n";

				if ((tex->imaflag & (TEX_CALCALPHA | TEX_USEALPHA)) || (tex->flag & TEX_NEGALPHA))
				{
					ts = "";
					if (tex->imaflag & TEX_CALCALPHA) ts += "calc_alpha ";
					if (tex->imaflag & TEX_USEALPHA) ts += "use_alpha ";
					if (tex->flag & TEX_NEGALPHA) ts += "neg_alpha";
					ostr << "\t\t<alpha_flag value=\"" << ts << "\" />\n";
				}

				ostr << "\t</modulator>\n";
				xmlfile << ostr.str();

			}
		}
		xmlfile << "</shader>\n\n";
	*/
	}
}

void aqsisFileRender_t::writeLamps()
{
	// all lamps
	for (int i=0;i<R.totlamp;i++)
	{
		LampRen* lamp = R.la[i];

		// Output the transform
        	RiTransformBegin();
                AQTransform lightTransform(lamp);
                lightTransform.writeLight();
		//writeLightTransform(lamp, 0);

		RtLightHandle lh = 0;

		RtToken lighttype = "lamp";
		if (lamp->type==LA_LOCAL)
			lighttype = "lamp";
		else if (lamp->type==LA_SPOT)
			lighttype = "spot";
		else if (lamp->type==LA_SUN)
			lighttype = "sun";
		else if (lamp->type==LA_AREA)
			lighttype = "area";
		else if (lamp->type==LA_HEMI)
			lighttype = "hemi";
		RtColor color;
		color[0] = lamp->r;
		color[1] = lamp->g;
		color[2] = lamp->b;
		RtFloat energy = lamp->energy;
		RtFloat dist = lamp->dist;
		RtFloat negative = (lamp->mode & LA_NEG)? 1.0f : 0.0f;
		RtFloat quad1 = lamp->ld1;
		RtFloat quad2 = lamp->ld2;
		RtFloat quad = (lamp->mode & LA_QUAD)? 1.0f : 0.0f;
		RtFloat sphere = (lamp->mode & LA_SPHERE)? 1.0f : 0.0f;
		RtFloat __nonspecular = (lamp->mode & LA_NO_SPEC)? 1.0f : 0.0f;
		RtFloat __nondiffuse = (lamp->mode & LA_NO_DIFF)? 1.0f : 0.0f;
		RtFloat __foglight = (lamp->mode & LA_HALO)? 1.0f : 0.0f;
		RtFloat spotSi = lamp->spotsi;
		RtFloat spotBl = lamp->spotbl;
		RtFloat haloInt = lamp->haint;
		RtFloat square = (lamp->mode & LA_SQUARE)? 1.0f : 0.0f;
		// shadowmap
		RtFloat samples = lamp->samp;
		RtFloat soft = lamp->soft;
		RtFloat bias = lamp->bias;
		RtFloat onlyshadow = (lamp->mode & LA_ONLYSHADOW) ? 1.0f : 0.0f;

		// find the shadow map name
		std::string shadowmap;
		if ( (R.r.mode & R_SHADOW) && (lamp->type == LA_SPOT) &&
			(lamp->mode & LA_SHAD) )
		{
			std::stringstream strMapName;
			strMapName << &getShadowName(lamp, R.r.AQ_relpath)[0]
				<< ".sm" << std::ends;
			shadowmap = strMapName.str();
		}
		const char*   shadowmaps[1];
		shadowmaps[0] = shadowmap.c_str();

		std::vector<RtToken> mapTypes;
		std::vector<RtToken> blendTypes;
		std::vector<RtFloat> texColors;
		std::vector<RtFloat> texFilterColors;
		std::vector<RtFloat> texContrasts;
		std::vector<RtFloat> texBrightnesses;
		std::vector<RtFloat> texColValues;
		std::vector<RtFloat> texNoiseDepths;
		std::vector<RtFloat> texNoiseTurbs;
		std::vector<RtFloat> texNoiseSizes;
		std::vector<RtFloat> texNoiseTypes;
		std::vector<RtFloat> texNoiseSTypes;
		std::vector<RtFloat> texOffsets;
		std::vector<RtFloat> texSizes;

		// blendermappers
		for (int m=0;m<8;m++) 
		{
			char* type = "";
			char* blendType = "mix";
			RtFloat texColValue = 0;
			RtColor texColor = {0.0f,0.0f,0.0f};
			RtColor texFilterColor = {0.0f,0.0f,0.0f};
			RtFloat texContrast = 0.0f;
			RtFloat texBrightness = 0.0f;
			RtFloat texNoiseDepth = 1.0f;
			RtFloat texNoiseTurb = 0.0f;
			RtFloat texNoiseSize = 1.0f;
			RtFloat texNoiseType = 1.0f;
			RtPoint texOffset = {1.0f, 1.0f, 1.0f};
			RtPoint texSize = {1.0f, 1.0f, 1.0f};
			RtFloat texNoiseSType = 0.0f;
			// Empty texture slots are left as emtpy string.
			MTex* mtex = lamp->mtex[m];
			if (mtex!=NULL && mtex->tex!=NULL)
			{
				Tex* tex = mtex->tex;
				map<string, pair<Material*, MTex*> >::const_iterator mtexL = used_textures.find(string(tex->id.name));
				if (mtexL!=used_textures.end())
				{
					switch(tex->type)
					{
						case TEX_CLOUDS:
							type = "clouds";
							break;
						case TEX_WOOD:
							type = "wood";
							break;
						case TEX_MARBLE:
							type = "marble";
							break;
						case TEX_MAGIC:
							type = "magic";
							break;
						case TEX_BLEND:
							type = "blend";
							break;
						case TEX_STUCCI:
							type = "stucci";
							break;
						case TEX_NOISE:
							type = "noise";
							break;
						case TEX_IMAGE:
							type = "image";
							break;
						case TEX_PLUGIN:
							type = "plugin";
							break;
						case TEX_ENVMAP:
							type = "envmap";
							break;
					}
					switch(mtex->blendtype)
					{
						case MTEX_MUL:
							blendType="mul";
							break;
						case MTEX_ADD:
							blendType="add";
							break;
						case MTEX_SUB:
							blendType="sub";
							break;
					}

					if(mtex->mapto & MAP_COL)
						texColValue = 1;

					texColor[0] = mtex->r;
					texColor[1] = mtex->g;
					texColor[2] = mtex->b;
					texFilterColor[0] = tex->rfac;
					texFilterColor[1] = tex->gfac;
					texFilterColor[2] = tex->bfac;
					texContrast = tex->contrast;
					texBrightness = tex->bright;
					texNoiseDepth = tex->noisedepth;
					texNoiseTurb = tex->turbul;
					texNoiseSize = tex->noisesize;
					texNoiseType = tex->noisetype;
					texOffset[0] = mtex->ofs[0];
					texOffset[1] = mtex->ofs[1];
					texOffset[2] = mtex->ofs[2];
					texSize[0] = mtex->size[0];
					texSize[1] = mtex->size[1];
					texSize[2] = mtex->size[2];
					texNoiseSType = tex->stype;
				}
			}
			mapTypes.push_back(type);
			blendTypes.push_back(blendType);
			texColors.push_back(texColor[0]);
			texColors.push_back(texColor[1]);
			texColors.push_back(texColor[2]);
			texFilterColors.push_back(texFilterColor[0]);
			texFilterColors.push_back(texFilterColor[1]);
			texFilterColors.push_back(texFilterColor[2]);
			texContrasts.push_back(texContrast);
			texBrightnesses.push_back(texBrightness);
			texNoiseTypes.push_back(texNoiseType);
			texNoiseDepths.push_back(texNoiseDepth);
			texNoiseTurbs.push_back(texNoiseTurb);
			texNoiseSizes.push_back(texNoiseSize);
			texColValues.push_back(texColValue);
			texSizes.push_back(texSize[0]);
			texSizes.push_back(texSize[1]);
			texSizes.push_back(texSize[2]);
			texOffsets.push_back(texOffset[0]);
			texOffsets.push_back(texOffset[1]);
			texOffsets.push_back(texOffset[2]);
			texNoiseSTypes.push_back(texNoiseSType);
		}

                lh = RiLightSource("BL_light", 
                "string lighttype", &lighttype,
                "uniform float energy", &energy,
                "uniform color colour", &color, 
                //"uniform float beamdistribution", &beamdistribution,
                "uniform float dist", &dist,
                "uniform float negative", &negative,
                "uniform float quad", &quad,
                "uniform float sphere", &sphere,
                "uniform float quad1", &quad1,
                "uniform float quad2", &quad2,
                "varying float __nonspecular", &__nonspecular,
                "varying float __nondiffuse", &__nondiffuse,
                "float __foglight", &__foglight,
		/* SHADOW & SPOT SECTION */
                "uniform float spotSi", &spotSi,
                "uniform float spotBl", &spotBl,
                "uniform float haloInt", &haloInt,
                "uniform float square", &square,
                "uniform string shadowmap", &shadowmaps,
                "uniform float samples", &samples,
                "uniform float soft", &soft,
                "uniform float bias", &bias,
                "uniform float onlyshadow", &onlyshadow,
		/* TEXTURE SECTION */
                "uniform string[8] texmaptype", &mapTypes[0], 
                "uniform string[8] texblendtype", &blendTypes[0],
                "uniform float[8] texColValue", &texColValues[0],
                "uniform color[8] texcolor", &texColors[0],
                "uniform color[8] texfiltercolor", &texFilterColors[0],
                "uniform float[8] texcontrast", &texContrasts[0],
                "uniform float[8] texbrightness", &texBrightnesses[0],
                "uniform float[8] texnoisedepth", &texNoiseDepths[0],
                "uniform float[8] texnoisesize", &texNoiseSizes[0],
                "uniform float[8] texnoiseturb", &texNoiseTurbs[0],
                "uniform float[8] texnoisetype", &texNoiseTypes[0],
                "uniform float[8] texnoiseStype", &texNoiseSTypes[0],
                "uniform vector[8] texofs", &texOffsets[0],
                "uniform vector[8] texsize", &texSizes[0],
		RI_NULL
                );
                
		RiTransformEnd();

		// if the light has the "LA_LAYER" attribute it only illuminates
		//  objects in the same layer.  these lights are added to the
		//  layerLights list so that they can be switched on and off
		//  as necessary
		if (lamp->mode & LA_LAYER) {
			layerLights[lh] = lamp->lay;
		} else if (lh >= 0) {
			RiIlluminate(lh, 1);
		}
	}
}



/**
 * Turns on all lamps from the given layer.
 *
 * Layer-specific lamps, which only illuminate a single layer of the Blender
 * scene, are stored in the layerLights map as they are created.  This
 * method turns these lights on or off (using RiIlluminate) depending upon 
 * the layer attribute.
 *
 * @param layer		Layer to turn on lights for (lights from all other
 *				layers are turned off).
 */
void aqsisFileRender_t::switchLayerLamps(int layer) 
{
	map<RtLightHandle, int>::const_iterator llighti;

	for (llighti=layerLights.begin(); llighti != layerLights.end();
		++llighti)
	{
		RtLightHandle lightHandle = llighti->first;
		int lightLayer = llighti->second;

		RiIlluminate(lightHandle, (lightLayer==layer));
	}
}



/**
 * Returns the name of a RIB file for a specified material.
 *
 * Consistent use of this method will allow us to mangle material names if
 * necessary.
 *
 * \param mat	        Material.
 * \param ribpath       Path to the RIB output directory (with terminating
 *                       path delimiter).
 * \param relative      true if the path is not to be included.
 */
std::string aqsisFileRender_t::getMaterialRIBName(const Material *mat,
        const string& ribpath, bool relative)
{
	std::string matname = mat->id.name;
	std::stringstream strName;
        if (!relative)
        {
                strName << ribpath;
        }
	strName << matname.c_str() << ".rib" << std::ends;

	return strName.str();
}



/**
 * Returns the name of a Dobby DSO shading engine parameter file for a
 * specified material.
 *
 * \param mat           Material.
 * \param ribpath       Path to the RIB output directory (with terminating
 *                        path delimiter).
 * \param relative      true if the path is not to be included.
 */
std::string aqsisFileRender_t::getMaterialDobbyName(const Material *mat,
        const string& ribpath, bool relative)
{
        std::string matname = mat->id.name;
        std::stringstream strName;
        if (!relative)
                strName << ribpath;
        strName << matname.c_str() << ".dobby" << std::ends;

        return strName.str();
}



/**
 * Creates the shadow map frame for a given lamp and a given frame.
 *
 * @param lamp		lamp to create the shadow map for
 */
void aqsisFileRender_t::writeShadowMap(LampRen* lamp)
{
	// find the frame number
	int frameNumber = G.scene->r.cfra;

	// add the lamp to our internal map of LampRen to integers
	//  for shadows (NB: I can't see a way to get at the object name
	//  from Blender, so this mapping is our way of identifying the
	//  lamps for now.)
	map<LampRen*, int>::const_iterator lighti;
	lighti = shadowDict.end();
	int lampNumber = lighti->second + 1;
	shadowDict[lamp] = lampNumber;

	// export the frame
	RiArchiveRecord(RI_COMMENT, " This is a shadow map frame.", RI_NULL);
	RiFrameBegin(frameNumber);

		std::stringstream strFrameName, strMapName;
		strFrameName << &getShadowName(lamp, R.r.AQ_relpath)[0] 
			<< ".z" << std::ends;
		strMapName << &getShadowName(lamp, R.r.AQ_relpath)[0]
			<< ".sm" << std::ends;
		RiDisplay(&strFrameName.str()[0], "zfile", RI_Z, RI_NULL);

		// if a framebuffer display is requested, then also fire up
		//  one of these (for now).  this is a developer aid
		/*
		if ( R.r.AQ_displaytype == AQ_DISP_FB )
		{
			std::stringstream fbFrameName;
			fbFrameName << "+" << &strFrameName.str()[0] 
				<< std::ends;
			RiDisplay( &fbFrameName.str()[0], RI_FRAMEBUFFER,
				RI_RGB, RI_NULL );
		}
		*/

		// find the required fov
		RtFloat fov = cos( lamp->spotsi );
		fov = atan( sqrt(2.0) * tan(fov) );
		fov *= 2.0 * 180.0 / PI;
		fov *= 1.03;	/* safety margin of 3 percent */
		RiProjection(RI_PERSPECTIVE, RI_FOV, &fov, RI_NULL);

		// set clipping
		RiClipping(lamp->clipsta, lamp->clipend);
		
		// the following are recommended settings from ARMan.
		RiPixelSamples(1,1);
		RiPixelFilter(RiBoxFilter, 1.0, 1.0);
		RtFloat jitter = 0;
		char* depthfilter = "midpoint";
		RiHider("hidden", "jitter", &jitter,
			"depthfilter", &depthfilter, RI_NULL);
		RiShadingRate(R.r.AQ_shadow_shadingrate);
		RiFormat(lamp->bufsize, lamp->bufsize, 1.0);
		
		// write out the light's transformation matrix
		RiArchiveRecord(RI_COMMENT, " Camera transformation", RI_NULL);
                AQTransform lightTransform(lamp);
                lightTransform.writeLightShadowMap();

		// write out the world block
		RiWorldBegin();
			writeAllObjects(1);
		RiWorldEnd();

	RiFrameEnd();

	// call to make the shadow map
	RiMakeShadow( &strFrameName.str()[0], &strMapName.str()[0], RI_NULL );
	
	RiArchiveRecord(RI_COMMENT, " End of shadow map frame.", RI_NULL);

}



/**
 * Returns a string for a shadow map name.  Append the current frame number
 * and then ".z" or ".sm" to this string to find the full shadow map name.
 *
 * \param lamp          Light.
 * \param relative      true if the RIB file path should not be included
 */
std::string aqsisFileRender_t::getShadowName(LampRen* lamp, bool relative)
{
	// find the lamp number
	int lampNumber = shadowDict[lamp];

	std::stringstream strName;
        if (!relative)
        {
                strName << ribpath;
        }
	strName << "ShadowMap_lamp" << lampNumber << std::ends;

	return strName.str();
}



/**
 * Writes a material include out to a the RIB stream.
 *
 * \param mat   Material to write.
 */
void aqsisFileRender_t::writeMaterial(Material *mat)
{
        assert(mat);

        if (R.r.AQ_shadebug == AQ_SHADEBUG_OFF)
        {
                std::string matname = mat->id.name;
                if (matname.empty()) 
                        RiSurface("matte", RI_NULL);
                else
                        RiReadArchive(
                                &getMaterialRIBName(mat, ribpath,
                                        R.r.AQ_relpath)[0], 0, RI_NULL);
        }
        else if (R.r.AQ_shadebug == AQ_SHADEBUG_SHOWNORMALS)
        {
                RiSurface("BL_shadebug_shownormals", RI_NULL);
        }
        else if (R.r.AQ_shadebug == AQ_SHADEBUG_SHOWGRIDS)
        {
                RiSurface("BL_shadebug_showgrids", RI_NULL);
        }
        else if (R.r.AQ_shadebug == AQ_SHADEBUG_SHOWUV)
        {
                RiSurface("BL_shadebug_showuv", RI_NULL);
        }
        else if (R.r.AQ_shadebug == AQ_SHADEBUG_SHOWMIST)
        {
                RtFloat sta = 0, di = 10;
                World *world = G.scene->world;
                Camera *cam = reinterpret_cast<Camera*>(
                        maincam_obj->data);
                if (world)
                {
                        sta = world->miststa;
                        di  = world->mistdist;
                }
                else if (cam)
                {
                        sta = cam->clipsta;
                        di  = cam->clipend - sta;
                }
                RiSurface("BL_shadebug_showmist",
                        "uniform float sta", &sta,
                        "uniform float di",  &di,
                        RI_NULL);
        }
}

