/**
 *
 */

#include <assert.h>
#include <iostream>
#include <sstream>
#include "ri.h"
#include "aqsis_Object.h"

using namespace std;



/**
 * Adds rendering data to this object.
 *
 * This base-class method simply checks that the object that v refers to is
 * the same as the object that the AQObject was constructed with.
 *
 * \param v     rendering data chunk
 */
void AQObject::addData(VlakRen* v)
{
        assert(v);
        assert(m_obj);
        assert(v->ob == m_obj);

        if (!m_material)
        {
                m_material = v->mat;
        }
#if 0
        else
        {
                if (m_material != v->mat)
                {
                        cerr << "Export of an object with multiple materials!"
                             << endl;
                }
        }
#endif
}



/**
 * Increments the current sub-frame of the animation.
 */
void AQObject::incrementSubFrame()
{
        // increment the number of sub-frames
        ++m_subframe;
}



/**
 * Writes an object to file.
 *
 * This method sets up the RenderMan context for the export, and then calls
 * the internWriteObject() to export the actual object itself.
 *
 * \param ribpath       path to the RIB file
 * \param mblur         motion blur enabling flag
 */
void AQObject::writeObject(const string& ribpath, bool mblur)
{
        // store the original context handle
        RtContextHandle oldRI = RiGetContext();

        // start a new RIB file for the geometry of this object
        RiBegin(&getFileName(ribpath,false)[0]);

        // comment this file to let the user know exactly what it is
        RiArchiveRecord(RI_COMMENT,
                " The RenderMan (R) Interface Procedures and Protocol are:",
                RI_NULL);
        RiArchiveRecord(RI_COMMENT,
                "       Copyright 1988, 1989, Pixar", RI_NULL);
        RiArchiveRecord(RI_COMMENT,
                "        All Rights Reserved", RI_NULL);
        RiArchiveRecord(RI_COMMENT, AQ_EXPORTER_NOTIFY, RI_NULL);
        stringstream strComment;
        strComment << " Object: " << getName() << ends;
        RiArchiveRecord(RI_COMMENT, &strComment.str()[0], RI_NULL);

        // call the internal protected method which is used to write the object
        internWriteObject(mblur);

        // close the current rib file
        RiEnd();

        // reset the original RIB context handle
        RiContext(oldRI);
}



/**
 * Returns the file name associated with this object.
 *
 * \param ribpath       Output directory for RIB files (with terminating path
 *                       delimiter.
 * \param relative      True if the output directory should not be included.
 *
 * \return      file name
 */
string AQObject::getFileName(const string& ribpath, bool relative) const
{
        std::stringstream strName;

        if (!relative)
        {
                strName << ribpath;
        }
        strName << getName() << ".rib" << ends;

        return strName.str();
}

