/**
 *
 */
#ifndef __AQSIS_OBJECT_H
#define __AQSIS_OBJECT_H
#ifdef __cplusplus

extern "C" {
#include "DNA_object_types.h"   // for Object
#include "render_types.h"       // for VlakRen
}
#include <assert.h>
#include <string>
#include <vector>
#include "aqsis_defines.h"
#include "aqsis_Transform.h"

using namespace std;


/**
 * Base class for objects to be rendered that receive VlakRen objects as data
 * for rendering.
 */
class AQObject
{
        //------------------------------------------------------ Public Methods
        public:
                // constructor and destructor
                AQObject(Object* obj) :
                        m_material(NULL),
                        m_subframe(0),
                        m_obj(obj)
                {
                        assert(m_obj);
                }
                virtual ~AQObject() {}
                // addition of export data
                virtual void addData(VlakRen* v);
                virtual void incrementSubFrame();
                // exporter methods
                virtual void writeObject(const string& ribpath, bool mblur);
                // query methods
                Object* getBlenderObject() const { return m_obj; }
                int getNSubFrames() const { return m_subframe; }
                virtual const char* getName() const { return m_obj->id.name; }
                virtual string getFileName(const string& ribpath, 
                        bool relative) const;
                virtual Material* getMaterial() const { return m_material; }
        //--------------------------------------------------- Protected Methods
        protected:
                /// Override to provide object writing functionality
                virtual void internWriteObject(bool mblur) {}
        //--------------------------------------------------- Protected Members
        protected:
                /// Blender material
                Material* m_material;
                /// Sub-frame number
                int m_subframe;
        //----------------------------------------------------- Private Members
        private:
                /// Pointer to the Blender object that this AQObject wraps.
                Object* m_obj;
};



#endif // __cplusplus
#endif // __AQSIS_OBJECT_H

