#ifndef __AQSIS_RENDER_H
#define __AQSIS_RENDER_H

#ifdef __cplusplus
extern "C" {
#endif

#include "ri.h"

#include "MEM_guardedalloc.h"
#include "IMB_imbuf_types.h"

#include "DNA_action_types.h"
#include "DNA_armature_types.h"
#include "DNA_camera_types.h"
#include "DNA_constraint_types.h"
#include "DNA_curve_types.h"
#include "DNA_effect_types.h"
#include "DNA_group_types.h"
#include "DNA_ika_types.h"
#include "DNA_image_types.h"
#include "DNA_key_types.h"
#include "DNA_lamp_types.h"
#include "DNA_lattice_types.h"
#include "DNA_material_types.h"
#include "DNA_meta_types.h"
#include "DNA_mesh_types.h"
#include "DNA_object_types.h"
#include "DNA_packedFile_types.h"
#include "DNA_radio_types.h"
#include "DNA_scene_types.h"
#include "DNA_screen_types.h"
#include "DNA_sound_types.h"
#include "DNA_space_types.h"
#include "DNA_texture_types.h"
#include "DNA_userdef_types.h"
#include "DNA_vfont_types.h"
#include "DNA_view3d_types.h"
#include "DNA_world_types.h"

#include "BKE_global.h"
#include "BKE_image.h"
#include "BKE_effect.h"
#include "render.h"

/* useful matrix & vector operations */
#include "MTC_matrixops.h"
#include "MTC_vectorops.h"

#include "BLI_blenlib.h"

/* need error(), so extern declare here */
extern void error (char *fmt, ...);

#ifdef __cplusplus
}
#endif

#ifdef __cplusplus
#include <iostream>
#include <iomanip>
#include <sstream>
#include <fstream>
#include <map>
#include <string>
#include <vector>

#include "aqsis_Transform.h"
#include "aqsis_Object.h"
#include "aqsis_Mesh.h"

class aqsisRender_t
{
        //------------------------------------------------------ Public Methods
	public:
		aqsisRender_t() :
                        maincam_obj(NULL),
                        mainCamLens(-1)
                {}
		virtual ~aqsisRender_t() {}

		void lastSubFrame(bool last);
		bool exportSubFrame();
		bool doRender();
		void addDupliMtx(Object* obj);
		bool objectKnownData(Object* obj);

        //--------------------------------------------------- Protected Methods
protected:
		bool getAllMatTexObs();
		void addTransformationMatrix(Object* obj);

		virtual bool initExport()=0;
		virtual bool finishExport()=0;
		virtual void writeFrame()=0;
		virtual void writeCamera()=0;
		virtual void writeAllObjects(int shadowpass)=0;
		virtual void writeMaterialsAndModulators()=0;
		virtual void writeLamps()=0;
		virtual void switchLayerLamps(int layer)=0;
		virtual void writeShadowMap(LampRen* lamp)=0;
		//virtual std::string getShadowName(LampRen* lamp)=0;
		void clearAll();
                static bool isParticleSystem(Object* obj);

                // debugging stuff
                bool checkData();
                static string objStringType(const Object* obj);

        //--------------------------------------------------- Protected Members
	protected:
                
                // main camera object
		Object* maincam_obj;

                // Blender lens value of the main camera
		float mainCamLens;

                // true if we're in the last rendering pass
		bool lastPass;

                // map of Blender objects to AQObjects
                std::map<Object*, AQObject*> m_objects;

                // map of object transforms.  object pointers are mapped to
                //  a transform stack, which contains multiple transforms for
                //  each object
                std::map<Object*, AQTransformStack> objTransforms;

		// map of string names of materials to the materials
		//  themselves
		std::map<std::string, Material*> used_materials;

		// map of string names of textures to the textures
		//  themselves
		std::map<std::string, std::pair<Material*, MTex*> > 
			used_textures;
		
		// map of linked objects to an assigned "source" object
		std::map<Object*, Object*> dup_srcob;

		// map of object data to original objects.  this changes from
                //  frame to frame, and shouldn't be used in the final
                //  export
		std::map<void*, Object*> objectData;

		// for lights that only illuminate a single layer, this maps
		//  the light handle to the layer number
		std::map<RtLightHandle, int> layerLights;

		// the following maps lamp objects to an integer, so that we
		//  can keep track of shadow maps
		std::map<LampRen*, int> shadowDict;

};

#endif


#endif /*__AQSIS_RENDER_H */
