/**
 * BL_surface.sl
 */

#include "util.h"
#include "surface_functions.h"

surface BL_surface (
        uniform string plist      = "";
        uniform float  recvshadow = 1;  /* needed by lights */
) {

	/* General stuff */
	normal Nf = faceforward (normalize(N), I);
	vector Vn = -normalize(I);

        /*******************
         * DIFFUSE SHADING *
         *******************/
        uniform color diffusecolor = dobby_get(plist, "shaders", 
                "diffusecolor");
        uniform float diffusemode = dobby_get(plist, "shaders", "diffusemode");
        uniform float Ktranslu = dobby_get(plist, "shaders", "Ktranslu");
        uniform float Kd_ref = dobby_get(plist, "shaders", "Kd_ref");
        color diffusefinal = 0, translufinal = 0;
        if (FEQ(diffusemode, D_LAMBERT))
        {
                diffusefinal = lambertian_diffuse(plist, Nf, diffusecolor,
                        Kd_ref);
                if (Ktranslu > EPSILON)
                {
                        translufinal = lambertian_diffuse(
                                plist, -Nf, diffusecolor, Ktranslu);
                }
        }
        else if (FEQ(diffusemode, D_ORENNAYAR))
        {
                diffusefinal = oren_nayar_diffuse(plist, Nf, Vn, diffusecolor,
                        Kd_ref);
                if (Ktranslu > EPSILON)
                {
                        translufinal = oren_nayar_diffuse(
                                plist, -Nf, Vn, diffusecolor, Ktranslu);
                }
        }
        else if (FEQ(diffusemode, D_TOON))
        {
                diffusefinal = toon_diffuse(plist, Nf, Vn, diffusecolor,
                        Kd_ref);
                if (Ktranslu > EPSILON)
                {
                        translufinal = toon_diffuse(
                                plist, -Nf, Vn, diffusecolor, Ktranslu);
                }
        }

        /********************
         * SPECULAR SHADING *
         ********************/
        uniform color specularcolor = dobby_get(plist, "shaders", 
                "specularcolor");
        uniform float specularmode  = dobby_get(plist, "shaders", 
                "specularmode");
        color specularfinal;
        if (FEQ(specularmode, S_TOON))
        {
                specularfinal = toon_specular(plist, Nf, Vn, specularcolor);
        }
        else if (FEQ(specularmode, S_BLINN))
        {
                specularfinal = blinn_specular(plist, Nf, Vn, specularcolor);
        }
        else if (FEQ(specularmode, S_PHONG))
        {
                specularfinal = phong_specular(plist, Nf, Vn, specularcolor);
        }
        else if (FEQ(specularmode, S_COOKTORRENCE))
        {
                specularfinal = cook_torrence_specular(
                        plist, Nf, Vn, specularcolor);
        }

        /************
         * EMISSION *
         ************/
        uniform float Kemit = dobby_get(plist, "shaders", "Kemit");
        color emitfinal = Kemit * diffusecolor;

        /* Final surface color */
        Ci = diffusefinal + translufinal + specularfinal + emitfinal;

        /* Specular transparency */
        uniform float specTra = dobby_get(plist, "shaders", "specTra");
        float spm = max(comp(specularfinal,0),
                        comp(specularfinal,1),
                        comp(specularfinal,2));
        spm = min(1, spm * specTra);
        uniform float Kalpha = dobby_get(plist, "shaders", "Kalpha");
        float alpha = (1-spm) * Kalpha + spm;
        
        /* Final surface opacity */
        uniform float refractionmode = dobby_get(plist, "shaders", 
                "refractionmode");
        if (FEQ(refractionmode, 1))
        {
                Oi = alpha;
                Ci *= Oi;
        }
        else
        {
                Ci *= alpha;
                Oi = 1;
        }
}

