/**
 * context.h
 */
#ifndef CONTEXT_H
#define CONTEXT_H

#include <map>
#include <string>

#include "colorband.h"
#include "parameters.h"

//-----------------------------------------------------------------------------

#define DECLARE_BOOL_PARAMETER(group,param)                     \
private:                                                        \
bool m_cfl_ ## group ## __ ## param;                            \
bool m_val_ ## group ## __ ## param;                            \
public:                                                         \
bool group ## __ ## param ()                                    \
{                                                               \
        if ( !( m_cfl_ ## group ## __ ## param ) )              \
        {                                                       \
                m_val_ ## group ## __ ## param =                \
                        (getInt( #group, #param ) != 0);        \
                m_cfl_ ## group ## __ ## param = true;          \
        }                                                       \
        return m_val_ ## group ## __ ## param;                  \
}

#define INIT_BOOL_PARAMETER(group,param)                        \
m_cfl_ ## group ## __ ## param ( false ),                       \
m_val_ ## group ## __ ## param ( false )

//-----------------------------------------------------------------------------

#define DECLARE_INT_PARAMETER(group,param)                      \
private:                                                        \
bool m_cfl_ ## group ## __ ## param;                            \
int  m_val_ ## group ## __ ## param;                            \
public:                                                         \
int group ## __ ## param ()                                     \
{                                                               \
        if ( !( m_cfl_ ## group ## __ ## param ) )              \
        {                                                       \
                m_val_ ## group ## __ ## param =                \
                        getInt( #group, #param );               \
                m_cfl_ ## group ## __ ## param = true;          \
        }                                                       \
        return m_val_ ## group ## __ ## param;                  \
}

#define INIT_INT_PARAMETER(group,param)                         \
m_cfl_ ## group ## __ ## param ( false ),                       \
m_val_ ## group ## __ ## param ( 0 )

//-----------------------------------------------------------------------------

#define DECLARE_FLOAT_PARAMETER(group,param)                    \
private:                                                        \
bool  m_cfl_ ## group ## __ ## param;                           \
float m_val_ ## group ## __ ## param;                           \
public:                                                         \
float group ## __ ## param ()                                   \
{                                                               \
        if ( !( m_cfl_ ## group ## __ ## param ) )              \
        {                                                       \
                m_val_ ## group ## __ ## param =                \
                        getFloat( #group, #param );             \
                m_cfl_ ## group ## __ ## param = true;          \
        }                                                       \
        return m_val_ ## group ## __ ## param;                  \
}

#define INIT_FLOAT_PARAMETER(group,param)                       \
m_cfl_ ## group ## __ ## param ( false ),                       \
m_val_ ## group ## __ ## param ( 0.0f )

//-----------------------------------------------------------------------------

/**
 * \class DY_Context
 *
 * All Dobby rendering occurs in a specific context, which is indexed
 * according to a filename.  The DY_Context class contains all data which is
 * specific to a particular rendering context.
 */
class DY_Context
{
        //------------------------------------------------------ Public Methods
        public:
                static DY_Context *createContext(const string& filename);
                ~DY_Context() {}
        //----------------------------------------------------- Shading Methods
        public:
                void  colorramp_diffuse_illum(float result[3], float shader,
                        float energy[3], float dpnorm, float diffuse_base[3]);
                void  colorramp_specular_illum(float result[3], float shader,
                        float energy[3], float dpnorm, float specular_base[3]);
                float getFloat(const string& group, const string& param) const;
                int   getInt(const string& group, const string& param) const;
                void  getTriple(float result[3], const string& group,
                        const string& param) const;
        //----------------------------------------------- Particular Parameters
        public:
                DECLARE_BOOL_PARAMETER  ( ramp_col, use );
                DECLARE_INT_PARAMETER   ( ramp_col, input );
                DECLARE_FLOAT_PARAMETER ( ramp_col, factor );
                DECLARE_INT_PARAMETER   ( ramp_col, blend );
                DECLARE_BOOL_PARAMETER  ( ramp_spec, use );
                DECLARE_INT_PARAMETER   ( ramp_spec, input );
                DECLARE_FLOAT_PARAMETER ( ramp_spec, factor );
                DECLARE_INT_PARAMETER   ( ramp_spec, blend );
        //----------------------------------------------- Miscellaneous Methods
        public:
                void constructColorband(const string& group, 
                        ColorBand *cb) const;
        //--------------------------------------------------- Protected Methods
        protected:
                /**
                 * Do not use this constructor.  The standard usage scenario
                 * is to use the createContext() static method, which can
                 * return NULL if an error occurs.
                 */
                DY_Context() :
                        INIT_BOOL_PARAMETER  ( ramp_col, use ),
                        INIT_INT_PARAMETER   ( ramp_col, input ),
                        INIT_FLOAT_PARAMETER ( ramp_col, factor ),
                        INIT_INT_PARAMETER   ( ramp_col, blend ),
                        INIT_BOOL_PARAMETER  ( ramp_spec, use ),
                        INIT_INT_PARAMETER   ( ramp_spec, input ),
                        INIT_FLOAT_PARAMETER ( ramp_spec, factor ),
                        INIT_INT_PARAMETER   ( ramp_spec, blend )
                {}
        //--------------------------------------------------- Protected Members
        protected:
                DY_ParamManager m_pman; ///< Parameter manager.
};


/**
 * \class DY_ContextManager
 *
 * The top-level class which keeps track of all Dobby contexts.
 */
class DY_ContextManager
{
        //------------------------------------------------------ Public Methods
        public:
                DY_ContextManager() {}
                ~DY_ContextManager() 
                {
                        // delete the contexts that we manage
                        map<string,DY_Context*>::iterator i;
                        for (i = m_contexts.begin(); i != m_contexts.end(); 
                                ++i)
                        {
                                delete i->second;
                        }
                }
                DY_Context *getContext(const string& filename);
        //--------------------------------------------------- Protected Members
        protected:
                map<string,DY_Context*> m_contexts;      ///< List of contexts.
};



#endif // CONTEXT_H
