/**
 * interface.cpp
 */

#include <iostream>

#include "shadeop.h"
#include "context.h"
#include "util.h"

//---- FUNCTIONS LIST ---------------------------------------------------------
/**

color dobby_colorramp_diffuse_illum(string paramfile;
        float shader; color energy; float dpnorm; color diffuse_base)

color dobby_colorramp_specular_illum(string paramfile;
        float shader; color energy; float dpnorm; color specular_base)

float dobby_feq(float a, b;)

{ float | color } dobby_get(string paramfile, group, name;)

**/

//---- MACROS -----------------------------------------------------------------

#define GET_CONTEXT(context) \
{ \
        DY_ContextManager *cm = static_cast<DY_ContextManager*>(initdata); \
        assert(cm); \
        STRING_DESC *params = static_cast<STRING_DESC*>(argv[1]); \
        context = cm->getContext(params->s); \
        if (!context) \
                exit(0); \
}

//---- INIT / CLEANUP ---------------------------------------------------------

extern "C" SHADEOP_INIT(dobby_init)
{
        DY_ContextManager *cm = new DY_ContextManager();
        assert(cm);
        return ((void*)cm);
}

extern "C" SHADEOP_CLEANUP(dobby_cleanup)
{
        DY_ContextManager *cm = static_cast<DY_ContextManager*>(initdata);
        assert(cm);
        delete cm;
}

//---- dobby_colorramp_diffuse_illum ------------------------------------------

extern "C" {
        SHADEOP_TABLE(dobby_colorramp_diffuse_illum) =
        {
                { "color dobby_colorramp_diffuse_illum(string, float, " \
                        "color, float, color)",
                  "dobby_init",
                  "dobby_cleanup"
                },
                { "", "", "" }
        };
}

extern "C" SHADEOP(dobby_colorramp_diffuse_illum)
{
        float *result       =  static_cast<float*>(argv[0]); // color
        // parameters file / context = argv[1]
        float  shader       = *static_cast<float*>(argv[2]);
        float *energy       =  static_cast<float*>(argv[3]); // color
        float  dpnorm       = *static_cast<float*>(argv[4]);
        float *diffuse_base =  static_cast<float*>(argv[5]);

        DY_Context *context;
        GET_CONTEXT(context);

        context->colorramp_diffuse_illum(result, shader, energy, dpnorm, 
                diffuse_base);

        return 0;
}

//---- dobby_colorramp_specular_illum -----------------------------------------

extern "C" {
        SHADEOP_TABLE(dobby_colorramp_specular_illum) =
        {
                { "color dobby_colorramp_specular_illum(string, float, " \
                        "color, float, color)",
                  "dobby_init",
                  "dobby_cleanup"
                },
                { "", "", "" }
        };
}

extern "C" SHADEOP(dobby_colorramp_specular_illum)
{
        float *result        =  static_cast<float*>(argv[0]); // color
        // parameters file / context = argv[1]
        float  shader        = *static_cast<float*>(argv[2]);
        float *energy        =  static_cast<float*>(argv[3]); // color
        float  dpnorm        = *static_cast<float*>(argv[4]);
        float *specular_base =  static_cast<float*>(argv[5]);

        DY_Context *context;
        GET_CONTEXT(context);

        context->colorramp_specular_illum(result, shader, energy, dpnorm,
                specular_base);

        return 0;
}

//---- dobby_feq --------------------------------------------------------------

extern "C" {
        SHADEOP_TABLE(dobby_feq) =
        {
                { "float dobby_feq(float, float)", 
                  "dobby_init", 
                  "dobby_cleanup" },
                { "", "", "" }
        };
}

extern "C" SHADEOP(dobby_feq)
{
        float *result = static_cast<float*>(argv[0]); // float
        float *a      = static_cast<float*>(argv[1]);
        float *b      = static_cast<float*>(argv[2]);

        *result = ( (*a + (EPSILON) >= *b) && (*a - (EPSILON) <= *b) );
        return 0;
}

//---- dobby_get --------------------------------------------------------------

extern "C" {
        SHADEOP_TABLE(dobby_get) =
        {
                { "float dobby_get_float(string, string, string)", 
                  "dobby_init", "dobby_cleanup" },
                { "color dobby_get_triple(string, string, string)",
                  "dobby_init", "dobby_cleanup" },
                { "", "", "" }
        };
}

extern "C" SHADEOP(dobby_get_float)
{
        float       *result = static_cast<float*>(argv[0]);
        // parameters file / context = argv[1]
        STRING_DESC *group  = static_cast<STRING_DESC*>(argv[2]);
        STRING_DESC *name   = static_cast<STRING_DESC*>(argv[3]);

        DY_Context *context;
        GET_CONTEXT(context);
        
        *result = context->getFloat(group->s, name->s);
        return 0;
}

extern "C" SHADEOP(dobby_get_triple)
{
        float       *result = static_cast<float*>(argv[0]); // color
        // parameters file / context = argv[1]
        STRING_DESC *group  = static_cast<STRING_DESC*>(argv[2]);
        STRING_DESC *name   = static_cast<STRING_DESC*>(argv[3]);
        
        DY_Context *context;
        GET_CONTEXT(context);

        context->getTriple(result, group->s, name->s);
        return 0;
}

