/**
 * texture_functions.h
 *
 * Texturing functions for the Blender surface shaders.
 */
#ifndef TEXTURE_FUNCTIONS_H
#define TEXTURE_FUNCTIONS_H

#include "util.h"

/* Key types */
/* from DNA_key_types.h */
#define KEY_LINEAR              0
#define KEY_CARDINAL            1
#define KEY_BSPLINE             2

/* Ramp constants */
/* from DNA_material_types.h */
#define MA_RAMP_IN_SHADER       0
#define MA_RAMP_IN_ENERGY       1
#define MA_RAMP_IN_NOR          2
#define MA_RAMP_IN_RESULT       3
#define MA_RAMP_BLEND           0
#define MA_RAMP_ADD             1
#define MA_RAMP_MULT            2
#define MA_RAMP_SUB             3

/* Diffuse color ramp parameters */
#define RAMP_COL_PARAM_DECL_DEFAULTS \
        uniform float  ramp_col_input      = 0;         \
        uniform float  ramp_col_blend      = 0;         \
        uniform float  ramp_col_factor     = 0;         \
        uniform float  ramp_col_tot        = 0;         \
        uniform float  ramp_col_ipotype    = 0;         \
        uniform float  ramp_col_pos[16]    =            \
                {0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0};      \
        uniform float  ramp_col_alpha[16]  =            \
                {0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0};      \
        uniform color  ramp_col_color[16]  =            \
                {0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0};
#define RAMP_COL_PARAM_DECL \
        uniform float  ramp_col_input, ramp_col_blend, ramp_col_factor, \
                       ramp_col_tot, ramp_col_ipotype; \
        uniform float  ramp_col_pos[16]; \
        uniform float  ramp_col_alpha[16]; \
        uniform color  ramp_col_color[16];
#define RAMP_COL_PARAM_SET \
        ramp_col_input, ramp_col_blend, ramp_col_factor, ramp_col_tot, \
        ramp_col_ipotype, ramp_col_pos, ramp_col_alpha, ramp_col_color

/* Specular color ramp parameters */
#define RAMP_SPEC_PARAM_DECL_DEFAULTS \
        uniform float  ramp_spec_input     = 0;         \
        uniform float  ramp_spec_blend     = 0;         \
        uniform float  ramp_spec_factor    = 0;         \
        uniform float  ramp_spec_tot       = 0;         \
        uniform float  ramp_spec_ipotype   = 0;         \
        uniform float  ramp_spec_pos[16]   =            \
                {0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0};      \
        uniform float  ramp_spec_alpha[16] =            \
                {0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0};      \
        uniform color  ramp_spec_color[16] =            \
                {0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0};
#define RAMP_SPEC_PARAM_DECL \
        uniform float ramp_spec_input, ramp_spec_blend, ramp_spec_factor, \
                      ramp_spec_tot, ramp_spec_ipotype; \
        uniform float ramp_spec_pos[16]; \
        uniform float ramp_spec_alpha[16]; \
        uniform color ramp_spec_color[16];
#define RAMP_SPEC_PARAM_SET \
        ramp_spec_input, ramp_spec_blend, ramp_spec_factor, ramp_spec_tot, \
        ramp_spec_ipotype, ramp_spec_pos, ramp_spec_alpha, ramp_spec_color



/**
 * Copy of set_four_ipo from key.c
 */
void set_four_ipo(
        float d;
        output float data[4];
        float type;
) {
        if (FEQ(type,KEY_LINEAR))
        {
                data[IX(0)] = 0;
                data[IX(1)] = 1-d;
                data[IX(2)] = d;
                data[IX(3)] = 0;
        }
        else if (FEQ(type,KEY_CARDINAL))
        {
                float d2 = d*d;
                float d3 = d2*d;
                float fc = 0.71;
                
                data[IX(0)] = -fc*d3 + 2*fc*d2 - fc*d;
                data[IX(1)] = (2-fc)*d3 + (fc-3)*d2 + 1;
                data[IX(2)] = (fc-2)*d3 + (3-2*fc)*d2 + fc*d;
                data[IX(3)] = fc*d3 - fc*d2;
        }
        else if (FEQ(type,KEY_BSPLINE))
        {
                float d2 = d*d;
                float d3 = d2*d;

                data[IX(0)] = -0.1666*d3 + 0.5*d2 - 0.5*d + 0.16666;
                data[IX(1)] = 0.5*d3 - d2 + 0.6666;
                data[IX(2)] = -0.5*d3 + 0.5*d2 + 0.5*d + 0.1666;
                data[IX(3)] = 0.1666*d3;
        }
}



/**
 * Copy of do_colorband from texture.c
 */
float do_colorband(
        float tot;
        float ipotype;
        float pos[16];
        float alpha[16];
        color col[16];
        float in;
        output color color_out;
        output float alpha_out;
) {
        float ret;

        if (FEQ(tot,0))
        {
                ret = 0;
        }
        else if (FEQ(tot,1))
        {
                color_out = col[IX(0)];
                alpha_out = alpha[IX(0)];

                ret = 1;
        }
        else
        {
                float win = (in < pos[IX(0)]) ? pos[IX(0)] : in;
                /* set i1 to the index of the first pos that is
                 * greater than win */
                float i1;
                for (i1 = 0; i1 < tot; i1 += 1)
                {
                        if (pos[IX(i1)] > win)
                        {
                                break;
                        }
                }
                if (i1 > (tot-1))
                        i1 = tot-1;
                if (win > pos[IX(i1)])
                        win = pos[IX(i1)];
                float i2 = i1-1;
                float denom = pos[IX(i1)] - pos[IX(i2)];
                /* handle the case of a denominator <= 0.  This shouldn't
                 * happen! */
                if (denom <= 0)
                        denom = 0.0001;
                float fac = (pos[IX(i1)] - win) / denom;

                if (FEQ(ipotype,2))
                {
                        float i0 = (i1 >= (tot-1)) ? i1 : (i1+1);
                        float i3 = (i1 < 2) ? i2 : (i2-1);

                        float t[4];
                        set_four_ipo(fac, t, KEY_BSPLINE);
                        
                        float rgba[4];
                        uniform float idx;
                        for (idx=0; idx < 3; idx = idx + 1)
                        {
                                rgba[idx] = t[IX(3)]*comp(col[i3],idx) +
                                            t[IX(2)]*comp(col[i2],idx) +
                                            t[IX(1)]*comp(col[i1],idx) +
                                            t[IX(0)]*comp(col[i0],idx);
                                setcomp(color_out, idx, 
                                        clamp(rgba[IX(idx)],0,1));
                        }
                        rgba[IX(3)] = t[IX(3)]*alpha[IX(i3)] +
                                      t[IX(2)]*alpha[IX(i2)] +
                                      t[IX(1)]*alpha[IX(i1)] +
                                      t[IX(0)]*alpha[IX(i0)];
                        alpha_out = clamp(rgba[IX(3)], 0, 1);
                }
                else
                {
                        if (FEQ(ipotype,1))
                                fac = fac*fac * (3 - 2*fac);
                        float mfac = 1-fac;

                        color_out = clamp(mfac*col[IX(i1)] + 
                                          fac*col[IX(i2)],0,1);
                        alpha_out = clamp(mfac*alpha[IX(i1)] + 
                                          fac*alpha[IX(i2)],0,1);
                }

                ret = 1;
        }

        return ret;
}



#endif /* TEXTURE_FUNCTIONS_H */
