/*
 * $Id: sector.h 55 2003-09-01 19:01:49Z hos $
 *
 * ***** BEGIN GPL/BL DUAL LICENSE BLOCK *****
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version. The Blender
 * Foundation also sells licenses for use in proprietary software under
 * the Blender License.  See http://www.blender.org/BL/ for information
 * about this.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software Foundation,
 * Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 *
 * The Original Code is Copyright (C) 2001-2002 by NaN Holding BV.
 * All rights reserved.
 *
 * The Original Code is: all of this file.
 *
 * Contributor(s): none yet.
 *
 * ***** END GPL/BL DUAL LICENSE BLOCK *****
 */

/* Version: $Id: sector.h 55 2003-09-01 19:01:49Z hos $ */
#ifndef SECTOR_H
#define SECTOR_H

#include "game.h"

#define LF_MAXBUF	128
#define LF_MAXLIFE	128

#define TOLER 0.000076

/* dface->flag */
#define DF_HILITE		1

/* dface->edcode: (zie ook mface) ook voor vertices! */
#define DF_V1			1
#define DF_V2			2
#define DF_V3			4
#define DF_V4			8
#define DF_V1V2			16
#define DF_V2V3			32
#define DF_V3V1			64
#define DF_V3V4			64
#define DF_V4V1			128

/* sn->flag */
#define FH_SECTOR	1
#define FH_PROP		2
#define FH_DYNA		3


#define DTIME	0.02
#define IDTIME	50.0
#define TOTKEY 32

/* ipo->type: gedeeltelijk flags */
#define	IP_LOC		1
#define	IP_ROT		2
#define	IP_SIZE		4
#define IP_OBCOL	8
#define	IP_FROMOB	15

/* lf->type */
#define LF_LINK	1


/* lf->flag1: buttons, sgistuff */

#define LF_CALC_MATRIX	2
#define LF_GHOST_OTHER	8
#define LF_DRAWNEAR		16
#define LF_AERO_AXIS_Y	256
#define LF_DO_FH		4096

/* lf->dflag : starten altijd op nul */
#define LF_TEMPLIFE		1
#define LF_DYNACHANGED	2
#define LF_OMEGA		4
#define LF_DONTDRAW		8
#define LF_TRIP_POS		16
#define LF_TRIP_NEG		32
#define LF_NO_DAMAGE	64
#define LF_ROT_LOCAL	128

/* lf->state: lokale vars: namen willekeurig? */
#define LF_STATE	1
#define LF_DAMAGE	2
#define LF_ACTION	3
#define LF_FRAME	4


typedef struct DFace {
		/* alleen gedurende simulatie */
	float *v1, *v2, *v3, *v4;
	float dist;
	Material *ma;
	float no[3];
	short proj;
	char flag, edcode;
	unsigned short ocx, ocy, ocz, rt;
	
} DFace;

typedef struct pIpo {
	void *next, *prev;

	short type, nr_keys;	/* keys staan achter ipodata */
	short nr_elems, sta;
	short elemsize, rt;
	int offs[3];			/* zodat loc-ipo's shortjes blijven */
} pIpo;


typedef struct Life {
	ID id;

	Object *sector;
	Object *ob;
	
	float oldloc[3], loc[3], speed[3];		/* loc ook in object zetten */
	float oldloc1[3], loc1[3], speed1[3];	/* lokale sector co's */
	float startloc[3], startrot[3];
	float rot[3], rotspeed[3];
	float oldimat[4][4];
	float mass, frict, rotfrict, axsize, frictfac;
	float r, g, b;
	float aero, localaxsize;
	
	float omega[3], force[3];
	float dloc[3], drot[3];
	
	Material *contact;
	Object *collision, *from;
	DFace *floor;
	Mesh *oldmesh;

	char type, lay;
	short flag1, pad;
	short pad2;
	int flag;
	
	short timer, sfra, cfra, dflag;		/* voor init/afhandeling ipoos, dflag: zit niet aan buttons */
	short state[4];						/* lokale variables */

	float colloc[3];					/* collision loc */
	float floorloc[3];
	
	LBuf links;							/* tijdens simul: de kinderen, in volgorde */
	
	ListBase ipo;
	
} Life;



/* sector.c */

Object *find_sector(float *vec, float *local);
void add_to_lbuf(LBuf *lbuf, Object *ob);
void free_lbuf(LBuf *lbuf);
void del_from_lbuf(LBuf *lbuf, Object *ob);
Object *find_sector(float *vec, float *local);
void life_in_sector_test(Object *ob);
void init_dynamesh(Object *ob, Mesh *me);
void end_dynamesh(Mesh *me);
int test_visibility(float *lookat, float *from, Life *lf, Object *se);
void add_dyna_life(Object *ob);
void add_dupli_life(Object *ob, Object *from, int time);
void del_dupli_life(Object *ob);
int cliptest_sector(float *vec, float *size, float *mat);
void build_sectorlist(Object *cam);
void sector_simulate(void);
void del_dupli_lifes(void);
void ApplyMatrix(void *mat, float *old, float *new);

#endif /* SECTOR_H */
