/**
 * $Id: console.c 114 2003-09-23 12:32:57Z phase $
 *
 * ***** BEGIN GPL/BL DUAL LICENSE BLOCK *****
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version. The Blender
 * Foundation also sells licenses for use in proprietary software under
 * the Blender License.  See http://www.blender.org/BL/ for information
 * about this.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software Foundation,
 * Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 *
 * The Original Code is written by Rob Haarsma (phase).
 *
 * Contributor(s): none yet.
 *
 * ***** END GPL/BL DUAL LICENSE BLOCK *****
 * Everything for drawing a console.
 */

#include <stdlib.h>
#include <stdarg.h>

#include <math.h>
#ifndef _WIN32
#include <unistd.h>
#else
#include <io.h>
#include "BLI_winstuff.h"
#endif   

#include "MEM_guardedalloc.h"	/* vprintf, etc ??? */
#include "PIL_time.h"			/* PIL_check_seconds_timer */

#include "DNA_text_types.h"
#include "DNA_space_types.h"
#include "DNA_screen_types.h"
#include "DNA_userdef_types.h"

#include "BKE_global.h"		/* G */
#include "BKE_main.h"		/* G.main */
#include "BKE_library.h"	/* free_libblock */
#include "BKE_sca.h"		/* free_text_controllers */
#include "BKE_text.h"		/* add_text_console */
#include "BKE_utildefines.h"

#include "BLI_blenlib.h"	/* BLI_addtail */

#include "BPY_extern.h"		/* BPY_clear_bad_scriptlinks */

#include "BIF_drawtext.h"	/* pop_spave_text, unlink_text */
#include "BSE_filesel.h"	/* filesel_prevspace */
#include "BIF_screen.h"		/* curarea */
#include "BIF_space.h"		/* newspace */
#include "BIF_console.h"


static int prev_headertype = 0;

Text *get_text_console(void)
{
	ID *id;
	Text *text;
	
	id= G.main->text.first;
	while(id) {
		text = (Text *)id;
		if(text && (text->flags &= TXT_CONSOLE)) {
			return(text);
		}
		id= id->next;
	}
	return 0;
}


void set_console_textsize(void)
{
	Text *text;
	TextLine *tmp;
	int linecount = 0;

	text = get_text_console();

	if(!text) return;

	tmp= text->lines.first;

	while(tmp) {
		tmp= tmp->next;
		linecount++;
	}

	while(linecount > U.console_buffer) {
		txt_delete_line_console(text, text->lines.first);
		linecount--;
	}

	tmp= text->lines.last;
	text->curl = tmp;
	text->sell = tmp;

	if(G.curscreen->console == 1) {
		ScrArea *sa;
		SpaceText *st;
		for (sa= G.curscreen->areabase.first; sa; sa= sa->next) {
			st= sa->spacedata.first;
			if (st && st->spacetype==SPACE_TEXT) {
				text= st->text;
				if(text && (text->flags &= TXT_CONSOLE)) {
					txt_move_eol(text, 0);
					pop_space_text (st);
					scrarea_queue_redraw(sa);
				}
			}
		}
	}
}


ScrArea *find_lowest_area(void)
{
	ScrArea *sa, *big=0;

	sa= G.curscreen->areabase.first;

	while(sa) {
		if(sa->v1->vec.x == 0 && sa->v1->vec.y == 0)
			big= sa;

/*		BIF_printf("cwx %d, cwy %d, cvx %d, cvy %d\n", sa->winx, sa->winy, sa->v1->vec.x, sa->v1->vec.y);
*/
		sa= sa->next;
	}
	
	if(big) return big;
	else return curarea;
}


void BIF_printf(char *console, ...)
{
	char buf[2048];
	va_list ap;

	static double prevtime=0;
	double newtime;

	va_start(ap, console);
	vsprintf(buf, console, ap);
	va_end(ap);

	if(G.curscreen && U.console_out) {
		ScrArea *sa;
		SpaceText *st;
		Text *text;
		TextLine *tmp;
		int linecount = 0;

		text = (Text *)get_text_console();
		if(text == 0) {
			text = (Text *)add_text_console();
		}

		if(text && (text->flags &= TXT_CONSOLE)) {

			tmp = txt_new_line(buf);

			BLI_addtail(&text->lines, tmp);
			
			text->curl = tmp;
			text->sell = tmp;
//			text->curc = 0;
//			text->selc = 0;

			tmp= text->lines.first;
			while(tmp) {
				tmp= tmp->next;
				linecount++;
			}

			if(linecount > U.console_buffer) {
				txt_delete_line_console(text, text->lines.first);
			}
			
			if(U.console_out == 2) printf(buf);

			if(G.curscreen->console == 1) {
				for (sa= G.curscreen->areabase.first; sa; sa= sa->next) {
					st= sa->spacedata.first;
					if (st && st->spacetype==SPACE_TEXT) {
						text= st->text;
						if(text && (text->flags &= TXT_CONSOLE)) {
							txt_move_eol(text, 0);
							pop_space_text (st);

							/* to prevent slowdown with printf floods */
							newtime = PIL_check_seconds_timer();
							if ((newtime - prevtime) > 0.05) { 
								scrarea_queue_redraw(sa);
								prevtime = newtime;
							}
						}
					}
				}
			}
		}
	} else {
		printf(buf);
	}
	return;
}


void popup_console(void)
{
	if((G.curscreen->console == 0) && U.console_out) {
	   if(get_text_console()) {
			/* zoek en popup */
			display_console();
		} else {
			/* open console textfile en popup */
			open_console();
		}
	}
}


void display_console(void)
{
	ScrArea *sa, *so;
	SpaceText *st;
	Text *text;
	
	so = curarea;

	sa = find_lowest_area();

	/* find lower SPACE_BUTS area, otherwise get last window */
/*
	if(curarea->spacetype==SPACE_BUTS) sa= curarea;
	else {
		// look for area 
		sa= G.curscreen->areabase.first;
		while(sa) {
			if(sa->spacetype==SPACE_BUTS) break;
			sa= sa->next;
		}
	}
	
	if(sa==0) sa= G.curscreen->areabase.last;
*/
	prev_headertype = sa->headertype;
	
	if(sa!=curarea) areawinset(sa->win);

	newspace(sa, SPACE_TEXT);
	text = get_text_console();
	st = sa->spacedata.first;
	st->text = text;
	pop_space_text (st);

//	scrarea_change_headertype(sa, 0);

	G.curscreen->console = 1;

	areawinset(so->win);
}


void open_console(void)
{
	ScrArea *sa, *so;
	SpaceText *st;

	// remember current window
	so = curarea;
	
	if(get_text_console()) return;

	sa = find_lowest_area();

	/* find lower SPACE_BUTS area, otherwise get last window */
/*	sa= G.curscreen->areabase.first;
	while(sa) {
		if(sa->spacetype==SPACE_BUTS) break;
		sa= sa->next;
	}
	
	if(sa==0) sa= G.curscreen->areabase.last;
*/	
	if(sa!=curarea) areawinset(sa->win);

	// remember headertype
	prev_headertype = sa->headertype;

	newspace(sa, SPACE_TEXT);

	st= sa->spacedata.first;

	st->text = get_text_console();
	if(!st->text) st->text = (Text *)add_text_console();
	pop_space_text (st);

	G.curscreen->console = 1;
	
	// hide header for console
//	scrarea_change_headertype(sa, 0);

	scrarea_queue_winredraw(sa);

	// return focus to current window
	areawinset(so->win);
}


void hide_console(void)
{
	if((G.curscreen && U.console_out) || G.curscreen->console) {
		ScrArea *sa, *so;
		SpaceText *st;
		Text *text;

		so = curarea;

		for (sa= G.curscreen->areabase.first; sa; sa= sa->next) {
			st= sa->spacedata.first;
			if (st && st->spacetype==SPACE_TEXT) {
				text= st->text;
				if(text && (text->flags &= TXT_CONSOLE)) {
					if(sa!=curarea) areawinset(sa->win);

					G.curscreen->console = 0;

					// return to previous headertype and window
//					scrarea_change_headertype(sa, prev_headertype);

					filesel_prevspace();
					areawinset(so->win);
//					return;
				}
			}
		}
	} else {
		return;
	}
}


void delete_console(void)
{
	if(get_text_console()) {
		ScrArea *sa, *so;
		SpaceText *st;
		Text *text;

		so = curarea;
		
		for (sa= G.curscreen->areabase.first; sa; sa= sa->next) {
			st= sa->spacedata.first;
					
			if (st && st->spacetype==SPACE_TEXT) {
				text= st->text;

				if(text && (text->flags &= TXT_CONSOLE)) {
					if(sa!=curarea) areawinset(sa->win);

					BPY_clear_bad_scriptlinks(text);
					free_text_controllers(text);
					
					unlink_text(text);
					free_libblock(&G.main->text, text);
				}
			}
		}
		
		G.curscreen->console = 0;

		// set previous headertype and window
//		scrarea_change_headertype(sa, prev_headertype);

		filesel_prevspace();
		areawinset(so->win);
	}
}




/* just a reminder on how to toggle bits in the global struct (in BKE_global.h)
 short special1, special2;
 G.special1

#define G_HOLO					1
#define G_CONSOLEISVISIBLE		2

		G.curscreen->console = 1;
		G.special1 |= G_CONSOLEISVISIBLE;

		G.curscreen->console = 0;
		G.special1 ^= G_CONSOLEISVISIBLE;

  		if(G.curscreen->console == 1) {
		if(G.special1 & G_CONSOLEISVISIBLE) {
		if( !(G.special1 & G_CONSOLEISVISIBLE)) popup_console();

*/

/* another reminder: how to scan all area's of all screens

Text *get_text_console(void)
{
	bScreen *sc;
		
	sc= G.main->screen.first;
	while(sc) {
		ScrArea *sa= sc->areabase.first;
		while(sa) {
			SpaceLink  *sl= sa->spacedata.first;
			while(sl) {
				if(sl->spacetype==SPACE_TEXT) {
					SpaceText *st = (SpaceText *) sl;
					Text *text;
					text= st->text;
					if(text && (text->flags &= TXT_CONSOLE)) {
						return(text);
					}
				}
				sl= sl->next;
			}
			sa= sa->next;
		}
		sc= sc->id.next;
	}
	return 0;
}
*/

