#ifdef RENDERBORK

/**
 * $Id: renderman.c 18 2003-08-15 05:03:32Z hos $
 *
 * ***** BEGIN GPL/BL DUAL LICENSE BLOCK *****
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version. The Blender
 * Foundation also sells licenses for use in proprietary software under
 * the Blender License.  See http://www.blender.org/BL/ for information
 * about this.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software Foundation,
 * Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 *
 * The Original Code is Copyright (C) 2001-2002 by NaN Holding BV.
 * All rights reserved.
 *
 * The Original Code is: all of this file.
 *
 * Contributor(s): none yet.
 *
 * ***** END GPL/BL DUAL LICENSE BLOCK *****
 */

/* Renderman code written by Christoffer Green (christoffer.green@sofe.nu) */

#include <ctype.h> /* isdigit, isspace */
#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include <fcntl.h>
#include <string.h>

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#ifndef WIN32 
#include <unistd.h>
#else
#include "BLI_winstuff.h"
#include <io.h>
#endif

#include "MEM_guardedalloc.h"

#include "DNA_object_types.h"
#include "DNA_mesh_types.h"
#include "DNA_material_types.h"
#include "DNA_lamp_types.h"
#include "DNA_curve_types.h"
#include "DNA_image_types.h"
#include "DNA_camera_types.h"
#include "DNA_scene_types.h"
#include "DNA_view3d_types.h"
#include "DNA_userdef_types.h"
#include "DNA_texture_types.h"

#include "BKE_bad_level_calls.h"
#include "BKE_utildefines.h"
#include "BLI_blenlib.h"
#include "BLI_arithb.h"
#include "BLI_editVert.h"

#include "BKE_global.h"
#include "BKE_main.h"
#include "BKE_mesh.h"
#include "BKE_library.h"
#include "BKE_global.h"
#include "BKE_object.h"
#include "BKE_material.h"
#include "BKE_exotic.h"
#include "BKE_screen.h"
#include "BKE_displist.h"
#include "BKE_curve.h"

#include "BSE_headerbuttons.h"

#include "BPY_extern.h"

/* My own includes */
#include <renderman.h>
#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <sys/stat.h>

void write_shaders(); /* one shader created for every material */
void write_mesh(Base *base, int subdiv); /* processing of meshes */
int  write_shadow(Object*ob); /* create shadows */


FILE *fp; /* the master rib file */
FILE *sfp; /* master shadowmap creator */
int renderer;

#ifdef WIN32
int rwindows = 1;
#else
int rwindows = 0;
#endif


char masterBase [1024];
char masterRibFilePath [1024]; /* this is where all the stuff ends up */
char baseDir [1024]; /* renderman exported data goes here */
char shaderDir [1024]; /* exported renderman shaders go here */
char meshDir [1024]; /* exported meshes go here */
char imageDir [1024]; /* exported meshes go here */
char shadowDir [1024]; /* exported shadowfiles go here */
char textureDir [1024]; /* exported texture files go here */

void renderman(int animation, int renderer2) /* everything starts here */
{
	char *pwd = getenv("PWD");
	char *shadowpath[1024];
	char *dir[1024];

	sprintf(dir, "%s", U.renderdir);

#ifdef WIN32
chdir("c:\\\\");
sprintf(masterBase, "blenderExport\\");
sprintf(masterRibFilePath, "blenderExport\\renderman\\default.rib");
sprintf(baseDir, "blenderExport\\renderman\\");
sprintf(shaderDir, "blenderExport\\renderman\\shaders\\");
sprintf(meshDir, "blenderExport\\renderman\\Meshes\\");
sprintf(imageDir, "blenderExport\\renderman\\images\\");
sprintf(shadowDir, "blenderExport\\renderman\\shadows\\");
sprintf(textureDir, "blenderExport\\renderman\\textures\\");
#else
 chdir(dir);
 sprintf(masterBase, "blenderExport/");
 sprintf(masterRibFilePath, "blenderExport/renderman/default.rib");
 sprintf(baseDir, "blenderExport/renderman/");
 sprintf(shaderDir, "blenderExport/renderman/shaders/");
 sprintf(meshDir, "blenderExport/renderman/Meshes/");
 sprintf(imageDir, "blenderExport/renderman/images/");
 sprintf(shadowDir, "blenderExport/renderman/shadows/");
 sprintf(textureDir, "blenderExport/renderman/textures/");
#endif

	renderer = renderer2;
	

	BLI_recurdir_fileops(masterBase);
	BLI_recurdir_fileops(baseDir);
	BLI_recurdir_fileops(shaderDir);
	BLI_recurdir_fileops(meshDir);
	BLI_recurdir_fileops(imageDir);
	BLI_recurdir_fileops(shadowDir);
	BLI_recurdir_fileops(textureDir);

/*
#else
#ifdef __APPLE__
	 OS X is weird ...
	 FIXME: change to a directory the user specified instead of the user's home 
	if (pwd) chdir(pwd);
	else chdir(BLI_gethome());
#endif  __APPLE__ 
	mkdir("blenderExport", S_IRWXU);
	mkdir(baseDir, S_IRWXU);
	mkdir(shaderDir,S_IRWXU);
	mkdir(meshDir, S_IRWXU);
	mkdir(imageDir, S_IRWXU);
	mkdir(shadowDir, S_IRWXU);
	mkdir(textureDir, S_IRWXU);
#endif 
*/

	open_file(); /* open the default rib file */
	sprintf(shadowpath, "%sshadow.rib", baseDir);
	fprintf(fp, "\t\tReadArchive \"%s\"\n", shadowpath);

	if (animation == 1) /* animation or still? */
	{
		int curframe = G.scene->r.cfra; /* curentframe buffer used to reset the framenumber after an anim */
		G.scene->r.cfra = G.scene->r.sfra; /* set the curent frame to the start frame */
		for((G.scene->r.cfra)=(G.scene->r.sfra); (G.scene->r.cfra)<=(G.scene->r.efra); (G.scene->r.cfra)++) /* loop through all frames */
		{
			update_for_newframe(); /* makes blender aware that a framechange has happened */
			exec_functions(); /* starts the exporting */
		}
		G.scene->r.cfra = curframe;
		update_for_newframe();
	}
	else
	{
		exec_functions(); /* starts the exporting */
	}


	fclose(fp); /* close the main ribfile */
	fclose(sfp); /* close the main shadow file */
	exec_renderman_renderer(renderer); /* starts the renderer */

	return;
}

void exec_functions() /* this is where the real export starts */
{
	fprintf(fp, "FrameBegin %i\n", G.scene->r.cfra); /* mostly used for anims, but not extremly important even there */
	write_shaders();
	write_start_header(); /* image resolution.. etc.. */
	write_objects(0, fp); /* export the scene objects, meshes, curves, nurbs etc... */
	write_end_header(); /* just a WorldEnd, same as </html> */
	fprintf(fp, "FrameEnd\n"); /* ends the frame */
}

void open_file() 
{
	char *shadowpath[1024];
	sprintf(shadowpath, "%sshadow.rib", baseDir);
	if ( (fp = fopen(masterRibFilePath, "w")) == NULL) /* open the main rib file */
	{
		fprintf(stderr, "Error opening file.\n");
	}
	if ( (sfp = fopen(shadowpath, "w")) == NULL) /* open the shadowmap rib file */
	{
		fprintf(stderr, "Error opening master shadow file.\n");
	}
}

void write_camera(Object *ob, FILE *fjp)
{
	Mat4Invert(ob->imat, ob->obmat);
	fprintf(fjp, "Transform [%f %f %f %f %f %f %f %f %f %f %f %f %f %f %f %f]\n",
	ob->imat[0][0], ob->imat[0][1],
	 - ob->imat[0][2], ob->imat[0][3],
	ob->imat[1][0], ob->imat[1][1],
	- ob->imat[1][2], ob->imat[1][3],
	ob->imat[2][0], ob->imat[2][1],
	- ob->imat[2][2], ob->imat[2][3],
	ob->imat[3][0], ob->imat[3][1],
	- ob->imat[3][2], ob->imat[3][3]);    
}

void write_start_header()
{
	float fov = 0.0;
	float x = G.scene->r.xsch;
	int y = G.scene->r.ysch;
	float factor = (y/x);
	float fov2 = (float)(360.0 * atan(factor * 16.0 / fov) / 3.14);
	double sr;
	int ps;

	Object *ob;
	Camera *cam;
	Base *base = G.scene->base.first;

	if (G.scene->r.renderer == R_3DELIGHT) /* 3delight */
	{
		/*fprintf(fp, "Option \"searchpath\" \"display\" [\"$DELIGHT/displays/\"]\n");*/
		fprintf(fp, "Option \"searchpath\" \"shader\" [\"$DL_SHADERS_PATH:blenderExport/renderman/shaders/\"]\n");
		fprintf(fp, "Option \"searchpath\" \"texture\" [\"$DL_TEXTURES_PATH:blenderExport/renderman/textures/:blenderExport/renderman/shadows/\"]\n");
	}
	if (G.scene->r.renderer == R_AQSIS) /* Aqsis */
	{
		/*fprintf(fp, "Option \"searchpath\" \"display\" [\"$DELIGHT/displays/\"]\n");*/
		fprintf(fp, "Option \"searchpath\" \"shader\" [\"$AQSIS_SHADERS_PATH:blenderExport/renderman/shaders/\"]\n");
		fprintf(fp, "Option \"searchpath\" \"texture\" [\"$AQSIS_TEXTURES_PATH:blenderExport/renderman/textures/:blenderExport/renderman/shadows/\"]\n");
	}

	else if (G.scene->r.renderer == R_PRMAN)
	{
		/*fprintf(fp, "Option \"searchpath\" \"display\" [\"$DELIGHT/displays/\"]\n");*/
		/*fprintf(fp, "Option \"searchpath\" \"shader\" [\"/usr/local/prman/lib/shaders/:c:\Program Files\Pixar\prman-11.0.1\lib\shaders:$DELIGHT/shaders/:./%s\"]\n", shaderDir);*/
	}

	if ((G.scene->r.rendermanWindow != 1) && (G.scene->r.rendermanFile == 1))
	{
		fprintf(fp, "Display \"image%i.tiff\" \"file\" \"rgb\"\n", G.scene->r.cfra);
	}
	else if ((G.scene->r.rendermanWindow == 1) && (G.scene->r.rendermanFile == 1))
	{
		fprintf(fp, "Display \"image%i.tiff\" \"file\" \"rgb\"\n", G.scene->r.cfra);
		fprintf(fp, "Display \"+window\" \"framebuffer\" \"rgb\"\n");
	}
	else
	{
		fprintf(fp, "Display \"window\" \"framebuffer\" \"rgb\"\n");
	}
	fprintf(fp, "Format %i %i 1\n", G.scene->r.xsch, G.scene->r.ysch);
	fprintf(fp, "PixelFilter \"gaussian\" 2 2\n");
	/*fprintf(fp, "Option \"limits\" \"geommemory\" [8000]\n");*/
	/*fprintf(fp, "Option \"limits\" \"gridsize\" 16\n");*/

	if (G.scene->r.mode & R_OSA)
	{
		if (G.scene->r.osa == 5)
		{
			ps = 2;		sr = 0.6;
		}
		else if (G.scene->r.osa == 8)
		{
			ps = 3;		sr = 0.35;
		}
		else if (G.scene->r.osa == 11)
		{
			ps = 5;		sr = 0.25;
		}
		else
		{
			ps = 10;	sr = 0.1;
		}
	}
	else
	{
		ps = 1;		sr = 10;
	}
	fprintf(fp, "PixelSamples %i %i\n", ps, ps);
	if (G.scene->r.rendermanDof == 1)
	{
		fprintf(fp, "DepthOfField %f %f %f\n", G.scene->r.rendermanStop, G.scene->r.rendermanLength, G.scene->r.rendermanDistance);
	}

	while(base)
	{
		ob = base->object;
		if(ob->type == OB_CAMERA)
		{
			cam = ob->data;
			fov = cam->lens;
			break;
		}

		base = base->next;
	}
	fprintf(fp, "Projection \"perspective\" \"fov\" [%f]\n", (360.0 * atan(factor * 16.0 / fov) / 3.14));

	write_camera(ob, fp);
	fprintf(fp, "ShadingRate %f\n", sr);
	if (G.scene->r.renderer == R_3DELIGHT) /* 3delight */
	{
		fprintf(fp, "Option \"trace\" \"int maxdepth\" [5]\n");
	}
	fprintf(fp, "WorldBegin\n");
	fprintf(fp, "\tAttributeBegin\n");
}

int write_shadow(Object *ob)
{
	FILE *shf;
	Lamp *la;

	char shadowname[1024];

	la = ob->data;
	sprintf(shadowname, "%s%s/%i.rib", shadowDir, ob->id.name, G.scene->r.cfra);
	if ((shf = fopen(shadowname, "w")) == NULL)
	{
		fprintf(stderr, "Error opening shadowfile file %s.\n", shadowname);
		return 0;
	}

	fprintf(shf, "FrameBegin %i\n", G.scene->r.cfra);
	fprintf(shf, "ShadingRate 1\n");
	fprintf(shf, "PixelFilter \"box\" 1 1\n");
	fprintf(shf, "PixelSamples 3 3\n");
	fprintf(shf, "Clipping %f %f\n", la->clipsta, la->clipend);

	fprintf(shf, "Display \"%s%s%i\" \"zfile\" \"z\"\n", shadowDir, ob->id.name, G.scene->r.cfra);
	/*fprintf(shf, "Display \"+shadowmap\" \"framebuffer\" \"rgb\" \"uniform string autoclose\" \"off\"\n");*/

	fprintf(shf, "Format %i %i 1\n", la->bufsize, la->bufsize); 
	fprintf(shf, "Projection \"perspective\" \"fov\" [%f]\n", la->spotsize);

	write_camera(ob, shf);

	fprintf(shf, "WorldBegin\n");
	write_objects(1, shf);
	fprintf(shf, "WorldEnd\n");
	fprintf(shf, "FrameEnd\n");

	fprintf(sfp, "ReadArchive \"%s\"\n",shadowname);
	fprintf(sfp, "MakeShadow \"%s%s%i\" \"./%s%s%i.shd\"\n", shadowDir, ob->id.name, G.scene->r.cfra, shadowDir, ob->id.name, G.scene->r.cfra);
	fclose(shf);
}

void write_bml(Object *ob)
{
	FILE *sf;
	char bmlPath[FILE_MAXDIR+FILE_MAXFILE];
	char systemExec[1024];
	
	printf("hello3\n");

	sprintf(bmlPath, "%sbml.sl", shaderDir);
	if ((sf = fopen(bmlPath, "w")) == NULL)
	{
		fprintf(stderr, "Error opening light shader file.\n");
	}

	fprintf(sf, "light\n");
	fprintf(sf, "bml(\n");
		fprintf(sf, "float  intensity = 1;\n");
		fprintf(sf, "color  lightcolor = 1;\n");
		fprintf(sf, "point  from = point \"shader\" (0,0,0);\n");	/* light position */
		fprintf(sf, "point  to = point \"shader\" (0,0,1);\n");
		fprintf(sf, "float  coneangle = radians(30);\n");
		fprintf(sf, "float  conedeltaangle = radians(5);\n");
		fprintf(sf, "float  beamdistribution = 2;\n");
		fprintf(sf, "string shadowname = \"\";\n");
		fprintf(sf, "float samples = 128;\n");
		fprintf(sf, "float blur = 0.2;\n");
		fprintf(sf, "float  shadowbias = 0.5;\n");
	fprintf(sf, ")\n");
	fprintf(sf, "{\n");
		fprintf(sf, "float  atten, cosangle;\n");
		fprintf(sf, "uniform vector A = (to - from) / length(to - from);\n");
		fprintf(sf, "uniform float cosoutside= cos(coneangle),\n");
		fprintf(sf, "cosinside = cos(coneangle-conedeltaangle);\n");
		fprintf(sf, "illuminate( from, A, coneangle ) {\n");
		fprintf(sf, "cosangle = L.A / length(L);\n");
		fprintf(sf, "atten = pow(cosangle, beamdistribution) / (L.L);\n");
		fprintf(sf, "atten *= smoothstep( cosoutside, cosinside, cosangle );\n");
		fprintf(sf, "if (shadowname != \"\")\n");
		fprintf(sf, "{\n");
			fprintf(sf, "atten *= 1-shadow(shadowname, Ps, \"samples\", samples, \"blur\", blur, \"bias\", shadowbias);\n");
		fprintf(sf, "}\n");
		fprintf(sf, "Cl = atten * intensity * lightcolor;\n");
		fprintf(sf, "}\n");
	fprintf(sf, "}\n");
	fclose(sf);
	if (G.scene->r.renderer == R_3DELIGHT)
	{
		sprintf(systemExec, "shaderdl -d %s %s", shaderDir, bmlPath);
	}
	else if (G.scene->r.renderer == R_PIXIE)
	{
		sprintf(systemExec, "sdrc -I %s %s", shaderDir, bmlPath);
	}
	else if (G.scene->r.renderer == R_PRMAN)
	{
		sprintf(systemExec, "shader -I%s %s", shaderDir, bmlPath);
	}
	else if (G.scene->r.renderer == R_AQSIS)
	{
		sprintf(systemExec, "aqsl -i %s %s", shaderDir, bmlPath);
	}

	system(systemExec);
}

void write_objects(int shadow, FILE *of)
{
	Object *ob;
	Base *base;
	Lamp *la;
	char changedName[256];
	char *loc;
	char *ch = ".";
	Material *ma;

	/* Write Lamps */
	printf("hello4\n");
	base = G.scene->base.first;
	if (shadow == 0)
	{
		while(base)
		{
			if(base->lay & G.scene->lay)
			{
				ob = base->object;
				where_is_object(ob);
				if(base->object->type == OB_LAMP)
				{
					int negative = 1;
					float tox, toy, toz;
					la = ob->data;
					/* printf("lamp %i\n", la->mode);*/

					tox = - ob->obmat[2][0] + ob->obmat[3][0];
					toy = - ob->obmat[2][1] + ob->obmat[3][1];
					toz = - ob->obmat[2][2] + ob->obmat[3][2];
					/*fprintf(fp, "\tLightSource \"bml\" 0 \"float shadowbias\" 0.25 \"float blur\" 0.0 \"float samples\" 1 \"coneangle\" 0.636138503785 \"conedeltaangle\" 0.383766233921 \"from\" [-6.49563694 -2.76116538048 27.5270900726] \"to\" [-6.27830979228 -2.75023511797 26.5510523915] \"intensity\" 8785.71438789 \"lightcolor\" [1.0 1.0 1.0]\n");*/

					if ((la->mode & LA_SHAD) && (G.scene->r.mode & R_SHADOW))
					{
						fprintf(fp, "\tAttribute \"light\" \"shadows\" [\"on\"]\n");
					}
					else
					{
						/*fprintf(fp, "\tAttribute \"light\" \"shadows\" [\"off\"]\n");*/
					}
					if (la->mode & LA_NEG)
					{
						negative = -1;
					}
					if (la->type == LA_LOCAL)
					{
						fprintf(of, "\tLightSource \"pointlight\" 0 \"from\" [%f %f %f] \"intensity\" %f\n\t \"lightcolor\" [%f %f %f]\n",
						ob->loc[0], ob->loc[1], ob->loc[2], (la->energy * 150 * negative), la->r, la->g, la->b);
					}
					else if (la->type == LA_SPOT)
					{
						write_bml(ob);
						if ((la->mode & LA_SHAD) && (G.scene->r.mode & R_SHADOW))
						{
							write_shadow(ob);
						}
						fprintf(of, "\tLightSource \"bml\" 0 \"from\" [%f %f %f] \"to\" [%f %f %f] \"conedeltaangle\" %f \"coneangle\" %f \"intensity\" %f\n \"lightcolor\" [%f %f %f] \"string shadowname\" [\"%s%s.shd\"] \"float blur\" %f \"float shadowbias\" %f \"int samples\" %d\n",
						ob->loc[0], ob->loc[1], ob->loc[2], tox, toy, toz, (la->spotblend / 4), (la->spotsize * 3.14 / 360), (la->energy * 180 * negative), la->r, la->g, la->b, shadowDir, ob->id.name, la->soft / 200, la->bias, la->samp * 10);
					}
				}
			}
			base = base->next;
		}
	}

	/* Write Meshes */
	base = G.scene->base.first;
	while(base)
	{
		if(base->lay & G.scene->lay)
		{
			ob = base->object;
			where_is_object(ob);
			if(ob->type == OB_MESH)
			{
				int subdiv = 0;
				Mesh *me = ob->data;
				fprintf(of, "\tAttributeBegin\n");
				fprintf(of, "\tAttribute \"visibility\" \"integer trace\" [1]\n");
				fprintf(of, "\t\tAttribute \"visibility\" \"string transmission\" [\"opaque\"]\n");
				fprintf(fp, "\t\tAttribute \"bound\" \"displacement\" 3.5\n");
				/*fprintf(fp, "\tAttribute \"visibility\" \"camera\" 1 \"transmission\" [\"opaque\"] \"trace\" 1\n");*/
				fprintf(of, "\t\tAttribute \"trace\" \"integer displacements\" [0]\n");
				/*fprintf(fp, "\tAttribute \"limits\" \"integer eyesplits\" [2]\n");*/
				fprintf(of, "\t\tTransform [%f %f %f %f %f %f %f %f %f %f %f %f %f %f %f %f]\n",
				ob->obmat[0][0], ob->obmat[0][1], ob->obmat[0][2], ob->obmat[0][3],
				ob->obmat[1][0], ob->obmat[1][1], ob->obmat[1][2], ob->obmat[1][3],
				ob->obmat[2][0], ob->obmat[2][1], ob->obmat[2][2], ob->obmat[2][3],
				ob->obmat[3][0], ob->obmat[3][1], ob->obmat[3][2], ob->obmat[3][3]);
				if ((me->subdivr && (me->flag & ME_SUBSURF)))
				{
					subdiv = 1;
				}
				else
				{
					subdiv = 0;
				}
				printf("hello6\n");
				write_mesh(base, subdiv);

				ma = G.main->mat.first;
				ma = give_current_material(ob, 1);

				if (ma != NULL)
				{
					fprintf(fp, "\t\tColor [%f %f %f]\n", ma->r, ma->g, ma->b);
					fprintf(fp, "\t\tOpacity [%f %f %f]\n", ma->alpha, ma->alpha, ma->alpha);
				}
				printf("hello1\n");
				if (ma != NULL) 
				{
					if (ma->rendermanExternalDisplacementShader == 1)
					{
						fprintf(fp, "\t\tDisplacement \"%s\"\n", ma->rendermanExternalDisplacementShaderName);
					}
					if (ma->rendermanExternalSurfaceShader == 1)
					{
						printf("hello5\n");
						fprintf(fp, "\t\tSurface \"%s\"\n", ma->rendermanExternalSurfaceShaderName);
					}
					else
					{
						sprintf(changedName, "%s", ma->id.name);
						/* no . allowed in shader names, replace the . with an a */
						loc = strstr(changedName, ch);
						if (loc != NULL)
						{
							*loc = 97; /* 97 = a */
						}
						/*printf("changed name: %s\n", changedName);*/
						fprintf(fp, "\t\tSurface \"%s%i\"\n", changedName, G.scene->r.cfra);
					}
				}
				else {
					fprintf(fp, "\t\tSurface \"plastic\"\n");
				}
				fprintf(of, "\t\tReadArchive \"%s%s%i\"\n",meshDir, ob->id.name, G.scene->r.cfra);
				fprintf(of, "\tAttributeEnd\n");
			}
			printf("hello2\n");
		}
		base = base->next;
	}
	/* Write Curves */
	/*base = G.scene->base.first;
	while(base)
	{
		if(base->lay & G.scene->lay)
		{
			ob = base->object;
			where_is_object(ob);
			if(ob->type == OB_CURVE)
			{
				fprintf(fp, "\tAttributeBegin\n");
				write_shader(ob);
				fprintf(fp, "\t\tTransform [%f %f %f %f %f %f %f %f %f %f %f %f %f %f %f %f]\n",
				ob->obmat[0][0], ob->obmat[0][1], ob->obmat[0][2], ob->obmat[0][3],
				ob->obmat[1][0], ob->obmat[1][1], ob->obmat[1][2], ob->obmat[1][3],
				ob->obmat[2][0], ob->obmat[2][1], ob->obmat[2][2], ob->obmat[2][3],
				ob->obmat[3][0], ob->obmat[3][1], ob->obmat[3][2], ob->obmat[3][3]);
				Curve *cu = ob->data;
				write_curve(base);
				fprintf(fp, "\tAttributeEnd\n");
			}
		}
		base = base->next;
	}*/
}

void write_texture(Material *ma, FILE *sf)
{
	MTex *mtex;
	char newImagePath[255];
	char newImagePath2[255];
	char systemExec[255];
	char fullPath[FILE_MAXDIR+FILE_MAXFILE];
	int i;

	for (i = 0; i < 8; i++)
	{
		mtex = ma->mtex[i];

		if (mtex == NULL)
		{
			break;
		}

		strcpy(fullPath, (char *)mtex->tex->ima->name);
		BLI_convertstringcode(fullPath, G.sce, G.scene->r.cfra);
		sprintf(newImagePath, "%s%s", textureDir, (char *)mtex->tex->id.name);
		if (G.scene->r.renderer == R_3DELIGHT) /* 3delight*/

		{
			sprintf(systemExec, "tdlmake -filter gaussian -mode periodic -filterwidth 4 \"%s\" %s.tdl\n", fullPath, newImagePath);
		}
		else if (G.scene->r.renderer == R_PIXIE) /* Pixie */
		{
			sprintf(systemExec, "texmake -filter gaussian -mode periodic -filterwidth 4 \"%s\" %s.tdl\n", fullPath, newImagePath);
		}
		else if (G.scene->r.renderer == R_PRMAN) /* Pixie */
		{
			sprintf(systemExec, "txmake -filter gaussian -mode periodic -filterwidth 4 \"%s\" %s.tdl\n", fullPath, newImagePath);
		}
		else if (G.scene->r.renderer == R_AQSIS) /* Aqsis */
		{
			sprintf(systemExec, "teqser -filter=gaussian -swrap=periodic  -twrap=periodic -swidth=4 -twidth=4 \"%s\" %s.tdl\n", fullPath, newImagePath);
		}

		system(systemExec);

		if (mtex->texco & TEXCO_REFL)
		{
			fprintf(sf, "\t{\n");
			fprintf(sf, "\t\tnormal Nf = normalize(faceforward(N, I));\n");
			fprintf(sf, "\t\tvector R = normalize(reflect(I, N));\n");
			fprintf(sf, "\t\tCt = color ReflMap(\"%s.tdl\", R, 0.0);\n", (char *)mtex->tex->id.name);
			fprintf(sf, "\t}\n");
			fprintf(sf, "\n");
		}
		else
		{
			fprintf(sf, "\tCt = color texture(\"%s.tdl\", s, t, \"width\", 0.8, \"fill\",-1);\n", (char *)mtex->tex->id.name);
			fprintf(sf, "\n");
		}
		
		if ((mtex != NULL) && (mtex->tex->type & TEX_IMAGE) && (ma->mtex[0]->tex->ima->name != NULL))
		{
			fprintf(sf, "\tBaseTextureColor = mix(BaseTextureColor, Ct, %f);\n", mtex->colfac);
			fprintf(sf, "\n");
		}
		if (mtex->mapto & MAP_NORM)
		{
		}
	}
}


void write_shaders()
{
	/*printf("Number of materials: %i\n", ob->totcol);*/
	Material *ma;
	MTex *mtex;
	int i;

	ma = G.main->mat.first;
	while(ma) {
/*	if (give_current_material(ob, 1) != NULL)
	{*/
		/*ma = give_current_material(ob, 1);*/
		mtex = ma->mtex[0];
		if (ma)
		{
			FILE *sf;
			int materialType = 0;
			float hard = (1-(ma->har / 255.0)) / 6.5;
			char sname[100];
			char changedName[256];
			char *loc;
			char *ch = ".";
			char newImagePath[255];
			char newImagePath2[255];
			char systemExec[255];

			sprintf(changedName, "%s%i", ma->id.name, G.scene->r.cfra);
			/* no . allowed in shader names, replace the . with an a */
			loc = strstr(changedName, ch);
			if (loc != NULL)
			{
				*loc = 97; /* 97 = a */
			}

			/*printf("changed name: %s\n", changedName);*/
			/*printf("original name name: %s\n", ob->id.name);*/
			/* Open Surface Shader */
			sprintf(sname, "./%s%s.sl",shaderDir, changedName);
			if ((sf = fopen(sname, "w")) == NULL)
			{
				fprintf(stderr, "Error opening material shader .sl file.\n");
				return;
			}

			if ((ma->mtex[0] != NULL) && (mtex->tex->type & TEX_IMAGE) && ma->mtex[0]->tex->ima->name)
			{
				sprintf(newImagePath, "%s%s", textureDir, ma->mtex[0]->tex->id.name);
				sprintf(newImagePath2, "%s%s.tdl", textureDir, ma->mtex[0]->tex->id.name);
			}
			else
			{
				sprintf(newImagePath2, "");
			}
			
			/*****************************/
			/* Start Surface Shader Code */
			/*****************************/
			fprintf(sf, "surface\n");
			fprintf(sf, "%s(float Ks=%f, Kd=%f, Kt=%f, Ka=%f, roughness=%f; color specularcolor=1;)\n", changedName, ma->spec * 2, ma->ref, ma->rendermanRefract, ma->emit, hard);
			fprintf(sf, "{\n");

				fprintf(sf, "\tcolor BaseTextureColor = 1;\n");
				fprintf(sf, "\tcolor Ct;\n");
				fprintf(sf, "\tcolor gi;\n");
				fprintf(sf, "\tcolor reflect;\n");
				fprintf(sf, "\n");

				fprintf(sf, "\tcolor ReflMap(string reflname; point P; float blur)\n");
				fprintf(sf, "\t{\n");
					fprintf(sf, "\t\tpoint Pndc = transform(\"NDC\", P);\n");
					fprintf(sf, "\t\tfloat x = xcomp(Pndc), y = ycomp(Pndc);\n");
					fprintf(sf, "\t\treturn color texture(reflname, x, y, \"blur\", blur);\n");
				fprintf(sf, "\t}\n");

				write_texture(ma, sf);

				fprintf(sf, "\tcolor MaterialPlastic(normal Nf; color Cs, Os, specularcolor, BaseTextureColor; float Ka, Kd, Ks, roughness;)\n");
				fprintf(sf, "\t{\n");
					fprintf(sf, "\t\textern vector I;\n");
					fprintf(sf, "\t\tvarying vector V = normalize(-I);\n");
					fprintf(sf, "\t\n");

					/****************/
					/* gi occlusion */
					/****************/
					fprintf(sf, "\t\tfloat occlusionSL()\n");
					fprintf(sf, "\t\t{\n");
						if (ma->rendermanOcclusion == 1)
						{
								fprintf(sf, "\t\t\treturn (occlusion(P, Ng, %i) *  -%f);\n", ma->rendermanOcclusionSample, ma->rendermanOcclusionRemove);
						}
						else
						{
							fprintf(sf, "\t\t\treturn 0;\n");
						}
					fprintf(sf, "\t\t}\n");
					fprintf(sf, "\n");

					/* Code for normal gi */
					fprintf(sf, "\t\tcolor gi()\n");
					fprintf(sf, "\t\t{\n");
						if (ma->rendermanGI == 1)
						{
							if (ma->rendermanHdriName != "")
							{
								fprintf(sf, "\t\t\treturn indirectdiffuse(P, Ng, %i, \"environmentmap\", \"%s\") * %f;\n", ma->rendermanGISample, ma->rendermanHdriName, ma->rendermanGIColor);
							}
							else
							{
								fprintf(sf, "\t\t\treturn indirectdiffuse(P, Ng, %i) * %f;\n", ma->rendermanGISample, ma->rendermanGIColor);
							}
						}
						else
						{
							fprintf(sf, "\t\t\treturn color(0, 0, 0);\n");
						}
					fprintf(sf, "\t\t}\n");
					fprintf(sf, "\n");

					/* raytracing, reflections, refractions*/
					fprintf(sf, "\t\tcolor raytraceRefract()\n");
					fprintf(sf, "\t\t{\n");
						if (ma->rendermanRaytrace == 1)
						{

						
							fprintf(sf, "\t\t\tvector	R;\n");
							fprintf(sf, "\t\t\tvector	In;\n");
							fprintf(sf, "\t\t\tcolor	refract = 0;\n");
							
							fprintf(sf, "\t\t\tN	=	normalize(N);\n");
							fprintf(sf, "\t\t\tIn	=	normalize(I);\n");

							fprintf(sf, "\t\t\tif ((In.N) < 0) {\n");
								fprintf(sf, "\t\t\t\tR	=	refract(In,N,1/%f);\n", ma->rendermanIOR);
								fprintf(sf, "\t\t\t\trefract	=	Ks*specular(N,-In,roughness);\n");
							fprintf(sf, "\t\t\t} else {\n");
								fprintf(sf, "\t\t\t\tR	=	refract(In,-N,%f);\n", ma->rendermanIOR);
								fprintf(sf, "\t\t\t\trefract	=	0;\n");
							fprintf(sf, "\t\t\t}\n");

							fprintf(sf, "\t\t\trefract += trace(P,R);\n");

							fprintf(sf, "\t\t\treturn refract;\n");
						}
						else
						{
							fprintf(sf, "\t\t\treturn color(0, 0, 0);\n");
						}
					fprintf(sf, "\t\t}\n");
					fprintf(sf, "\n");
					fprintf(sf, "\t\tcolor raytraceReflect()\n");
					fprintf(sf, "\t\t{\n");
						if (ma->rendermanRaytrace == 1)
						{
							fprintf(sf, "\t\t\treturn trace(P, reflect (I, Nf)) * %f;\n", ma->rendermanReflect);
						}
						else
						{
							fprintf(sf, "\t\t\treturn color(0, 0, 0);\n");
						}
					fprintf(sf, "\t\t}\n");
					fprintf(sf, "\n");

					fprintf(sf, "\t\tcolor translucent()\n");
					fprintf(sf, "\t\t{\n");
						if (ma->rendermanTranslucent == 1)
						{
							fprintf(sf, "\t\t\treturn ((Kd*diffuse(-Nf))*%f);\n", ma->rendermanTranslucentColor);
						}
						else
						{
							fprintf(sf, "\t\t\treturn color(0, 0, 0);\n");
						}
					fprintf(sf, "\t\t}\n");
					fprintf(sf, "\n");
					
					fprintf(sf, "\t\treturn Cs * BaseTextureColor * (Ka*ambient() + (Kd*diffuse(Nf)) + translucent()) + occlusionSL() + gi() + raytraceReflect() + Ks*specularcolor*specular(Nf, V, roughness);\n");
				fprintf(sf, "\t}\n");
				fprintf(sf, "\tnormal Nf = faceforward(normalize(N), I);\n");
				fprintf(sf, "\tCi = MaterialPlastic(Nf, Cs, Os, specularcolor, BaseTextureColor, Ka, Kd, Ks, roughness);\n");
				fprintf(sf, "\tOi = Os;\n");
				fprintf(sf, "\tCi *= Oi;\n");
			fprintf(sf, "}\n");

			fclose(sf);

			if (G.scene->r.renderer == R_3DELIGHT) /* 3delight */
			{
				sprintf(sname, "shaderdl -d %s %s%s.sl",shaderDir, shaderDir, changedName);
			}
			else if (G.scene->r.renderer == R_PIXIE) /* Pixie */
			{
				sprintf(sname, "sdrc -I %s %s%s.sl", shaderDir, shaderDir, changedName);
			}
			else if (G.scene->r.renderer == R_PRMAN) /* Pixie */
			{
				sprintf(sname, "shader -I%s %s%s.sl", shaderDir, shaderDir, changedName);
			}
			else if (G.scene->r.renderer == R_AQSIS) /* Aqsis  */
			{
				sprintf(sname, "aqsl -i %s %s%s.sl", shaderDir, shaderDir, changedName);
			}

			system(sname);
			printf("%s\n", sname);
			/*system("/usr/local/prman/bin/shader -o blenderExport/renderman/shaders/bms.slo blenderExport/renderman/shaders/bms.sl");*/
		}
		ma = ma->id.next;
	}
}

/*void write_curve(Base *base)
{
	int a;

	Curve *cu;
	BPoint *cpoint;
	cu = base->object->data;
	cpoint = cu->mface;

	fprintf(fp, "\t\tCurves \"cubic\"");

}*/

void write_mesh(Base *base, int subdiv)
{
	int a;
	int i, bigLoop;
	float r, g, b, cola;
	unsigned int *mcol;
	char meshFilename[255];
	DispList *dl;
	DispListMesh *meM;
	FILE *mf;
	Mesh *me;
	MVert *mvert;
	MFace *mface;
	TFace *tface;
	Object *ob;
	float *extverts;

	me = base->object->data;
	mface = me->mface;
	ob = base->object;

	
	sprintf(meshFilename, "%s%s%i", meshDir, ob->id.name, G.scene->r.cfra);

	mface = me->mface;

	if ( (mf = fopen(meshFilename, "w")) == NULL)
	{
		fprintf(stderr, "Error opening meshfile.\n");
	}

	if (subdiv == 0)
	{
		fprintf(mf, "\t\tPointsPolygons [");
	}
	else
	{
		fprintf(mf, "\t\tSubdivisionMesh \"catmull-clark\" [");
	}
	for(a = 0; a < me->totface; a++, mface++)
	{
		if(mface->v4 == 0)
		{
			fprintf(mf, "3 ");
		}
		else
		{
			fprintf(mf, "4 ");
		}
	}

	fprintf(mf, "] [");
	mface = me->mface;
	for(a = 0; a < me->totface; a++, mface++)
	{
		if(mface->v4 == 0)
		{
			fprintf(mf, "%d %d %d ", mface->v1, mface->v2, mface->v3);
		}
		else
		{
			fprintf(mf, "%d %d %d %d ", mface->v1, mface->v2, mface->v3, mface->v4);
		}
	}

	if (subdiv == 0)
	{
		fprintf(mf, "]\"P\" [");
	}
	else
	{
		fprintf(mf, "] [\"interpolateboundary\"] [0 0] [] [] \"P\" [");
	}

	dl = find_displist(&ob->disp, DL_VERTS);
	if(dl)
	{
		extverts = dl->verts; 
		for(a = 0; a < me->totvert; a++, mvert++)
		{
			fprintf(mf, "%f %f %f ", extverts[0], extverts[1], extverts[2]);
			extverts++;
			extverts++;
			extverts++;
		}
	}
	else {
		mvert = me->mvert;
		for(a = 0; a < me->totvert; a++, mvert++)
		{
			fprintf(mf, "%f %f %f ", mvert->co[0], mvert->co[1], mvert->co[2]);
		}
	}

	fprintf(mf, "]\n");

	mface = me->mface;
	if (mface->flag == 1)
	{
		mvert = me->mvert;
		fprintf(mf, "\"N\" [");
		for(a = 0; a < me->totvert; a++, mvert++)
		{
			fprintf(mf, "%f %f %f ", mvert->no[0] / 32767.0, mvert->no[1] / 32767.0, mvert->no[2] / 32767.0);
		}
		fprintf(mf, "]\n");
	}
	/* write vertex colors */
	mcol = mcol_to_vcol(me);
	a = me->totvert;
	if(mcol)
	{
		fprintf(mf, "\"Cs\" [");
		while(a--)
		{
			mcol_to_rgba(*mcol, &r, &g, &b, &cola);
			fprintf(mf, "%f %f %f ", r, g, b);
			mcol++;
		}
		fprintf(mf, "]\n");
	}

	if (me->tface)
	{
		fprintf(mf, "\"st\" [");
		bigLoop = 0;
		while (bigLoop < me->totvert)
		{
			tface = me->tface;
			mface = me->mface;
			for(i = 0; i < me->totface; i++, mface++, tface++)
			{
				if (mface->v1 == bigLoop)
				{
					fprintf(mf, "%f %f \n", tface->uv[0][0], 1.0 - tface->uv[0][1]);
					break;
				}
				else if (mface->v2 == bigLoop)
				{
					fprintf(mf, "%f %f \n", tface->uv[1][0], 1.0 - tface->uv[1][1]);
					break;
				}
				else if (mface->v3 == bigLoop)
				{
					fprintf(mf, "%f %f \n", tface->uv[2][0], 1.0 - tface->uv[2][1]);
					break;
				}
				else if ((mface->v4 == bigLoop) && (mface->v4 != 0))
				{
					fprintf(mf, "%f %f \n", tface->uv[3][0], 1.0 - tface->uv[3][1]);
					break;
				}
			}
			bigLoop++;
		}
		fprintf(mf, "]\n");
	}

	
	fclose(mf);
}

void write_end_header()
{
	fprintf(fp, "\nAttributeEnd\n");
	fprintf(fp, "WorldEnd\n");
}

void exec_renderman_renderer(int renderer)
{
/*	FILE *fp;

	fp = fopen("renderdl", "r");
	if(fp){
		printf("succes opening renderdl\n");
		fclose(fp);
	}
	fp = fopen("blenderExport/renderman/default.rib", "r");
	if(fp){
		printf("succes opening rib\n");
		fclose(fp);
	}
*/
	char rendercmd[1024];

	if (G.scene->r.renderer == R_3DELIGHT)
	{
		sprintf(rendercmd, "renderdl %s", masterRibFilePath);
	}
	else if (G.scene->r.renderer == R_PIXIE) /* pixie */
	{
		sprintf(rendercmd, "rndr %s", masterRibFilePath);
	}
	else if (G.scene->r.renderer == R_PRMAN)
	{
		sprintf(rendercmd, "prman %s", masterRibFilePath);
	}
	else if (G.scene->r.renderer == R_AQSIS) /* Aqsis */
	{
		sprintf(rendercmd, "aqsis %s", masterRibFilePath);
	}
	system(rendercmd);
}

#endif
