/**
 * ***** BEGIN GPL/BL DUAL LICENSE BLOCK *****
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version. The Blender
 * Foundation also sells licenses for use in proprietary software under
 * the Blender License.  See http://www.blender.org/BL/ for information
 * about this.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software Foundation,
 * Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 *
 * The Original Code is Copyright (C) 2001-2002 by NaN Holding BV.
 * All rights reserved.
 *
 * The Original Code is: all of this file.
 *
 * Contributor(s): None yet.
 *
 * ***** END GPL/BL DUAL LICENSE BLOCK *****
 */

/* 
 * Saves info from the RenderData struct to file
 *
 */
#include <string.h>
#include <stdlib.h>

#include "DNA_listBase.h"
#include "DNA_userdef_types.h"
#include "DNA_scene_types.h"

#include "BKE_global.h"		/* G */
#include "BKE_utildefines.h"

#include "BLI_blenlib.h"
#include "BLI_linklist.h"	/* linknode */

#include "BIF_language.h"
#include "BIF_space.h"		/* allqueue() */
#include "BIF_toolbox.h"	/* error() */
#include "BIF_rendersettings.h"

#include "MEM_guardedalloc.h"	/* vprintf, etc ??? */

#include "mydevice.h"		/* REDRAWALL */

#include "BMF_Api.h"

void save_rendersettings(void)
{
FILE *set;

if( (set = fopen(G.scene->r.setfile, "w")) == NULL){
	printf("opening file failed");
}
else{
fprintf(set, "#***Blender-Render-setting-file****\n");

/* Special modes */
fprintf(set, "\n#Special-modes\n");
	if(G.scene->r.scemode & R_DOSEQ)
		fprintf(set, "Do_sequence: 1\n");
	else
		fprintf(set, "Do_sequence: 0\n");

	if(G.scene->r.mode & R_UNIFIED)
		fprintf(set, "Do_unified: 1\n");
	else	
		fprintf(set, "Do_unified: 0\n");
	
	if(G.scene->r.mode & R_PANORAMA)
		fprintf(set, "Do_panorama: 1\n");
	else	
		fprintf(set, "Do_panorama: 0\n");
	
	
/* Size */
fprintf(set, "\n#Size\n");
	fprintf(set, "Size_(perc): %d\n", G.scene->r.size);
	fprintf(set, "Sizex: %d\n", G.scene->r.xsch);
	fprintf(set, "Sizey: %d\n", G.scene->r.ysch);
	fprintf(set, "Aspx: %d\n", G.scene->r.xasp);
	fprintf(set, "Aspy: %d\n", G.scene->r.yasp);
	fprintf(set, "Xparts: %d\n", G.scene->r.xparts);
	fprintf(set, "Yparts: %d\n", G.scene->r.yparts);
	fprintf(set, "Winpos: %d\n", G.scene->r.winpos);
	

/* OSA */
fprintf(set, "\n#Antialiasing\n");
	if(G.scene->r.mode & R_OSA)
		fprintf(set, "Do_osa: 1\n");
	else	
		fprintf(set, "Do_osa: 0\n");
	fprintf(set, "OSA: %d\n", G.scene->r.osa);

/* SHADOW */
fprintf(set, "\n#Shadow\n");
	if(G.scene->r.mode & R_SHADOW)
		fprintf(set, "Do_shadow: 1\n");
	else	
		fprintf(set, "Do_shadow: 0\n");

/* MBLUR */
fprintf(set, "\n#Mblur\n");
	if(G.scene->r.mode & R_MBLUR)
		fprintf(set, "Do_mblur: 1\n");
	else	
		fprintf(set, "Do_mblur: 0\n");
	fprintf(set, "Blurfactor: %f\n", G.scene->r.blurfac);
	
/* Output */
fprintf(set, "\n#Output\n");
	fprintf(set, "Format: ");
	if(G.scene->r.imtype == R_TARGA)
		fprintf(set, "TARGA\n");
	else if	(G.scene->r.imtype == R_IRIS)
		fprintf(set, "IRIS\n");
	else if	(G.scene->r.imtype == R_IRIZ)
		fprintf(set, "IRIS+Z\n");
	else if	(G.scene->r.imtype == R_HAMX)
		fprintf(set, "HAMX\n");
	else if	(G.scene->r.imtype == R_FTYPE)
		fprintf(set, "FTYPE\n");
	else if	(G.scene->r.imtype == R_JPEG90)
		fprintf(set, "JPEG\n");
	else if	(G.scene->r.imtype == R_RAWTGA)
		fprintf(set, "TGARAW\n");
	else if	(G.scene->r.imtype == R_AVIRAW)
		fprintf(set, "AVIRAW\n");
	else if	(G.scene->r.imtype == R_AVIJPEG)
		fprintf(set, "AVIJPEG\n");
	else if	(G.scene->r.imtype == R_AVICODEC)
		fprintf(set, "AVICODEC\n");
	else if	(G.scene->r.imtype == R_QUICKTIME)
		fprintf(set, "QUICKTIME\n");
	else
		fprintf(set, "UNKNOWN\n");
	
	fprintf(set, "Quality: %d\n", G.scene->r.quality);
		
	if(G.scene->r.scemode & R_EXTENSION)
		fprintf(set, "Do_extensions: 1\n");
	else
		fprintf(set, "Do_extensions: 0\n");

	fprintf(set, "Colormode: ");		
	if(G.scene->r.planes & R_PLANES24)
		fprintf(set, "RGB\n");
	else if	(G.scene->r.planes & R_PLANES32)
		fprintf(set, "RGBA\n");
	else if	(G.scene->r.planes & R_PLANESBW)
		fprintf(set, "BW\n");
	else	
		fprintf(set, "UNKNOWN\n");		
	
/* BackBuff/FrontBuff */
fprintf(set, "\n#Backbuffer/Frontbuffer\n");
	if(G.scene->r.bufflag & R_BACKBUF)
		fprintf(set, "Do_backbuffimage: 1\n");
	else
		fprintf(set, "Do_backbuffimage: 0\n");
	
	if(G.scene->r.bufflag & R_BACKBUFANIM)
		fprintf(set, "Do_backbuffanim: 1\n");
	else
		fprintf(set, "Do_backbuffanim: 0\n");

	if(G.scene->r.bufflag & R_FRONTBUF)
		fprintf(set, "Do_frontbuffimage: 1\n");
	else
		fprintf(set, "Do_frontbuffimage: 0\n");
	
	if(G.scene->r.bufflag & R_FRONTBUFANIM)
		fprintf(set, "Do_frontbuffanim: 1\n");
	else
		fprintf(set, "Do_frontbuffanim: 0\n");	
		

/* EDGE */
fprintf(set, "\n#Edge-Rendering\n");
	if(G.scene->r.mode & R_EDGE)
		fprintf(set, "Do_edge: 1\n");
	else	
		fprintf(set, "Do_edge: 0\n");
	fprintf(set, "Edgeint: %d\n", G.scene->r.edgeint);
	fprintf(set, "Edge_R: %f\n", G.scene->r.edgeR);
	fprintf(set, "Edge_G: %f\n", G.scene->r.edgeG);
	fprintf(set, "Edge_B: %f\n", G.scene->r.edgeB);


/* FIELDS */
fprintf(set, "\n#Fields\n");
	if(G.scene->r.mode & R_FIELDS)
		fprintf(set, "Do_fields: 1\n");
	else	
		fprintf(set, "Do_fields: 0\n");
		
	if(G.scene->r.mode & R_FIELDSTILL)
		fprintf(set, "Do_fieldstill: 1\n");
	else	
		fprintf(set, "Do_fieldstill: 0\n");
	
	if(G.scene->r.mode & R_ODDFIELD)
		fprintf(set, "Do_fieldodd: 1\n");
	else	
		fprintf(set, "Do_fieldodd: 0\n");

/* ANIMATION */
fprintf(set, "\n#Animation\n");
	fprintf(set, "Frs_sec: %d\n", G.scene->r.frs_sec);
	/*fprintf(set, "Start_frame: %d\n", G.scene->r.sfra);*/
	/*fprintf(set, "End_frame: %d\n", G.scene->r.efra);*/


/* UNIFIED */
fprintf(set, "\n#Unified-renderer-options\n");
	fprintf(set, "Gamma: %f\n", G.scene->r.gamma);
	fprintf(set, "Postmul: %f\n", G.scene->r.postmul);
	fprintf(set, "Postgamma: %f\n", G.scene->r.postgamma);
	fprintf(set, "Postadd: %f\n", G.scene->r.postadd);
	fprintf(set, "Postigamma: %f\n", G.scene->r.postigamma);	
	fprintf(set, "Antishift: %d\n", G.scene->r.same_mat_redux);
	
/* Various */
fprintf(set, "\n#Various\n");
	if(G.scene->r.mode & R_BORDER)
		fprintf(set, "Do_border: 1\n");
	else	
		fprintf(set, "Do_border: 0\n");
		
	if(G.scene->r.mode & R_MOVIECROP)
		fprintf(set, "Do_crop: 1\n");
	else	
		fprintf(set, "Do_crop: 0\n");
	
	if(G.scene->r.mode & R_GAMMA)
		fprintf(set, "Do_gamma: 1\n");
	else	
		fprintf(set, "Do_gamma: 0\n");
		

/* Background options */
fprintf(set, "\n#Background-options\n");
	if(G.scene->r.alphamode == R_ADDSKY)
		fprintf(set, "Addsky: 1\n");
	else	
		fprintf(set, "Addsky: 0\n");
		
	if(G.scene->r.alphamode == R_ALPHAPREMUL)
		fprintf(set, "Alphapremul: 1\n");
	else	
		fprintf(set, "Alphapremul: 0\n");
		
	if(G.scene->r.alphamode == R_ALPHAKEY)
		fprintf(set, "Alphakey: 1\n");
	else	
		fprintf(set, "Alphakey: 0\n");
}

if(set)
fclose(set);

}	

	
void load_rendersettings(void)
{
FILE *set;
char setting[160], str[30];
short comment=0, s;
float f;

if( (set = fopen(G.scene->r.setfile, "r")) == NULL){
	printf("opening file failed");
}
else{
	
	while(!feof(set)){
		comment = 0;
		fscanf(set, "%s", setting);		
		if (!strncmp(setting, "#", sizeof(char))){	
			comment = 1;			
		}
		else{			
			if(!strcmp(setting,"Do_sequence:")){
				fscanf(set, "%d", &s);
				if(s)					
					G.scene->r.scemode |= R_DOSEQ;				
				else
					G.scene->r.scemode &= ~R_DOSEQ;
			}
			else if(!strcmp(setting,"Do_unified:")){
				fscanf(set, "%d", &s);
				if(s)					
					G.scene->r.mode |= R_UNIFIED;				
				else
					G.scene->r.mode &= ~R_UNIFIED;
			}
			else if(!strcmp(setting,"Do_panorama:")){
				fscanf(set, "%d", &s);
				if(s)					
					G.scene->r.mode |= R_PANORAMA;				
				else
					G.scene->r.mode &= ~R_PANORAMA;
			}
			else if(!strcmp(setting,"Size_(perc):")){
				fscanf(set, "%d", &s);
				G.scene->r.size = s; 				
			}
			else if(!strcmp(setting,"Sizex:")){
				fscanf(set, "%d", &s);
				G.scene->r.xsch = s; 				
			}			
			else if(!strcmp(setting,"Sizey:")){
				fscanf(set, "%d", &s);
				G.scene->r.ysch = s; 				
			}
			else if(!strcmp(setting,"Aspx:")){
				fscanf(set, "%d", &s);
				G.scene->r.xasp = s; 				
			}
			else if(!strcmp(setting,"Aspy:")){
				fscanf(set, "%d", &s);
				G.scene->r.yasp = s; 				
			}
			else if(!strcmp(setting,"Xparts:")){
				fscanf(set, "%d", &s);
				G.scene->r.xparts = s; 				
			}
			else if(!strcmp(setting,"Yparts:")){
				fscanf(set, "%d", &s);
				G.scene->r.yparts = s; 				
			}
			else if(!strcmp(setting,"Winpos:")){
				fscanf(set, "%d", &s);
				G.scene->r.winpos = s; 				
			}
			else if(!strcmp(setting,"Do_osa:")){
				fscanf(set, "%d", &s);
				if(s)					
					G.scene->r.mode |= R_OSA;				
				else
					G.scene->r.mode &= ~R_OSA;
			}
			else if(!strcmp(setting,"OSA:")){
				fscanf(set, "%d", &s);
				G.scene->r.osa = s; 				
			}
			else if(!strcmp(setting,"Do_shadow:")){
				fscanf(set, "%d", &s);
				if(s)					
					G.scene->r.mode |= R_SHADOW;				
				else
					G.scene->r.mode &= ~R_SHADOW;
			}
			else if(!strcmp(setting,"Do_mblur:")){
				fscanf(set, "%d", &s);
				if(s)					
					G.scene->r.mode |= R_MBLUR;				
				else
					G.scene->r.mode &= ~R_MBLUR;
			}
			else if(!strcmp(setting,"Blurfactor:")){
				fscanf(set, "%f", &f);
				G.scene->r.blurfac = s; 				
			}
			else if(!strcmp(setting,"Format:")){
				fscanf(set, "%s", str);				
				if(!strcmp(str,"TARGA")){
					G.scene->r.imtype = R_TARGA;									
				}
				else if(!strcmp(str,"IRIS")){
					G.scene->r.imtype = R_IRIS;					
				}
				else if(!strcmp(str,"IRIZ")){										
					G.scene->r.imtype = R_IRIZ;					
				}
				else if(!strcmp(str,"HAMX")){										
					G.scene->r.imtype = R_HAMX;					
				}
				else if(!strcmp(str,"FTYPE")){								
					G.scene->r.imtype = R_FTYPE;									
				}
				else if(!strcmp(str,"JPEG")){									
					G.scene->r.imtype = R_JPEG90;									
				}
				else if(!strcmp(str,"TGARAW")){									
					G.scene->r.imtype = R_RAWTGA;									
				}
				else if(!strcmp(str,"AVIRAW")){										
					G.scene->r.imtype = R_AVIRAW;									
				}
				else if(!strcmp(str,"AVIJPEG")){								
					G.scene->r.imtype = R_AVIJPEG;									
				}
				else if(!strcmp(str,"AVICODEC")){										
					G.scene->r.imtype = R_AVICODEC;									
				}
				else if(!strcmp(str,"QUICKTIME")){								
					G.scene->r.imtype = R_QUICKTIME;
				}				
			}
			else if(!strcmp(setting,"Quality:")){
				fscanf(set, "%d", &s);
				G.scene->r.quality = s; 				
			}
			else if(!strcmp(setting,"Do_extensions:")){
				fscanf(set, "%d", &s);
				if(s)					
					G.scene->r.scemode |= R_EXTENSION;				
				else
					G.scene->r.scemode &= ~R_EXTENSION;
			}
			else if(!strcmp(setting,"Colormode:")){
				fscanf(set, "%s", str);
				if(!strcmp(str,"RGB")){								
					G.scene->r.planes |=R_PLANES24;
				}
				else if(!strcmp(str,"RGBA")){								
					G.scene->r.planes |=R_PLANES32;
				}
				else if(!strcmp(str,"BW")){								
					G.scene->r.planes |=R_PLANESBW;
				}
			}
			else if(!strcmp(setting,"Do_backbuffimage:")){
				fscanf(set, "%d", &s);
				if(s)					
					G.scene->r.bufflag |= R_BACKBUF;				
				else
					G.scene->r.bufflag &= ~R_BACKBUF;
			}
			else if(!strcmp(setting,"Do_backbuffanim:")){
				fscanf(set, "%d", &s);
				if(s)					
					G.scene->r.bufflag |= R_BACKBUFANIM;				
				else
					G.scene->r.bufflag &= ~R_BACKBUFANIM;
			}
			else if(!strcmp(setting,"Do_frontbuffimage:")){
				fscanf(set, "%d", &s);
				if(s)					
					G.scene->r.bufflag |= R_FRONTBUF;				
				else
					G.scene->r.bufflag &= ~R_FRONTBUF;
			}
			else if(!strcmp(setting,"Do_frontbuffanim:")){
				fscanf(set, "%d", &s);
				if(s)					
					G.scene->r.bufflag |= R_FRONTBUFANIM;				
				else
					G.scene->r.bufflag &= ~R_FRONTBUFANIM;
			}
			else if(!strcmp(setting,"Do_edge:")){
				fscanf(set, "%d", &s);
				if(s)					
					G.scene->r.mode |= R_EDGE;				
				else
					G.scene->r.mode &= ~R_EDGE;
			}
			else if(!strcmp(setting,"Edgeint:")){
				fscanf(set, "%d", &s);
				G.scene->r.edgeint = s; 				
			}
			else if(!strcmp(setting,"Edge_R:")){
				fscanf(set, "%f", &f);
				G.scene->r.edgeR = f; 				
			}
			else if(!strcmp(setting,"Edge_G:")){
				fscanf(set, "%f", &f);
				G.scene->r.edgeG = f; 				
			}
			else if(!strcmp(setting,"Edge_B:")){
				fscanf(set, "%f", &f);
				G.scene->r.edgeB = f; 				
			}
			else if(!strcmp(setting,"Do_fields:")){
				fscanf(set, "%d", &s);
				if(s)					
					G.scene->r.mode |= R_FIELDS;				
				else
					G.scene->r.mode &= ~R_FIELDS;
			}
			else if(!strcmp(setting,"Do_fieldstill:")){
				fscanf(set, "%d", &s);
				if(s)					
					G.scene->r.mode |= R_FIELDSTILL;				
				else
					G.scene->r.mode &= ~R_FIELDSTILL;
			}
			else if(!strcmp(setting,"Do_fieldodd:")){
				fscanf(set, "%d", &s);
				if(s)					
					G.scene->r.mode |= R_ODDFIELD;				
				else
					G.scene->r.mode &= ~R_ODDFIELD;
			}
			else if(!strcmp(setting,"Gamma:")){
				fscanf(set, "%f", &f);
				G.scene->r.gamma = f; 				
			}
			else if(!strcmp(setting,"Postmul:")){
				fscanf(set, "%f", &f);
				G.scene->r.postmul = f; 				
			}
			else if(!strcmp(setting,"Postgamma:")){
				fscanf(set, "%f", &f);
				G.scene->r.postgamma = f; 				
			}
			else if(!strcmp(setting,"Postadd:")){
				fscanf(set, "%f", &f);
				G.scene->r.postadd = f; 				
			}
			else if(!strcmp(setting,"Postigamma:")){
				fscanf(set, "%f", &f);
				G.scene->r.postigamma = f; 				
			}
			else if(!strcmp(setting,"Antishift:")){
				fscanf(set, "%d", &s);
				G.scene->r.same_mat_redux = s; 				
			}
			else if(!strcmp(setting,"Do_border:")){
				fscanf(set, "%d", &s);
				if(s)					
					G.scene->r.mode |= R_BORDER;				
				else
					G.scene->r.mode &= ~R_BORDER;
			}
			else if(!strcmp(setting,"Do_crop:")){
				fscanf(set, "%d", &s);
				if(s)					
					G.scene->r.mode |= R_MOVIECROP;				
				else
					G.scene->r.mode &= ~R_MOVIECROP;
			}
			else if(!strcmp(setting,"Do_gamma:")){
				fscanf(set, "%d", &s);
				if(s)					
					G.scene->r.mode |= R_GAMMA;				
				else
					G.scene->r.mode &= ~R_GAMMA;
			}
			else if(!strcmp(setting,"Addsky:")){
				fscanf(set, "%d", &s);
				if(s)					
					G.scene->r.alphamode = R_ADDSKY;				
				else
					G.scene->r.alphamode  = R_ADDSKY;
			}
			else if(!strcmp(setting,"Alphapremul:")){
				fscanf(set, "%d", &s);
				if(s)					
					G.scene->r.alphamode |= R_ALPHAPREMUL;				
				else
					G.scene->r.alphamode &= ~R_ALPHAPREMUL;
			}
			else if(!strcmp(setting,"Alphakey:")){
				fscanf(set, "%d", &s);
				if(s)					
					G.scene->r.alphamode |= R_ALPHAKEY;				
				else
					G.scene->r.alphamode &= ~R_ALPHAKEY;
			}
			else{
				fscanf(set, "%s", str);
			}
						
		}
	}
}

if(set)
fclose(set);

}
