/**
 * $Id: KX_BlenderPolyMaterial.cpp 780 2004-08-05 01:58:38Z kester $
 * ***** BEGIN GPL/BL DUAL LICENSE BLOCK *****
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version. The Blender
 * Foundation also sells licenses for use in proprietary software under
 * the Blender License.  See http://www.blender.org/BL/ for information
 * about this.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software Foundation,
 * Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 *
 * The Original Code is Copyright (C) 2001-2002 by NaN Holding BV.
 * All rights reserved.
 *
 * The Original Code is: all of this file.
 *
 * Contributor(s): none yet.
 *
 * ***** END GPL/BL DUAL LICENSE BLOCK *****
 */
 
#include "GL/glew.h"

#include "KX_BlenderPolyMaterial.h"

#include "BKE_mesh.h"
#include "BKE_global.h"
#include "BKE_image.h"

extern "C" {
#include "BDR_drawmesh.h"
}

#include "DNA_material_types.h"
#include "DNA_texture_types.h"
#include "DNA_image_types.h"
#include "DNA_mesh_types.h"

#include "IMB_imbuf_types.h"

#include "MEM_guardedalloc.h"

#include "RAS_LightObject.h"

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#ifdef MULTITEXTURE
class KX_Texture
{
	MTex *m_mtex;
	unsigned int bind;
	
public:	
	KX_Texture(MTex *mtex) :
		m_mtex(mtex),
		bind(0)
	{
	}
	~KX_Texture()
	{
		if (bind)
			glDeleteTextures(1, &bind);
	}

	int GetTextureType()
	{
		bool alpha = m_mtex->tex->imaflag & TEX_USEALPHA;
		bool compress = GLEW_ARB_texture_compression;
		
		if (m_mtex->texflag & MTEX_RGBTOINT)
		{
			if (alpha)
				return compress?GL_COMPRESSED_LUMINANCE_ALPHA_ARB:GL_LUMINANCE_ALPHA;
			else
				return compress?GL_COMPRESSED_LUMINANCE_ARB:GL_LUMINANCE;
		} else {
			if (!alpha)
				return compress?GL_COMPRESSED_RGB_ARB:GL_RGB;
		}
		
		return compress?GL_COMPRESSED_RGBA_ARB:GL_RGBA;
	}

};

class KX_Material
{
	Material *m_material;
	KX_Texture *m_tex[8];

public:
	KX_Material(Material *mat) :
		m_material(mat)
	{
		for (unsigned int i = 0 ; i < 8 ; i++)
		{
			MTex *mtex = m_material->mtex[i];
			if (mtex && !(m_material->septex & (1<<i)))
				m_tex[i] = new KX_Texture(m_material->mtex[i]);
			else
				m_tex[i] = NULL;
		}
	}
	
	~KX_Material()
	{
		for (unsigned int i = 0; i < 8 ; i++)
			delete m_tex[i];
	}

};

struct KX_BlenderPolyMaterial::MTexCache
{
	MTex *m_lastmtex;
	
	Image *m_curpage;
	bool m_curmode;    //< Tile mode
	int m_curtileXRep;
	int m_curtileYRep;
	int m_curtile;    //< Animation frame
	
	float m_texwindx;
	float m_texwindy;
	float m_texwinsx;
	float m_texwinsy;
	
	int m_textureType; // GL_RGBA, GL_COMPRESSED_RGBA, GL_LUMINANCE etc.
	int m_mipMap;      // mipmap quality: GL_LINEAR_MIPMAP_LINEAR etc
	bool m_doMipMap;   // mipmaps enabled
	
	int m_blendType; // TexEnv mode
	bool m_texgen; // texco gen mode
	
	MTexCache() :
		m_lastmtex(NULL),
		m_curpage(NULL),
		m_curmode(false),
		m_curtileXRep(0),
		m_curtileYRep(0),
		m_curtile(-1),
		m_texwindx(0.0),
		m_texwindy(0.0),
		m_texwinsx(0.0),
		m_texwinsy(0.0),
		m_textureType(GL_RGBA),
		m_mipMap(GL_LINEAR_MIPMAP_LINEAR),
		m_doMipMap(true)
	{
	}

	static bool QueryExtensions()
	{
		if (!GLEW_ARB_multitexture)
		{
			printf("Materials: ARB_multitexture not supported!\n");
			return false;
		}
		
		if (!GLEW_ARB_texture_env_combine)
		{
			printf("Materials: ARB_texture_env_combine not supported!\n");
			return false;
		}
		
		if (!GLEW_ARB_texture_cube_map)
		{
			printf("Materials: ARB_texture_cube_map not supported!\n");
			return false;
		}
		
		return true;
	}

	void disable()
	{
		if(m_lastmtex==0) return;
		
		m_lastmtex= 0;
		m_curtile= 0;
		m_curpage= 0;
		if(m_curmode) {
			glMatrixMode(GL_TEXTURE);
			glLoadIdentity();
			glMatrixMode(GL_MODELVIEW);
		}
		m_curmode= 0;
		m_curtileXRep=0;
		m_curtileYRep=0;
		
		//bc.set_blendmode(0);
		
		glDisable(GL_TEXTURE_2D);
		glDisable(GL_TEXTURE_GEN_S);
		glDisable(GL_TEXTURE_GEN_T);
	
		return;
	}

	void bindimage(unsigned int *bind, unsigned int *rect, int rectw, int recth)
	{
		glGenTextures(1, (GLuint*) bind);
		
		if(G.f & G_DEBUG) {
			printf("Image: %s\n", m_curpage->id.name+2);
			printf("name: %d, w: %d\n", *bind, rectw);
			printf("tile: %d, mode: %d\n", m_curtile, m_curmode);
		}
		glBindTexture( GL_TEXTURE_2D, (GLuint) *bind);
	
		if (!m_doMipMap)
		{
			glTexImage2D(GL_TEXTURE_2D, 0,  m_textureType,  rectw, recth, 0, GL_RGBA, GL_UNSIGNED_BYTE, rect);
			glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
			glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
		} else
		{
			gluBuild2DMipmaps(GL_TEXTURE_2D, m_textureType, rectw, recth, GL_RGBA, GL_UNSIGNED_BYTE, rect);
			glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, m_mipMap);
			glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
		}
	}

	bool SetTexEnv()
	{
		bool mapped = false;
		if (m_lastmtex->mapto & MAP_COL)
		{
			GLint colmode = m_lastmtex->texflag & MTEX_NEGATIVE ? GL_ONE_MINUS_SRC_COLOR : GL_SRC_COLOR;
			glTexEnvi(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_COMBINE_ARB);
			switch (m_lastmtex->blendtype)
			{
				case MTEX_BLEND:
					glTexEnvi(GL_TEXTURE_ENV, GL_COMBINE_RGB_ARB, GL_INTERPOLATE_ARB);
					glTexEnvi(GL_TEXTURE_ENV, GL_SOURCE2_RGB_ARB, GL_TEXTURE);
					glTexEnvi(GL_TEXTURE_ENV, GL_OPERAND2_RGB_ARB, m_lastmtex->tex->flag & TEX_NEGALPHA ? GL_SRC_ALPHA : GL_ONE_MINUS_SRC_ALPHA);
					break;
				case MTEX_MUL:
					glTexEnvi(GL_TEXTURE_ENV, GL_COMBINE_RGB_ARB, GL_MODULATE);
					break;
				case MTEX_ADD:
					glTexEnvi(GL_TEXTURE_ENV, GL_COMBINE_RGB_ARB, GL_ADD);
					break;
				case MTEX_SUB:
					glTexEnvi(GL_TEXTURE_ENV, GL_COMBINE_RGB_ARB, GL_SUBTRACT_ARB);
					break;
			}
		
			glTexEnvi(GL_TEXTURE_ENV, GL_SOURCE1_RGB_ARB, GL_TEXTURE);
			glTexEnvi(GL_TEXTURE_ENV, GL_OPERAND1_RGB_ARB, colmode);
			glTexEnvi(GL_TEXTURE_ENV, GL_SOURCE0_RGB_ARB, GL_PREVIOUS_ARB);
			glTexEnvi(GL_TEXTURE_ENV, GL_OPERAND0_RGB_ARB, GL_SRC_COLOR);
			mapped = true;
		}
		
		if (m_lastmtex->mapto & MAP_ALPHA)
		{
			glTexEnvi(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_COMBINE_ARB);
			switch (m_lastmtex->blendtype)
			{
				case MTEX_BLEND:
					glTexEnvi(GL_TEXTURE_ENV, GL_COMBINE_ALPHA_ARB, GL_INTERPOLATE_ARB);
					glTexEnvi(GL_TEXTURE_ENV, GL_SOURCE2_ALPHA_ARB, GL_TEXTURE);
					glTexEnvi(GL_TEXTURE_ENV, GL_OPERAND2_ALPHA_ARB, m_lastmtex->tex->flag & TEX_NEGALPHA ? GL_SRC_ALPHA : GL_ONE_MINUS_SRC_ALPHA);
					break;
				case MTEX_MUL:
					glTexEnvi(GL_TEXTURE_ENV, GL_COMBINE_ALPHA_ARB, GL_MODULATE);
					break;
				case MTEX_ADD:
					glTexEnvi(GL_TEXTURE_ENV, GL_COMBINE_ALPHA_ARB, GL_ADD);
					break;
				case MTEX_SUB:
					glTexEnvi(GL_TEXTURE_ENV, GL_COMBINE_ALPHA_ARB, GL_SUBTRACT_ARB);
					break;
			}
		
			glTexEnvi(GL_TEXTURE_ENV, GL_SOURCE1_ALPHA_ARB, GL_TEXTURE);
			glTexEnvi(GL_TEXTURE_ENV, GL_OPERAND1_ALPHA_ARB, GL_SRC_ALPHA);
			glTexEnvi(GL_TEXTURE_ENV, GL_SOURCE0_ALPHA_ARB, GL_PREVIOUS_ARB);
			glTexEnvi(GL_TEXTURE_ENV, GL_OPERAND0_ALPHA_ARB, GL_ONE_MINUS_SRC_ALPHA);
			mapped = true;
		}
		
		return mapped;
	}

	void SetTexRepeat()
	{
		int extend = m_lastmtex->tex->extend;
		switch (extend)
		{
			case TEX_EXTEND:
			case TEX_CLIP:
			case TEX_CLIPCUBE:
			{
				float border[4] = {0.0, 0.0, 0.0, 0.0};
				glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP);
				glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP);
				glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_R, GL_CLAMP);
				glTexParameterfv(GL_TEXTURE_2D, GL_TEXTURE_BORDER_COLOR, border);
				break;
			}
			case TEX_REPEAT:
				glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_REPEAT);
				glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_REPEAT);
				glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_R, GL_REPEAT);
				break;
		}
	}

	bool SetTextureType()
	{
		bool alpha = m_lastmtex->tex->imaflag & TEX_USEALPHA;
		bool compress = GLEW_ARB_texture_compression;
		
		if (m_lastmtex->texflag & MTEX_RGBTOINT)
		{
			if (alpha)
				m_textureType = compress?GL_COMPRESSED_LUMINANCE_ALPHA_ARB:GL_LUMINANCE_ALPHA;
			else
				m_textureType = compress?GL_COMPRESSED_LUMINANCE_ARB:GL_LUMINANCE;
		} else {
			if (alpha)
				m_textureType = compress?GL_COMPRESSED_RGBA_ARB:GL_RGBA;
			else
				m_textureType = compress?GL_COMPRESSED_RGB_ARB:GL_RGB;
		}
		
		return m_textureType != GL_RGBA;
	}

	/* REMEMBER!  Changes here must go into my_set_tpage() as well */
	bool set(RAS_IRasterizer* rasty, MTex *mtex)
	{	
		/* disable */
		if(mtex==0 || mtex->tex == 0 || mtex->tex->type != TEX_IMAGE) 
		{
			disable();
			return false;
		}
		m_lastmtex = mtex;
		
		Tex *tex = mtex->tex;
		Image *ima= tex->ima;
	
	
		bool tilemode= tex->extend == TEX_REPEAT;
		int tileXRep = 0;
		int tileYRep = 0;
		if (ima)
		{
			tileXRep = ima->xrep;
			tileYRep = ima->yrep;
		}
	/*
		if(ima==m_curpage && tilemode==m_curmode && m_curtileXRep==tileXRep && m_curtileYRep == tileYRep) return ima!=0;
	
		if(tilemode!=m_curmode || m_curtileXRep!=tileXRep || m_curtileYRep != tileYRep)
		{
			glMatrixMode(GL_TEXTURE);
			glLoadIdentity();
			
			if(tilemode && ima!=0)
				glScalef(ima->xrep, ima->yrep, 1.0);
	
			glMatrixMode(GL_MODELVIEW);
		}
	*/
		if(ima==0 || ima->ok==0) {
			glDisable(GL_TEXTURE_2D);
			
			m_curpage= 0;
			m_curmode= tilemode;
			m_curtileXRep = tileXRep;
			m_curtileYRep = tileYRep;
	
			return false;
		}
	
		if(ima->ibuf==0) {
			load_image(ima, IB_rect, G.sce, /*framenum*/ 0);
			
			if(ima->ibuf==0) {
				ima->ok= 0;
				
				m_curtile= 0;
				m_curpage= 0;
				m_curmode= tilemode;
				m_curtileXRep = tileXRep;
				m_curtileYRep = tileYRep;
				
				glDisable(GL_TEXTURE_2D);
				return false;
			}
			
		}
	
		if(ima->tpageflag & IMA_TWINANIM) m_curtile= ima->lastframe;
	
		int tpx = 0, tpy = 0;
		unsigned int *rect = NULL;
		unsigned int *bind = NULL;
	#if 0	
		if(tilemode) {
	
			if(ima->repbind==0) make_repbind(ima);
			
			if(m_curtile>=ima->totbind) m_curtile= 0;
			
			/* this happens when you change repeat buttons */
			if(ima->repbind) bind= ima->repbind+m_curtile;
			else bind= &ima->bindcode;
			
			if(*bind==0) {
				
				fTexwindx= ima->ibuf->x/ima->xrep;
				fTexwindy= ima->ibuf->y/ima->yrep;
				
				if(m_curtile>=ima->xrep*ima->yrep) m_curtile= ima->xrep*ima->yrep-1;
		
				fTexwinsy= m_curtile / ima->xrep;
				fTexwinsx= m_curtile - fTexwinsy*ima->xrep;
		
				fTexwinsx*= fTexwindx;
				fTexwinsy*= fTexwindy;
		
				tpx= fTexwindx;
				tpy= fTexwindy;
	
				rect= ima->ibuf->rect + int(fTexwinsy)*ima->ibuf->x + int(fTexwinsx);
			}
		}
		else 
	#endif
		{
			bind= &ima->bindcode;
			
			if(*bind==0) {
				tpx= ima->ibuf->x;
				tpy= ima->ibuf->y;
				rect= ima->ibuf->rect;
			}
		}
	
		m_curpage= ima;
		m_curmode= tilemode;
		m_curtileXRep = tileXRep;
		m_curtileYRep = tileYRep;
	
		if(*bind==0)
		{
			int rectw = tpx, recth = tpy;
			unsigned int *tilerect= NULL, *scalerect= NULL;
	
			/*
			* Maarten:
			* According to Ton this code is not needed anymore. It was used only
			* in really old Blenders.
			* Reevan:
			* Actually it is needed for backwards compatibility.  Simpledemo 6 does not display correctly without it.
			*/
	#if 0
			if (m_curmode) {
				int y;
				tilerect= (unsigned int*) MEM_mallocN(rectw*recth*sizeof(*tilerect), "tilerect");
				for (y=0; y<recth; y++) {
					unsigned int *rectrow= &rect[y*ima->ibuf->x];
					void *tilerectrow= &tilerect[y*rectw];
						
					memcpy(tilerectrow, rectrow, rectw*sizeof(*rectrow));
				}
					
				rect= tilerect;
			}
	#endif
			if (!is_pow2(rectw) || !is_pow2(recth)) {
				rectw= smaller_pow2(rectw);
				recth= smaller_pow2(recth);
				
				scalerect= (unsigned int *) MEM_mallocN(rectw*recth*sizeof(*scalerect), "scalerect");
				gluScaleImage(GL_RGBA, tpx, tpy, GL_UNSIGNED_BYTE, rect, rectw, recth, GL_UNSIGNED_BYTE, scalerect);
				rect= scalerect;
			}
			
			m_doMipMap = tex->imaflag & TEX_MIPMAP;
					
			SetTextureType();
			bindimage(bind, rect, rectw, recth);
			
			if (tilerect)
				MEM_freeN(tilerect);
			if (scalerect)
				MEM_freeN(scalerect);
		} else 
			glBindTexture( GL_TEXTURE_2D, *bind);
		
		if (!SetTexEnv())
		{
			glDisable(GL_TEXTURE_2D);
			return false;
		}
		
		SetTexRepeat();
		SetTexCoords(rasty);
		/* Enable or disable reflection mapping */
		if (ima && (ima->flag & IMA_REFLECT))
		{
			glTexGeni(GL_S, GL_TEXTURE_GEN_MODE, GL_SPHERE_MAP);
			glTexGeni(GL_T, GL_TEXTURE_GEN_MODE, GL_SPHERE_MAP);
	
			glEnable(GL_TEXTURE_GEN_S);
			glEnable(GL_TEXTURE_GEN_T);
			rasty->SetTexCoords(RAS_IRasterizer::RAS_TEXCO_GEN);
		}
		
		SetTexMatrix();
	
		glEnable(GL_TEXTURE_2D);
		return true;
	}

	void SetTexMatrix()
	{
		float tmatrix[] = { 0.0, 0.0, 0.0, 0.0,
				0.0, 0.0, 0.0, 0.0,
				0.0, 0.0, 0.0, 0.0,
				m_lastmtex->ofs[0], m_lastmtex->ofs[1], m_lastmtex->ofs[2], 1.0 };
		
		if (m_lastmtex->projx)
			tmatrix[ m_lastmtex->projx*4 - 4] = m_lastmtex->size[0];
		if (m_lastmtex->projy)
			tmatrix[ m_lastmtex->projy*4 - 3] = m_lastmtex->size[1];
		if (m_lastmtex->projz)
			tmatrix[ m_lastmtex->projz*4 - 2] = m_lastmtex->size[2];
		
		glMatrixMode(GL_TEXTURE);
		glMultMatrixf(tmatrix);
		glMatrixMode(GL_MODELVIEW);
	}

	bool SetTexCoords(RAS_IRasterizer* rasty)
	{
		bool texgen = false;
		switch (m_lastmtex->texco)
		{
			default:
				printf("SetTexCoords: %d\n", m_lastmtex->texco);
				return false;
			case TEXCO_ORCO:
				rasty->SetTexCoords(RAS_IRasterizer::RAS_TEXCO_ORCO);
				break;
			case TEXCO_REFL:
				glTexGeni(GL_S, GL_TEXTURE_GEN_MODE, GL_REFLECTION_MAP_ARB);
				glTexGeni(GL_T, GL_TEXTURE_GEN_MODE, GL_REFLECTION_MAP_ARB);
				glTexGeni(GL_R, GL_TEXTURE_GEN_MODE, GL_REFLECTION_MAP_ARB);
				glTexGeni(GL_Q, GL_TEXTURE_GEN_MODE, GL_REFLECTION_MAP_ARB);
				
				texgen = true;
				rasty->SetTexCoords(RAS_IRasterizer::RAS_TEXCO_GEN);
				break;
			case TEXCO_NORM:
				glTexGeni(GL_S, GL_TEXTURE_GEN_MODE, GL_NORMAL_MAP_ARB);
				glTexGeni(GL_T, GL_TEXTURE_GEN_MODE, GL_NORMAL_MAP_ARB);
				glTexGeni(GL_R, GL_TEXTURE_GEN_MODE, GL_NORMAL_MAP_ARB);
				glTexGeni(GL_Q, GL_TEXTURE_GEN_MODE, GL_NORMAL_MAP_ARB);
				
				texgen = true;
				rasty->SetTexCoords(RAS_IRasterizer::RAS_TEXCO_GEN);
				break;
			case TEXCO_UV:
				rasty->SetTexCoords(RAS_IRasterizer::RAS_TEXCO_UV1);
				break;
			case TEXCO_WINDOW:
			{
				float pmatrix[16];
				glGetFloatv(GL_PROJECTION_MATRIX, pmatrix);
				float splane[4] = {pmatrix[0] + pmatrix[3], pmatrix[4] + pmatrix[7], pmatrix[8] + pmatrix[11], pmatrix[12] + pmatrix[15]};
				float tplane[4] = {pmatrix[3] - pmatrix[1], pmatrix[7] - pmatrix[5], pmatrix[11] - pmatrix[9], pmatrix[15] - pmatrix[13]};
				float rplane[4] = {pmatrix[2] + pmatrix[3], pmatrix[6] + pmatrix[7], pmatrix[10] + pmatrix[11], pmatrix[14] + pmatrix[15]};
				glTexGeni(GL_S,  GL_TEXTURE_GEN_MODE, GL_EYE_LINEAR);
				glTexGenfv(GL_S, GL_EYE_PLANE, splane);
				glTexGeni(GL_T,  GL_TEXTURE_GEN_MODE, GL_EYE_LINEAR);
				glTexGenfv(GL_T, GL_EYE_PLANE, tplane);
				glTexGeni(GL_R,  GL_TEXTURE_GEN_MODE, GL_EYE_LINEAR);
				glTexGenfv(GL_R, GL_EYE_PLANE, rplane);
				
				texgen = true;
				rasty->SetTexCoords(RAS_IRasterizer::RAS_TEXCO_GEN);
				break;
			}
			case TEXCO_GLOB:
			{
				float mmatrix[16];
				glGetFloatv(GL_MODELVIEW_MATRIX, mmatrix);
				glMatrixMode(GL_TEXTURE);
				glMultMatrixf(mmatrix);
				glMatrixMode(GL_MODELVIEW_MATRIX);
				rasty->SetTexCoords(RAS_IRasterizer::RAS_TEXCO_ORCO);
				break;
			}	
	/*TODO:		case TEXCO_OBJECT:
				
				glTexGeni(GL_S, GL_TEXTURE_GEN_MODE, GL_OBJECT_LINEAR);
				glTexGenfv(GL_S, GL_OBJECT_PLANE, splane);
				glTexGeni(GL_T, GL_TEXTURE_GEN_MODE, GL_OBJECT_LINEAR);
				glTexGenfv(GL_T, GL_OBJECT_PLANE, tplane);
				
				texgen = true;
			case TEXCO_LAVECTOR:
			case TEXCO_VIEW:
			case TEXCO_STICKY:
			case TEXCO_WINDOW: */
		}
		if (texgen)
		{
			glEnable(GL_TEXTURE_GEN_S);
			glEnable(GL_TEXTURE_GEN_T);
			glEnable(GL_TEXTURE_GEN_R);
		} else
		{
			glDisable(GL_TEXTURE_GEN_S);
			glDisable(GL_TEXTURE_GEN_T);
			glDisable(GL_TEXTURE_GEN_R);
		}
		m_texgen = texgen;
		return true;
	}
};
#endif /* MULTITEXTURE */

/* (n&(n-1)) zeros the least significant bit of n */
static int is_pow2(int num) {
	return ((num)&(num-1))==0;
}
static int smaller_pow2(int num) {
	while (!is_pow2(num))
		num= num&(num-1);
	return num;	
}


KX_BlenderPolyMaterial::KX_BlenderPolyMaterial(const STR_String &texname,
											   bool ba,
											   Material *material,
											   int tile,
											   int tilexrep,
											   int tileyrep,
											   int mode,
											   bool smooth,
											   bool transparant,
											   bool zsort,
											   int subsurf,
											   bool linearsubsurf,
											   int lightlayer,
											   bool bIsTriangle,
											   void* clientobject,
											   struct TFace* tface)
		: RAS_IPolyMaterial(texname,
							false,
							material?STR_String(material->id.name):"",
							tile,
							tilexrep,
							tileyrep,
							mode,
							smooth,
							transparant,
							zsort,
							subsurf,
							linearsubsurf,
							lightlayer,
							bIsTriangle,
							clientobject),
		m_tface(tface),
		m_material(material),
		m_vprogram(0),
		m_fprogram(0),
		m_shader(0)
{
	for (unsigned int i = 0; i < 8 ; i++)
		m_textures[i] = 0;
}

KX_BlenderPolyMaterial::~KX_BlenderPolyMaterial()
{
	delete m_vprogram;
	delete m_fprogram;
	for (unsigned int i = 0; i < 8 ; i++)
		delete m_textures[i];
}

/**
 *  YUK:  This is copied from KX_PythonInit.cpp.
 *
 *  Will do proper object when material moves to Ketsji and gets
 *  python methods.
 */
struct shader_type
{
	RAS_IRasterizer::IShader::Target m_target;
	STR_String m_shader;
	STR_String m_name;
	shader_type(RAS_IRasterizer::IShader::Target target, STR_String shader, STR_String name)
		: m_target(target),
		  m_shader(shader),
		  m_name(name)
	{
	}
};

extern std::vector<shader_type> shader;
extern bool newshader;

bool KX_BlenderPolyMaterial::ActivateShader(RAS_IRasterizer *rasty) const
{
	if (!newshader)
	{
		if (m_shader)
		{
			if (m_shader->isValid())
				m_shader->Enable();
			else
			{
				m_shader->Disable();
				return false;
			}
		}
		return m_shader != NULL;
	}
	
	delete m_shader;
	
	m_shader = rasty->NewShader();
	newshader = false;
	if (!m_shader)
		return false; // GLSL not supported.
	
	std::vector<shader_type>::iterator shit = shader.begin();
	for (; shit != shader.end(); ++shit)
	{
//		m_shader->LoadShader((*shit).m_target, (*shit).m_shader, (*shit).m_name);
	}
	shader.clear();
	if (m_shader->LinkShader())
	{
		m_shader->Enable();
		return true;
	}
	
	m_shader->Disable();
	return false;
}

void KX_BlenderPolyMaterial::ActivateVertexShader(RAS_IRasterizer *rasty) const
{
	extern bool newvprogram;
	extern STR_String vprogram;
	if (newvprogram)
	{
		delete m_vprogram;
		
		m_vprogram = rasty->NewProgram(vprogram, "Vertex Program");
		newvprogram = false;
	} 
	

	if (!m_vprogram)
	{
		if (!rasty->Lighting()->IsEnabled(0))
			return;
/**
 *  Vertex Shader
 *  Based on example shader for Gizmo3D by David Luotonen & Anders Mod�
 *
 *  Inputs:  vertex.position
 *           vertex.color                  Vertex colour
 *           vertex.texcoord[0]            UV coordinates
 * 
 *  Outputs: result.color                  Vertex colour
 *           result.color.secondary        Vertex light diffuse contributions
 *           result.texcoord[0]            UV coordinates
 *           result.texcoord[1]            Vertex normal
 *           result.texcoord[2]            View vector
 *           result.texcoord[3+n]          Light vector n
 */
		STR_String temp;
		STR_String vshader(
"!!ARBvp1.0\n"
"\n"
"PARAM mvinv[4]={state.matrix.modelview.invtrans};\n"
"PARAM mvp[4]  ={state.matrix.mvp};\n"
"\n"
"# Transform & output position\n"
"DP4     result.position.x, mvp[0], vertex.position;\n"
"DP4     result.position.y, mvp[1], vertex.position;\n"
"DP4     result.position.z, mvp[2], vertex.position;\n"
"DP4     result.position.w, mvp[3], vertex.position;\n"
"\n"
"# Transform the normal to eye space.\n"
"TEMP EyeNormal;\n"
"DP3     EyeNormal.x, mvinv[0], vertex.normal;\n"
"DP3     EyeNormal.y, mvinv[1], vertex.normal;\n"
"DP3     EyeNormal.z, mvinv[2], vertex.normal;\n"
"\n");
		unsigned int light;
		unsigned int numlights = rasty->Lighting()->GetNumLights();
		RAS_IRasterizer::ILighting *lighting = rasty->Lighting();

		for (light = 0; light < rasty->GetPerPixelLights() && lighting->IsEnabled(light); light++)
		{
			if (light == 0) // First per pixel light.
			{
				vshader += 
"PARAM m[4]={state.matrix.modelview};\n"
"\n"
"TEMP EyeVertex;\n"
"\n"
"#Vertex position in eye space...\n"
"DP4     EyeVertex.x, m[0], vertex.position;\n"
"DP4     EyeVertex.y, m[1], vertex.position;\n"
"DP4     EyeVertex.z, m[2], vertex.position;\n"
"\n"
"\n"
"MOV     result.texcoord[2], -EyeVertex;\n"
"MOV     result.texcoord[1], EyeNormal;\n";
			}
			
			switch (lighting->Light(light)->m_type)
			{
				default:
				case RAS_LightObject::LIGHT_SPOT:
				case RAS_LightObject::LIGHT_NORMAL:
					vshader += temp.Format(
"SUB     result.texcoord[%d], state.light[%d].position, EyeVertex;\n", light + 3, light);
					break;
				case RAS_LightObject::LIGHT_SUN:
					vshader += temp.Format(
"MOV     result.texcoord[%d], state.light[%d].position;\n"
					, light + 3, light);
					break;
			}
		}
		
		// Add the rest of the lights as per vertex lights.
		for (light = rasty->GetPerPixelLights(); light < numlights && lighting->IsEnabled(light); light++)
		{
			if (light == rasty->GetPerPixelLights()) // First vertex light
				vshader +=
"# Compute diffuse dot products\n"
"TEMP  dots, att, colour;\n"
"ALIAS lightVec = dots;\n";
			switch (lighting->Light(light)->m_type)
			{
				case RAS_LightObject::LIGHT_SPOT:
					vshader += temp.Format(
// Diffuse component
"SUB   lightVec, state.light[%d].position, EyeVertex;\n"
"DP3   att, lightVec, lightVec;\n"
"RSQ   lightVec.w, att.w;\n"
"MUL   lightVec, lightVec.w, lightVec;\n" // dots = normalised lightvec
"DST   att, att, lightVec.w;\n" // attenuation = [1.0, d, d^2, 1/d] where d = |lightVec|
"DP3   dots.w, EyeNormal, lightVec;\n"
"DP3   att.x, att, state.light[%d].attenuation;\n"
"RCP   att.x, att.x;\n"

"# Spot attenuation.\n"
"DP3   att.y, state.light[%d].spot.direction, -lightVec;\n"       // y = spot-lightvec angle
"SGE   att.z, att.y, state.light[%d].spot.direction.w;\n" // z = 1.0 if inside cone
"MOV   att.w, state.light[%d].attenuation.w;\n"    // w = spot exponent
"LIT   att.z, att.zyyw;\n"                 // z = y^w if inside cone, 0.0 if outside
"MUL   att.x, att.x, att.z;\n"

"MUL   dots.w, att.x, dots.w;\n"
	, light, light, light, light, light);
					break;

				case RAS_LightObject::LIGHT_SUN:
					vshader += temp.Format(
// Diffuse component
"DP3   dots.w, EyeNormal, state.light[%d].position;\n", light);
					break;
				default:
				case RAS_LightObject::LIGHT_NORMAL:
					vshader += temp.Format(
// Diffuse component
"SUB   lightVec, state.light[%d].position, EyeVertex;\n"
"DP3   att, lightVec, lightVec;\n"
"RSQ   lightVec.w, att.w;\n"
"MUL   lightVec, lightVec.w, lightVec;\n"             // normalise lightvec
"DST   att, att, lightVec.w;\n"                       // att = [1.0, d, d^2, 1/d] where d = |lightVec|
"DP3   dots.w, EyeNormal, lightVec;\n"                // dots.w = diffuse
"DP3   att.x, att, state.light[%d].attenuation;\n"    
"RCP   att.x, att.x;\n"                               // att.x = 1/attenuation
"MUL   dots.w, att.x, dots.w;\n"                      // dots.w = attenuated diffuse.
	, light, light);
					break;
			}

			STR_String dest("colour");
			if (!lighting->IsEnabled(light + 1))
				dest = "result.color.secondary"; // Last per vertex light.
			
			if (light == rasty->GetPerPixelLights()) // First per vertex light
			{
				vshader += temp.Format(
"MUL   %s,                 dots.w, state.lightprod[%d].diffuse;\n", dest.ReadPtr(), light);
				
			}
			else
			{
				vshader += temp.Format(
"MAD   %s,                 dots.w, state.lightprod[%d].diffuse, colour;\n", dest.ReadPtr(), light);
			}
		}
		if (light == rasty->GetPerPixelLights())
		{
			vshader +=
"MOV     result.color.secondary, {0.0, 0.0, 0.0, 1.0};\n"; // No per vertex lights.
		}
		vshader +=
"MOV     result.texcoord[0], vertex.texcoord[0];\n"
"MOV     result.color, vertex.color;\n"
"END\n";
		STR_String name;
		name.Format("Vertex Shader %d/%d", rasty->GetPerPixelLights(), light);
		m_vprogram = rasty->NewProgram(vshader, name);
	} 
	
	if (m_vprogram)
	{
		if (m_vprogram->isValid())
			m_vprogram->Enable();
		else
			m_vprogram->Disable();
	}
}

void KX_BlenderPolyMaterial::ActivateFragmentShader(RAS_IRasterizer *rasty) const 
{
	extern bool newfprogram;
	extern STR_String fprogram;
	if (newfprogram)
	{
		delete m_fprogram;
			
		m_fprogram = rasty->NewProgram(fprogram, "Fragment Program");
		newfprogram = false;
	} 

	if (!m_fprogram)
	{
		if (!rasty->Lighting()->IsEnabled(0))
			return;

/*
 *  Fragment Shader
 *  Based on example shader for Gizmo3D by David Luotonen & Anders Mod�
 *
 *  Input: vertex.color               Vertex colour
 *         vertex.color.secondary     Additional diffuse light
 *         vertex.texcoord[0]         UV coordinates
 *         vertex.texcoord[1]         Vertex normal
 *         vertex.texcoord[2]         View vector
 *         vertex.texcoord[3+n]       Light Vector n.
 */
		STR_String fshader(
"!!ARBfp1.0\n"
"\n"
"OPTION ARB_precision_hint_nicest;\n"
"\n"
"TEMP normal, lightVec, viewVec, reflVec, attenuation, finalCol, lightContrib, tex;\n"
"TEMP ambient;\n"
"PARAM  unit = {0.0, 0.5, 1.0, 2.0};\n"
"\n"
// Normalise view directions
"DP3             viewVec.w, fragment.texcoord[2], fragment.texcoord[2];\n"
"RSQ             viewVec.w, viewVec.w;\n"
"MUL             viewVec.xyz, fragment.texcoord[2], viewVec.w;\n");

		fshader +=
// Normalise normal
"DP3             normal.w, fragment.texcoord[1], fragment.texcoord[1];\n"
"RSQ             normal.w, normal.w;\n"
"MUL             normal.xyz, fragment.texcoord[1], normal.w;\n";

		if (m_drawingmode & 4) // Texture
		{
			// FIXME: proper check for sphere map.
			if (glIsEnabled(GL_TEXTURE_GEN_S))
			{
				fshader += 
// Compute reflection vector.
"DP3             reflVec, viewVec, normal;\n"
"MUL             reflVec, reflVec, 2.0;\n"
"MAD             reflVec, reflVec, normal, -viewVec;\n"

// tex.y = reflVec . y'         -1 <= tex.y <= 1
"DP3             tex.y, reflVec, unit.xyxx;\n"
"DP3             tex.x, reflVec, unit.yxxx;\n"
// tex.x = 0.5*tex.x + 0.5      0 <= tex.x <= 1
"MAD             tex, 0.5, tex, 0.5;\n"

"TEX             tex, tex, texture[0], 2D;\n"
"MUL             tex, fragment.color, tex;\n"
"\n";
			}
			else
			{
				fshader += 
"TEX             tex, fragment.texcoord[0], texture[0], 2D;\n"
"MUL             tex, fragment.color, tex;\n"
"\n";
			}
		}
		else
		{
			fshader +=
"MOV             tex, fragment.color;\n";
		}

		RAS_IRasterizer::ILighting *lighting = rasty->Lighting();
		STR_String temp;
		unsigned int light;
		for (light = 0; light < rasty->GetPerPixelLights() && lighting->IsEnabled(light); light++)
		{
			unsigned int texco = 3 + light;
			fshader += temp.Format(
"DP3             attenuation, fragment.texcoord[%d], fragment.texcoord[%d];\n"       // attenuation = |lightVec|^2
"RSQ             lightVec.w, attenuation.w;\n"
"MUL             lightVec, fragment.texcoord[%d], lightVec.w;\n", texco, texco, texco); // normalise lightvec

			fshader +=
"\n"
"#Diffuse\n"
"DP3_SAT lightContrib.x, normal, lightVec;\n"
"\n"
"#Specular\n"
"# Phong:\n"
//"DP3             reflVec, lightVec, normal;\n"
"MUL             reflVec, lightContrib.x, unit.w;\n"
"MAD             reflVec, reflVec, normal, -lightVec;\n"
"\n"
"DP3 lightContrib.y, reflVec, viewVec;\n"
"\n"
"# Blinn:\n"
"#       ADD             reflVec, lightVec, viewVec;     # reflVec == Half-angle.\n"
"#       DP3             reflVec.w, reflVec, reflVec;\n"
"#       RSQ             reflVec.w, reflVec.w;\n"
"#       MUL             reflVec.xyz, reflVec, reflVec.w;\n"
"#       DP3_SAT             lightContrib.y, reflVec, normal;\n"
"\n"
"MOV             lightContrib.w, state.material.shininess.x;\n"
"\n"
"#Accelerates lighting computations\n"
"LIT     lightContrib, lightContrib;\n";
			if (lighting->Light(light)->m_type != RAS_LightObject::LIGHT_SUN)
			{
					fshader += temp.Format(
"#Calculate attenuation.\n"
"DST             attenuation, attenuation, lightVec.w;\n" // attenuation = [1.0, d, d^2, 1/d] where d = |lightVec|
"DP3             attenuation.x, attenuation, state.light[%d].attenuation;\n"
"RCP             attenuation.x, attenuation.x;\n"
					, light);
				if (lighting->Light(light)->m_type == RAS_LightObject::LIGHT_SPOT)
				{
					fshader += temp.Format(
"# Spot attenuation.\n"
"DP3             attenuation.y, state.light[%d].spot.direction, -lightVec;\n"       // y = spot-lightvec angle
"SGE             attenuation.z, attenuation.y, state.light[%d].spot.direction.w;\n" // z = 1.0 if inside cone
"MOV             attenuation.w, state.light[%d].attenuation.w;\n"    // w = spot exponent
"LIT             attenuation.z, attenuation.zyyw;\n"                 // z = y^w if inside cone, 0.0 if outside
"MUL             attenuation.x, attenuation.x, attenuation.z;\n"
					, light, light, light);          
				}
					fshader +=
"MUL             lightContrib, attenuation.x, lightContrib;\n";
			}
			if (light == 0)
			{
				fshader += temp.Format(
//"MUL             finalCol, lightContrib.y, state.lightprod[%d].diffuse;\n" // DEBUG: ignore per vertex contributions.
"MAD             finalCol, lightContrib.y, state.lightprod[%d].diffuse, fragment.color.secondary;\n"
"MUL             ambient, state.lightprod[%d].ambient, tex;\n"
"MAD             ambient, lightContrib.z, state.lightprod[%d].specular, ambient;\n"
					, light, light, light);

			}
			else
			{
				fshader += temp.Format(
"MAD             finalCol, lightContrib.y, state.lightprod[%d].diffuse, finalCol;\n"
"MAD             ambient, state.lightprod[%d].ambient, tex, ambient;\n"
"MAD             ambient, lightContrib.z, state.lightprod[%d].specular, ambient;\n"
					, light, light, light);
			}
		}
		
			fshader += 
"\n"
"MAD             result.color.rgb, finalCol, tex, ambient;\n"      // Multiply diffuse by texture & add ambient/specular
"MOV             result.color.a, tex.a; #, fragment.color.a;\n"    // Pass alpha through.
"END\n";
		STR_String name;
		name.Format("Fragment Shader %d/%d", rasty->GetPerPixelLights(), light);
		m_fprogram = rasty->NewProgram(fshader, name);
	} 
	
	if (m_fprogram)
	{
		if (m_fprogram->isValid())
			m_fprogram->Enable();
		else
			m_fprogram->Disable();
	}
}

void KX_BlenderPolyMaterial::ConvertMaterial(RAS_IRasterizer* rasty) const
{
	unsigned int t;
	unsigned int i;
	if (m_textures[0])
		return;
	// TODO
	for ( i = 0, t = 0; i < 8; i++)
	{
		MTex* mtex = m_material->mtex[i];
		if (!mtex)
			continue;
			
		if ((m_material->septex & (1<<i))) // Tex channel disabled
			continue;
		
		m_textures[t] = CreateTexture(rasty, mtex);
		if (m_textures[t])
			t++;
	}
}

RAS_IRasterizer::ITexture* KX_BlenderPolyMaterial::CreateTexture(RAS_IRasterizer* rasty, MTex* mtex) const
{
	/* disable */
	if(mtex==0 || mtex->tex == 0 || mtex->tex->type != TEX_IMAGE) 
		return NULL;
		
	Tex *tex = mtex->tex;
	Image *ima= tex->ima;
	
	if(ima==0 || ima->ok==0) 
	{
		return NULL;
	}
	
	if(ima->ibuf==0) 
	{
		load_image(ima, IB_rect, G.sce, /*framenum*/ 0);
			
		if(ima->ibuf==0) {
			ima->ok= 0;
				
			return NULL;
		}
	
	}
	
	unsigned int *bind= &ima->bindcode;
	
	if(*bind==0) {
		int tpx= ima->ibuf->x;
		int tpy= ima->ibuf->y;
		unsigned int *rect= ima->ibuf->rect;
			
		int rectw = tpx, recth = tpy;
		unsigned int *tilerect= NULL, *scalerect= NULL;
	
		if (!is_pow2(rectw) || !is_pow2(recth)) {
			rectw= smaller_pow2(rectw);
			recth= smaller_pow2(recth);
			
			scalerect= (unsigned int *) MEM_mallocN(rectw*recth*sizeof(*scalerect), "scalerect");
			gluScaleImage(GL_RGBA, tpx, tpy, GL_UNSIGNED_BYTE, rect, rectw, recth, GL_UNSIGNED_BYTE, scalerect);
			rect= scalerect;
		}
		
		RAS_IRasterizer::ITexture *texture = rasty->NewTexture();
		//m_normap->SetMipMaps(tex->imaflag & TEX_MIPMAP);
		texture->SetMipMap(RAS_IRasterizer::ITexture::LINEAR, RAS_IRasterizer::ITexture::LINEAR);
		texture->SetCompression(false);
		texture->SetTextureFormat(mtex->texflag & MTEX_RGBTOINT, mtex->tex->imaflag & TEX_USEALPHA);
		
		texture->BindImage(rect, rectw, recth);
		
		if (tilerect)
			MEM_freeN(tilerect);
		if (scalerect)
			MEM_freeN(scalerect);
		return texture;
	} 
	else 
	{
		return rasty->NewTexture(*bind);
	}
	
	return NULL;
}

bool KX_BlenderPolyMaterial::Activate(RAS_IRasterizer* rasty, TCachingInfo& cachingInfo) const 
{
	unsigned int i;
	unsigned int unit = 0;
	bool multipass = false;
	
	if (GetCachingInfo() != cachingInfo)
	{
		if (!cachingInfo)
			set_tpage(NULL);
		
		cachingInfo = GetCachingInfo();

		if (m_material)
			ConvertMaterial(rasty);
		
		if ((m_drawingmode & 4)&& (rasty->GetDrawingMode() == RAS_IRasterizer::KX_TEXTURED))
		{
			bool blend = false;
			unit = 0;
			update_realtime_texture(m_tface, rasty->GetTime());
			rasty->SetTexUnit(0);
			rasty->EnableTextures(true);
			set_tpage(m_tface);
			rasty->SetTexCoords(RAS_IRasterizer::RAS_TEXCO_UV1);
			unit++;
			
			for (i = 0; i < 8 && m_textures[i]; i++)
			{
				rasty->SetTexUnit(unit);
				rasty->SetTexCoords(RAS_IRasterizer::RAS_TEXCO_UV1);
				m_textures[i]->Bind();
				rasty->EnableTextures(true);
				unit++;
			}
			
			for (; unit < rasty->GetNumTextureUnits(); unit++)
			{
				rasty->SetTexUnit(unit);
				rasty->SetTexCoords(RAS_IRasterizer::RAS_TEXCO_DISABLE);
			}
		}
		else
		{
			rasty->SetTexUnit(0);
			set_tpage(NULL);
		}
		
		if (ActivateShader(rasty))
		{
			while (unit-- > 0)
			{
				STR_String uniform;
				int loc = m_shader->GetUniformLocation(uniform.Format("texture%d", unit));
				if (loc >= 0)
					m_shader->SetUniform(loc, unit);
			}
		}
		else
		{
			if (GLEW_ARB_shading_language_100 && GLEW_ARB_shader_objects && GLEW_ARB_vertex_shader && GLEW_ARB_fragment_shader)
			{
				glUseProgramObjectARB(0);
			}
			
			if (GLEW_ARB_fragment_program && GLEW_ARB_vertex_program && rasty->GetPerPixelLights())
			{
				rasty->SetTexUnit(0);
				if (GetDrawingMode()&16) // Light
				{
					ActivateVertexShader(rasty);
					ActivateFragmentShader(rasty);
				} else {
					glDisable(GL_VERTEX_PROGRAM_ARB);
					glDisable(GL_FRAGMENT_PROGRAM_ARB);
				}
			}
		}
		
		if(m_drawingmode & RAS_IRasterizer::KX_TWOSIDE)
		{
			rasty->SetCullFace(false);
		}
		else
		{
			rasty->SetCullFace(true);
		}

		if (m_drawingmode & RAS_IRasterizer::KX_LINES) {
			rasty->SetLines(true);
		}
		else {
			rasty->SetLines(false);
		}
		rasty->SetSubsurf(m_subsurf, m_linearsubsurf);
	}

	rasty->SetSpecularity(m_specular[0],m_specular[1],m_specular[2],m_specularity);
	rasty->SetShinyness(m_shininess);
	if (m_material)
	{
		rasty->SetDiffuse(m_diffuse[0], m_diffuse[1],m_diffuse[2], m_material->alpha);
		rasty->SetAmbient(MT_Vector4(m_diffuse[0], m_diffuse[1],m_diffuse[2], m_material->amb));
		//rasty->SetEmit(MT_Vector4(m_diffuse[0], m_diffuse[1],m_diffuse[2], m_material->emit));
	} else {
		rasty->SetDiffuse(m_diffuse[0], m_diffuse[1],m_diffuse[2], 1.0);
	}
	
	return multipass;
}

bool KX_BlenderPolyMaterial::ActivateNextPass(RAS_IRasterizer *rasty, unsigned int pass) const
{
	if (!pass)
		return true;
		
	if (!m_material)
		return false;
	bool dopass = false;
#ifdef MULTITEXTURE
	unsigned int i, unit;
	int blendmode = -1;
	for (i = m_unit, unit = 0; i < 8 && unit < rasty->GetNumTextureUnits(); i++)
	{
		MTex* mtex = m_material->mtex[i];
		if (!mtex)
			continue;
			
		if ((m_material->septex & (1<<i)))
			continue;
			
		rasty->SetTexUnit(unit);
		if (mtexs[unit].set(rasty, mtex))
		{
			if (blendmode < 0)
			{
				const int blend2tf[] = {TF_ALPHA, TF_MUL, TF_ADD, TF_SUB};
				blendmode = blend2tf[mtex->blendtype];
			}
			unit++;
		}
	}
	m_unit = i;
	dopass = unit;
	for (; unit < rasty->GetNumTextureUnits(); unit++)
	{
		rasty->SetTexUnit(unit);
		rasty->SetTexCoords(RAS_IRasterizer::RAS_TEXCO_DISABLE);
	}
	if (dopass)
	{
		bc.set_blendmode(blendmode);
		rasty->SetSpecularity(0.0, 0.0, 0.0, 0.0);
		rasty->SetShinyness(0.0);
		rasty->SetDiffuse(0.0, 0.0, 0.0, 0.0);
		rasty->SetAmbient(MT_Vector4(0.0, 0.0, 0.0, 0.0));
		rasty->SetEmit(MT_Vector4(0.0, 0.0, 0.0, 0.0));
	}
#endif /* MULTITEXTURE */
	return dopass;
}

#ifdef MULTITEXTURE
std::vector<KX_BlenderPolyMaterial::MTexCache> KX_BlenderPolyMaterial::mtexs;
KX_BlenderPolyMaterial::BlendCache KX_BlenderPolyMaterial::bc;
#endif
