/**
 * $Id: KX_BlenderPolyMaterial.h 647 2004-07-02 12:28:37Z kester $
 *
 * ***** BEGIN GPL/BL DUAL LICENSE BLOCK *****
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version. The Blender
 * Foundation also sells licenses for use in proprietary software under
 * the Blender License.  See http://www.blender.org/BL/ for information
 * about this.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software Foundation,
 * Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 *
 * The Original Code is Copyright (C) 2001-2002 by NaN Holding BV.
 * All rights reserved.
 *
 * The Original Code is: all of this file.
 *
 * Contributor(s): none yet.
 *
 * ***** END GPL/BL DUAL LICENSE BLOCK *****
 */
#ifndef __KX_BLENDERPOLYMATERIAL
#define __KX_BLENDERPOLYMATERIAL

#include "RAS_MaterialBucket.h"
#include "RAS_IRasterizer.h"

struct TFace;
struct Material;
struct MTex;
extern "C" int set_tpage(TFace* tface);		/* Worst hack ever */

class KX_BlenderPolyMaterial : public RAS_IPolyMaterial
{
//	class MTexCache;
//	class BlendCache;
	/** Blender texture face structure. */
	TFace* m_tface;
	Material* m_material;
//	mutable unsigned int m_unit;
	
//	static std::vector<MTexCache> mtexs;
//	static BlendCache bc;

	mutable RAS_IRasterizer::IProgram *m_vprogram, *m_fprogram;
	mutable RAS_IRasterizer::IShader  *m_shader;
	mutable RAS_IRasterizer::ITexture *m_textures[8];
	
	void ActivateVertexShader(RAS_IRasterizer *rasty) const;
	void ActivateFragmentShader(RAS_IRasterizer *rasty) const;
	bool ActivateShader(RAS_IRasterizer *rasty) const;
	
	void ActivateNorMTex(RAS_IRasterizer *rasty) const;
	
	RAS_IRasterizer::ITexture* CreateTexture(RAS_IRasterizer *rasty, MTex *mtex) const;
	void ConvertMaterial(RAS_IRasterizer* rasty) const;
public:
	
	KX_BlenderPolyMaterial(const STR_String &texname,
		bool ba,
		Material* matname,
		int tile,
		int tilexrep,
		int tileyrep,
		int mode,
		bool smooth,
		bool transparant,
		bool zsort,
		int subsurf,
		bool linearsubsurf,
		int lightlayer,
		bool bIsTriangle,
		void* clientobject,
		struct TFace* tface);
	virtual ~KX_BlenderPolyMaterial();
	
	/**
	 * Returns the caching information for this material,
	 * This can be used to speed up the rasterizing process.
	 * @return The caching information.
	 */
	virtual TCachingInfo GetCachingInfo(void) const;

	/**
	 * Activates the material in the (OpenGL) rasterizer.
	 * On entry, the cachingInfo contains info about the last activated material.
	 * On exit, the cachingInfo should contain updated info about this material.
	 * @param rasty			The rasterizer in which the material should be active.
	 * @param cachingInfo	The information about the material used to speed up rasterizing.
	 */
	virtual bool Activate(RAS_IRasterizer* rasty, TCachingInfo& cachingInfo) const;
	virtual bool ActivateNextPass(RAS_IRasterizer *rasty, unsigned int pass) const;

	/**
	 * Returns the Blender texture face structure that is used for this material.
	 * @return The material's texture face.
	 */
	TFace* GetTFace(void) const;
protected:
private:
};


inline TFace* KX_BlenderPolyMaterial::GetTFace(void) const
{
	return m_tface;
}

inline RAS_IPolyMaterial::TCachingInfo KX_BlenderPolyMaterial::GetCachingInfo(void) const
{
	return GetTFace();
}

#endif // __KX_BLENDERPOLYMATERIAL

