/**
 * $Id: KX_BlenderRenderTools.cpp 719 2004-07-10 03:18:57Z kester $
 * ***** BEGIN GPL/BL DUAL LICENSE BLOCK *****
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version. The Blender
 * Foundation also sells licenses for use in proprietary software under
 * the Blender License.  See http://www.blender.org/BL/ for information
 * about this.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software Foundation,
 * Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 *
 * The Original Code is Copyright (C) 2001-2002 by NaN Holding BV.
 * All rights reserved.
 *
 * The Original Code is: all of this file.
 *
 * Contributor(s): none yet.
 *
 * ***** END GPL/BL DUAL LICENSE BLOCK *****
 */

#include "KX_BlenderRenderTools.h"

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include "GL/glew.h"

#include "DNA_texture_types.h"
#include "DNA_image_types.h"
#include "DNA_text_types.h"

#include "BKE_global.h"
#include "BKE_main.h"
#include "BKE_image.h"

#include "IMB_imbuf_types.h"

#include "MEM_guardedalloc.h"

#include "RAS_IRenderTools.h"
#include "RAS_IRasterizer.h"
#include "RAS_LightObject.h"
#include "RAS_ICanvas.h"

// next two includes/dependencies come from the shadow feature
// it needs the gameobject and the sumo physics scene for a raycast
#include "KX_GameObject.h"

//#include "KX_BlenderPolyMaterial.h"
#include "KX_PolygonMaterial.h"
#include "Value.h"

#include "KX_BlenderGL.h" // for text printing
#include "STR_String.h"
#include "RAS_BucketManager.h" // for polymaterial (needed for textprinting)

#include "SM_Scene.h"
#include "SumoPhysicsEnvironment.h"
#include "KX_SumoPhysicsController.h"
#include "KX_Scene.h"

extern "C" int set_tpage(TFace* tface);		/* Worst hack ever */

KX_BlenderRenderTools::KX_BlenderRenderTools()
{
}

/**
ProcessLighting performs lighting on objects. the layer is a bitfield that contains layer information.
There are 20 'official' layers in blender.
A light is applied on an object only when they are in the same layer.
OpenGL has a maximum of 8 lights (simultaneous), so 20 * 8 lights are possible in a scene.
*/

int	KX_BlenderRenderTools::ProcessLighting(RAS_IRasterizer* rasty, int layer) 
{
	int result = false;

	if (layer < 0)
	{
		rasty->Lighting()->DisableLighting();
		//DisableOpenGLLights();
		result = false;
	} else
	{
		if (m_clientobject)
		{
			if (applyLights(rasty, layer))
			{
				rasty->Lighting()->EnableLighting();
				result = true;
			} else
			{
				rasty->Lighting()->DisableLighting();
				result = false;
			}

			
		}
	}
	return result;
	
	
}


void KX_BlenderRenderTools::BeginFrame(RAS_IRasterizer* rasty)
{
	m_clientobject = NULL;
	m_lastblenderobject = NULL;
	m_lastblenderlights = false;
	m_lastlayer = -1;
	m_lastlighting = false;
	m_modified = true;
	rasty->Lighting()->DisableLighting();
}
	

void KX_BlenderRenderTools::applyTransform(RAS_IRasterizer* rasty,double* oglmatrix,int objectdrawmode )
{
	/* FIXME:
	blender: intern/moto/include/MT_Vector3.inl:42: MT_Vector3 operator/(const
	MT_Vector3&, double): Assertion `!MT_fuzzyZero(s)' failed. 
	
	Program received signal SIGABRT, Aborted. 
	[Switching to Thread 16384 (LWP 1519)] 
	0x40477571 in kill () from /lib/libc.so.6 
	(gdb) bt 
	#7  0x08334368 in MT_Vector3::normalized() const () 
	#8  0x0833e6ec in KX_BlenderRenderTools::applyTransform(RAS_IRasterizer*, double*, int) () 
	*/

	if (objectdrawmode & RAS_IPolyMaterial::BILLBOARD_SCREENALIGNED ||
		objectdrawmode & RAS_IPolyMaterial::BILLBOARD_AXISALIGNED)
	{
		// rotate the billboard/halo
		//page 360/361 3D Game Engine Design, David Eberly for a discussion
		// on screen aligned and axis aligned billboards
		// assumed is that the preprocessor transformed all billboard polygons
		// so that their normal points into the positive x direction (1.0 , 0.0 , 0.0)
		// when new parenting for objects is done, this rotation
		// will be moved into the object
		
		MT_Point3 objpos (oglmatrix[12],oglmatrix[13],oglmatrix[14]);
		MT_Point3 campos = rasty->GetCameraPosition();
		MT_Vector3 dir = (campos - objpos).safe_normalized();
		MT_Vector3 up(0,0,1.0);

		KX_GameObject* gameobj = (KX_GameObject*) this->m_clientobject;
		// get scaling of halo object
		MT_Vector3  size = gameobj->GetSGNode()->GetLocalScale();
		
		bool screenaligned = (objectdrawmode & RAS_IPolyMaterial::BILLBOARD_SCREENALIGNED)!=0;//false; //either screen or axisaligned
		if (screenaligned)
		{
			up = (up - up.dot(dir) * dir).safe_normalized();
		} else
		{
			dir = (dir - up.dot(dir)*up).safe_normalized();
		}

		MT_Vector3 left = dir.normalized();
		dir = (left.cross(up)).normalized();

		// we have calculated the row vectors, now we keep
		// local scaling into account:

		left *= size[0];
		dir  *= size[1];
		up   *= size[2];
		double maat[16]={
			left[0], left[1],left[2], 0,
				dir[0], dir[1],dir[2],0,
				up[0],up[1],up[2],0,
				0,0,0,1};
			glTranslated(objpos[0],objpos[1],objpos[2]);
			glMultMatrixd(maat);
			
	} else
	{
		if (objectdrawmode & RAS_IPolyMaterial::SHADOW)
		{
			// shadow must be cast to the ground, physics system needed here!
			MT_Point3 frompoint(oglmatrix[12],oglmatrix[13],oglmatrix[14]);
			KX_GameObject *gameobj = (KX_GameObject*) this->m_clientobject;
			MT_Vector3 direction = MT_Vector3(0,0,-1);

			
			direction.normalize();
			direction *= 100000;

			MT_Point3 topoint = frompoint + direction;
			MT_Point3 resultpoint;
			MT_Vector3 resultnormal;

			//todo:
			//use physics abstraction
			KX_Scene* kxscene = (KX_Scene*) m_auxilaryClientInfo;
			SumoPhysicsEnvironment *spe = dynamic_cast<SumoPhysicsEnvironment *>( kxscene->GetPhysicsEnvironment());
			SM_Scene *scene = spe->GetSumoScene();
			KX_SumoPhysicsController *spc = dynamic_cast<KX_SumoPhysicsController *>( gameobj->GetPhysicsController());
			KX_GameObject *parent = gameobj->GetParent();
			if (!spc && parent)
				spc = dynamic_cast<KX_SumoPhysicsController *>(parent->GetPhysicsController());
			if (parent)
				parent->Release();
			SM_Object *thisObj = spc?spc->GetSumoObject():NULL;
			
			if (scene->rayTest(thisObj, frompoint, topoint, resultpoint, resultnormal))
			{
				MT_Vector3 left(oglmatrix[0],oglmatrix[1],oglmatrix[2]);
				MT_Vector3 dir = -(left.cross(resultnormal)).safe_normalized();
				left = (dir.cross(resultnormal)).safe_normalized();
				// for the up vector, we take the 'resultnormal' returned by the physics
				
				double maat[16]={
					        left[0],        left[1],        left[2], 0,
					         dir[0],         dir[1],         dir[2], 0,
					resultnormal[0],resultnormal[1],resultnormal[2], 0,
					              0,              0,              0, 1};
				glTranslated(resultpoint[0],resultpoint[1],resultpoint[2]);
				//glMultMatrixd(oglmatrix);
				glMultMatrixd(maat);
			} else
			{
				glMultMatrixd(oglmatrix);
			}
		} else
		{

			// 'normal' object
			glMultMatrixd(oglmatrix);
		}
	}
}


/**
Render Text renders text into a (series of) polygon, using a texture font,
Each character consists of one polygon (one quad or two triangles)
*/
void	KX_BlenderRenderTools::RenderText(int mode,RAS_IPolyMaterial* polymat,float v1[3],float v2[3],float v3[3],float v4[3])
{
		
	STR_String mytext = ((CValue*)m_clientobject)->GetPropertyText("Text");
	
//	KX_BlenderPolyMaterial* blenderpoly = (KX_BlenderPolyMaterial*)polymat;
	KX_PolygonMaterial* blenderpoly = static_cast<KX_PolygonMaterial*>(polymat);
	struct TFace* tface = blenderpoly->GetTFace();
	
	BL_RenderText( mode,mytext,mytext.Length(),tface,v1,v2,v3,v4);
	
}



KX_BlenderRenderTools::~KX_BlenderRenderTools()
{
};
	
	
void	KX_BlenderRenderTools::EndFrame(RAS_IRasterizer*)
{
}
	


/**
 * Rendering text using 2D bitmap functionality.  
 */
void KX_BlenderRenderTools::RenderText2D(RAS_TEXT_RENDER_MODE mode,
										 const char* text,
										 int xco,
										 int yco,									 
										 int width,
										 int height)
{
	switch (mode) {
	case RAS_IRenderTools::RAS_TEXT_PADDED: {
		STR_String tmpstr(text);
		BL_print_gamedebug_line_padded(tmpstr.Ptr(),xco,yco,width,height);
		break;
	}
	default: {
		STR_String tmpstr(text);
		BL_print_gamedebug_line(tmpstr.Ptr(),xco,yco,width,height);
	}
	}
}

	

void KX_BlenderRenderTools::PushMatrix()
{
	glPushMatrix();
}

void KX_BlenderRenderTools::PopMatrix()
{
	glPopMatrix();
}

#if 0
int KX_BlenderRenderTools::applyLights(RAS_IRasterizer *rasty, int objectlayer, const KX_MeshSlot &ms)
{
	/* Rank by:
	 1.	Light layer
	 2.	Light frustum in camera
	 3.	Light type SPOT > SUN > NORMAL
	 4.	attenuated(brightness, Distance(light, object))
	*/
}
#endif

int	KX_BlenderRenderTools::applyLights(RAS_IRasterizer *rasty, int objectlayer)
{
// taken from blender source, incompatibility between Blender Object / GameObject	
	
	RAS_IRasterizer::ILighting *lighting = rasty->Lighting();
	int count;
	int numLights = lighting->GetNumLights();
	
	std::vector<struct	RAS_LightObject*>::iterator lit = m_lights.begin();
	
	lighting->BeginLighting();
	for (lit = m_lights.begin(), count = 0; (lit != m_lights.end()) && (count < numLights); ++lit)
	{
		RAS_LightObject* lightdata = (*lit);
		if (lightdata->m_layer & objectlayer)
		{
			lighting->Enable(count, lightdata);
			count++;
		}
	}
	for (int i = count; i < numLights; i++)
		lighting->Disable(i);
	lighting->EndLighting();

	return count;
}

int KX_BlenderRenderTools::SetTPage(struct TFace* tface) const
{
	return set_tpage(tface);
}

/* (n&(n-1)) zeros the least significant bit of n */
static int is_pow2(int num) {
	return ((num)&(num-1))==0;
}
static int smaller_pow2(int num) {
	while (!is_pow2(num))
		num= num&(num-1);
	return num;	
}

RAS_IRasterizer::ITexture* KX_BlenderRenderTools::CreateTexture(RAS_IRasterizer* rasty, MTex* mtex) const
{
	/* disable */
	if(mtex==0 || mtex->tex == 0 || mtex->tex->type != TEX_IMAGE) 
		return NULL;
		
	Tex *tex = mtex->tex;
	Image *ima= tex->ima;
	
	if(ima==0 || ima->ok==0) 
	{
		return NULL;
	}
	
	if(ima->ibuf==0) 
	{
		load_image(ima, IB_rect, G.sce, /*framenum*/ 0);
			
		if(ima->ibuf==0) {
			ima->ok= 0;
				
			return NULL;
		}
	
	}
	
	unsigned int *bind= &ima->bindcode;
	
	if(*bind==0) {
		int tpx= ima->ibuf->x;
		int tpy= ima->ibuf->y;
		unsigned int *rect= ima->ibuf->rect;
			
		int rectw = tpx, recth = tpy;
		unsigned int *tilerect= NULL, *scalerect= NULL;
	
		if (!is_pow2(rectw) || !is_pow2(recth)) {
			rectw= smaller_pow2(rectw);
			recth= smaller_pow2(recth);
			
			scalerect= (unsigned int *) MEM_mallocN(rectw*recth*sizeof(*scalerect), "scalerect");
			gluScaleImage(GL_RGBA, tpx, tpy, GL_UNSIGNED_BYTE, rect, rectw, recth, GL_UNSIGNED_BYTE, scalerect);
			rect= scalerect;
		}
		
		RAS_IRasterizer::ITexture *texture = rasty->NewTexture();
		//m_normap->SetMipMaps(tex->imaflag & TEX_MIPMAP);
		texture->SetMipMap(RAS_IRasterizer::ITexture::LINEAR, RAS_IRasterizer::ITexture::LINEAR);
		texture->SetCompression(false);
		texture->SetTextureFormat(mtex->texflag & MTEX_RGBTOINT, mtex->tex->imaflag & TEX_USEALPHA);
		
		texture->BindImage(rect, rectw, recth);
		
		if (tilerect)
			MEM_freeN(tilerect);
		if (scalerect)
			MEM_freeN(scalerect);
		return texture;
	} 
	else 
	{
		return rasty->NewTexture(*bind);
	}
	
	return NULL;
}


RAS_IPolyMaterial* KX_BlenderRenderTools::CreateBlenderPolyMaterial(
		const STR_String &texname,
		bool ba,
		Material *mat,
		int tile,int tilexrep,int tileyrep,
		int mode,
		bool smooth,
		bool transparant, bool zsort, 
		int subsurf, bool linear_subsurf,
		int lightlayer, 
		bool bIsTriangle, 
		void* clientobject,
		void* tface)
{
	return new KX_PolygonMaterial(
		texname,
		ba,
		mat,
		tile, tilexrep, tileyrep,
		mode,
		smooth,
		transparant, zsort,
		subsurf, linear_subsurf, 
		lightlayer,
		bIsTriangle,
		clientobject,
		(struct TFace*)tface,
		this);
/*	return new KX_BlenderPolyMaterial(
		texname,
		ba,
		mat,
		tile, tilexrep, tileyrep,
		mode,
		smooth,
		transparant, zsort,
		subsurf, linear_subsurf, 
		lightlayer,
		bIsTriangle,
		clientobject,
		(struct TFace*)tface);*/
}

STR_String KX_BlenderRenderTools::GetText(const STR_String &name)
{
	Text *text = (Text*)G.main->text.first;
	
	while (text)
	{
		if (name == STR_String(text->id.name+2))
		{
			STR_String data;
			TextLine *line = (TextLine*) text->lines.first;
			while (line)
			{
				data += line->line + STR_String('\n');
				line = line->next;
			}
			return data;
		}
		text = (Text*)text->id.next;
	}
	
	return "";
}
