/**
 * $Id: GPC_PolygonMaterial.cpp 677 2004-07-04 13:22:53Z kester $
 *
 * ***** BEGIN GPL/BL DUAL LICENSE BLOCK *****
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version. The Blender
 * Foundation also sells licenses for use in proprietary software under
 * the Blender License.  See http://www.blender.org/BL/ for information
 * about this.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software Foundation,
 * Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 *
 * The Original Code is Copyright (C) 2001-2002 by NaN Holding BV.
 * All rights reserved.
 *
 * The Original Code is: all of this file.
 *
 * Contributor(s): none yet.
 *
 * ***** END GPL/BL DUAL LICENSE BLOCK *****
 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include "GL/glew.h"

#include "GPC_PolygonMaterial.h"
#include "MT_Vector3.h"
#include "RAS_IRasterizer.h"
#include "RAS_IRenderTools.h"

/* This list includes only data type definitions */
#include "DNA_object_types.h"
#include "DNA_material_types.h"
#include "DNA_image_types.h"
#include "DNA_lamp_types.h"
#include "DNA_group_types.h"
#include "DNA_scene_types.h"
#include "DNA_camera_types.h"
#include "DNA_property_types.h"
#include "DNA_text_types.h"
#include "DNA_sensor_types.h"
#include "DNA_controller_types.h"
#include "DNA_actuator_types.h"
#include "DNA_mesh_types.h"
#include "DNA_meshdata_types.h"
#include "DNA_view3d_types.h"
#include "DNA_world_types.h"

#include "BKE_global.h"
#include "BKE_image.h"
#include "BKE_mesh.h"

#include "MEM_guardedalloc.h"

#include "IMB_imbuf_types.h"
/* end of blender include block */


GPC_PolygonMaterial::GPC_PolygonMaterial(const STR_String& texname, 
			bool ba, 
			const STR_String& matname,
			int tile, int tileXrep, int tileYrep, 
			int mode, 
			bool smooth,
			bool transparant, bool zsort,
			int subsurf, bool linearsubsurf,
			int lightlayer, 
			bool bIsTriangle, 
			void* clientobject, 
			void* tpage,
			RAS_IRenderTools* rendertools) :
			RAS_IPolyMaterial(texname, 
				ba, 
				matname, 
				tile, tileXrep, tileYrep, 
				mode,
				smooth,
				transparant, zsort, 
				subsurf, linearsubsurf, 
				lightlayer, 
				bIsTriangle, 
				clientobject), 
			m_tface((struct TFace*)tpage),
			m_rendertools(rendertools)
{
	// clear local caching info
	rendertools->SetTPage(0);
}


GPC_PolygonMaterial::~GPC_PolygonMaterial(void)
{
}


bool GPC_PolygonMaterial::Activate(RAS_IRasterizer* rasty, TCachingInfo& cachingInfo) const
{
	if (GetCachingInfo() != cachingInfo)
	{
		if (!cachingInfo)
		{
			m_rendertools->SetTPage(0);
		}
		cachingInfo = GetCachingInfo();

		if ((m_drawingmode & 4)&& (rasty->GetDrawingMode() == RAS_IRasterizer::KX_TEXTURED) )
		{
			update_realtime_texture((struct TFace*) m_tface, rasty->GetTime());
			rasty->SetTexUnit(0);
			m_rendertools->SetTPage(m_tface);
			rasty->EnableTextures(true);
			rasty->SetTexCoords(RAS_IRasterizer::RAS_TEXCO_UV1);
		} else
		{
			m_rendertools->SetTPage(0);
			rasty->EnableTextures(false);
		}
	
		//TF_TWOSIDE == 512, todo, make this a ketsji enum
		if(m_drawingmode & 512) {
			rasty->SetCullFace(false);
		}
			
		else
		{
			rasty->SetCullFace(true);//glEnable(GL_CULL_FACE);
			//else glDisable(GL_CULL_FACE);
		}
	}
	rasty->SetSpecularity(m_specular[0],m_specular[1],m_specular[2],m_specularity);
	rasty->SetShinyness(m_shininess);
	rasty->SetDiffuse(m_diffuse[0], m_diffuse[1],m_diffuse[2], 1.0);
	
	return false;
}

bool GPC_PolygonMaterial::ActivateNextPass(RAS_IRasterizer *rasty, unsigned int pass) const
{
	return !pass;
}

