/*
 * $Id: KX_Frustum.h 721 2004-07-10 11:54:00Z kester $
 *
 * ***** BEGIN GPL/BL DUAL LICENSE BLOCK *****
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version. The Blender
 * Foundation also sells licenses for use in proprietary software under
 * the Blender License.  See http://www.blender.org/BL/ for information
 * about this.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software Foundation,
 * Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 *
 * The Original Code is Copyright (C) 2001-2002 by NaN Holding BV.
 * All rights reserved.
 *
 * The Original Code is: all of this file.
 *
 * Contributor(s): none yet.
 *
 * ***** END GPL/BL DUAL LICENSE BLOCK *****
 * Camera in the gameengine. Cameras are also used for views.
 */

#ifndef __KX_FRUSTUM_H__
#define __KX_FRUSTUM_H__


#include "MT_Transform.h"
#include "MT_Matrix3x3.h"
#include "MT_Matrix4x4.h"
#include "MT_Vector3.h"
#include "MT_Point3.h"

class KX_IFrustum
{
	/**
	 *  True if the parent has changed.
	 */
	bool         m_dirty;
	/**
	 * true if the frustum planes have been normalised.
	 */
	bool         m_normalised;
	
	/**
	 * true if frustum culling
	 */
	bool         m_frustum_culling;
	/**
	 * View Frustum clip planes.
	 */
	MT_Vector4   m_planes[6];
	
	/**
	 * The centre point of the frustum.
	 */
	MT_Point3    m_frustum_centre;
	MT_Scalar    m_frustum_radius;
	bool         m_set_frustum_centre;
protected:
	virtual const MT_Matrix4x4& GetProjectionMatrix() const = 0;
	virtual float GetCameraNear() const = 0;
	virtual float GetCameraFar() const = 0;
	/**
	 * Extracts the camera clip frames from the projection and world-to-camera matrices.
	 */
	void ExtractClipPlanes();
	
	/**
	 * Normalise the camera clip frames.
	 */
	void NormaliseClipPlanes();
	/**
	 * Extracts the bound sphere of the view frustum.
	 */
	void ExtractFrustumSphere();

public:
	typedef enum { INSIDE, INTERSECT, OUTSIDE } ;

	KX_IFrustum()
		: m_dirty(true),
		  m_normalised(false),
		  m_frustum_culling(true),
		  m_set_frustum_centre(false)

	{
	}
	
	virtual ~KX_IFrustum()
	{
	}
	
	virtual MT_Transform GetWorldToCamera() const = 0;
	virtual MT_Transform GetCameraToWorld() const = 0;
	virtual const MT_Vector3 GetWorldScaling() const = 0;
	virtual const MT_Point3 GetWorldPosition() const = 0;
	
	/**
	 * Tests if the given sphere is inside this camera's view frustum.
	 *
	 * @param centre The centre of the sphere, in world coordinates.
	 * @param radius The radius of the sphere.
	 * @return INSIDE, INTERSECT, or OUTSIDE depending on the sphere's relation to the frustum.
	 */
	int SphereInsideFrustum(const MT_Point3& centre, const MT_Scalar &radius);
	/**
	 * Tests the given eight corners of a box with the view frustum.
	 *
	 * @param box a pointer to eight MT_Point3 representing the world coordinates of the corners of the box.
	 * @return INSIDE, INTERSECT, or OUTSIDE depending on the box's relation to the frustum.
	 */
	int BoxInsideFrustum(const MT_Point3 *box);
	/**
	 * Tests the given point against the view frustum.
	 * @return true if the given point is inside or on the view frustum; false if it is outside.
	 */
	bool PointInsideFrustum(const MT_Point3& x);
	/**
	 * Sets the frustum has changed.
	 */
	void SetDirty();
	/**
	 * True if the camera is frustum culling.
	 */
	bool GetFrustumCulling() const;
	void SetFrustumCulling(bool cull);
};

#endif //__KX_FRUSTUM_H__

