/**
 * $Id: KX_Light.h 753 2004-07-22 01:48:28Z kester $
 *
 * ***** BEGIN GPL/BL DUAL LICENSE BLOCK *****
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version. The Blender
 * Foundation also sells licenses for use in proprietary software under
 * the Blender License.  See http://www.blender.org/BL/ for information
 * about this.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software Foundation,
 * Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 *
 * The Original Code is Copyright (C) 2001-2002 by NaN Holding BV.
 * All rights reserved.
 *
 * The Original Code is: all of this file.
 *
 * Contributor(s): none yet.
 *
 * ***** END GPL/BL DUAL LICENSE BLOCK *****
 */
#ifndef __KX_LIGHT
#define __KX_LIGHT

#include "RAS_IRasterizer.h"
#include "RAS_LightObject.h"
#include "RAS_Rect.h"
#include "RAS_FramingManager.h"

#include "KX_GameObject.h"
#include "KX_Frustum.h"

class KX_LightObject : public KX_GameObject, public KX_IFrustum
{
	Py_Header;
protected:
	RAS_LightObject		m_lightobj;
	class RAS_IRenderTools*	m_rendertools;	//needed for registering and replication of lightobj
	bool			m_set_projection;
	
	static char		doc[];

public:
	KX_LightObject(void* sgReplicationInfo,SG_Callbacks callbacks,class RAS_IRenderTools* rendertools,const struct RAS_LightObject&	lightobj, PyTypeObject *T = &Type);
	virtual ~KX_LightObject();
	virtual CValue*		GetReplica();
	RAS_LightObject*	GetLightData() { return &m_lightobj;}
	
	bool		hasValidProjectionMatrix() const;
	void		SetProjectionMatrix(const MT_Matrix4x4& proj);
	const MT_Matrix4x4& GetProjectionMatrix() const;
	void		SetModelviewMatrix(const MT_Matrix4x4& mvmat);
	
	MT_Transform	GetWorldToCamera() const;
	MT_Transform	GetCameraToWorld() const;

	RAS_IRasterizer::ITexture*	GetShadowMap(unsigned int index = 0);
	void				SetShadowMap(RAS_IRasterizer::ITexture* shadow, unsigned int index = 0);
	
	float	GetLens() const { return m_lightobj.m_lens; }
	float	GetCameraNear() const { return m_lightobj.m_clipsta; }
	float	GetCameraFar() const { return m_lightobj.m_clipend; }
	
	const MT_Vector3	GetWorldScaling() const;
	const MT_Point3		GetWorldPosition() const;
	
	virtual PyObject* _getattr(const STR_String& attr); /* lens, near, far, projection_matrix */
	virtual int       _setattr(const STR_String& attr, PyObject *pyvalue);
};

#endif //__KX_LIGHT

