/**
 * $Id: KX_PolygonMaterial.h 797 2004-08-15 00:21:57Z kester $
 *
 * ***** BEGIN GPL/BL DUAL LICENSE BLOCK *****
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version. The Blender
 * Foundation also sells licenses for use in proprietary software under
 * the Blender License.  See http://www.blender.org/BL/ for information
 * about this.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software Foundation,
 * Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 *
 * The Original Code is Copyright (C) 2001-2002 by NaN Holding BV.
 * All rights reserved.
 *
 * The Original Code is: all of this file.
 *
 * Contributor(s): none yet.
 *
 * ***** END GPL/BL DUAL LICENSE BLOCK *****
 */
#ifndef __KX_POLYGONMATERIAL_H__
#define __KX_POLYGONMATERIAL_H__

#include "PyObjectPlus.h"

#include "RAS_MaterialBucket.h"
#include "RAS_IRasterizer.h"

struct TFace;
struct Material;
struct MTex;

/**
 *  Material class.
 *
 *  This holds the shader, textures and python methods for setting the render state before
 *  rendering.
 */
class KX_PolygonMaterial : public PyObjectPlus, public RAS_IPolyMaterial
{
	Py_Header;
private:
	//GLSL Support
	struct Shader;
	class Attribute;
	class Uniform;
	template<class T> class VecUniform;
	template<class T> class MatUniform;
	class GenUniform;
	
	// ARB_vertex_program/ARB_fragment_program support
	struct Program;
	
	class Private;
	
	mutable class Private*	m;
	
	/** Blender texture face structure. */
	TFace*			m_tface;
	Material*		m_material;
	RAS_IRenderTools*	m_rendertools;
	
	PyObject*		m_pymaterial;

	//void ActivateVertexShader(RAS_IRasterizer *rasty) const;
	//void ActivateFragmentShader(RAS_IRasterizer *rasty) const;
	bool ActivateShader(RAS_IRasterizer *rasty) const;
	bool ActivatePrograms(RAS_IRasterizer *rasty) const;
	
	void ConvertMaterial(RAS_IRasterizer* rasty) const;
	
	void DisableTextures(RAS_IRasterizer *rasty, unsigned int unit) const;
	void ActivateMaterialProperties(RAS_IRasterizer *rasty) const;
	void ActivateUVTexture(RAS_IRasterizer *rasty, unsigned int &unit) const;
	bool ActivateShadowMap(RAS_IRasterizer *rasty, unsigned int li, unsigned int unit) const;
	void ActivateShaderTexture(RAS_IRasterizer *rasty, unsigned int &unit) const;
	bool ActivateLightAmbientPass(RAS_IRasterizer *rasty) const;
	bool ActivateLightDiffusePass(RAS_IRasterizer *rasty) const;
	bool ActivateLightSpecularPass(RAS_IRasterizer *rasty) const;
	bool ActivateLightTexturePass(RAS_IRasterizer *rasty) const;
	bool ActivateShaderPass(RAS_IRasterizer *rasty, unsigned int pass) const;
	bool ActivateProgramPass(RAS_IRasterizer *rasty, unsigned int pass) const;
	bool ActivateShadowPass(RAS_IRasterizer *rasty, unsigned int pass) const;

public:
	
	KX_PolygonMaterial(const STR_String &texname,
		bool ba,
		Material* matname,
		int tile,
		int tilexrep,
		int tileyrep,
		int mode,
		bool smooth,
		bool transparant,
		bool zsort,
		int subsurf,
		bool linearsubsurf,
		int lightlayer,
		bool bIsTriangle,
		void* clientobject,
		struct TFace* tface,
		class RAS_IRenderTools* rendertools,
		PyTypeObject *T = &Type);
	virtual ~KX_PolygonMaterial();
	
	/**
	 * Returns the caching information for this material,
	 * This can be used to speed up the rasterizing process.
	 * @return The caching information.
	 */
	virtual TCachingInfo GetCachingInfo(void) const;

	/**
	 * Activates the material in the (OpenGL) rasterizer.
	 * On entry, the cachingInfo contains info about the last activated material.
	 * On exit, the cachingInfo should contain updated info about this material.
	 * @param rasty			The rasterizer in which the material should be active.
	 * @param cachingInfo	The information about the material used to speed up rasterizing.
	 */
	virtual bool Activate(RAS_IRasterizer* rasty, TCachingInfo& cachingInfo) const;
	virtual bool ActivateNextPass(RAS_IRasterizer *rasty, unsigned int pass) const;
	virtual void ActivateMeshSlot(RAS_IRasterizer* rasty, const KX_MeshSlot &ms) const;

	/**
	 * Returns the Blender texture face structure that is used for this material.
	 * @return The material's texture face.
	 */
	TFace* GetTFace(void) const;
	
	
	KX_PYMETHOD_DOC(KX_PolygonMaterial, loadShader);
	KX_PYMETHOD_DOC(KX_PolygonMaterial, setUniform);
	KX_PYMETHOD_DOC(KX_PolygonMaterial, genUniform);
	KX_PYMETHOD_DOC(KX_PolygonMaterial, bindAttribute);
	KX_PYMETHOD_DOC(KX_PolygonMaterial, setCustomMaterial);
	KX_PYMETHOD_DOC(KX_PolygonMaterial, loadProgram);
	
	virtual PyObject* _getattr(const STR_String& attr);
	virtual int       _setattr(const STR_String& attr, PyObject *pyvalue);
};


inline TFace* KX_PolygonMaterial::GetTFace(void) const
{
	return m_tface;
}

inline RAS_IPolyMaterial::TCachingInfo KX_PolygonMaterial::GetCachingInfo(void) const
{
	return (void*) this;
}

#endif // __KX_POLYGONMATERIAL_H__

