/**
 * $Id: BL_Quaternion.vs 839 2004-09-11 23:57:19Z kester $
 * ***** BEGIN GPL/BL DUAL LICENSE BLOCK *****
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version. The Blender
 * Foundation also sells licenses for use in proprietary software under
 * the Blender License.  See http://www.blender.org/BL/ for information
 * about this.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software Foundation,
 * Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 *
 * The Original Code is Copyright (C) 2001-2002 by NaN Holding BV.
 * All rights reserved.
 *
 * The Original Code is: all of this file.
 *
 * Contributor(s): none yet.
 *
 * ***** END GPL/BL DUAL LICENSE BLOCK *****
 *
 * Shader Quaternion Routines
 *
 * Real part is q.w
 *
 * TARGET: Any processor
 */

 
vec4 mult(const vec4 q, const vec3 w)
{
	// return vec4(q.w*w.xyz + w.w*q.xyz + cross(q.xyz, w.xyz), q.w*w.w - dot(q.xyz, w.xyz)); with w.w = 0
	return vec4(q.w*w.xyz + cross(q.xyz, w.xyz), -dot(q.xyz, w.xyz));
}

vec4 mult(const vec3 w, const vec4 q)
{
	// return vec4(q.w*w.xyz + w.w*q.xyz + cross(w.xyz, q.xyz), q.w*w.w - dot(w.xyz, q.xyz)); with w.w = 0
	return vec4(q.w*w.xyz + cross(q.xyz, w.xyz), -dot(w.xyz, q.xyz)); 
}

vec4 mult(const vec4 q1, const vec4 q2)
{
	return vec4(q1.w*q2.xyz + q2.w*q1.xyz + cross(q1.xyz, q2.xyz), q1.w*q2.w - dot(q1.xyz, q2.xyz));
}

vec4 conjugate(const vec4 q) 
{
    return vec4(-q.x, -q.y, -q.z, q.w);
}

vec3 transform(const vec4 q, const vec3 w)
{
	vec4 t = mult(q, w);
	// mul(t, conjugate(q)).xyz
	return -t.w*q.xyz + q.w*w.xyz + cross(w.xyz, -q.xyz);
}
