/**
 * $Id: Blender.vs 780 2004-08-05 01:58:38Z kester $
 * ***** BEGIN GPL/BL DUAL LICENSE BLOCK *****
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version. The Blender
 * Foundation also sells licenses for use in proprietary software under
 * the Blender License.  See http://www.blender.org/BL/ for information
 * about this.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software Foundation,
 * Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 *
 * The Original Code is Copyright (C) 2001-2002 by NaN Holding BV.
 * All rights reserved.
 *
 * The Original Code is: all of this file.
 *
 * Contributor(s): none yet.
 *
 * ***** END GPL/BL DUAL LICENSE BLOCK *****
 *
 *  Useful Shading utilites.
 *
 *  TARGET: Any Processor.
 */

/*
 * Compute the length^2, 1/length & normalised vector.
 *
 * This is identical to the normalize standard function, but saves the
 * vector length byproducts.
 */
vec3 BL_Normalize(const vec3 vector, out float len2, out float invlen)
{
	len2 = dot(vector, vector);
	invlen = inversesqrt(len2);
	return invlen*vector;
}

/*
 * Computes a reflection vector, assuming ndoti has already been calculated.
 */
vec3 BL_Reflect(const vec3 normal, const vec3 incident, const float ndoti)
{
	return 2.0*ndoti*normal - incident;
}

/*
 * Returns vec2(diffuse, spec)
 */
vec2 BL_GLLighting(const vec3 normal, const vec3 lightvec, const vec3 viewvec)
{
	float diffuse = max(dot(normal, lightvec), 0.0);
	vec3 refl = BL_Reflect(normal, lightvec, diffuse);
	return vec2(diffuse, pow(max(dot(refl, viewvec), 0.0), gl_FrontMaterial.shininess));
}


/**
 *  Generate texture coordinates from eye linear texgen.
 *
 *  See glTexGen GL_EYE_LINEAR
 */
vec4 BL_EyeLinear(const int texunit, const vec4 point)
{
	return vec4(dot(gl_EyePlaneS[texunit], dot(gl_ModelViewMatrix, point)),
		dot(gl_EyePlaneT[texunit], dot(gl_ModelViewMatrix, point)),
		dot(gl_EyePlaneR[texunit], dot(gl_ModelViewMatrix, point)),
		dot(gl_EyePlaneQ[texunit], dot(gl_ModelViewMatrix, point)));
}

/**
 *  Generate texture coordinates from object linear texgen.
 *
 *  See glTexGen GL_OBJECT_LINEAR
 */
vec4 BL_ObjectLinear(const int texunit, const vec4 point)
{
	return vec4(dot(gl_ObjectPlaneS[texunit], point),
		dot(gl_ObjectPlaneT[texunit], point),
		dot(gl_ObjectPlaneR[texunit], point),
		dot(gl_ObjectPlaneQ[texunit], point));
}

