/**
 * $Id: Lighting.fs 839 2004-09-11 23:57:19Z kester $
 * ***** BEGIN GPL/BL DUAL LICENSE BLOCK *****
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version. The Blender
 * Foundation also sells licenses for use in proprietary software under
 * the Blender License.  See http://www.blender.org/BL/ for information
 * about this.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software Foundation,
 * Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 *
 * The Original Code is Copyright (C) 2001-2002 by NaN Holding BV.
 * All rights reserved.
 *
 * The Original Code is: all of this file.
 *
 * Contributor(s): none yet.
 *
 * ***** END GPL/BL DUAL LICENSE BLOCK *****
 *
 * Lighting Functions.
 *
 * TARGET: Fragment Processor (Any)
 */

#define SPOTSMOOTH 0.99


vec3 BL_LightAttenuationConstants(const int light)
{
	return vec3(gl_LightSource[light].constantAttenuation,
				gl_LightSource[light].linearAttenuation,
				gl_LightSource[light].quadraticAttenuation);
}

float BL_Attenuation(const int light, const float dist2, const float invdist)
{
	return 1.0/dot(BL_LightAttenuationConstants(light), vec3(1.0, dist*invdist, dist2));
}

/**
 *  Generates a spotlight circle
 *
 *  Light = GL light to use.
 *  lightvec = lightpos - vertex in eye coordinates.
 *
 */
float BL_SpotAttenuation(const int light, const vec3 lightvec)
{
	float x = dot(gl_LightSource[light].spotDirection, -lightvec);
	return step(gl_LightSource[light].spotCosCutoff, x)
		* pow(x, gl_LightSource[light].spotExponent);
}

/**
 *  Generates a spotlight circle, with a smooth transition.
 *
 *  Light = GL light to use.
 *  lightvec = lightpos -> vertex in eye coordinates.
 *
 */
float BL_SmoothSpotAttenuation(const int light, const vec3 lightvec)
{
	float x = dot(gl_LightSource[light].spotDirection, -lightvec);
	return smoothstep(SPOTSMOOTH*gl_LightSource[light].spotCosCutoff, gl_LightSource[light].spotCosCutoff, x)
		* pow(x, gl_LightSource[light].spotExponent);
}

