# $Id: KX_Light.py 732 2004-07-11 05:31:22Z kester $
# Documentation for Light game objects.
from KX_GameObject import *

class KX_Light(KX_GameObject):
	"""
	A Light object.

	Example:
	
	# Turn on a red alert light.
	import GameLogic
	
	co = GameLogic.getCurrentController()
	light = co.getOwner()
	
	light.energy = 1.0
	light.colour = [1.0, 0.0, 0.0]
	
	@group Constants: NORMAL, SPOT, SUN
	@ivar SPOT:   A spot light source. See attribute 'type'
	@ivar SUN:    A point light source with no attenuation. See attribute 'type'
	@ivar NORMAL: A point light source. See attribute 'type'
	
	@ivar type:            The type of light - must be SPOT, SUN or NORMAL
	@ivar layer:           The layer mask that this light affects object on.
	@type layer:           bitfield
	@ivar energy:          The brightness of this light. 
	@type energy:          float
	@ivar distance:        The maximum distance this light can illuminate. (SPOT and NORMAL lights only)
	@type distance:        float
	@ivar colour:          The colour of this light. Black = [0.0, 0.0, 0.0], White = [1.0, 1.0, 1.0]
	@type colour:          list [r, g, b]
	@ivar color:           Synonym for colour.
	@ivar lin_attenuation: The linear component of this light's attenuation. (SPOT and NORMAL lights only)
	@type lin_attenuation: float
	@ivar quad_attenuation: The quadratic component of this light's attenuation (SPOT and NORMAL lights only)
	@type quad_attenuation: float
	@ivar spotsize:        The cone angle of the spot light, in degrees. (float) (SPOT lights only)
	                       0.0 <= spotsize <= 180.0. Spotsize = 360.0 is also accepted. 
	@ivar spotblend:       Specifies the intensity distribution of the spot light. (float) (SPOT lights only)
	                       Higher values result in a more focused light source.
	                       0.0 <= spotblend <= 1.0.
	@ivar shadow:          Enables shadow mapping for this light.  Shadow mapping is only available on spot lights.
	@type shadow:          boolean
	@ivar resolution:      Shadow map resolution.  Must by a power of two.
	@type resolution:      integer
	@ivar lens:            The lights lens value for rendering the shadow map.
	@type lens:            float
	@ivar near:            The near clip distance for shadow mapping.
	@type near:            float
	@ivar far:             The far clip distance for shadow mapping.
	@type far:             float
	@ivar projection_matrix: The projection matrix for shadow mapping.
	@type projection_matrix: Matrix 4x4
	@ivar modelview_matrix: The model-view matrix for shadow mapping. (Read only)
	@type modelview_matrix: Matrix 4x4
	
	
	"""
