/**
 * $Id: RAS_VAOpenGLRasterizer.cpp 820 2004-08-26 06:00:21Z kester $
 * ***** BEGIN GPL/BL DUAL LICENSE BLOCK *****
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version. The Blender
 * Foundation also sells licenses for use in proprietary software under
 * the Blender License.  See http://www.blender.org/BL/ for information
 * about this.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software Foundation,
 * Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 *
 * The Original Code is Copyright (C) 2001-2002 by NaN Holding BV.
 * All rights reserved.
 *
 * The Original Code is: all of this file.
 *
 * Contributor(s): none yet.
 *
 * ***** END GPL/BL DUAL LICENSE BLOCK *****
 */
#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include "GL/glew.h"

#include "RAS_VAOpenGLRasterizer.h"

#include "STR_String.h"
#include "RAS_TexVert.h"
#include "MT_CmMatrix4x4.h"
#include "RAS_IRenderTools.h" // rendering text

RAS_VAOpenGLRasterizer::RAS_VAOpenGLRasterizer(RAS_ICanvas* canvas)
:RAS_OpenGLRasterizer(canvas)
{
	PrintGLInfo();
}



RAS_VAOpenGLRasterizer::~RAS_VAOpenGLRasterizer()
{
}

bool RAS_VAOpenGLRasterizer::PrintGLInfo()
{
	static bool debug = true;
	if (!debug)
		return false;
		
	debug = false;
	std::cout << "   Vertex Arrays Enabled." << std::endl;
	if (GLEW_ARB_vertex_buffer_object)
		std::cout << "   ARB_vertex_buffer_objects enabled." << std::endl;
	else
		std::cout << "   ARB_vertex_buffer_objects not supported." << std::endl; 
}

bool RAS_VAOpenGLRasterizer::Init(void)
{
	
	bool result = RAS_OpenGLRasterizer::Init();
	
	if (result)
	{
		glEnableClientState(GL_VERTEX_ARRAY);
		glDisableClientState(GL_TEXTURE_COORD_ARRAY);
		glEnableClientState(GL_NORMAL_ARRAY);
		glDisableClientState(GL_COLOR_ARRAY);
		if (GLEW_ARB_vertex_program)
			glEnableVertexAttribArrayARB(1);

		glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
	}

	return result;
}



void RAS_VAOpenGLRasterizer::SetDrawingMode(int drawingmode)
{
	m_drawingmode = drawingmode;

		switch (m_drawingmode)
	{
	case KX_BOUNDINGBOX:
		{
		}
	case KX_WIREFRAME:
		{
			glDisable (GL_CULL_FACE);
			break;
		}
	case KX_TEXTURED:
		{
		}
	case KX_SHADED:
		{
			glEnableClientState(GL_COLOR_ARRAY);
		}
	case KX_SOLID:
		{
			break;
		}
	default:
		{
		}
	}
}



void RAS_VAOpenGLRasterizer::Exit()
{
	glDisableClientState(GL_VERTEX_ARRAY);
	glDisableClientState(GL_TEXTURE_COORD_ARRAY);
	glDisableClientState(GL_COLOR_ARRAY);
	glDisableClientState(GL_NORMAL_ARRAY);
	if (GLEW_ARB_vertex_program)
		glDisableVertexAttribArrayARB(1);

	RAS_OpenGLRasterizer::Exit();
}

class RAS_VAOpenGLRasterizer::VAGLDrawHandle : public RAS_IDrawHandle
{
	bool		m_modified;
	bool		m_vertexmodified;
	bool		m_indexmodified;

public:
	VAGLDrawHandle()
		: RAS_IDrawHandle(),
		m_modified(true)
	{
	}
	
	virtual ~VAGLDrawHandle()
	{
	}
	
	virtual const GLvoid* BufferOffset(const void *offset, const void *) const
	{
		return offset;
	}
	
	virtual void BindVertexArray(unsigned int)
	{
		if (GLEW_ARB_vertex_buffer_object)
			glBindBufferARB(GL_ARRAY_BUFFER_ARB, 0);
	}
	
	virtual void BindIndexArray(unsigned int)
	{
		if (GLEW_ARB_vertex_buffer_object)
			glBindBufferARB(GL_ELEMENT_ARRAY_BUFFER_ARB, 0);
	}
	
	bool IsModified()
	{
		return m_modified;
	}
	
	void SetModified(bool modified)
	{
		m_modified = modified;
	}
	
	virtual bool IsVertexModified()
	{
		return m_vertexmodified;
	}
	
	virtual void SetVertexModified(bool modified)
	{
		m_vertexmodified = modified;
		m_modified = modified;
	}
	
	virtual bool IsIndexModified()
	{
		return m_indexmodified;
	}
	
	virtual void SetIndexModified(bool modified)
	{
		m_indexmodified = modified;
		m_modified = modified;
	}
};

class RAS_VAOpenGLRasterizer::VAGLBufferHandle : public RAS_VAOpenGLRasterizer::VAGLDrawHandle
{
	GLuint		*m_varrays;
	GLuint		*m_iarrays;
	unsigned int	m_size; // # of arrays.
	bool		m_error;
	GLenum		m_vertexdrawmode; // GL_STATIC_DRAW_ARB, GL_STREAM_DRAW_ARB
	GLenum		m_indexdrawmode; // GL_STATIC_DRAW_ARB, GL_STREAM_DRAW_ARB
public:	
	VAGLBufferHandle(unsigned int size)
		: VAGLDrawHandle(),
		  m_varrays(new GLuint[size]),
		  m_iarrays(new GLuint[size]),
		  m_size(size),
		  m_error(false),
		  m_vertexdrawmode(GL_STATIC_DRAW_ARB),
		  m_indexdrawmode(GL_STATIC_DRAW_ARB)
	{
		glGenBuffersARB(size, m_varrays);
		glGenBuffersARB(size, m_iarrays);
	}
	
	virtual ~VAGLBufferHandle()
	{
		glDeleteBuffersARB(m_size, m_varrays);
		glDeleteBuffersARB(m_size, m_iarrays);
		delete [] m_varrays;
		delete [] m_iarrays;
	}
	
	bool IsValid()
	{
		return !m_error;
	}
	
	void BufferVertexData(unsigned int index, unsigned int size, const RAS_TexVert *data)
	{
		assert(data);
		assert(index < m_size);
		assert(m_varrays[index]);
		glBindBufferARB(GL_ARRAY_BUFFER_ARB, m_varrays[index]);
		glBufferDataARB(GL_ARRAY_BUFFER_ARB, size*sizeof(RAS_TexVert), data, m_vertexdrawmode);
		SetVertexModified(false);
	}
	
	void BufferIndexData(unsigned int index, unsigned int size, const unsigned short *data)
	{
		assert(data);
		assert(index < m_size);
		assert(m_iarrays[index]);
		glBindBufferARB(GL_ELEMENT_ARRAY_BUFFER_ARB, m_iarrays[index]);
		glBufferDataARB(GL_ELEMENT_ARRAY_BUFFER_ARB, size*sizeof(unsigned short), data, m_indexdrawmode);
		SetIndexModified(false);
	}
	
	virtual void BindVertexArray(unsigned int index)
	{
		assert(index < m_size);
		glBindBufferARB(GL_ARRAY_BUFFER_ARB, m_varrays[index]);
	}
	
	virtual void BindIndexArray(unsigned int index)
	{
		assert(index < m_size);
		glBindBufferARB(GL_ELEMENT_ARRAY_BUFFER_ARB, m_iarrays[index]);
	}
	
	GLuint GetVBuffer(unsigned int index)
	{
		return m_varrays[index];
	}
	
	GLuint GetIBuffer(unsigned int index)
	{
		return m_iarrays[index];
	}
	
	virtual const GLvoid* BufferOffset(const void *offset, const void *base) const
	{
		assert(sizeof(const void *) == sizeof(unsigned));
		assert(sizeof(GLvoid*) == sizeof(const void *));
		return (GLvoid*)(unsigned(offset) - unsigned(base));
	}
	
	virtual void SetIndexModified(bool modified)
	{
		// Someone modified the vertex data!
		// Go into streaming mode.
		if (modified)
		{
			m_indexdrawmode = GL_STREAM_DRAW_ARB;
		}
		
		VAGLDrawHandle::SetIndexModified(modified);
	}

	virtual void SetVertexModified(bool modified)
	{
		// Someone modified the vertex data!
		// Go into streaming mode.
		if (modified)
		{
			m_vertexdrawmode = GL_STREAM_DRAW_ARB;
		}
		
		VAGLDrawHandle::SetVertexModified(modified);
	}
};


RAS_IDrawHandle* RAS_VAOpenGLRasterizer::IndexPrimitives( RAS_IDrawHandle *renderhandle,
							const vecVertexArray& vertexarrays,
							const vecIndexArrays & indexarrays,
							int mode,
							class RAS_IPolyMaterial*,
							class RAS_IRenderTools*,
							bool useObjectColor,
							const MT_Vector4& rgbacolor)
{
	static const GLsizei vtxstride = sizeof(RAS_TexVert);
	GLenum drawmode;
	switch (mode)
	{
	case 0:
		{
		drawmode = GL_TRIANGLES;
		break;
		}
	case 2:
		{
		drawmode = GL_QUADS;
		break;
		}
	case 1:	//lines
		{
		}
	default:
		{
		drawmode = GL_LINES;
		break;
		}
	}
	
	if (drawmode != GL_LINES)
	{
		if (useObjectColor)
		{
			glDisableClientState(GL_COLOR_ARRAY);
			glColor4d(rgbacolor[0], rgbacolor[1], rgbacolor[2], rgbacolor[3]);
		} else
		{
			glColor4d(0,0,0,1.0);
			glEnableClientState(GL_COLOR_ARRAY);
		}
	}
	else
	{
		glColor3d(0,0,0);
	}
	
	const RAS_TexVert* vertexarray;
	unsigned int numindices, vt;
	VAGLDrawHandle *h = NULL;
	if (renderhandle)
		h = static_cast<VAGLDrawHandle*>(renderhandle);
	
	if (GLEW_ARB_vertex_buffer_object)
	{
		VAGLBufferHandle *buffer;
		if (!renderhandle)
		{
			glGetError();
			renderhandle = h = buffer = new VAGLBufferHandle(vertexarrays.size());
			for (vt = 0; vt < vertexarrays.size(); vt++)
			{
				const KX_VertexArray & vertarray = (*vertexarrays[vt]);
				buffer->BufferVertexData(vt, vertarray.size(), &(vertarray[0]));
				
				const KX_IndexArray & indexarray = (*indexarrays[vt]);
				if (!indexarray.size())
					continue;
				buffer->BufferIndexData(vt, indexarray.size(), &(indexarray[0]));
			}
			buffer->SetModified(false);
			GLenum err = glGetError();
			if (err)
			{
				std::cout << "Creating buffer object failed." << gluErrorString(err) << std::endl;
				delete buffer;
				buffer = NULL;
				h = NULL;
			}
		}
		else
		{
			buffer = dynamic_cast<VAGLBufferHandle*>(renderhandle);
		
			if (buffer && buffer->IsModified())
			{
				glGetError();
				for (vt = 0; vt < vertexarrays.size(); vt++)
				{
					if (buffer->IsVertexModified())
					{
						const KX_VertexArray & vertarray = (*vertexarrays[vt]);
						buffer->BufferVertexData(vt, vertarray.size(), &(vertarray[0]));
					}
					
					if (buffer->IsIndexModified())
					{
						const KX_IndexArray & indexarray = (*indexarrays[vt]);
						if (!indexarray.size())
							continue;
						buffer->BufferIndexData(vt, indexarray.size(), &(indexarray[0]));
					}
				}
				buffer->SetModified(false);
				GLenum err = glGetError();
				if (err)
				{
					std::cout << "Recreating buffer object failed. " << gluErrorString(err) << std::endl;
					delete buffer;
					buffer = NULL;
					h = NULL;
				}
			}
		}
		glBindBufferARB(GL_ARRAY_BUFFER_ARB, 0);
		glBindBufferARB(GL_ELEMENT_ARRAY_BUFFER_ARB, 0);
	}
	
	if (!h)
		h = new VAGLDrawHandle();
		
	// use glDrawElements to draw each vertexarray
	for (vt=0;vt<vertexarrays.size();vt++)
	{
		vertexarray = &((*vertexarrays[vt]) [0]);
		const KX_IndexArray & indexarray = (*indexarrays[vt]);
		numindices = indexarray.size();
		
		if (!numindices)
			continue;
		
		h->BindVertexArray(vt);
		
		TexCoordPointer(h, vertexarray);
		if (drawmode != GL_LINES && !useObjectColor)
			glColorPointer(4,  GL_UNSIGNED_BYTE, vtxstride, h->BufferOffset(vertexarray->getRGBA(), vertexarray));
		glNormalPointer(   GL_FLOAT,         vtxstride, h->BufferOffset(vertexarray->getNormal(), vertexarray));
		if (GLEW_ARB_vertex_program)
			glVertexAttribPointerARB(1, 4, GL_FLOAT, true, vtxstride, h->BufferOffset(vertexarray->getTangent(), vertexarray)); // vertex.weight
		glVertexPointer(3, GL_FLOAT,         vtxstride, h->BufferOffset(vertexarray->getLocalXYZ(), vertexarray));
		//glLockArraysEXT(0,numverts);
		// here the actual drawing takes places
		h->BindIndexArray(vt);
		glDrawElements(drawmode,numindices,GL_UNSIGNED_SHORT,h->BufferOffset(&(indexarray[0]),&(indexarray[0])));
		//glUnlockArraysEXT();
	}
	if (GLEW_ARB_vertex_buffer_object)
	{
		glBindBufferARB(GL_ARRAY_BUFFER_ARB, 0);
		glBindBufferARB(GL_ELEMENT_ARRAY_BUFFER_ARB, 0);
	}
	return h;
}

void RAS_VAOpenGLRasterizer::EnableTextures(bool enable)
{
	if (enable)
		glEnableClientState(GL_TEXTURE_COORD_ARRAY);
	else
		glDisableClientState(GL_TEXTURE_COORD_ARRAY);
	RAS_OpenGLRasterizer::EnableTextures(enable);
}

void	RAS_VAOpenGLRasterizer::SetTexUnit(unsigned int unit)
{
	if (GLEW_ARB_multitexture && unit < m_texunits.size())
	{
		glActiveTextureARB(GL_TEXTURE0 + unit);
		glClientActiveTextureARB(GL_TEXTURE0 + unit);
		m_activetexture = unit;
	}
}

void RAS_VAOpenGLRasterizer::TexCoordPointer(VAGLDrawHandle *h, const RAS_TexVert *tv)
{
	if (!GLEW_ARB_multitexture)
	{
		if (!m_texunits[0].enable)
			return;
	
		switch (m_texunits[0].texco)
		{
			case RAS_TEXCO_GEN:
				// GL is generating tex cos.
				glDisableClientState(GL_TEXTURE_COORD_ARRAY);
				break;
			case RAS_TEXCO_ORCO:
				glEnableClientState(GL_TEXTURE_COORD_ARRAY);
				glTexCoordPointer(3, GL_FLOAT, sizeof(RAS_TexVert), h->BufferOffset(tv->getLocalXYZ(), tv));
				break;
/*			RAS_TEXCO_OBJECT,
			RAS_TEXCO_LAVECTOR,
			RAS_TEXCO_VIEW,
			RAS_TEXCO_STICKY,
			RAS_TEXCO_WINDOW,
			case RAS_TEXCO_GLOB:*/
				// not implemented...
			case RAS_TEXCO_UV1:
				glEnableClientState(GL_TEXTURE_COORD_ARRAY);
				glTexCoordPointer(2, GL_FLOAT, sizeof(RAS_TexVert), h->BufferOffset(tv->getUV1(), tv));
				break;

			case RAS_TEXCO_DISABLE:
				m_texunits[0].enable = false;
				glDisableClientState(GL_TEXTURE_COORD_ARRAY);
				glDisable(GL_TEXTURE_2D);
				break;
		}
		return;
	}

	for (unsigned int unit = 0; unit < m_texunits.size(); unit++)
	{
		if (!m_texunits[unit].enable)
			break;
		
		glClientActiveTextureARB(GL_TEXTURE0 + unit);
		switch (m_texunits[unit].texco)
		{
			case RAS_TEXCO_GEN:
				// GL is generating tex cos.
				glDisableClientState(GL_TEXTURE_COORD_ARRAY);
				break;
			case RAS_TEXCO_ORCO:
				glEnableClientState(GL_TEXTURE_COORD_ARRAY);
				glTexCoordPointer(3, GL_FLOAT, sizeof(RAS_TexVert), h->BufferOffset(tv->getLocalXYZ(), tv));
				break;
/*			RAS_TEXCO_OBJECT,
			RAS_TEXCO_LAVECTOR,
			RAS_TEXCO_VIEW,
			RAS_TEXCO_STICKY,
			RAS_TEXCO_WINDOW,
			case RAS_TEXCO_GLOB:*/
				// not implemented...
			case RAS_TEXCO_UV1:
				glEnableClientState(GL_TEXTURE_COORD_ARRAY);
				glTexCoordPointer(2, GL_FLOAT, sizeof(RAS_TexVert), h->BufferOffset(tv->getUV1(), tv));
				break;

			case RAS_TEXCO_DISABLE:
				m_texunits[unit].enable = false;
				glActiveTextureARB(GL_TEXTURE0 + unit);
				glClientActiveTextureARB(GL_TEXTURE0 + unit);
				glDisableClientState(GL_TEXTURE_COORD_ARRAY);
				glDisable(GL_TEXTURE_2D);
				break;
		}
	}
}
