/**
 * $Id: OdePhysicsController.h 63 2003-09-09 02:54:07Z hos $
 *
 * ***** BEGIN GPL/BL DUAL LICENSE BLOCK *****
 *
 * The contents of this file may be used under the terms of either the GNU
 * General Public License Version 2 or later (the "GPL", see
 * http://www.gnu.org/licenses/gpl.html ), or the Blender License 1.0 or
 * later (the "BL", see http://www.blender.org/BL/ ) which has to be
 * bought from the Blender Foundation to become active, in which case the
 * above mentioned GPL option does not apply.
 *
 * The Original Code is Copyright (C) 2002 by NaN Holding BV.
 * All rights reserved.
 *
 * The Original Code is: all of this file.
 *
 * Contributor(s): none yet.
 *
 * ***** END GPL/BL DUAL LICENSE BLOCK *****
 */
#ifndef __ODE_PHYSICSCONTROLLER_H
#define __ODE_PHYSICSCONTROLLER_H

#include <vector>

#include <ode/ode.h>

#include "PHY_IPhysicsController.h"
#include "KX_ConvertPhysicsObject.h"
#include "PHY_Pro.h"
#include "OdeSensor.h"

class RAS_MeshObject;
/**
	ODE Physics Controller, a special kind of a PhysicsController.
	A Physics Controller is a special kind of Scene Graph Transformation Controller.
	Each time the scene graph get's updated, the controller get's a chance
	in the 'Update' method to reflect changes.
*/

class ODEPhysicsController : public PHY_IPhysicsController, public KX_ICollisionSensor
							 
{


public:
	ODEPhysicsController(	dSpaceID space,
						dWorldID world,
						RAS_MeshObject *meshobj,
						struct	PHY_ShapeProps*		shapeprops,
						struct	PHY_MaterialProps*	smmaterial,
						struct	KX_ObjectProperties*	objprop,
						class	PHY_IMotionState*	motionstate
	);
/*	ODEPhysicsController(
		bool dyna,
		bool fullRigidBody,
		bool phantom,
		class PHY_IMotionState* motionstate,
		dSpaceID space,
		dWorldID world,
		float	mass,
		float	friction,
		float	restitution,
		bool	implicitsphere,
		float	center[3],
		float	extends[3],
		float radius); */

	virtual ~ODEPhysicsController();
	
		// kinematic methods
	virtual void		RelativeTranslate(float dlocX,float dlocY,float dlocZ,bool local);
	virtual void		RelativeRotate(const float drot[9],bool local);
	virtual	void		getOrientation(float &quatImag0,float &quatImag1,float &quatImag2,float &quatReal);
	virtual	void		setOrientation(float quatImag0,float quatImag1,float quatImag2,float quatReal);
	virtual	void		setPosition(float posX,float posY,float posZ);
	virtual	void		setScaling(float scaleX,float scaleY,float scaleZ);
	
	// physics methods
	virtual void		ApplyTorque(float torqueX,float torqueY,float torqueZ,bool local);
	virtual void		ApplyForce(float forceX,float forceY,float forceZ,bool local);
	virtual void		SetAngularVelocity(float ang_velX,float ang_velY,float ang_velZ,bool local);
	virtual void		SetLinearVelocity(float lin_velX,float lin_velY,float lin_velZ,bool local);
	virtual void		applyImpulse(float attachX,float attachY,float attachZ, float impulseX,float impulseY,float impulseZ);
	virtual void		SetActive(bool active){};
	virtual void		SuspendDynamics();
	virtual void		RestoreDynamics();


	/**  
		reading out information from physics
	*/
	virtual void		GetLinearVelocity(float& linvX,float& linvY,float& linvZ);
	/** 
		GetVelocity parameters are in geometric coordinates (Origin is not center of mass!).
	*/
	virtual void		GetVelocity(const float posX,const float posY,const float posZ,float& linvX,float& linvY,float& linvZ); 
	virtual	float		getMass();
	virtual	void		getReactionForce(float& forceX,float& forceY,float& forceZ);
	virtual	void		setRigidBody(bool rigid);
		
	
	virtual	void		PostProcessReplica(class PHY_IMotionState* motionstate,class PHY_IPhysicsController* parentctrl);
	
	// \todo remove next line !
	virtual void			SetSimulatedTime(float time);
	

	virtual	void		WriteDynamicsToMotionState() {};
	virtual void	WriteMotionStateToDynamics(bool nondynaonly);

	/** 
		call from Scene Graph Node to 'update'.
	*/
	virtual bool	SynchronizeMotionStates(float time);

		// clientinfo for raycasts for example
	virtual	void*				getClientInfo() { return m_clientInfo;}
	virtual	void				setClientInfo(void* clientinfo) {m_clientInfo = clientinfo;};
	void*						m_clientInfo;

	dBodyID				GetOdeBodyId() { return m_bodyId; }

//	float	getFriction() { return m_friction;}
//	float	getRestitution() { return m_restitution;}

	void attachCollisionSensor(KX_ICollisionSensor *sens);
	ODESensor *newSphereSensor(KX_ICollisionSensor *sens, float margin);
	ODESensor *newRaySensor(KX_ICollisionSensor *sens, int axis, float length);
//	ODESensor *new
	
	virtual void	HandleCollision(ODESensor *o1, ODESensor *o2);

	float getFriction() { return m_material->m_friction; }
	float getRestitution() { return m_material->m_restitution; }
private:

	bool						m_firstTime;

	// data to calculate fake velocities for kinematic objects (non-dynas)
	bool						m_bKinematic;
	bool						m_bPrevKinematic;

	
	float						m_lastTime;
	bool						m_OdeSuspendDynamics;
	class	PHY_IMotionState*			m_MotionState;

	//Ode specific members
	dBodyID				m_bodyId;
	//dGeomID				m_geomId;
	dSpaceID			m_space;
	dWorldID			m_world;
	struct	PHY_ShapeProps*		m_shape;
	struct	PHY_MaterialProps*	m_material;
	struct	KX_ObjectProperties	m_objprop;

	std::vector<ODESensor*>			m_OdeSensors;
	std::vector<KX_ICollisionSensor*>		m_OdeCollisionSensors;

	RAS_MeshObject *m_meshobj;
};

#endif //__ODE_PHYSICSCONTROLLER_H

