/**
* $Id: OdeSensor.h 133 2003-10-05 02:56:47Z kester $
*
 * ***** BEGIN GPL/BL DUAL LICENSE BLOCK *****
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version. The Blender
 * Foundation also sells licenses for use in proprietary software under
 * the Blender License.  See http://www.blender.org/BL/ for information
 * about this.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software Foundation,
 * Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 *
 * The Original Code is Copyright (C) 2001-2002 by NaN Holding BV.
 * All rights reserved.
 *
 * The Original Code is: all of this file.
 *
 * Contributor(s): none yet.
 *
 * ***** END GPL/BL DUAL LICENSE BLOCK *****
*/
#ifndef ODE_SENSOR_H
#define ODE_SENSOR_H

#include <ode/ode.h>

class ODEPhysicsController;

class ODESensor;

class KX_ICollisionSensor {
public:
	virtual void HandleCollision(ODESensor *o1, ODESensor *o2) = 0;

};


class ODESensor {
protected:
	ODEPhysicsController *m_ctrl;
	KX_ICollisionSensor *m_sens;

	dSpaceID m_space;
	dGeomID m_geom;

	bool	m_body;
	bool	m_recurse;
	int	m_debug;
	void	*m_client;

private:
	float	m_rayDepth;

public:
	ODESensor(dSpaceID space, ODEPhysicsController *ctrl, KX_ICollisionSensor *sens);
	virtual ~ODESensor();

	virtual ODEPhysicsController *getController(void) { return m_ctrl; }

	virtual void setCollisionSensor(KX_ICollisionSensor *sens) { m_sens = sens; }

	dGeomID SphereSensor(float radius);
	void setRadius(float radius);

	dGeomID CCylinderSensor(float radius, float length);
	dGeomID BoxSensor(float dx, float dy, float dz);
	dGeomID RaySensor(int axis, float length);
	float	RayDepth(void) {
		return m_rayDepth;
	}

	virtual void setBody(bool body) { m_body = body ; }
	virtual bool getBody(void) { return m_body ; }

	virtual void setBodyID(dBodyID body);

	virtual bool spaceRecurse(void) { return m_recurse; };

	void setClientObject(void *client) { m_client = client ; }
	void *getClientObject(void) { return m_client ; }

	void setDebug(bool debug) { m_debug = debug; }

	virtual void setPosition(float x, float y, float z);
	virtual void setRotation(dMatrix3 &R);
	virtual void setQuaternion(dQuaternion &q);
	virtual void setSpace(dSpaceID s);

	virtual void HandleCollision(ODESensor *obj2);
};


#endif

