#include <stdlib.h>
#include <stdio.h>
#include <string.h>

#include "DNA_hemesh_types.h" 
#include "DNA_mesh_types.h"
#include "DNA_meshdata_types.h"
#include "DNA_object_types.h"
#include "DNA_material_types.h"

#include "BKE_global.h"
#include "BKE_library.h"
#include "BKE_main.h"
#include "BKE_displist.h"
#include "BKE_utildefines.h"
#include "BKE_dmesh.h"
#include "BKE_DFDeform.h"
#include "BKE_object.h"
#include "BKE_mesh.h"
#include "BKE_material.h"

#include "BIF_meshHE.h"
#include "BLI_arithb.h"
#include "BLI_parray.h"
#include "he_error.h"
#include "he_editmesh.h"
#include "math.h"

void free_hemesh(HE_Mesh *me)
{
	if (me->verts) MEM_freeN(me->verts);
	if (me->faces) MEM_freeN(me->faces);
	if (me->half_edges) MEM_freeN(me->half_edges);
	if (me->iedges) MEM_freeN(me->iedges);
//if (me->bb) MEM_freeN(me->bb);
	if (me->dl.first) freedisplist(&me->dl);
	if (me->dmesh) free_DMesh(me->dmesh);
}


void unlink_hemesh(HE_Mesh *me)
{
	int i;

	if (!me->mat) return;

	for (i=0; i<me->totcol; i++)
	{
		me->mat[i]->id.us--;
	}
}


void *add_hemesh(void)
{
	HE_Mesh *me = alloc_libblock(&G.main->he_mesh, ID_HME, "He_Mesh");
	me->dl.first = me->dl.last = NULL;
	me->bb = unit_boundbox();
	me->totcol = 0;
	me->subdivLevels = 1;
	return me;
}


void set_hemesh(Object *ob, HE_Mesh *me)
{
	HE_Mesh *old=NULL;

	if(ob==NULL) return;

	if(ob->type==OB_HEMESH)
	{
		old= ob->data;
		old->id.us--;
		ob->data= me;
		id_us_plus((ID *)me);
	}
	test_object_materials((ID *)me);
}


HE_Mesh *duplicate_hemesh(HE_Mesh *oldme, int do_libblock)
{
	HE_Mesh *me;
	if (do_libblock) me = alloc_libblock(&G.main->he_mesh, ID_HME, "He_Mesh");
	else me = MEM_callocN(sizeof(HE_Mesh), "HE_Mesh no libblock");

	if (oldme->verts) me->verts = MEM_dupallocN(oldme->verts);
	if (oldme->iedges) me->iedges= MEM_dupallocN(oldme->iedges);
	if (oldme->half_edges) me->half_edges = MEM_dupallocN(oldme->half_edges);
	if (oldme->faces) me->faces = MEM_dupallocN(oldme->faces);
	if (oldme->mat) me->mat = MEM_dupallocN(oldme->mat);

	me->totverts = oldme->totverts;
	me->totfaces = oldme->totfaces;
	me->totedges = oldme->totedges;
	me->totiedges = oldme->totiedges;
	me->totcol = oldme->totcol;
	me->f = oldme->f;
	me->f1 = oldme->f1;
	me->subdivLevels = oldme->subdivLevels;

	return me;
}


HE_Mesh *mesh_to_hemesh(Mesh *oldme, int do_libblock)
{
	HE_Mesh *me;
	HE_EditMesh *em;
	HE_EditVert *vlist[4];
	HE_EditFace *efa;
	MVert *mvert;
	MFace *mef;
	int i;
	HE_EditVert **vert_table = MEM_callocN(sizeof(void*)*oldme->totvert, "vert table mesh-to-hemesh");

/*SOOO glad I removed G.he_editMesh dependency from most
  of the code!*/
	em = MEM_callocN(sizeof(HE_EditMesh), "HE_EditMesh");
	G.he_editMesh = em;

/*set fast vert lookup*/
	mvert = oldme->mvert;
	for (i=0; i<oldme->totvert; i++, mvert++)
	{
		vert_table[i] = HE_addvertlist(em, mvert->co);
	}

	mef = oldme->mface;
	for (i=0; i<oldme->totface; i++, mef++)
	{
		if (!mef->v3) continue;

		vlist[0] = vert_table[mef->v1];
		vlist[1] = vert_table[mef->v2];
		vlist[2] = vert_table[mef->v3];
		break;
		if (mef->v4)
		{
			vlist[3] = vert_table[mef->v4];
			efa = HE_addfacelist(em, vlist, (int) 4);
		} else efa = HE_addfacelist(em, vlist, (int) 3);

		efa->mat = mef->mat_nr;
		efa->f = mef->flag & SELECT;
		if (mef->flag & ME_SMOOTH) efa->f |= HEM_SMOOTH;

		if (HE_CatchError(HE_NON_MANIFOLD))
		{
			while (HE_CatchError(0));			  /*clear stack*/
			he_free_HEEditMesh(em);
			return NULL;
		}
	}

	if (do_libblock) alloc_libblock(&G.main->he_mesh, ID_HME, "He_Mesh");
	else me = MEM_callocN(sizeof(HE_Mesh), "HE_Mesh no libblock");
	memset(me, 0, sizeof(HE_Mesh));

	check_faces(em);
	he_load_HEEditMesh(me, em);
	he_free_HEEditMesh(em);

	if (oldme->mat)
	{
		me->mat = MEM_dupallocN(oldme->mat);
		me->totcol = oldme->totcol;
		for (i=0; i<me->totcol; i++)
		{
			id_us_plus((ID*)me->mat[i]);
		}
	}

	if (vert_table) MEM_freeN(vert_table);
	return me;
}


/* the rules of faces-to-MFaces for HE MEsh:
	1. Four-vert quads turn into one face.
	2. Five-vert ngons turn into three triangles.
	3. More then five-vert ngons form a center-vert
	   and makes triangles around that.*/
/*face_nors is pointer to pointer to face normal array*/
Mesh *hemesh_to_mesh(HE_Mesh *oldme, int do_libblock, float **face_nors)
{
	PointerArray varray = {NULL, 0};
	PointerArray farray = {NULL, 0};
	PointerArray nfarray = {NULL, 0};

	HE_Face *hef;
	HE_Edge *hed;

	Mesh *me;
	MFace *mef;
	MVert *v1;

	int i, j, totv = 0, totf = 0, len, mat_nr;
	float cent[3], *nor;
	unsigned int vlist[6000];

#if 0
	if (do_libblock) me = add_mesh();
	else me = MEM_callocN(sizeof(Mesh), "Mesh from hemesh");

	totv = oldme->totverts;

	totv = 0;
	totf = 0;
	for (i=0; i<oldme->totverts; i++)
	{
		v1 = MEM_callocN(sizeof(MVert), "MVert from hevert");
		VECCOPY(v1->co, oldme->verts[i].co);
		VECCOPY(v1->no, oldme->verts[i].no);
		PA_AddToArray(&varray, v1, i);
		totv++;
	}

	for (i=0; i<oldme->totfaces; i++)
	{
		hef = &oldme->faces[i];
		mat_nr = hef->mat_index;
		nor = hef->nor;

		hed = &oldme->half_edges[hef->edge];
		len = 0;
		cent[0] = cent[1] = cent[2] = 0;
		do
		{
			vlist[len] = hed->v1;
			VecAddf(cent, cent, oldme->verts[hed->v1].co);
			len++;
			hed = &oldme->half_edges[hed->next];
		} while (hed != &oldme->half_edges[hef->edge]);
		VecMulf(cent, 1.0f/((float)len));		  /*calc center coord*/

		if (len==3)
		{
			mef = MEM_callocN(sizeof(MFace), "MFace from he face");
			mef->v1 = vlist[0];
			mef->v2 = vlist[1];
			mef->v3 = vlist[2];
			mef->v4 = 0;
			mef->mat_nr = mat_nr;

			PA_AddToArray(&farray, mef, totf);
			PA_AddToArray(&nfarray, nor, totf);
			if (hef->f & HEM_SMOOTH) mef->flag |= ME_SMOOTH;
			test_index_mface(mef, 3);
			totf++;
		}
		else if (len==4)
		{
			mef = MEM_callocN(sizeof(MFace), "MFace from he face");
			mef->v1 = vlist[0];
			mef->v2 = vlist[1];
			mef->v3 = vlist[2];
			mef->v4 = vlist[3];
			mef->mat_nr = mat_nr;

			test_index_mface(mef, 4);
			if (hef->f & HEM_SMOOTH) mef->flag |= ME_SMOOTH;
			PA_AddToArray(&farray, mef, totf);
			PA_AddToArray(&nfarray, nor, totf);
			totf++;

		}
		else if (len==5)
		{
			mef = MEM_callocN(sizeof(MFace), "MFace from he face");
			mef->v1 = vlist[0];
			mef->v2 = vlist[1];
			mef->v3 = vlist[2];
			mef->v4 = 0;
			mef->mat_nr = mat_nr;

			test_index_mface(mef, 3);
			if (hef->f & HEM_SMOOTH) mef->flag |= ME_SMOOTH;
			PA_AddToArray(&farray, mef, totf);
			PA_AddToArray(&nfarray, nor, totf);
			totf++;

			mef = MEM_callocN(sizeof(MFace), "MFace from he face");
			mef->v1 = vlist[2];
			mef->v2 = vlist[3];
			mef->v3 = vlist[4];
			mef->v4 = 0;
			mef->mat_nr = mat_nr;

			test_index_mface(mef, 3);
			if (hef->f & HEM_SMOOTH) mef->flag |= ME_SMOOTH;
			PA_AddToArray(&farray, mef, totf);
			PA_AddToArray(&nfarray, nor, totf);
			totf++;

			mef = MEM_callocN(sizeof(MFace), "MFace from he face");
			mef->v1 = vlist[4];
			mef->v2 = vlist[0];
			mef->v3 = vlist[2];
			mef->v4 = 0;
			mef->mat_nr = mat_nr;

			test_index_mface(mef, 3);
			if (hef->f & HEM_SMOOTH) mef->flag |= ME_SMOOTH;
			PA_AddToArray(&farray, mef, totf);
			PA_AddToArray(&nfarray, nor, totf);
			totf++;
		}										  /*make a tri fan*/
		else
		{
			v1 = MEM_callocN(sizeof(MVert), "MVertcenterHEf");
			VECCOPY(v1->co, cent);
			VECCOPY(v1->no, hef->nor);
			PA_AddToArray(&varray, v1, totv);
			totv++;

			for (j=0; j<len; j++)
			{
				mef = MEM_callocN(sizeof(MFace), "MFace from he face");
				mef->v1 = vlist[j];
				if (j != len-1) mef->v2 = vlist[j+1];
				else mef->v2 = vlist[0];
				mef->v3 = totv - 1;
				mef->v4 = 0;
				mef->mat_nr = mat_nr;

				test_index_mface(mef, 3);
				if (hef->f & HEM_SMOOTH) mef->flag |= ME_SMOOTH;
				PA_AddToArray(&farray, mef, totf);
				PA_AddToArray(&nfarray, nor, totf);
				totf++;
			}
		}
	}

	me->mvert = MEM_callocN((totv+1)*sizeof(MVert), "MVert from verts");
	me->mface = MEM_callocN((totf+1)*sizeof(MFace), "MFace from faces");
	me->totvert = totv;
	me->totface = totf;
	if (face_nors) *face_nors = MEM_callocN(sizeof(float*)*(totf*3), "face_nors in hemesh.c");

	memset(me->mvert, 0, sizeof(me->mvert));
	for (i=0; i<totv; i++)
	{
		memcpy(&me->mvert[i], varray.array[i], sizeof(MVert));
		MEM_freeN(varray.array[i]);
	}

	mef = me->mface;
	for (i=0; i<totf; i++)
	{
		memcpy(&mef[i], farray.array[i], sizeof(MFace));
		if (face_nors)
		{
			float *fptr = *face_nors;
			VecCopyf(&fptr[i*3], ((float*)nfarray.array[i]));
		}
		MEM_freeN(farray.array[i]);
	}

	PA_FreeArray(&farray);
	PA_FreeArray(&varray);
	PA_FreeArray(&nfarray);

	me->subdiv = me->subdivr = oldme->subdivLevels;
	if (oldme->f & HEM_SUBSURF)
	{
		printf("subsurf!\n");
		me->flag |= ME_SUBSURF;
	}

/*copy materials*/
	me->totcol = oldme->totcol;
	me->mat = MEM_callocN(sizeof(Material*)*oldme->totcol, "material array");
	for (i=0; i<oldme->totcol; i++)
	{
		me->mat[i] = oldme->mat[i];
		id_us_plus((ID*)me->mat[i]);
	}
#endif
//make_orco_mesh(me);
	return me;
}
