#include "MEM_guardedalloc.h"

#include "BMF_Api.h" 

#include "IMB_imbuf.h"

#include "MTC_matrixops.h"

#include "DNA_camera_types.h"
#include "DNA_curve_types.h"
#include "DNA_constraint_types.h" // for drawing constraint
#include "DNA_effect_types.h"
#include "DNA_ipo_types.h"
#include "DNA_lamp_types.h"
#include "DNA_lattice_types.h"
#include "DNA_material_types.h"
#include "DNA_mesh_types.h"
#include "DNA_meshdata_types.h"
#include "DNA_meta_types.h"
#include "DNA_object_types.h"
#include "DNA_space_types.h"
#include "DNA_scene_types.h"
#include "DNA_screen_types.h"
#include "DNA_view3d_types.h"
#include "DNA_world_types.h"

#include "BKE_utildefines.h"
#include "BKE_curve.h"
#include "BKE_constraint.h" // for the get_constraint_target function
#include "BKE_deform.h"		// lattice_modifier()
#include "BKE_displist.h"
#include "BKE_effect.h"
#include "BKE_global.h"
#include "BKE_ipo.h"
#include "BKE_lattice.h"
#include "BKE_mesh.h"
#include "BKE_material.h"
#include "BKE_mball.h"
#include "BKE_object.h"
#include "BKE_deform.h"
#include "BKE_hemesh.h"

#include "BIF_gl.h"
#include "BIF_mywindow.h"
#include "BIF_screen.h"
#include "BIF_space.h"
#include "BIF_editarmature.h"
#include "BIF_editmesh.h"
#include "BIF_glutil.h"
#include "BIF_resources.h"
#include "BIF_editmode_undo.h"
#include "BIF_meshHE.h"

#include "BDR_drawmesh.h"
#include "BDR_drawobject.h"
#include "BDR_editobject.h"

#include "BSE_view.h"
#include "BSE_drawview.h"
#include "BSE_trans_types.h"

#include "blendef.h"
#include "mydevice.h"
#include "nla.h"

#include "BLI_blenlib.h"
#include "BLI_arithb.h"
#include "BLI_editVert.h"

#include "he_editmesh.h"
#include "he_error.h"
#include "math.h"


//speedup tests. . .
//#define MEM_callocN(size, str) calloc(1, size)
//#define MEM_freeN(ptr) ;

/*********** Undo -- origional version ****************/
void *heeditmeshToUndoMesh(void)
{
	heUndoMesh *undo = MEM_callocN(sizeof(heUndoMesh), "heUndoMesh");
	int i;
	int totedges=0, totverts=0, totfaces=0, totiedges=0;
	int tothash=0, curhash = 0;
	HE_EditMesh *em = G.he_editMesh;
	HE_EditVert *eve;
	HE_EditFace *efa;
	HE_EditEdge *eed;
	HE_EditEdgeInfo *ieed;
	HE_HashEdge *hash;

	/*all ->f1 flags of all HE_Edit* will store the index*/
	tothash = 0;
	for (eve=em->verts.first; eve; eve=eve->next) {
		eve->f1 = totverts;
		for (hash=eve->hashedges.first; hash; hash=hash->next) {
			tothash++;
		}
		totverts++;
	}
	undo->totv = totverts;
	undo->tothash = tothash;

 	for (eed=em->edges.first; eed; eed=eed->next) {
 		eed->f1 = totedges;
		totedges++;
	}
	undo->tothe = totedges;

	for (ieed=em->einfo.first; ieed; ieed=ieed->next) {
		ieed->f1 = totiedges; 
		totiedges++;
	}
	undo->totie = totiedges;

	for (efa=em->faces.first; efa; efa=efa->next) {
		efa->f1 = totfaces;
		totfaces++;
	}
	undo->totf = totfaces;


	undo->faces = MEM_callocN(totfaces*sizeof(heUndoFace), "heUndoFace");
	undo->half_edges = MEM_callocN(totedges*sizeof(heUndoHalfEdge), "heUndoFace");
	undo->iedges = MEM_callocN(totiedges*sizeof(heUndoEdgeInfo), "heUndoFace");
	undo->verts = MEM_callocN(totverts*sizeof(heUndoVert), "heUndoFace");
	undo->hashes = MEM_callocN(tothash*sizeof(heUndoHashEdge), "heUndoHashEdge");

	i = 0;

	curhash = 0;
	for (eve=em->verts.first; eve; i++, eve=eve->next) {
		VECCOPY(undo->verts[i].co, eve->co);
		VECCOPY(undo->verts[i].no, eve->no);
		undo->verts[i].f = eve->f;
		undo->verts[i].h = eve->h;
		undo->verts[i].f1 = i;
		if (eve->edge) {
			undo->verts[i].edge = &undo->half_edges[eve->edge->f1];
		} else undo->verts[i].edge = NULL;
		undo->verts[i].index = i;	 
	}

	i = 0;
	for (eed=em->edges.first; eed; i++, eed=eed->next) {
		VECCOPY(undo->half_edges[i].no, eed->no);
		undo->half_edges[i].v1 = &(undo->verts[eed->v1->f1]);
		undo->half_edges[i].v2 = &(undo->verts[eed->v2->f1]);
		undo->half_edges[i].enext = &(undo->half_edges[eed->enext->f1]);
		undo->half_edges[i].eprev = &(undo->half_edges[eed->eprev->f1]);
		if (eed->pair) undo->half_edges[i].pair = &(undo->half_edges[eed->pair->f1]);
		else undo->half_edges[i].pair = NULL;

		if (eed->ieed)
			undo->half_edges[i].ieed = &(undo->iedges[eed->ieed->f1]);

		undo->half_edges[i].face = &(undo->faces[eed->face->f1]);
		undo->half_edges[i].index = i;
	}

	i = 0;
	for (ieed=em->einfo.first; ieed; i++, ieed=ieed->next) {
		undo->iedges[i].f = ieed->f;
		undo->iedges[i].h = ieed->f;
		undo->iedges[i].e1 = &undo->half_edges[ieed->e1->f1];
		if (ieed->e2) undo->iedges[i].e2 = &(undo->half_edges[ieed->e2->f1]);
		undo->iedges[i].crease = ieed->crease;
		undo->iedges[i].seam = ieed->seam;
		undo->iedges[i].f1 = i;
		undo->iedges[i].index = i;
	}  

	i = 0;
 	for (efa=em->faces.first; efa; i++, efa=efa->next) {
		undo->faces[i].f = efa->f;
  		undo->faces[i].h = efa->h;
  		VECCOPY(undo->faces[i].cent, efa->cent);
  		VECCOPY(undo->faces[i].no, efa->nor);
		undo->faces[i].edge = &(undo->half_edges[efa->edge->f1]);
		undo->faces[i].f1 = i;
		undo->faces[i].index = i;
		undo->faces[i].mat = efa->mat;
	}

	return undo;
}

void freeundoHE(heUndoMesh *undo)
{
	if (undo->verts) MEM_freeN(undo->verts);
	if (undo->half_edges) MEM_freeN(undo->half_edges);
	if (undo->iedges) MEM_freeN(undo->iedges);
	if (undo->faces) MEM_freeN(undo->faces);
	if (undo->hashes) MEM_freeN(undo->hashes);
	MEM_freeN(undo);

} 

/*#define MAXBLOCK	0x7ffdefff
  never mind all this. . .
  #define callochedge(i)	(i<em->totbhe?&em->blockhe[i]:MEM_callocN(sizeof(HE_EditEdge), "HE_EditEdge"));
  #define callociedge(i)	(i<em->totbie?&em->blockie[i]:MEM_callocN(sizeof(HE_EditEdgeInfo), "HE_EditEdgeInfo"));
  #define callocvert(i)	(i<em->totbv?&em->blockv[i]:MEM_callocN(sizeof(HE_EditVert), "HE_EditVert"));
  #define callocface(i)	(i<em->totbf?&em->blockf[i]:MEM_callocN(sizeof(HE_EditFace), "HE_EditFace"));
*/

HE_EditMesh *undomeshToHEEditMesh(heUndoMesh *undo)
{
	HE_EditMesh *em = MEM_callocN(sizeof(HE_EditMesh), "HE_EditMesh from undo");
	HE_EditFace *efa;
	HE_EditEdge *eed;
	HE_EditVert *eve, *vlist[7000];
	heUndoHalfEdge *ueed;
	heUndoFace *uefa;
	heUndoVert *ueve;
	int i, len;

	/*free old mesh. . .*/
	free_HEEditMesh();
	G.he_editMesh = em;

	/*make verts. . .*/
	for (i=0; i<undo->totv; i++) {
		ueve = &undo->verts[i];
		eve = HE_addvertlist(em, ueve->co);
		VECCOPY(eve->no, ueve->no);
		eve->f = ueve->f;
		ueve->neve = eve;
	}

	/*make faces. . .*/
	for (i=0; i<undo->totf; i++) {
		uefa = &undo->faces[i];
		ueed = uefa->edge;
		len = 0;
		do {
			vlist[len] = ueed->v1->neve;
			len++;
			ueed=ueed->enext;
		} while (ueed != uefa->edge);
		efa = HE_addfacelist(em, vlist, len);
		efa->f = uefa->f;
		efa->mat = uefa->mat;

		ueed = uefa->edge;
		eed = efa->edge;
		do {
			eed->ieed->f = ueed->ieed->f;
			eed->ieed->seam = ueed->ieed->seam;
			eed->ieed->crease = ueed->ieed->crease;
			VECCOPY(eed->no, ueed->no);
			eed=eed->enext;
			ueed=ueed->enext;
		} while (eed != efa->edge);	 
	}

	return em;
}
#if 0
HE_EditMesh *bugged_yet_fastundomeshToHEEditMesh(heUndoMesh *undo)
{
	HE_EditMesh *em = MEM_callocN(sizeof(HE_EditMesh), "HE_EditMesh from Undo");
 	HE_EditVert *eve;
	HE_EditEdge *eed;
	HE_EditEdgeInfo *ieed;
	HE_EditFace *efa;

	heUndoHalfEdge *ueed;
	heUndoFace *uefa;
	heUndoVert *ueve;

	HE_HashEdge *hash;

	int i, curhash=0;

	free_HEEditMesh();
	G.he_editMesh = em;

	em->blockv = MEM_callocN(sizeof(HE_EditVert)*undo->totv, "undo to heeditmesh verts");
	em->totbv = undo->totv;
	em->blockhe = MEM_callocN(sizeof(HE_EditEdge)*undo->tothe, "undo to heeditmesh halfedges");
	em->totbhe = undo->tothe;
	em->blockf = MEM_callocN(sizeof(HE_EditFace)*undo->totf, "undo to edithe faces");
	em->totbf = undo->totf;
	em->blockie = MEM_callocN(sizeof(HE_EditEdgeInfo)*undo->totie, "undo to edithemesh");
	em->totbie = undo->totie;
	em->blockhash = MEM_callocN(sizeof(HE_HashEdge)*undo->tothash, "undo to edgehash");
	em->curbhash = em->totbhash = undo->tothash;

	curhash = 0;

	/*clear all new pointers. . .*/
	for (i=0; i<undo->totf; i++) {
		uefa = &undo->faces[i];
		uefa->nefa = NULL;
		ueed = uefa->edge;
		do {
			ueed->need = NULL;
			ueed->v1->neve = NULL;
			if (ueed->v2) ueed->v2->neve = NULL;
			ueed->ieed->nieed = NULL;
			ueed=ueed->enext;
		} while (ueed != uefa->edge);
	}

	/*its vert time!*/
	for (i=0; i<undo->totv; i++) {
		ueve = &undo->verts[i];
		eve = &em->blockv[ueve->index];
		VECCOPY(eve->co, ueve->co);
		VECCOPY(eve->no, ueve->no);
		eve->f = ueve->f;
		eve->fast = 1;
		if (ueve->edge) eve->edge = &em->blockhe[ueve->edge->index];
		BLI_addtail(&em->verts, eve);
		ueve->neve = eve;
	}
	curhash = 0;
	/*Now restore data, starting with faces. . .*/
	for (i=0; i<undo->totf; i++) {
		uefa = &undo->faces[i];
		efa = &em->blockf[uefa->index];
		VECCOPY(efa->nor, uefa->no);
		VECCOPY(efa->cent, uefa->cent);
		efa->edge = &em->blockhe[uefa->edge->index];
		efa->f = uefa->f;
		efa->fast = 1;

		/*restore this face's half-edges and verts. . .*/
		ueed = uefa->edge;
		do {
			eed = &em->blockhe[ueed->index];
			if (ueed->index >= em->totbhe) printf("!!!!!!!!em->totbhe: %d, eed->index: %d\n", em->totbhe, ueed->index);
			eed->fast = 1;

			/*generate hash half-edges. . .*/
			if (curhash < em->totbhash) hash = &em->blockhash[curhash];
			else { printf("making new h\n"); hash = MEM_callocN(sizeof(HE_HashEdge), "HE_HashEdge");}
			hash->eed = eed;
			hash->fast = 1;
			eve = &em->blockv[ueed->v1->index];
			BLI_addtail(&eve->hashedges, hash);

			/*no hash edge may be in more then one vert list!*/
			eve = &em->blockv[ueed->v2->index];
			if (curhash < em->totbhash) hash = &em->blockhash[curhash];
			else { printf("making new h\n"); hash = MEM_callocN(sizeof(HE_HashEdge), "HE_HashEdge");}
			hash->eed = eed;
			hash->fast = 1;
			BLI_addtail(&eve->hashedges, hash);
			curhash++;

			/*now check the info edge. . .*/
			ieed = &em->blockie[ueed->ieed->index];
			if (!ueed->ieed->nieed) { /*restore info edge. . .*/
				ieed->e1 = &em->blockhe[ueed->ieed->e1->index];
				if (ueed->ieed->e2) ieed->e2 = &em->blockhe[ueed->ieed->e2->index];
				ieed->crease = ueed->ieed->crease;
				ieed->seam = ueed->ieed->seam;
				ieed->f = ueed->ieed->f;
				ieed->fast = 1;
				ueed->ieed->nieed = ieed;
				BLI_addtail(&em->einfo, ieed);
			}

			/*restore this half-edge. . .*/
			eed->face = efa;
			eed->ieed = ieed;
			if (ueed->pair) eed->pair = &em->blockhe[ueed->pair->index];

			//printf("max he: %d, ueed->pair->index: %d\n", em->totbhe, ueed->pair->index);
			eed->v1 = &em->blockv[ueed->v1->index];
			if (ueed->v2) eed->v2 = &em->blockv[ueed->v2->index];
			eed->enext = &em->blockhe[ueed->enext->index];
			eed->eprev = &em->blockhe[ueed->eprev->index];

			/*resore vert #1. . .*/
			if (!ueed->v1->neve) {
				eve = &em->blockv[ueed->v1->index];
				VECCOPY(eve->co, ueed->v1->co);
				VECCOPY(eve->no, ueed->v1->no);
				eve->f = ueed->v1->f;
				eve->fast = 1;
				if (ueed->v1->edge) eve->edge = &em->blockhe[ueed->v1->edge->index];
				BLI_addtail(&em->verts, eve);
				ueed->v1->neve = eve;
			}
			/*resore vert #2. . .*/
			if (!ueed->v2->neve) {
				eve = &em->blockv[ueed->v2->index];
				VECCOPY(eve->co, ueed->v2->co);
				VECCOPY(eve->no, ueed->v2->no);
				eve->f = ueed->v2->f;
				eve->fast = 1;
				if (ueed->v2->edge) eve->edge = &em->blockhe[ueed->v2->edge->index];
				BLI_addtail(&em->verts, eve);
				ueed->v2->neve = eve;
			}

			BLI_addtail(&em->edges, eed);
			ueed=ueed->enext;
		} while (ueed != uefa->edge);
		BLI_addtail(&em->faces, efa);
	}

	for (eed=em->edges.first; eed; eed=eed->next) {
		if (eed->pair) {
			printf("eed->pair->f1: %d\n", eed->pair->f1);
			printf("eed->pair->face->f: %d\n", eed->pair->face->f1);
		}
  		printf("eed->face->f: %d\n", eed->face->f);
	}
	return em;
}
#endif

void undo_push_hemesh(char *str)
{
   	undo_editmode_push(str, freeundoHE, undomeshToHEEditMesh, heeditmeshToUndoMesh);
} 

/*********** Mesh conversion stuff***/

void he_free_HEEditMesh(HE_EditMesh *em)
{
	HE_EditVert *eve, *vnext;
	HE_EditEdge *eed, *henext;
	HE_EditEdgeInfo *ieed, *ienext;
	HE_EditFace *efa, *fnext;
	HE_HashEdge *hash, *hnext;

	if (!em) return;

	/** delete each element, making sure it isn't in the 
		block memory first. . .*/
	for (eve=em->verts.first; eve; eve=vnext) {
		vnext = eve->next;
		for (hash=eve->hashedges.first; hash; hash=hnext) {
			hnext = hash->next;
			if (hash->fast==0) MEM_freeN(hash);
		}
		if (eve->fast==0) MEM_freeN(eve);
	}  
	for (eed=em->edges.first; eed; eed=henext) {
		henext = eed->next;
		if (eed->fast==0) MEM_freeN(eed);
	}
	for (ieed=em->einfo.first; ieed; ieed=ienext) {
		ienext = ieed->next;
		if (ieed->fast==0) MEM_freeN(ieed);
	}

	for (efa=em->faces.first; efa; efa=fnext) {
		fnext = efa->next;
		if (efa->fast==0) MEM_freeN(efa);
	}
	/*BLI_freelistN(&em->verts);
	  BLI_freelistN(&em->edges);
	  BLI_freelistN(&em->einfo);
	  BLI_freelistN(&em->faces);*/

	/*delete block-allocated memory. . .*/
	if (em->blockf) MEM_freeN(em->blockf);
	if (em->blockv) MEM_freeN(em->blockv);
	if (em->blockie) MEM_freeN(em->blockie);
	if (em->blockhe) MEM_freeN(em->blockhe);
	if (em->blockhash) MEM_freeN(em->blockhash);

	MEM_freeN(em);  
}

void free_HEEditMesh(void)
{
	he_free_HEEditMesh(G.he_editMesh);
	G.he_editMesh = NULL;
}

void he_set_blockmem(HE_EditMesh *em, int type, int tot)
{
	if (type=='v') {
		em->blockv = MEM_callocN(tot*sizeof(HE_EditVert), "blockv");
		em->curbv = 0;
		em->totbv = tot;
	}
 	else if (type=='h') {
		em->blockhe = MEM_callocN(tot*sizeof(HE_EditEdge), "blockv");
		em->curbhe = 0;
		em->totbhe = tot;
	}
	else if (type=='i') {
		em->blockie = MEM_callocN(tot*sizeof(HE_EditEdgeInfo), "blockv");
		em->curbie = 0;
		em->totbie = tot;
	}
	else if (type=='f') {
		em->blockf = MEM_callocN(tot*sizeof(HE_EditFace), "blockv");
		em->curbf = 0;
		em->totbf = tot;
	}
}

void he_load_HEEditMesh(HE_Mesh *me, HE_EditMesh *em)
{
	HE_EditVert *eve;
	HE_EditEdge *eed;
	HE_EditEdgeInfo *ieed;
	HE_EditFace *efa;

	HE_Vert *mvert;
	HE_Face *mface;
	HE_Edge *medge;
	HE_EdgeInfo *minfo;

	int totverts=0, i=0;
	int totfaces=0;
	int totedges=0;
	int totiedges=0;

	if (me->verts) MEM_freeN(me->verts);
	if (me->faces) MEM_freeN(me->faces);
	if (me->half_edges) MEM_freeN(me->half_edges);
	if (me->iedges) MEM_freeN(me->iedges);

	/*all ->f1 flags of all HE_Edit* will store the index*/
	for (eve=em->verts.first; eve; eve=eve->next) {
		eve->f1 = totverts;
		totverts++;
	}
	me->totverts = totverts;

 	for (eed=em->edges.first; eed; eed=eed->next) {
 		eed->f1 = totedges;
		totedges++;
	}
	me->totedges = totedges;

	for (ieed=em->einfo.first; ieed; ieed=ieed->next) {
		ieed->f1 = totiedges; 
		totiedges++;
	}
	me->totiedges = totiedges;

	for (efa=em->faces.first; efa; efa=efa->next) {
		efa->f1 = totfaces;
		totfaces++;
	}

	me->totfaces = totfaces;

	G.totvert = totverts;

	me->verts = mvert = MEM_callocN(totverts*sizeof(HE_Vert), "totHE_Vert");
	me->half_edges = medge = MEM_callocN(totedges*sizeof(HE_Edge), "totHE_Edge");
	me->iedges = minfo = MEM_callocN(totiedges*sizeof(HE_EdgeInfo), "totHE_EdgeInfo");
	me->faces = mface = MEM_callocN(totfaces*sizeof(HE_Face), "totHE_Face");

	/*we set some indices to -1 for indice reasons */
	i = 0;
	for (eve=em->verts.first; eve; i++, eve=eve->next) {
		mvert[i].f = eve->f;
		VECCOPY(mvert[i].co, eve->co);
		VECCOPY(mvert[i].no, eve->no);
		mvert[i].h = eve->h;
		if (eve->edge) mvert[i].edge = eve->edge->f1;
		else mvert[i].edge = -1;
 	}

	i = 0;
 	for (eed=em->edges.first; eed; i++, eed=eed->next) {
 		medge[i].v1 = eed->v1->f1;
 		VECCOPY(medge[i].no, eed->no);
 		medge[i].v2 = eed->v2->f1;
 		medge[i].iedge = eed->ieed->f1;
 		medge[i].face = eed->face->f1;
 		medge[i].next = eed->enext->f1;
 		medge[i].prev = eed->eprev->f1;
 		if (eed->pair) medge[i].pair = eed->pair->f1;
 		else medge[i].pair = -1;
	}   
 
	i = 0;
	for (efa=em->faces.first; efa; i++, efa=efa->next) {
  		mface[i].edge = efa->edge->f1;
  		VECCOPY(mface[i].nor, efa->nor);
  		VECCOPY(mface[i].dot, efa->cent);
  		mface[i].f = efa->f;
  		mface[i].h = efa->h;
  		mface[i].mat_index = efa->mat;
	} 

	i=0;
	for (ieed=em->einfo.first; ieed; i++, ieed=ieed->next) {
		minfo[i].v1 = ieed->e1->v1->f1;
		minfo[i].v2 = ieed->e1->v2->f1;
		minfo[i].h = ieed->h;
		minfo[i].f = ieed->f;
		minfo[i].seam = ieed->seam;
		minfo[i].crease = ieed->crease;
		minfo[i].e1= ((HE_EditEdge*)(ieed->e1))->f1;
		if (ieed->e2) minfo[i].e2= ((HE_EditEdge*)(ieed->e2))->f1;   
		else minfo[i].e2 = -1;
	}

	/*make all other objects remake their displists. . .
	  should find better method of doing this :) */
/*	DISABLED till full rewrite --> sce = G.scene;
	base = sce->base.first;
	while (base) {
	obj = base->object;
	if (obj->data==me && (obj != G.obedit)) {
	me->f |= HEM_NODISPLIST;
<<<<<<< HE_Base.c
	
=======
	//makeDispList(obj);
>>>>>>> 1.13
	me->f &= ~HEM_NODISPLIST;
	}
	base=base->next;
	}	*/
}

void load_HEEditMesh(void)
{
	he_load_HEEditMesh(G.obedit->data, G.he_editMesh);
}

void make_HEEditMesh(void)
{
	HE_Mesh *me = G.obedit->data;
	HE_Face *mef;
	HE_Edge *hed;
	HE_EditVert **vert_table;
	HE_EditVert *fvlist[1000];
	HE_EditFace *efa;
	HE_EditEdge *eed;
	HE_EditMesh *em = MEM_callocN(sizeof(HE_EditMesh), "HE_EditMesh");
	int i, len;

	vert_table = MEM_callocN(me->totverts * sizeof(HE_EditVert *), 
							 "he vert table");
	if (G.he_editMesh) free_HEEditMesh();
	G.he_editMesh = em;

	for (i=0; i<me->totverts; i++) {
		vert_table[i] = HE_addvertlist(em, me->verts[i].co);
		vert_table[i]->f = me->verts[i].f;
	}
	for (i=0; i<me->totfaces; i++) {
		len = 0;
		mef = &me->faces[i];
		hed = &me->half_edges[mef->edge];
		do {
			fvlist[len] = vert_table[hed->v1];
			len++;
			hed = &me->half_edges[hed->next];
		} while (hed != &me->half_edges[mef->edge]);
		efa = HE_addfacelist(em, fvlist, len);
		efa->f = mef->f;
		efa->mat = mef->mat_index;
		efa->h = mef->h;
		
		eed = efa->edge;
		hed = &me->half_edges[mef->edge];
		do {
			eed->v1->h = me->verts[hed->v1].h;
			eed->ieed->f = me->iedges[hed->iedge].f;
			eed->ieed->h = me->iedges[hed->iedge].h;
			eed->ieed->crease = me->iedges[hed->iedge].crease;
			eed->ieed->seam = me->iedges[hed->iedge].seam;
			VECCOPY(eed->no, hed->no);

			eed=eed->enext;
			hed = &me->half_edges[hed->next];
		} while (eed != efa->edge);
	}

	/*this forces half-edge correctness, but
	  will crash on certain meshes an theoretically should not
	  be needed here*/
	//check_faces();
	HE_recalcnormals();
}

#if 0
void makssse_HEEditMesh(void)
{
	HE_Mesh *me = G.obedit->data;
	HE_Edge *med;
 	/*if (!me) {
	  me = alloc_libblock(&G.main->he_mesh, ID_HME, "HE_Mesh");  
	  G.obedit->data = me;
	  } */	 
 	HE_EditEdge *eed, *edges;
 	HE_EditEdgeInfo *ieed, *iedges;
 	HE_EditVert *eve, *verts;
 	HE_EditFace *efa, *faces;
 	HE_HashEdge *hash;
	HE_EditMesh *em = MEM_callocN(sizeof(HE_EditMesh), "HE_EditMesh");
	G.he_editMesh = em;
 	int i;
  	/*allocate the data first. . .*/
  	em->blockf = faces = MEM_callocN(sizeof(HE_EditFace)*me->totfaces, "faces");
  	em->blockhe = edges = MEM_callocN(sizeof(HE_EditEdge)*me->totedges, "edges");
  	em->blockie = iedges = MEM_callocN(sizeof(HE_EditEdgeInfo)*me->totiedges, "iedges");
  	em->blockv = verts = MEM_callocN(sizeof(HE_EditVert)*me->totverts, "verts");
  	em->totbf = em->curbf = me->totfaces;
	em->totbhe = em->curbhe = me->totedges;
	em->totbv = em->curbv = me->totverts;
	em->totbie = em->curbie = me->totiedges;

  	for (i=0; i<me->totverts; i++) {
		eve = &verts[i];
		eve->fast = 1;
		VECCOPY(eve->co, me->verts[i].co);
		VECCOPY(eve->no, me->verts[i].no);
  		/*Remember, we don't store f1/2/3 flags!!!*/
  		eve->f = me->verts[i].f;
  		eve->hashedges.first = eve->hashedges.last = NULL;
  		BLI_addtail(&em->verts, eve);
		if (me->verts[i].edge != -1) eve->edge = &edges[me->verts[i].edge];
	}

	/* load edge info (creases). . .*/
	for (i=0; i<me->totiedges; i++) {
		ieed = &iedges[i];
		ieed->fast = 1;
		ieed->f = me->iedges[i].f;
		ieed->h = me->iedges[i].h;
		ieed->crease = me->iedges[i].crease;
		ieed->seam = me->iedges[i].seam;  
		BLI_addtail(&em->einfo, ieed);
		ieed->e1 = &edges[me->iedges[i].e1];
		if (me->iedges[i].e2 != -1) ieed->e2 = &edges[me->iedges[i].e2];

	} 

	/* load half-edges. . .*/
	for (i=0; i<me->totedges; i++) {
		med = &me->half_edges[i];
		eed = &edges[i];
		eed->fast = 1;
		eed->v1 = &verts[med->v1];
		eed->v2 = &verts[med->v2];

		hash = MEM_callocN(sizeof(HE_HashEdge), "hashedge_he");
		hash->eed = eed;
		BLI_addtail(&eed->v1->hashedges, hash);
		/*the number one rule of hash edges: ONLY 1 PER VERT!*/
		hash = MEM_callocN(sizeof(HE_HashEdge), "HE_HashEdge");
		memcpy(hash, eed->v1->hashedges.last, sizeof(HE_HashEdge));

		BLI_addtail(&eed->v2->hashedges, hash);

  		eed->ieed = &iedges[med->iedge];

		eed->enext = &edges[med->next];
		eed->eprev = &edges[med->prev];
		if (med->pair != -1) {
	  		eed->pair = &edges[med->pair];
		} else eed->pair = NULL;
		eed->face = &faces[med->face];
  		BLI_addtail(&em->edges, eed);
	}

	for (eed=em->edges.first; eed; eed=eed->next) {
		if (eed->pair) printf("eed->pair->face->f: %d", eed->pair->face->f);
	}
	printf("\n");

	/* load faces. . .*/
	for (i=0; i<me->totfaces; i++) {
		efa = &faces[i];
		efa->f = 1;
		efa->f = me->faces[i].f;
		efa->edge = &edges[me->faces[i].edge];
		efa->fast = 1;
		VECCOPY(efa->nor, me->faces[i].nor);
		BLI_addtail(&em->faces, efa);
	}

 	/*redo pair pointers*/
 	HE_remake_pairs();
 
 	/*redo normals and face dots. . .*/
 	HE_recalcnormals();  
 
}
#endif

void HE_clear_flag_all(int flag)
{
	HE_EditMesh *em = G.he_editMesh;
	HE_EditVert *eve;
	HE_EditEdgeInfo *ieed;
	HE_EditFace *efa;

	for (eve=em->verts.first; eve; eve=eve->next) eve->f &= ~flag;
   	for (ieed=em->einfo.first; ieed; ieed=ieed->next) ieed->f &= ~flag;
	for (efa=em->faces.first; efa; efa=efa->next) efa->f &= ~flag;

} 

/********************** geometry stuff ********************/
/***verts***/
#if 0
static HE_EditVert *find_vert(float *co)
{
	HE_EditMesh *em = G.he_editMesh;
	HE_EditVert *eve;
	for (eve=em->verts.first; eve; eve=eve->next) {
		if (eve->co[0]==co[0] && eve->co[1]==co[1] && eve->co[2]==co[2]) {
			return eve;
		}
	}
	return NULL;
}
#endif

HE_EditVert *HE_addvertlist3f(float x, float y, float z)
{
	HE_EditVert *eve;// = find_vert(co);
	HE_EditMesh *em = G.he_editMesh;
	if (!em) return NULL;
	if (em->curbv < em->totbv) {
		eve = &em->blockv[em->curbv];
		eve->fast = 1;
		em->curbv++;
	} else eve = MEM_callocN(sizeof(HE_EditVert), "HE_EditVert");
	eve->co[0] = x;
	eve->co[1] = y;
	eve->co[2] = z;

	BLI_addtail(&em->verts, eve);
	return eve;
} 

HE_EditVert *HE_addvertlist(HE_EditMesh *em, float *co)
{
	HE_EditVert *eve;
	if (!em) return NULL;
	if (em->curbv < em->totbv) {
		eve = &em->blockv[em->curbv];
		eve->fast = 1;
		em->curbv++;
	} else eve = MEM_callocN(sizeof(HE_EditVert), "HE_EditVert");
	VECCOPY(eve->co, co);   
	BLI_addtail(&em->verts, eve); 
	return eve;
} 

void HE_delete_vert(HE_EditMesh *em, HE_EditVert *eve)
{
	BLI_freelistN(&eve->hashedges);
	BLI_remlink(&em->verts, eve);
	MEM_freeN(eve);
}

HE_EditEdgeInfo *find_iedge(HE_EditMesh *em, HE_EditVert *v1, HE_EditVert *v2)
{
	HE_HashEdge *hash;
	HE_EditVert *cv1, *cv2;

	if (!em) return NULL;

	for (hash=v1->hashedges.first; hash; hash=hash->next) {
		cv1 = hash->eed->v1;
		bw_assert(hash != NULL);
		bw_assert(hash->eed != NULL);
		bw_assert(hash->eed->freed==0);

		if (hash->eed->freed==1) printf("ref to free'd hedge!\n");
		if (!hash->eed) continue;

		if (hash->eed->enext) cv2 = hash->eed->enext->v1;
		else cv2 = hash->eed->v2;

		if (cv1==v1 && cv2==v2) {
			return hash->eed->ieed;
		}

		if (cv1==v2 && cv2==v1) {
			//printf("Found a one! %d ", (int) hash->eed->ieed);
			return hash->eed->ieed;
		}

	}

	for (hash=v2->hashedges.first; hash; hash=hash->next) {
		cv1 = hash->eed->v1;
		if (hash->eed->enext) cv2 = hash->eed->enext->v1;
		else cv2 = hash->eed->v2;

		if (cv1==v1 && cv2==v2) {
			//printf("Found one! %d\n", (int) hash->eed->ieed);
			return hash->eed->ieed;
		}
		if (cv1==v2 && cv2==v1) {
			//printf("Found one! %d\n", (int) hash->eed->ieed);
			return hash->eed->ieed;
		}   
	}  

	return NULL;
}

void HE_ManualEdgeAdd(HE_EditMesh *em, HE_EditEdge *eed)
{
	HE_HashEdge *hash;

	if (!em) return;

	hash = MEM_callocN(sizeof(HE_HashEdge), "HashEdge");
	hash->eed = eed;
	BLI_addtail(&eed->v1->hashedges, hash);
	BLI_addtail(&em->edges, eed);
}

/*this function doesn't perform info edge lookup*/
HE_EditEdge *HE_addedgelist_nosearch(HE_EditMesh *em, HE_EditVert *v1, HE_EditVert *v2)
{
	HE_EditEdge *eed;
	HE_EditEdgeInfo *ieed;

	if (em->curbhe < em->totbhe) {
		eed = &em->blockhe[em->curbv];
		eed->fast = 1;
		em->curbhe++;
	} else eed = MEM_callocN(sizeof(HE_EditEdge), "HE_EditEdge");

	eed->v1 = v1;
	eed->v2 = v2;
	eed->fast = 0;
	if (!v1) {printf("error making half-edge: no vert\n"); return NULL;}

	ieed = MEM_callocN(sizeof(HE_EditEdgeInfo), "HE_EditEdgeInfo");
	ieed->e1 = eed;
	ieed->e2 = NULL;
	eed->pair = NULL;
	BLI_addtail(&em->einfo, ieed); //on hold. . .add_iedge(ieed);

 	if (eed->pair==eed) printf("addedgelistBad!! eed == eed->pair!!!!\n");

 	if (!eed->v1->edge) eed->v1->edge = eed;
 	eed->ieed = ieed;
 
	HE_ManualEdgeAdd(em, eed);
	return eed;
}

HE_EditEdge *HE_addedgelist(HE_EditMesh *em, HE_EditVert *v1, HE_EditVert *v2)
{   
	HE_EditEdge *eed;
	HE_EditEdgeInfo *ieed = find_iedge(em, v1, v2);

	if (em->curbhe < em->totbhe) {
		eed = &em->blockhe[em->curbv];
		eed->fast = 1;
		em->curbhe++;
	} else eed = MEM_callocN(sizeof(HE_EditEdge), "HE_EditEdge");

	eed->v1 = v1;
	eed->v2 = v2;
	eed->fast = 0;
	if (!v1) {printf("error making half-edge: no vert\n"); return NULL;}

	if (!ieed) { /* create info edge*/
		//printf("creating info edge\n");
		ieed = MEM_callocN(sizeof(HE_EditEdgeInfo), "HE_EditEdgeInfo");
		ieed->e1 = eed;
		ieed->e2 = NULL;
		eed->pair = NULL;
		BLI_addtail(&em->einfo, ieed);
 	} else {
 		if (ieed->e2) {
 			/*uh-oh, a non-manfold mesh!*/
 			HE_RaiseError(HE_NON_MANIFOLD);
 		} else {
	 		ieed->e2 = eed;
	 		ieed->e1->pair = eed;
	 		eed->pair = ieed->e1;
	 	}
 	}   

 	if (eed->pair==eed) printf("addedgelistBad!! eed == eed->pair!!!!\n");

 	if (!eed->v1->edge) eed->v1->edge = eed;
 	eed->ieed = ieed;
 
 	HE_ManualEdgeAdd(em, eed); //BLI_addtail(&em->edges, eed);
	return eed;
}

void he_delete_iedge(HE_EditMesh *em, HE_EditEdgeInfo *ieed)
{
	if (!em) return;
	BLI_remlink(&em->einfo, ieed);
	if (ieed->fast==0) MEM_freeN(ieed);
}


void HE_delete_halfedge(HE_EditMesh *em, HE_EditEdge *eed)
{
	HE_HashEdge *hash, *hnext;
	if (!em) return;

	/*assert time!*/
	bw_assert(eed->face);
	bw_assert(eed->ieed);
	bw_assert(eed->ieed->e1);
	bw_assert(eed->face->edge);
	bw_assert(eed->freed==0);

	if (eed->ieed->e1==eed->ieed->e2) printf("Bad, duplicate ieed!!!\n");
	/*update info edge. . .*/
	if (eed->ieed->e1==eed) {
		eed->ieed->e1 = eed->ieed->e2;
		eed->ieed->e2 = NULL;
	} else if (eed->ieed->e2==eed) eed->ieed->e2 = NULL;
	else printf("Error deleting half-edge!!\n");

	if (eed->ieed->e1==NULL) {
		//printf("Removing info edge.\n");
		he_delete_iedge(em, eed->ieed);
	}
	if (eed->pair) eed->pair->pair = NULL;
	//BAD! BAD! if (eed->face->edge==eed) eed->face->edge = eed->enext;
	if (eed->v1->edge==eed) {
		if (eed->pair && eed->pair->enext) eed->v1->edge = eed->pair->enext;
		else {
		    /*abuse hash edges to find edge. . .*/
		    HE_HashEdge *hash;
		    for (hash=eed->v1->hashedges.first; hash; hash=hash->next) {
		        if (hash->eed != eed && (hash->eed->v1==eed->v1)) {
		            eed->v1->edge = hash->eed;
		            break;
		        }    
		    }    
		}	
	}

	for (hash=eed->v1->hashedges.first; hash; hash=hnext) {
		hnext = hash->next;
		if (hash->eed == eed) {
			BLI_remlink(&eed->v1->hashedges, hash);
			if (hash->fast==0) MEM_freeN(hash);
		}
	}

	for (hash=eed->v2->hashedges.first; hash; hash=hnext) {
		hnext = hash->next;
		if (hash->eed == eed) {
			BLI_remlink(&eed->v2->hashedges, hash);
			if (hash->fast==0) MEM_freeN(hash);
		}
	}

	BLI_remlink(&em->edges, eed);
	if (eed->fast==0) MEM_freeN(eed);
}

/********  Faces *********/
#if 0
HE_EditFace *HExcvfds_addfacelist(HE_EditVert **vert_list, int len)
{
	HE_EditMesh *em = G.he_editMesh;
	HE_EditEdge *eed=NULL, *last_eed=NULL;
   	HE_EditFace *efa = MEM_callocN(sizeof(HE_EditFace), "HE_EditFace");
	float aver[3] = {0.0f, 0.0f, 0.0f};
	int aver_len=0, switch_loop = 0;
	int i;

	if (!em) return NULL;

	bw_assert(len>=3);
	if (len < 3) return NULL; /*bad face*/

	for (i=0; i<len; i++) {
		if (i != len-1) eed = HE_addedgelist(em, vert_list[i], vert_list[i+1]);
		else eed = HE_addedgelist(em, vert_list[i], vert_list[0]);
		eed->face = efa;

		if (i==0) efa->edge = eed;

		/*test to see if we need to switch half-edge order
	 	  in order to maintain half-edge integrety*/
	   	if (eed->pair != NULL && eed->v1 == eed->pair->v1) switch_loop = 1;

		if (last_eed) {
			eed->eprev = last_eed;
			last_eed->enext = eed;
		}

		last_eed = eed;
		VecAddf(aver, aver, vert_list[i]->co);
		aver_len++;
	}
	/*complete the loop. . .*/
	eed->enext = efa->edge;
	efa->edge->eprev = eed;

	/* switch edge loop if needed to maintain proper half-edge 
	 * data structure*/
 	if (switch_loop) switch_face(efa);

	/*computer face dot. . .*/
	VecMulf(aver, 1.0f/((float)aver_len));
	VECCOPY(efa->cent, aver);

	/*compute normal (bad method). . .*/
	CalcNormFloat(vert_list[0]->co, vert_list[1]->co, vert_list[2]->co, efa->nor);
	Normalise(efa->nor);

	/*and then add to mesh face list*/
	BLI_addtail(&em->faces, efa);
	return efa;
}
#endif

HE_EditFace *HE_addfacelist(HE_EditMesh *em, HE_EditVert **vert_list, int len)
{
	int i;
	HE_EditEdge *eed=NULL, *last_eed=NULL;
	HE_EditFace *efa = MEM_callocN(sizeof(HE_EditFace), "HE_EditFace");
	HE_EditVert *v1, *v2;
	float aver[3] = {0.0f, 0.0f, 0.0f};
	int aver_len=0;
	int switch_loop = 0;

	if (len < 3) {
		printf("internal error: Bad face (for ngon mesh)!\n");
		return NULL;
	}
	for (i=0; i<len; i++) {
		v1 = vert_list[i];

		if (i<len-1) v2 = vert_list[i+1];
		else v2 = vert_list[0];

		eed = HE_addedgelist(em, v1, v2);
		eed->face = efa;
		if (eed->pair != NULL && eed->v1 == eed->pair->v1) switch_loop = 1;

		if (last_eed) {
			last_eed->enext = eed;
			eed->eprev = last_eed;
		} else efa->edge = eed;   

		last_eed = eed;

		/*add to the average for facedot. . .*/
		VecAddf(aver, aver, v1->co);
		aver_len++;
	}   
	/* computer average for face dot. . . */
	VecMulf(aver, 1.0f/(float)aver_len);
	VECCOPY(efa->cent, aver);

	/*normal. . .*/
	CalcNormFloat(vert_list[0]->co, vert_list[1]->co, efa->cent, efa->nor);
	Normalise(efa->nor);

	/* complete the edge loop. . .*/
	eed->enext = efa->edge; 
	efa->edge->eprev = eed;

	/* switch edge loop if needed to maintain proper half-edge 
	 * data structure*/
	if (switch_loop) switch_face(efa);
	efa->switched = 0;

	BLI_addtail(&em->faces, efa);
	return efa;
}

void HE_delete_face(HE_EditMesh *em, HE_EditFace *efa)
{
	HE_EditEdge *eed, *next;
	int i=0;

	if (em->f_mover==efa) em->f_mover = NULL;

	eed = efa->edge;
	do {
		bw_assert((eed != eed->eprev) && (eed != eed->enext));
		next = eed->enext;
		HE_delete_halfedge(em, eed);
		eed=next;
		i++;
	} while (eed != efa->edge);

	BLI_remlink(&em->faces, efa);
	if (efa->fast==0) MEM_freeN(efa);
}

/******************* Selection Stuff *********************/

void HE_faceselect_flush(void)
{
	HE_EditMesh *em = G.he_editMesh;
	HE_EditVert *eve;
	HE_EditEdgeInfo *ieed;
	HE_EditEdge *eed;
	HE_EditFace *efa;

	for (eve=em->verts.first; eve; eve=eve->next) {
		/*eve->f1 is now number of users*/
		if (eve->f & SELECT) eve->f1 = 1;
		else eve->f1 = 0;
	} 

	for (efa=em->faces.first; efa; efa=efa->next) {
		eed = efa->edge;
		if (efa->f & SELECT) {
			do {
				eed->ieed->f |= SELECT;
				eed->v1->f1++;
				eed->enext->f1++;
				eed=eed->enext;
			} while (eed != efa->edge);
		}
	}

	for (ieed=em->einfo.first; ieed; ieed=ieed->next) {
		if (ieed->f & SELECT) {
			if (ieed->e1) {
				ieed->e1->v1->f1++;
				ieed->e1->enext->f1++;
			}
		}
 	}   
	for (eve=em->verts.first; eve; eve=eve->next) {
		/* if we have no users: deselect vert*/
		if (eve->f1 == 0) {
			eve->f &= ~SELECT;
		} else {
			eve->f |= SELECT;
		 	eve->f1 = 0;   
		}
	}	 
}

void HE_edgeselect_flush(void)
{
	HE_EditMesh *em = G.he_editMesh;
	HE_EditVert *eve;
	HE_EditEdgeInfo *ieed;

	for (eve=em->verts.first; eve; eve=eve->next) {
		if (eve->f & SELECT) eve->f1 = 1;
		else eve->f1 = 0;
	} 

	for (ieed=em->einfo.first; ieed; ieed=ieed->next) {
		if (!ieed->e1) continue;

		if (ieed->f & SELECT) {
			ieed->e1->v1->f1++;
			ieed->e1->enext->v1->f1++;


		}
 	}   
	for (eve=em->verts.first; eve; eve=eve->next) {
		/* if we have no users: deselect vert*/
		if (eve->f1 == 0) {
			eve->f &= ~SELECT;
		} else {
			eve->f |= SELECT;
		 	eve->f1 = 0;   
		}
	}	 

}

void HE_update_edges(HE_EditMesh *em)
{
	HE_EditEdgeInfo *ieed;

	for (ieed=em->einfo.first; ieed; ieed=ieed->next) {
		if (ieed->e1) {
			if ((ieed->e1->v1->f & SELECT)&&(ieed->e1->enext->v1->f & SELECT)) {
				ieed->f |= SELECT;
			} else if (!(ieed->e1->v1->f & SELECT) || !(ieed->e1->enext->v1->f & SELECT)) {
				ieed->f &= ~SELECT;
			}
		}   
 	}	  
}

/*this functions "sets" the select mode*/
void HE_selectmode_set(void)
{
	HE_EditMesh *em = G.he_editMesh;
	if (!G.he_editMesh) return;
	/*clear all mouse highlights first. . .*/
	he_clear_all_mover();

	HE_update_edges(em);
	HE_update_faces(em);
	HE_edgeselect_flush();
}

void HE_update_faces(HE_EditMesh *em)
{
	HE_EditEdge *eed;
	HE_EditFace *efa;
	int ok;

	for (efa=em->faces.first; efa; efa=efa->next) {
		ok = 1;
		eed = efa->edge;
  		do {
  			if ((eed->v1->f & SELECT)==0) {
  				ok = 0;
  				break;
  			}
  			/*we only call HE_select_face if we're in vertex selmode. . .*/
  			if (!(G.scene->selectmode & SCE_SELECT_VERTEX)) {
  				if (!(eed->ieed->f & SELECT)) {
  					ok = 0;
  					break;
  				}
  			}
			eed=eed->enext;
	 	} while (eed != efa->edge);
	 	if (ok) {
	 		HE_select_face(efa);
	 	} else {
			efa->f &= ~SELECT;
	  	}	   
 	}   

}


void HE_select_face(HE_EditFace *efa)
{
	HE_EditEdge *eed;

	eed = efa->edge;
	do {
		eed->ieed->f |= SELECT;
		eed->v1->f |= SELECT;
		eed = eed->enext;
	} while (eed != efa->edge);
	efa->f |= SELECT;
}

void HE_deselect_face(HE_EditFace *efa)
{
	HE_EditEdge *eed;
	efa->f &= ~SELECT;
	eed = efa->edge;
	do {
		if ((eed->pair && !(eed->pair->face->f & SELECT)) || !eed->pair) {
			eed->ieed->f &= ~SELECT;
			eed->v1->f &= ~SELECT;
		}
		eed=eed->enext;
	} while (eed != efa->edge);
}

void HE_clear_flag_faces(HE_EditMesh *em, int f)
{
	HE_EditFace *efa;

	for (efa=em->faces.first; efa; efa=efa->next) {
		efa->f &= ~f;
	}
}

void HE_clear_flag_edges(HE_EditMesh *em, int f)
{
	HE_EditEdgeInfo *ieed;

	for (ieed=em->einfo.first; ieed; ieed=ieed->next) {
		ieed->f &= ~f;
	}
}


void HE_clear_flag_verts(HE_EditMesh *em, int f)
{
	HE_EditVert *eve;

	for (eve=em->verts.first; eve; eve=eve->next) {
		eve->f &= ~f;
	}
}

void he_deselect_all(void)
{
	HE_EditMesh *em = G.he_editMesh;
	HE_EditVert *eve;
	HE_EditEdgeInfo *ieed;
	HE_EditFace *efa;

	for (eve=em->verts.first; eve; eve=eve->next) eve->f &= ~SELECT;
	for (ieed=em->einfo.first; ieed; ieed=ieed->next) ieed->f &= ~SELECT;
	for (efa=em->faces.first; efa; efa=efa->next) efa->f &= ~SELECT;
}

/*dummy func, will eventually reside in own file*/
void hemesh_evt(unsigned int evt, short val)
{
	if (evt==LEFTMOUSE && val==0) mouse_hemesh();
}

void mouse_hemesh_loop(void)
{
	HE_EditMesh *em;
	HE_EditFace **flist;
	HE_EditEdgeInfo *ieed, **ielist;
	int selstate;
	int i=0;
	int do_undo = 0;

	em = G.he_editMesh;
	if (!em) return;

	if (!(G.qual & LR_SHIFTKEY)) he_deselect_all();
	if (G.scene->selectmode & SCE_SELECT_FACE) {
		ieed = HE_findnearestedge(em, 75);
		if (ieed) {
			selstate = ieed->f & SELECT;
			flist = HE_GetFaceLoop(ieed);
			if (!flist) return;
			i = 0;
			while (flist[i]) {
				do_undo = 1;
				if (selstate)
					HE_deselect_face(flist[i]);
		  		else HE_select_face(flist[i]);
				i++;
			}
			MEM_freeN(flist);
			if (G.scene->selectmode != SCE_SELECT_FACE) HE_update_faces(em);
			HE_edgeselect_flush();
		}
	} else { /*do edge loop*/
		ieed = HE_findnearestedge(em, 75);
		if (ieed) {
			selstate = ieed->f & SELECT;
			ielist = HE_GetEdgeLoop(ieed);
			if (!ielist) return;
			i = 0;
			while (ielist[i]) {
				do_undo = 1;
				if (!selstate) {
					ielist[i]->f |= SELECT;
					ielist[i]->e1->v1->f |= SELECT;
					if (ielist[i]->e2) ielist[i]->e2->v1->f |= SELECT;
				} else {
					ielist[i]->f &= ~SELECT;
					ielist[i]->e1->v1->f &= ~SELECT;
					if (ielist[i]->e2) ielist[i]->e2->v1->f &= ~SELECT;
				}
				i++;
			}
			MEM_freeN(ielist);
		}
		HE_update_faces(em);
		HE_edgeselect_flush();
	}	  
	allqueue(REDRAWVIEW3D, 1);
	if (do_undo) BIF_undo_push("Loop Sel");
}

void HE_select_iedge(HE_EditEdgeInfo *ieed)
{
	ieed->f |= SELECT;
	ieed->e1->v1->f |= SELECT;
	ieed->e1->enext->v1->f |= SELECT;
} 
void HE_deselect_iedge(HE_EditEdgeInfo *ieed)
{
	ieed->f &= ~SELECT;
	ieed->e1->v1->f &= ~SELECT;
	ieed->e1->enext->v1->f &= ~SELECT;
}

/*int is type, returns HEVERT, HEEDGE, or HEFACE*/
int he_unified_findnearest(HE_EditMesh *em, void **ptr, int pdis)
{
	HE_EditVert *eve = NULL;
	HE_EditEdgeInfo *ieed = NULL;
	HE_EditFace *efa=NULL;
	float dv=-1, de=-1, df=-1, x, y;
	short mval[2] = {0, 0};

	getmouseco_areawin(mval);

	calc_hemesh_facedots_ext();
	calc_hemesh_vertdots_ext();

 	if (G.scene->selectmode & SCE_SELECT_VERTEX) {
		eve = HE_findnearestvert(em, pdis);
		if (eve) {
			x = eve->xs[0]-mval[0];
			y = eve->xs[1]-mval[1];
			dv = sqrt(x*x+y*y);
		}
	}

	if (G.scene->selectmode & SCE_SELECT_EDGE) {
		ieed = HE_findnearestedge(em, pdis);
		if (ieed) de = he_dist_mval_edge(mval, ieed->e1);
	}

	if (G.scene->selectmode & SCE_SELECT_FACE) {
 		efa = HE_findnearestface(em, pdis);
		if (efa) {
			x = efa->xs[0]-mval[0];
			y = efa->xs[1]-mval[1];
			df = sqrt(x*x+y*y);
		}
 	}

	/*we try all cases, because there are
	  only 3 elements to test*/
	if (eve && efa && ieed) {
		//printf("eve efa and ieed!\n");
		if (dv <= df <= de) {
			*ptr = eve;
			return HEVERT;
		} else if ((de <= dv <= df) && (dv < 10)) {
			*ptr = eve;
			return HEVERT;
		} else if (de <= df) {
			*ptr = ieed;
			return HEIEDGE;
		} else {
			*ptr = efa;
			return HEFACE;
		}
	} else if (eve && efa) {
		// printf("eve and efa!\n");
		if (dv <= df) {
			*ptr = eve;
			return HEVERT;
		} else {
			*ptr = efa;
			return HEFACE;
		}
	} else if (eve && ieed) {
		//printf("eve and ieed!!!\n");
		if (dv <= de || (dv < 10)) {
			*ptr = eve;
			return HEVERT;
		} else {
			*ptr = ieed;
			return HEIEDGE;
		}
 	} else if (ieed && efa) {
 		//printf("ieed and efa!\n");
		// printf("de: %f, df: %f\n", de, df);
		if (de <= df) {
			*ptr = ieed;
			return HEIEDGE;
		} else {
			*ptr = efa;
			return HEFACE;
		} 
	} else {
		if (eve) {
			*ptr = eve;
			return HEVERT;
		} else if (ieed) {
		 	*ptr = ieed;
		  	return HEIEDGE;
		} else if (efa) {
	  		*ptr = efa;
			return HEFACE;
		}
	}

	*ptr = NULL;
	return HENONE;
}

void mouse_hemesh_select(void)
{
	HE_EditMesh *em = G.he_editMesh;
	HE_EditFace *efa;
	HE_EditEdgeInfo *ieed;
	HE_EditVert *eve;
	void *ptr = NULL;
	int type, pdis;

	if ((G.scene->selectmode & SCE_SELECT_VERTEX) && G.scene->selectmode > SCE_SELECT_VERTEX) {
		pdis = FDIS;
 	}	else pdis = FDIS;
 
	type = he_unified_findnearest(em, &ptr, pdis);

	if (type==HEVERT) {
		eve = (HE_EditVert*) ptr;
		if (G.qual & LR_SHIFTKEY) {
			if (eve->f & SELECT) eve->f &= ~SELECT;
			else eve->f |= SELECT;
		} else {
			HE_clear_flag_all(SELECT);
			eve->f |= SELECT;
		} 
		BIF_undo_push("Select Vert");
	} else if (type==HEIEDGE) {
		ieed = (HE_EditEdgeInfo*) ptr;
		if (G.qual & LR_SHIFTKEY) {
			if (ieed->f & SELECT) ieed->f &= ~SELECT;
			else ieed->f |= SELECT;
			for (eve=em->verts.first; eve; eve=eve->next) eve->f1 = 0;
			for (ieed=em->einfo.first; ieed; ieed=ieed->next) {
				if (ieed->f & SELECT) {
					ieed->e1->v1->f1++;
					ieed->e1->v2->f1++;
				}
			}
			for (eve=em->verts.first; eve; eve=eve->next) {
 				if (eve->f1==0) eve->f &= ~SELECT;
 				else eve->f |= SELECT;
 			}
		} else {
			HE_clear_flag_all(SELECT);
			ieed->f |= SELECT;
			ieed->e1->v1->f |= SELECT;
			ieed->e1->v2->f |= SELECT;
		}
		BIF_undo_push("Select Edge");
	} else if (type==HEFACE) {
		efa = (HE_EditFace*) ptr;
		if (G.qual & LR_SHIFTKEY) {
			if (efa->f & SELECT) HE_deselect_face(efa);
			else HE_select_face(efa);
		} else {
			HE_clear_flag_all(SELECT);
			HE_select_face(efa);
		}
		BIF_undo_push("Select Face");
	}

	if (G.scene->selectmode & SCE_SELECT_VERTEX) {
		HE_update_edges(em);
		HE_update_faces(em);
	}
	if (G.scene->selectmode & SCE_SELECT_EDGE) HE_update_faces(em);

	if (type != HENONE) {
		allqueue(REDRAWVIEW3D, 0);
		rightmouse_transform();
	}
}

#if 0
void mouse_hemesh_selectfdsfds(void)
{
	HE_EditMesh *em = G.he_editMesh;
	HE_EditVert *eve = NULL;
	HE_EditEdgeInfo *ieed = NULL, *iedge = NULL;
	HE_EditFace *efa=NULL, *face = NULL;

	int pdis;
	if ((G.scene->selectmode & SCE_SELECT_VERTEX) && G.scene->selectmode > SCE_SELECT_VERTEX) {
		pdis = -1;
 	}	else pdis = FDIS;
 
	/*if(G.scene->selectmode & SCE_SELECT_VERTEX) {
	  if (pdis == -1)
	  eve = HE_findnearestvert(FVDIS);
	  else eve = HE_findnearestvert(pdis);
	  if (eve) {
	  if (G.qual & LR_SHIFTKEY) {
	  if (eve->f & SELECT) eve->f &= ~SELECT;
	  else eve->f |= SELECT;
	  } else {
	  HE_clear_flag_verts(SELECT);
	  eve->f |= SELECT;
	  }
	  HE_update_edges(em);
	  HE_update_faces(em);
	  allqueue(REDRAWVIEW3D, 1);
	  BIF_undo_push("Select Vert");
	  rightmouse_transform();
	  return;
	  }
	  }*/
	if (G.scene->selectmode & SCE_SELECT_VERTEX) {
		if (HET_FindAndSelectVert(75)) {
			allqueue(REDRAWVIEW3D, 1);
			BIF_undo_push("Select Vert");
			rightmouse_transform();
			return;
		}
	}

	if(G.scene->selectmode & SCE_SELECT_EDGE) {
		if (pdis == -1)
			ieed = HE_findnearestedge(em, FEDIS);
		else ieed = HE_findnearestedge(em, pdis);
		if (ieed) {
			if (G.qual & LR_SHIFTKEY) {
				if (ieed->f & SELECT) HE_deselect_iedge(ieed);
				else HE_select_iedge(ieed);
		 	} else {
		 		for (iedge=em->einfo.first; iedge; iedge=iedge->next) {
		 			if (iedge->f & SELECT) HE_deselect_iedge(iedge);
		 		}
				HE_select_iedge(ieed);
		 	}
			/*this put here on purpose, even though it should be in it's
		 	  own function (it may already, in fact).*/
	 	  	/*deselect verts based on edge users*/
			for (eve=em->verts.first; eve; eve=eve->next) eve->f1 = 0;
			for (ieed=em->einfo.first; ieed; ieed=ieed->next) {
				ieed->e1->v1->f1++;
				ieed->e1->v2->f1++;
			}
			for (eve=em->verts.first; eve; eve=eve->next) {
 				if (eve->f1==0) eve->f &= SELECT;
 			}
   			HE_update_faces(em);
   
			allqueue(REDRAWVIEW3D, 1);
			BIF_undo_push("Select Edge");
			rightmouse_transform();
			return;
		}
	} 

 	if(G.scene->selectmode & SCE_SELECT_FACE) {
 		if (pdis == -1)
	 		efa = HE_findnearestface(em, FFDIS);
		else efa = HE_findnearestface(em, pdis);
		if (efa) {
			if (G.qual & LR_SHIFTKEY) {
				if (efa->f & SELECT) HE_deselect_face(efa);
				else HE_select_face(efa);
			} else {
	 			for (face=em->faces.first; face; face=face->next) {
	 				HE_deselect_face(face);
				}
				HE_select_face(efa);
			}  
	 		allqueue(REDRAWVIEW3D, 0);
	 		HE_faceselect_flush();
	 		if (G.scene->selectmode != SCE_SELECT_FACE) HE_update_faces(em);
	 		BIF_undo_push("Select Face");
	 		rightmouse_transform();
		}
	}
}
#endif

void mouse_hemesh(void)
{
	if (G.qual & LR_ALTKEY) mouse_hemesh_loop();
	else mouse_hemesh_select();
}

void HE_recalcnormals(void)
{
	HE_EditMesh *em = G.he_editMesh;
	HE_EditFace *efa, **flist;
	HE_EditEdge *eed;
	HE_EditVert *eve;
	float cent[3] = {0, 0, 0};
	int i;
	float len;
	float nor[3];
	//check_faces();

	for (efa=em->faces.first; efa; efa=efa->next) {
		float new_nor[3];

		len = 0;
		cent[0] = cent[1] = cent[2] = 0;
		eed = efa->edge;
		do {
			VecAddf(cent, cent, eed->v1->co);
			len += 1.0f;
			eed = eed->enext;
		} while (eed != efa->edge);
		VecMulf(cent, 1.0f/len);
		VECCOPY(efa->cent, cent);

		if (len > 3) {
			CalcNormFloat4(efa->edge->v1->co, 
						   efa->edge->enext->v1->co, 
						   efa->edge->enext->enext->v1->co, 
						   cent,
						   new_nor);
 		} else {
			CalcNormFloat(efa->edge->v1->co, 
						  efa->edge->enext->v1->co, 
						  cent, 
						  new_nor);
		} 
  		VECCOPY(efa->nor, new_nor);
	} 
	for (eve=em->verts.first; eve; eve=eve->next) {
 		eve->f1 = 0;
 		eve->no[0] = eve->no[1] = eve->no[2] = 0;
	}

	for (efa=em->faces.first; efa; efa=efa->next) {
		eed = efa->edge;
		do {
			VecAddf(eed->v1->no, eed->v1->no, efa->nor);
			eed->v1->f1++;
			eed=eed->enext;
		} while (eed != efa->edge);
	}

	for (eve=em->verts.first; eve; eve=eve->next) {
		if (eve->f1) {
			VecMulf(eve->no, 1.0f/((float)eve->f1));
		} else {
			eve->no[0] = 0; eve->no[1] = 0; eve->no[2] = 1;
		}
	}

	#if 0
	for (eve=em->verts.first; eve; eve=eve->next) {
		i = 0;
		nor[0] = nor[1] = nor[2] = 1;
		flist = he_getFacesAroundVert(em, eve);
		if (!flist) continue;
		while (flist[i] != NULL) {
			VecAddf(nor, nor, flist[i]->nor);
			i++;
		}
		VecMulf(nor, 1.0f/((float)i));
		VECCOPY(eve->no, nor);
		MEM_freeN(flist);
	} 
	//check_faces();

	return;   
	/*calc vert normals. . .*/
	for (eed=em->edges.first; eed; eed=eed->next) {
		eed->f1 = 0;
		eed->no[0] = eed->no[1] = eed->no[2] = 0;
	}

	/*attempt at calc'ing half-edge normals. . .
	  for support of "hard" edges a la Wings3D.*/
	/*UPDATE: we now store shading normals in general in half-edges*/
	for (efa=em->faces.first; efa; efa=efa->next) {
		eed = efa->edge;
		do {
			if ((eed->ieed->f & HEM_HARD) || (eed->eprev->ieed->f & HEM_HARD)) {
				VecAddf(eed->no, eed->no, efa->nor);
				eed->f1++;
				if (eed->pair) {
					VecAddf(eed->no, eed->no, eed->pair->face->nor);
					eed->f1 += 1;
				}
				//VecSubf(eed->enext->no, eed->enext->no, efa->nor);
				//eed->enext->f1--;
		 	} else {
				flist = he_getFacesAroundVert(em, eed->v1);
				if (flist) {
					i = 0;
					while (flist[i]) {
						//if (eed->pair) {
						//	if (!(eed->ieed->f & HEM_HARD) || (flist[i] != eed->pair->face)) {
						//	   VecAddf(eed->no, eed->no, flist[i]->nor);
						//	   eed->f1++;
						//   }
						//} else {
						VecAddf(eed->no, eed->no, flist[i]->nor);
						eed->f1++;
						//}
						i++; 
					}  
					MEM_freeN(flist); 
				}
			}
			if (eed->f1) {
				VecMulf(eed->no, 1.0f/((float)eed->f1));
			}
			eed=eed->enext;
		} while (eed != efa->edge); 
	}


 	
 	#endif
}
/********** Mesh Adding Stuff *******/

void addPrimitiveHEMesh(int type)
{
	Object *ob;
	HE_Mesh *me;
	HE_EditMesh *em;
	HE_EditVert *v1, *v2, *v3, *v4;
	HE_EditVert *tv1, *tv2, *tv3, *tv4;
	HE_EditVert *vlist[5];

 	add_object_draw(OB_HEMESH);
 
	ob= BASACT->object;
	where_is_object(ob);

	me = (HE_Mesh*) add_hemesh();
	ob->type = OB_HEMESH;
	ob->data = me;

	enter_editmode();

	em = G.he_editMesh;
	/* It's cube time! */

	/* face 1*/
	v1 = HE_addvertlist3f(1, 0, 0);
	v2 = HE_addvertlist3f(1, 1, 0);
	v3 = HE_addvertlist3f(-1, 0, 0);
	v4 = HE_addvertlist3f(-1, -1, 0);

	vlist[0] = v1; vlist[1] = v2;
	vlist[2] = v3; vlist[3] = v4;

	HE_addfacelist(em, vlist, 4);

	//return;
	/* face 2*/
	tv1 = HE_addvertlist3f(1, 0, 1);
	tv2 = HE_addvertlist3f(1, 1, 1);
	tv3 = HE_addvertlist3f(-1, 0, 1);
	tv4 = HE_addvertlist3f(-1, -1, 1);

	vlist[0] = tv1; vlist[1] = tv2;
	vlist[2] = tv3; vlist[3] = tv4;
	HE_addfacelist(em, vlist, 4);

  	vlist[0] = tv1; vlist[1] = v1;
	vlist[2] = v2; vlist[3] = tv2;
	HE_addfacelist(em, vlist, 4);

	vlist[0] = tv2; vlist[1] = v2;
	vlist[2] = v3; vlist[3] = tv3;
	HE_addfacelist(em, vlist, 4);

	vlist[0] = tv3; vlist[1] = v3;
	vlist[2] = v4; vlist[3] = tv4;
	HE_addfacelist(em, vlist, 4);

	vlist[0] = tv4; vlist[1] = v4;
	vlist[2] = v1; vlist[3] = tv1;
	HE_addfacelist(em, vlist, 4);

	check_faces(em);
	HE_recalcnormals();

}


