#include <stdio.h>
#include "he_error.h" 
#include "MEM_guardedalloc.h" 
#include "BLI_blenlib.h"
#include "DNA_listBase.h"
#include "BIF_space.h"

/*simple (and bad) error handling system for EditTools*/
typedef struct error_str {
	struct error_str *next, *prev;
	char *str;
} error_str;

static ListBase error_stack = {NULL, NULL};
static ListBase string_stack = {NULL, NULL};

int HE_CheckError(void)
{
	if (error_stack.first) return 1;
	else return 0;
}

static char *he_error_string(void)
{
	error_str *err = string_stack.first; 
	char *str;
	if (err) {
		/*pop the stack*/
		BLI_remlink(&string_stack, err);
		str = err->str;
		MEM_freeN(err);
		return str;
	} else return "Error string not set!";
}

char *HE_ErrorCodeToString(int code)
{
	switch (code) {
		case HE_CANCELED: return "Tool Canceled";
		case HE_BAD_ARGUMENT_STRING: return "Bad argument string";
		case HE_UNDEFINED_ERROR: return "Undefined error";
		case HE_NON_MANIFOLD: return "Non-manifold mesh error";
		case HE_NO_MESHDATA: return "No editmesh!";
		case HE_NOTOOL: return "Missing tool!";
		case HE_BADINDEX: return "Bad index!";
		case HE_INVALID_SELECTION: return "Invalid selection!";
		case HE_SEL_MUST_BE_SAME: return "Selections must be the same!";
		case HE_CUSTOM_STRING: return he_error_string();
	}
	return "Invalid error code";
}

int HE_CatchError(int errorcode)
{
	he_toolerror_entry *entry = error_stack.first;
	/*return first error on 0*/
	if (errorcode==0 && entry != NULL) {
		BLI_remlink(&error_stack, entry);
		return entry->errorcode;
	}

	if (entry && entry->errorcode==errorcode) {
		BLI_remlink(&error_stack, entry);
		return 1;
	} else return 0;
}

void HE_PrintErrorStack(void)
{
	//he_toolerror_entry *entry;
	//printf("Tool Error Traceback:\n");
	//for (entry = error_stack.last; entry; entry=entry->link.prev) {
	//printf("--Tool %s raised error code %d.", entry->tool->name, entry->errorcode);
	//}
}

void HE_RaiseError(int errorcode) 
{
	he_toolerror_entry *entry = MEM_callocN(sizeof(he_toolerror_entry), "he_toolerror_entry *entry");
	//entry->tool = tool;
	//entry->arg_string = arg_string;
	entry->errorcode = errorcode;

	BLI_addhead(&error_stack, entry);
}

/*str is NOT mem_freeN'd!*/
void HE_RaiseErrorString(char *str)
{
	error_str *err = MEM_callocN(sizeof(error_str), "error_str");
	HE_RaiseError(HE_CUSTOM_STRING);
	err->str = str;
	BLI_addhead(&string_stack, err);
}
