#include "BIF_meshHE.h"

#include "MEM_guardedalloc.h" 

#include "BMF_Api.h"

#include "IMB_imbuf.h"

#include "MTC_matrixops.h"

#include "DNA_camera_types.h"
#include "DNA_curve_types.h"
#include "DNA_constraint_types.h" // for drawing constraint
#include "DNA_effect_types.h"
#include "DNA_ipo_types.h"
#include "DNA_lamp_types.h"
#include "DNA_lattice_types.h"
#include "DNA_material_types.h"
#include "DNA_mesh_types.h"
#include "DNA_meshdata_types.h"
#include "DNA_meta_types.h"
#include "DNA_object_types.h"
#include "DNA_space_types.h"
#include "DNA_scene_types.h"
#include "DNA_screen_types.h"
#include "DNA_view3d_types.h"
#include "DNA_world_types.h"

#include "BLI_blenlib.h"
#include "BLI_arithb.h"
#include "BLI_editVert.h"

#include "BKE_utildefines.h"
#include "BKE_curve.h"
#include "BKE_constraint.h" // for the get_constraint_target function
#include "BKE_deform.h"		// lattice_modifier()
#include "BKE_displist.h"
#include "BKE_effect.h"
#include "BKE_global.h"
#include "BKE_ipo.h"
#include "BKE_lattice.h"
#include "BKE_mesh.h"
#include "BKE_material.h"
#include "BKE_mball.h"
#include "BKE_object.h"

#include "BIF_gl.h"
#include "BIF_mywindow.h"
#include "BIF_screen.h"
#include "BIF_space.h"
#include "BIF_editarmature.h"
#include "BIF_editmesh.h"
#include "BIF_glutil.h"
#include "BIF_resources.h"

#include "BDR_drawmesh.h"
#include "BDR_drawobject.h"
#include "BDR_editobject.h"

#include "BSE_view.h"
#include "BSE_drawview.h"
#include "BSE_trans_types.h"

#include "blendef.h"
#include "mydevice.h"
#include "nla.h"

#include "BKE_deform.h"
#include "BKE_hemesh.h"

#include "he_editmesh.h"
#include "BLI_parray.h"

static int he_set_connected(HE_EditFace *efa)
{
	HE_EditEdge *eed;
	efa->f1 = 1;
	eed = efa->edge;
	do {
		if (eed->pair && eed->pair->face->f1==0) he_set_connected(eed->pair->face);
		eed=eed->enext;
	} while (eed != efa->edge);
	return 0;
}

int HE_getComponents(void)
{
	HE_EditMesh *em = G.he_editMesh;
	HE_EditFace *efa;
	int tot=0;

	if (!em) return 0;
	for (efa=em->faces.first; efa; efa=efa->next) efa->f1 = 0;

	for (efa=em->faces.first; efa; efa=efa->next) {
		if (efa->f1==0) {
			he_set_connected(efa);
			tot++;
		}
	}

	return tot;
}

static void he_tag_hole_edges(HE_EditEdge *pass)
{
	HE_EditEdge **elist;
	int i;

	if (pass->f1) return;

	pass->f1 = 1;
	pass = pass->enext;
	elist = he_getHEdgesAroundVert(G.he_editMesh, pass->v1);
	if (elist) {
		i = 0;
		while (elist[i]) {
			if (elist[i]->pair==NULL) {
				he_tag_hole_edges(elist[i]);
			}
			i++;
		} 
		MEM_freeN(elist);
	}
}

int HE_getHoles(void)
{
	HE_EditMesh *em = G.he_editMesh;
	HE_EditEdge *eed;
	int tot=0;

	if (!em) return 0;
	for (eed=em->edges.first; eed; eed=eed->next) eed->f1 = 0;

	for (eed=em->edges.first; eed; eed=eed->next) {
		if ((eed->pair==NULL) && eed->f1==0) {
			he_tag_hole_edges(eed);
			tot++;
		}
	}
	return tot;
}  
