#include <stdio.h>
#include "BIF_meshHE.h"
#include "BIF_meshHE_tools.h" 
#include "BIF_space.h"
#include "mydevice.h"
#include "MEM_guardedalloc.h" 

#include "BLI_arithb.h"
#include "BLI_blenlib.h"

#include "BKE_displist.h"
#include "BKE_hemesh.h"
#include "BKE_global.h"
#include "BKE_main.h"

#include "he_editmesh.h"
#include "he_error.h"
#include "math.h"


#if 0
/*prefix T means user Tool*/
char *itotab(int num) {
	static char buf[1000];
	int i;
	for (i=0; i<num; i++) {
		buf[i] = ' ';
	}
	buf[i] = 0;
	return buf;
}  
#endif

static int check_face_loops(HE_EditMesh *em, HE_EditFace *efa, int recur);

void check_faces(HE_EditMesh *em)
{
	HE_EditFace *efa;

	if (!em) return;

//	do {
	//   go_on = 0;
	for (efa=em->faces.first; efa; efa=efa->next) {
		if (check_face_loops(em, efa, 0)==2) break; // {
		//	   go_on = 1;
		//printf("Fixed some faces\n");
		// }
	}   
	//} while (go_on == 1);	 
}

static int check_face_loops(HE_EditMesh *em, HE_EditFace *efa, int recur)
{
  	static int tab=0;
	int retval=0;
	long i;

	HE_EditEdge *eed;

	if (!em) return 0;

	if (recur > 60000) {
		/*ok, looks like we're going to exceed recursion depth. . .*/
		printf("exceeded max recursion depth\n");
		HE_RaiseErrorString("Unsolvable Mesh\n");
		return 2;
	}
	eed = efa->edge;
	do {
		if (eed->pair != NULL && eed->v1 == eed->pair->v1) { 
			switch_face(eed->pair->face);
			eed->pair->face->f1 = 1;
			efa->f1 = 1;
			tab++;
			i = check_face_loops(em, eed->pair->face, recur+1);
			if (i==2) {
				retval = 2;
				break;
			}
			tab--;
			retval = 1;
		}
		i++;
		eed=eed->enext;
	} while (eed != efa->edge);
	// printf(strcat(itotab(tab), "Exiting check_face_loop. . .\n"));

	return retval;
}

void switch_face(HE_EditFace *efa)
{
	HE_EditEdge *eed, *next;
	HE_EditVert *v1;
	HE_HashEdge *hash, *hnext;

 	if (!efa) {
		printf("error switching faces: face does not exist\n");
		return;
	}
	if (efa->switched==0) efa->switched=1; else efa->switched=0;
 	printf("switcheroo\n");

	if (efa->switched) efa->switched = 0; 
	else efa->switched = 1;

	eed = efa->edge;
	do {
		next = eed->enext;
		/*Switch the hash references. . .*/
		for (hash=eed->v1->hashedges.first; hash; hash=hnext) {
			hnext = hash->next;
			if (hash->eed==eed) {
				BLI_remlink(&eed->v1->hashedges, hash);
				BLI_addtail(&eed->v2->hashedges, hash);
			}
		}

		v1 = eed->v1;
		eed->v1 = eed->v2;
		eed->v2 = v1;
		if (v1->edge==eed) {
			if (eed->pair) v1->edge = eed->pair;
			else v1->edge = NULL;
		}  
		eed->enext = eed->eprev;
		eed->eprev = next;
		eed=next;
	} while (eed != efa->edge);

}

void collapse_edge(HE_EditMesh *em, HE_EditEdge *eed)
{
	HE_EditEdge *next;
	if (eed->face->edge==eed) eed->face->edge = eed->enext;
	eed->eprev->enext = eed->enext;
	eed->enext->eprev = eed->eprev;
	if (eed->pair) {
		next = eed->pair->enext->enext;
		HE_delete_halfedge(em, eed->pair->enext);
		eed->pair->enext = next;
		eed->pair->enext->eprev = eed->pair;
		if (eed->pair->face->edge==eed->pair->enext) eed->pair->face->edge = eed->pair;
	}

	he_delete_iedge(em, eed->eprev->ieed);
	eed->eprev->ieed = eed->pair->ieed;
	eed->eprev->ieed->e1 = eed->eprev;
	eed->eprev->ieed->e2 = eed->pair;

	HE_delete_halfedge(em, eed);
}
/*this is sortof obsolete now. . .should recode into a flip normals
  of all connected faces function. . . */ 
void HE_recalc_facenormals(void)
{
	HE_EditMesh *em = G.he_editMesh;
	HE_EditFace *efa;
	if (!em) return;

	for (efa=em->faces.first; efa; efa=efa->next) {
		switch_face(efa);
	}

	HE_recalcnormals(); 
	DAG_object_flush_update(G.scene, G.obedit, OB_RECALC_DATA);

	BIF_undo_push("Recalc Face Normals");
	allqueue(REDRAWVIEW3D, 0);
}   

void deselectall_hemesh(void)
{
	HET_SelOrDeselAll();
} 



HE_EditVert *split_edge(HE_EditMesh *em, HE_EditEdge *eed)
{
	HE_EditVert *cent;
	HE_EditEdge *e1, *e2, *pair;
	float co[3];
	if (!em) return NULL;

	VecAddf(co, eed->v1->co, eed->enext->v1->co);
	VecMulf(co, 0.5f);
	cent = HE_addvertlist(em, co);
	cent->f = eed->v1->f;
	/*this next line prevents non-manifold exceptions 
 	  from being raise, as we mess around with the info edges
	  a lot in this function.*/
	eed->ieed->e1 = eed->ieed->e2 = NULL;

	e1 = HE_addedgelist(em, cent, eed->enext->v1);
	e1->eprev = eed;
	e1->enext = eed->enext;
	eed->enext->eprev = e1;
	eed->enext = e1;
	e1->ieed->f = eed->ieed->f;
	e1->ieed->crease = eed->ieed->crease;
	e1->ieed->seam = eed->ieed->seam;
	e1->face = eed->face;
	eed->ieed->e1 = eed;
	eed->v2 = cent;   

	if (eed->pair) {
		pair = eed->pair;
		pair->v2 = cent;
		e2 = HE_addedgelist(em, cent, pair->enext->v1);
		e2->eprev = pair;
		e2->enext = pair->enext;
		e2->ieed->crease = pair->ieed->crease;
		e2->ieed->seam = pair->ieed->seam;
		pair->enext->eprev = e2;
		pair->enext = e2;
		pair->ieed = e1->ieed;
		pair->ieed->e2 = pair;
		pair->v2 = cent;
		e2->face = pair->face;
		e1->pair = pair;
		e2->pair = eed;

		e2->f2 = 1;
		pair->f2 = 1;
		pair->pair = e1;
		eed->pair = e2;
	} else {
		eed->pair = NULL;
		e1->pair = NULL;
	}
	cent->edge = e1;
	e1->ieed->f1 = 1;
	eed->ieed->f1 = 1;
	eed->f2 = 1;
	eed->f3 = 1;
	e1->f2 = 1;

	return cent;
}
