#include "BIF_meshHE.h"
#include "BIF_meshHE_tools.h"

#include "MEM_guardedalloc.h" 

#include "BMF_Api.h"

#include "IMB_imbuf.h"

#include "MTC_matrixops.h"

#include "DNA_camera_types.h"
#include "DNA_curve_types.h"
#include "DNA_constraint_types.h" // for drawing constraint
#include "DNA_effect_types.h"
#include "DNA_ipo_types.h"
#include "DNA_lamp_types.h"
#include "DNA_lattice_types.h"
#include "DNA_material_types.h"
#include "DNA_mesh_types.h"
#include "DNA_meshdata_types.h"
#include "DNA_meta_types.h"
#include "DNA_object_types.h"
#include "DNA_space_types.h"
#include "DNA_scene_types.h"
#include "DNA_screen_types.h"
#include "DNA_view3d_types.h"
#include "DNA_world_types.h"

#include "BLI_blenlib.h"
#include "BLI_arithb.h"
#include "BLI_editVert.h"

#include "BKE_utildefines.h"
#include "BKE_curve.h"
#include "BKE_constraint.h" // for the get_constraint_target function
#include "BKE_deform.h"		// lattice_modifier()
#include "BKE_displist.h"
#include "BKE_effect.h"
#include "BKE_global.h"
#include "BKE_ipo.h"
#include "BKE_lattice.h"
#include "BKE_mesh.h"
#include "BKE_material.h"
#include "BKE_mball.h"
#include "BKE_object.h"

#include "BIF_gl.h"
#include "BIF_mywindow.h"
#include "BIF_screen.h"
#include "BIF_space.h"
#include "BIF_editarmature.h"
#include "BIF_editmesh.h"
#include "BIF_glutil.h"
#include "BIF_resources.h"
#include "BIF_mainqueue.h"

#include "BDR_drawmesh.h"
#include "BDR_drawobject.h"
#include "BDR_editobject.h"

#include "BSE_view.h"
#include "BSE_drawview.h"
#include "BSE_trans_types.h"

#include "blendef.h"
#include "mydevice.h"
#include "nla.h"

#include "BKE_deform.h"
#include "BKE_hemesh.h"

#include "he_editmesh.h"
#include "he_tools.h"

#include "math.h"
#include "time.h"

#define HE_EDITMODE	1
#define HE_VPAINT	2

/*we only allow one element to be drawn at once. . .*/
/*not that, at least in the case of edges you can specify 
  whether to blend edge highlight with edge selcolor or not.
  if ieed->hl==1, then do blended version, if ieed->hl==2,
  do normal blue highlight.*/

static void clear_face_mover(void);

static void clear_face_mover(void)
{
	HE_EditMesh *em = G.he_editMesh;
	HE_EditEdge *eed;
	if (!em) return;
	if (!em->f_mover) return;

	eed = em->f_mover->edge;
	do {
		eed->ieed->hl = 0;
		eed=eed->enext;
	} while (eed != em->f_mover->edge);   

	em->f_mover = NULL; 
}

void clear_all_mover(void) 
{
	HE_EditMesh *em = G.he_editMesh;
	if (!em) return;

	if (em->v_mover) {
		em->v_mover->hl = 0;
		em->v_mover = NULL;
	}
	if (em->e_mover) {
		em->e_mover->hl = 0;
		em->e_mover = NULL;
	}
	if (em->f_mover) {
		clear_face_mover();
		em->f_mover = NULL;
	}
}

void he_clear_all_mover(void)
{
	clear_all_mover();
}

static int he_vert_mouseover(HE_EditVert *eve)
{
	HE_EditMesh *em = G.he_editMesh;
	if (!em) return 0;

	if (eve) {
		if (em->v_mover) em->v_mover->hl = 0;
		/*clear other mesh element's highlights. . .*/
		if (em->e_mover) em->e_mover->hl = 0;
		clear_face_mover();

		eve->hl = 1;
		em->v_mover = eve;
		return 1;
	}

	return 0;   
}

static int he_edge_mouseover(HE_EditEdgeInfo *ieed)
{
	HE_EditMesh *em = G.he_editMesh;
	if (!em) return 0;

	if (ieed) {
		if (em->e_mover) em->e_mover->hl = 0;
		if (em->v_mover) em->v_mover->hl = 0;
		clear_face_mover();

		if (ieed->f & SELECT) ieed->hl = 1;
		else ieed->hl = 2;
		em->e_mover = ieed;
		return 1;
	}
	return 0;
}

static int he_face_mouseover(HE_EditFace *efa)
{
	HE_EditMesh *em = G.he_editMesh;
	HE_EditEdge *eed;
	static int last_sel = 0;

	if (!em) return 0;

	if (efa) {
		/*clear old highlights, both vert/edge/face. . .*/
		clear_all_mover();

		/*now set highlight flags. . */
		eed = efa->edge;
		do {
			if (efa->f & SELECT) eed->ieed->hl = 1;
			else eed->ieed->hl = 2;
			eed=eed->enext;
		} while (eed != efa->edge);
		em->f_mover = efa;
		last_sel = efa->f & SELECT;
		return 1;
	}
	return 0;
}

void he_mouseover(void)
{
	HE_EditMesh *em = G.he_editMesh;
	unsigned short evt, val;
	void *ptr;
	float lasttime = 0;
	float curtime = (float)clock() / CLOCKS_PER_SEC;
	short evt_list[5000][2];
	int i, len, pdis;

	if (curtime - lasttime < 0.1) {
		lasttime = curtime;
		return;
	} else lasttime = curtime;

	if ((G.scene->selectmode & SCE_SELECT_VERTEX) && G.scene->selectmode > SCE_SELECT_VERTEX) {
		pdis = FDIS;
 	}	else pdis = FDIS;
  
	i = he_unified_findnearest(em, &ptr, pdis);
	switch (i) {
		case HEVERT:
			he_vert_mouseover(ptr);
			force_draw(0);
			break;
		case HEIEDGE:
			he_edge_mouseover(ptr);
			force_draw(0);
			break;
		case HEFACE:
			he_face_mouseover(ptr);
			force_draw(0);
			break;
	}

	/*and, finally, empty the freakin event queue 
	  of all mousex/y events. . .*/
	len = 0;
	while (qtest()) {
		evt= extern_qread(&val);
		if (evt != MOUSEX && evt != MOUSEY) {
			evt_list[len][0] = evt;
			evt_list[len][1] = val;
			len++;
		}
	}

	/*add non-mousex/y events back to the queue*/
	for (i=0; i<len; i++) {
		mainqenter(evt_list[i][0], evt_list[i][1]);
	}
}

/*a "correct" duplicate function isn't possible until the transform
  system can be interfaced with the tool API.  thus, this not-so-good
  function is placed here, where it will be seen and serve as a reminder
  that this feature isn't finished yet! -- joeedh*/
void HET_Duplicate(void)
{
	/*this function works by duplicating faces only, right now*/
	HE_EditMesh *em = G.he_editMesh;
	HE_EditFace *efa, *newfa;
	HE_EditVert *vlist[6000], *eve, *v1;
	HE_EditEdge *eed, *eed2;
	int len;

	if (!em) return;

	for (eve=em->verts.first; eve; eve=eve->next) {
		eve->f1 = eve->f2 = 0;
		eve->newvert = NULL;
	}

	for (efa=em->faces.first; efa; efa=efa->next) {
		efa->f1 = 0;
	}
	for (efa=em->faces.first; efa; efa=efa->next) {
		/*remember: NEVER use BLI_addhead! easy solution I know,
		  but will mess up construction history!*/
		if ((efa->f & SELECT) && efa->f1==0) {
			eed = efa->edge;
			len = 0;
			do {
				if (!eed->v1->newvert) {
					eed->v1->newvert = HE_addvertlist(em, eed->v1->co);
					eed->v1->newvert->f = eed->v1->f;
					v1 = eed->v1->newvert;
				} else v1 = eed->v1->newvert;

				vlist[len] = v1;
				eed=eed->enext;
				len++;
			} while (eed != efa->edge);
			newfa = HE_addfacelist(em, vlist, len);
			newfa->f1 = 1;
			if (newfa) {
				/*copy face and edge stuff. . */
				newfa->f = efa->f;
				eed = efa->edge;
				eed2 = newfa->edge;
				do {
					eed2->ieed->f = eed->ieed->f;
					eed2->ieed->crease = eed2->ieed->crease;
					eed2->ieed->seam = eed2->ieed->seam;

					eed2=eed2->enext;
					eed=eed->enext;
				} while (eed != efa->edge); 
			}
		}
	}
	HE_clear_flag_all(SELECT);
	for (efa=em->faces.first; efa; efa=efa->next) {
		if (efa->f1==1) HE_select_face(efa);
	}
	HE_recalcnormals();
	BIF_undo_push("Duplicate");
}

void HET_EditModeEvent(unsigned short evt, int val)
{
	if (val) {
		switch (evt) {
			case MOUSEX:
			case MOUSEY:
				if (G.f & G_HEMOVER) he_mouseover();
				break;
			case CKEY:
			{
				HE_EditEdge **solid_list;
				HE_EditVert *eve;
				HE_EditMesh *em = G.he_editMesh;
				int i;
				for (eve=em->verts.first; eve; eve=eve->next) {
					if (eve->f & SELECT) eve->f1 = 1;
					else eve->f1 = 0;
				}
				for (eve=em->verts.first; eve; eve=eve->next) {
					if ((eve->f & SELECT) && eve->f1==1) {
						solid_list = he_getSolidsAroundVert(em, eve);
						if (solid_list) {
							i = 0;
							while (solid_list[i]) {
								solid_list[i]->ieed->f |= SELECT;
								i++;
								if (solid_list) MEM_freeN(solid_list);
							}
						}
					}
				}
				allqueue(REDRAWVIEW3D, 1);
				/*HE_Mesh *me = G.obedit->data;
				float euler;
   				int cca = HE_getComponents(), gn = HE_getHoles();
				if (!G.obedit) return;


				//if (cca==0) cca = 1;
				euler = (gn + me->totverts - me->totiedges + me->totfaces) / (cca==0?1:cca); //-gn==0?1:cca-gn);
				printf("euler result: %f\n", euler);
				printf("number of components: %d\n", cca);
				printf("number of genus: %d\n", gn);*/
				break;
			}
   			case DKEY:
   				if (G.qual & LR_SHIFTKEY) {
					HET_Duplicate();
					// transform('g');
				}
   				break;
			case RIGHTMOUSE:
				mouse_hemesh();
				if (G.f & G_HEMOVER) he_mouseover();
				break;
			case AKEY:
				break;
		 	case EKEY:
		 		HET_Extrude();
		 		break;
			case FKEY:
				/*"try" to execute Bridge tool*/
				HE_ExecToolFromString_NoReport("Bridge", NULL);
				HE_recalcnormals();
				allqueue(REDRAWVIEW3D, 0);
				break;
			case HKEY:
				if(G.qual==LR_CTRLKEY) add_hook();
				break;
	 		case YKEY:
	 			HET_Triangulate();
	 			break;
			case PKEY: /*toggle mouseover highlights. . .*/
				if (G.f & G_HEMOVER) {
					G.f &= ~G_HEMOVER;
					clear_all_mover();
				}
				else {
					G.f |= G_HEMOVER;
					he_mouseover();
				}
				allqueue(REDRAWVIEW3D, 1);
				break;
			case LKEY:
				/* for (eve=em->verts.first; eve; eve=eve->next) {
				   if (eve->f & SELECT) {
				   if (!v1) v1 = eve;
				   else v2 = eve;
				   }  ==This is a debug thing that=============
				   }	  ==prints the distance between two verts==
				   if (!v1 || !v2) break;
				   x = v1->co[0] - v2->co[0];
				   y = v1->co[1] - v2->co[1];
				   z = v1->co[2] - v2->co[2];
				   d = sqrt(x*x+y*y+z*z);
				   printf("disr: %f\n", d);*/
				HET_SelectLinkedVerts();
				break;
			case XKEY:
				HET_DelVert();
		   		break;   
		}
	}
}

void HET_Event(unsigned short evt, int val)
{
	HET_EditModeEvent(evt, val);
}
