#include <stdio.h>
#include "MEM_guardedalloc.h" 

#include "BIF_meshHE.h"

#include "BLI_arithb.h" 
#include "BLI_parray.h"
#include "BLI_blenlib.h"

#include "BKE_displist.h"
#include "BKE_hemesh.h"
#include "BKE_global.h"
#include "BKE_main.h"

#include "BIF_space.h"

#include "he_editmesh.h"
#include "mydevice.h"
#include "blendef.h"
#include "math.h"


void HE_remake_pairs(void)
{
	HE_EditMesh *em = G.he_editMesh;
	HE_EditEdge *eed;
	HE_EditEdgeInfo *ieed;

	if (!em) return;
	/*first delete ieed->e2 refs. . .*/
	for (ieed=em->einfo.first; ieed; ieed=ieed->next) {
		if (ieed->e1->pair) ieed->e1->pair->pair = NULL;
		ieed->e1->pair = NULL;
		if (ieed->e2) ieed->e2->pair = NULL;
		ieed->e2 = NULL;
	}
	//printf("p[airdsdsgs\n");
	for (eed=em->edges.first; eed; eed=eed->next) {
		if (eed->enext) {
			ieed = find_iedge(em, eed->v1, eed->enext->v1);
		} else ieed = find_iedge(em, eed->v1, eed->v2);

		if (ieed->e1 != eed) {
			eed->pair = ieed->e1;
			ieed->e1->pair = eed;
			ieed->e2 = eed;
			eed->ieed = ieed;
		}
	}
}

/** Fix missing info edges */
void HE_add_missing_ieeds(void)
{
	HE_EditMesh *em = G.he_editMesh;
	HE_EditEdgeInfo *ieed;
	HE_EditEdge *eed;
	if (!em) return;

	for (eed=em->edges.first; eed; eed=(HE_EditEdge*)eed->next)
	{
		if (!eed->ieed) {
			ieed = find_iedge(em, eed->v1, eed->enext->v1);
			if (!ieed) {
				ieed = MEM_callocN(sizeof(HE_EditEdgeInfo), "editedgeinfo");
				ieed->e1 = eed;
				eed->ieed = ieed;
			} else { 
				eed->ieed = ieed;
				ieed->e2 = eed;
			}
		}
	}
}

#if 0
static int he_face_len(HE_EditFace *efa)
{
	HE_EditEdge *eed;
	int i=0;
	eed = efa->edge;
	do {
		i++;
		eed=eed->enext;
	} while (eed != efa->edge);

	return i;
}
#endif

/** Pops a vert from a face.  
 * @return Whether or not the face had to be deleted,
 * as 0/1
 */ 
int HE_pop_vert_from_face(HE_EditFace *efa, HE_EditVert *eve)
{
	HE_EditMesh *em = G.he_editMesh;
	HE_EditEdge *eed, *prev;
	HE_EditEdgeInfo *ieed, *inext;

	eed = efa->edge;
	if (eed->enext->enext->enext==eed) {
		//printf("Deleting face.\n");
		HE_delete_face(em, efa);
		return 0;
	}
	do {
		if (eed->v1==eve) break;
		eed=eed->enext;
	} while (eed != efa->edge);

	eed->enext->eprev = eed->eprev;
	prev = eed->eprev;
	ieed = MEM_callocN(sizeof(HE_EditEdgeInfo), "ieed");
	ieed->f = prev->ieed->f;
	ieed->seam = prev->ieed->seam;
	ieed->crease = prev->ieed->crease;
	if (prev->ieed->e1==prev) prev->ieed->e1 = prev->ieed->e2;
	if (!prev->ieed->e1) he_delete_iedge(em, prev->ieed);
	prev->ieed = ieed;
	BLI_addtail(&em->einfo, ieed);
	prev->enext = eed->enext;
	HE_delete_halfedge(em, eed);

	for (ieed=em->einfo.first; ieed; ieed=inext) {
		inext = ieed->next;
		if (!ieed->e1) {
			if (ieed->e2) ieed->e1 = ieed->e2;
			else he_delete_iedge(em, ieed);
		}
	}
	return 1;   
}

/*-----"Get adjacent geometry around vertices" iterators begin here-----*/

HE_EditEdge **he_getSolidsAroundVert(HE_EditMesh *em, HE_EditVert *eve)
{
	HE_EditEdge *eed, *last, **elist;
	HE_EditFace **flist;
	HE_HashEdge *hash;
	PointerArray arr = {NULL, 0};
	PointerArray all = {NULL, 0};
	int cursor=0, totcursor=0, i, stop=0;
#if 0
	
	/* stag "solid tag" is use ONLY by this function!
	 * it is allowed because of how important this function is!*/
	while (!stop) {
		elist = he_getHEdgesAroundVert(em, eve);
		if (elist) {
			PA_AddToArray(&arr, eve->edge, cursor);
			cursor++;
			i = 0;
			while (elist[i]) {
				elist[i]->stag = 1;
				/*for un-tagging afterward*/
				PA_AddToArray(&all, elist[i], totcursor);
				totcursor++;
				i++;
			}
		}
		stop = 1;
		for (hash=eve->hashedges.first; hash; hash=hash->next) {
			if (hash->eed->v1==eve && (hash->eed != eve->edge)) {
				if (!hash->eed->stag) {
					stop = 0;
					eve->edge = hash->eed;
				}
			}
		}
	}
	/*un-tag half-edges. . .*/
	for (i=0; i<totcursor; i++) {
		bw_assert(all.array[i]);
		eed = (HE_EditEdge*) all.array[i];
		eed->stag = 0;		
	}
	if (all.array) MEM_freeN(all.array);
	PA_AddToArray(&arr, NULL, cursor+1);
#endif
	return (HE_EditEdge*)arr.array;
}
HE_EditEdge **he_getHEdgesAroundVert(HE_EditMesh *em, HE_EditVert *eve)
{
	HE_EditEdge *eed, *last;
	PointerArray arr = {NULL, 0};
	int cursor=0;
#if 0
	if (!em) return NULL; //protect;  
	if (!eve->edge) return NULL; // protect;
	/*first we rewind. . .*/
	eed = eve->edge;
	last = eed;

	do {
		eed = eed->eprev->pair;
		if (!eed) { 
			eed = last;
			PA_AddToArray(&arr, eed->eprev, cursor);
			cursor++;
			break; 
		}
		last = eed;
	} while (eed != last && eed != eve->edge); 
	
	if (!eed) eed = eve->edge;
	eve->edge = eed;
	last = NULL;
	/*main loop starts here!*/
	do {
		PA_AddToArray(&arr, eed, cursor);
		cursor++;
		if (!eed->pair) {
			cursor++;
			break;
		} 
		else eed=eed->pair->enext;
		if (last==eed) {
			break;
			cursor++;
		}
		last = eed;
	} while (eed != eve->edge); 

	PA_AddToArray(&arr, NULL, cursor+1);
#endif
	return (HE_EditEdge**) arr.array;
}

HE_EditFace **he_getFacesAroundVert(HE_EditMesh *em, HE_EditVert *eve)
{
	HE_EditEdge *eed, *last;
	PointerArray arr = {NULL, 0};
	int cursor=0;
#if 0
	if (!em) return NULL; // protect;
	if (!eve->edge) return NULL; // protect; /*same as NULL*/
	for (eed=em->edges.first; eed; eed=eed->next) eed->f2 = 0;

	/*first we rewind. . .*/
	eed = eve->edge;
	last = eed;
	do {
		eed = eed->eprev->pair;
		if (!eed) {
			eed = last;
			break;
		}
		last = eed;
	} while (eed != last && eed != eve->edge);
	
	if (!eed) eed = eve->edge;
	eve->edge = eed;
	last = NULL;
	/*main loop starts here!*/
	do {
		PA_AddToArray(&arr, eed->face, cursor);  

		eed->f2 = 67;
		/* - iterators begin here - */
		if (!eed->pair) {
			cursor++;
			break;
		} else eed=eed->pair->enext;
		if (last==eed) {
			cursor++;
			break;
		}
		cursor++;

		last = eed;
	} while (eed != eve->edge); 

	PA_AddToArray(&arr, NULL, cursor+1);
#endif
	return (HE_EditFace**) arr.array;
}

HE_EditVert **he_getVertsAroundVert(HE_EditMesh *em, HE_EditVert *eve)
{
	HE_EditEdge *eed, *last;
	int cursor=0;
	PointerArray arr = {NULL, 0};
#if 0
	if (!em) return NULL; // protect;
	if (!eve->edge) return NULL; // protect; /*same as NULL*/

	/*first we rewind. . .*/
	eed = eve->edge;
	last = eed;

	do {
		eed = eed->eprev->pair;
		if (!eed) { 
			eed = last;
			PA_AddToArray(&arr, eed->eprev->v1, cursor);
			//PA_AddToArray(&arr, eed->enext->v1, cursor);
			cursor++;
			//cursor++;
			break; 
		}
		last = eed;
	} while (eed != last && eed != eve->edge); 
	if (!eed) eed = eve->edge;
	eve->edge = eed;
	last = NULL;
	/*main loop starts here!*/
	do {
		/*reallocate memory if necassary, although it should 
		  never really happen :)*/

		if (eed->v1==eve) PA_AddToArray(&arr, eed->enext->v1, cursor);
		else PA_AddToArray(&arr, eed->v1, cursor);
		cursor++;

		/* - iterators begin here - */
		if (!eed->eprev->pair) {
		}
		if (!eed->pair) {
			break;
		} 
		else eed=eed->pair->enext;
		if (last==eed) break;

		last = eed;
	} while (eed != eve->edge); 

	PA_AddToArray(&arr, NULL, cursor+1);
#endif
	return (HE_EditVert**) arr.array;
}
