#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "BIF_meshHE.h"

#include "MEM_guardedalloc.h" 

#include "BMF_Api.h"

#include "IMB_imbuf.h"

#include "MTC_matrixops.h"

#include "DNA_camera_types.h"
#include "DNA_curve_types.h"
#include "DNA_constraint_types.h" // for drawing constraint
#include "DNA_effect_types.h"
#include "DNA_ipo_types.h"
#include "DNA_lamp_types.h"
#include "DNA_lattice_types.h"
#include "DNA_material_types.h"
#include "DNA_mesh_types.h"
#include "DNA_meshdata_types.h"
#include "DNA_meta_types.h"
#include "DNA_object_types.h"
#include "DNA_space_types.h"
#include "DNA_scene_types.h"
#include "DNA_screen_types.h"
#include "DNA_view3d_types.h"
#include "DNA_world_types.h"

#include "BLI_blenlib.h"
#include "BLI_arithb.h"
#include "BLI_editVert.h"

#include "BKE_utildefines.h"
#include "BKE_curve.h"
#include "BKE_constraint.h" // for the get_constraint_target function
#include "BKE_deform.h"		// lattice_modifier()
#include "BKE_displist.h"
#include "BKE_effect.h"
#include "BKE_global.h"
#include "BKE_ipo.h"
#include "BKE_lattice.h"
#include "BKE_mesh.h"
#include "BKE_material.h"
#include "BKE_mball.h"
#include "BKE_object.h"

#include "BIF_gl.h"
#include "BIF_mywindow.h"
#include "BIF_screen.h"
#include "BIF_space.h"
#include "BIF_editarmature.h"
#include "BIF_editmesh.h"
#include "BIF_glutil.h"
#include "BIF_resources.h"

#include "BDR_drawmesh.h"
#include "BDR_drawobject.h"
#include "BDR_editobject.h"

#include "BSE_view.h"
#include "BSE_drawview.h"
#include "BSE_trans_types.h"

#include "blendef.h"
#include "mydevice.h"
#include "nla.h"

#include "BKE_deform.h"
#include "BKE_hemesh.h"

#include "he_editmesh.h"
#include "math.h"


/*****function prototypes. . .*********/

void he_selectvert_border_f(short min[2], short max[2], int selmode);
void he_selectvert_border(short minr[2], short maxr[2], int selmode);
void he_selectedge_border(short min[2], short max[2], int selmode);
void he_selectface_border(short min[2], short max[2], int selmode);
typedef struct _i_arraypointer {int len; int *array;} i_arraypointer;

static i_arraypointer *he_read_bordersel(short min[2], short max[2]);
static unsigned int *he_read_backbuf(short xmin, short ymin, short xmax, short ymax);

/*******************backbuffer select drawing functions. . .*****************/
void he_draw_bbsedges(void *data)
{
	HE_EditMesh *em = G.he_editMesh;
	HE_Mesh *me = G.obedit->data;
	HE_EditEdgeInfo *ieed;
	unsigned int i;
	int subsurf = 0;
	if (!em) return;

	if ((me->f & HEM_SUBSURF) && me->subdivLevels > 0) subsurf = 1;
	if (G.f & G_EDITDEFORM) subsurf = 1;

	bglPolygonOffset(5.0);
	glDisable(GL_DITHER);
	glDisable(GL_LIGHTING);
	glEnable(GL_DEPTH_TEST);

	drawHEEMeshSolid(2);

	if (subsurf) {
		HED_drawSubSurfEdges(1, 255);
		return;
	}
	glBegin(GL_LINES);
	for (ieed=em->einfo.first, i=1; ieed; i++, ieed=ieed->next) {
		if (ieed->h) continue;
		set_framebuffer_index_color(i);
		glVertex3fv(ieed->e1->v1->co);
		glVertex3fv(ieed->e1->enext->v1->co);
	}
	glEnd();
}

void he_draw_bbsverts(void *data)
{
	HE_EditMesh *em = G.he_editMesh;
	HE_Mesh *me = G.obedit->data;
	HE_EditVert *eve;
	unsigned int i, subsurf=0;
	if (!em) return;
	if ((me->f & HEM_SUBSURF) && me->subdivLevels > 0) subsurf = 1;
	if (G.f & G_EDITDEFORM) subsurf = 1;

	bglPolygonOffset(5.0);
	glDisable(GL_DITHER);

	drawHEEMeshSolid(2);
	glPointSize(BIF_GetThemeValuef(TH_VERTEX_SIZE)); 
	bglBegin(GL_POINTS);
	for (eve=em->verts.first, i=1; eve; i++, eve=eve->next) {
		if (eve->h) continue;
		set_framebuffer_index_color(i);
		if (subsurf) bglVertex3fv(eve->ssco);
		else bglVertex3fv(eve->co);
	}
	bglEnd();
}

void he_draw_bbsfaces(void *data)
{
	drawHEEMeshSolid(1);
}

/*takes poly as argument*/
/*static unsigned int *he_read_backbuf_poly(int *poly, int len, short x, short y)
  {
  int i;
  short min, max;
  glBegin(GL_POLYGON);
  glEnd();
  }*/
/* reads full rect, converts indices */
static unsigned int *he_read_backbuf(short xmin, short ymin, short xmax, short ymax)
{
	unsigned int *dr, *buf;
	int a;
	short xminc, yminc, xmaxc, ymaxc;

	/* clip */
	if(xmin<0) xminc= 0; else xminc= xmin;
	if(xmax>=curarea->winx) xmaxc= curarea->winx-1; else xmaxc= xmax;
	if(xminc > xmaxc) return NULL;

	if(ymin<0) yminc= 0; else yminc= ymin;
	if(ymax>=curarea->winy) ymaxc= curarea->winy-1; else ymaxc= ymax;
	if(yminc > ymaxc) return NULL;

	buf= MEM_mallocN( (xmaxc-xminc+1)*(ymaxc-yminc+1)*sizeof(int), "sample rect");

#ifdef __APPLE__
	glReadBuffer(GL_AUX0);
#endif
	glReadPixels(curarea->winrct.xmin+xminc, curarea->winrct.ymin+yminc, (xmaxc-xminc+1), (ymaxc-yminc+1), GL_RGBA, GL_UNSIGNED_BYTE, buf);
	glReadBuffer(GL_BACK);

	if(G.order==B_ENDIAN) IMB_convert_rgba_to_abgr((xmaxc-xminc+1)*(ymaxc-yminc+1), buf);

	a= (xmaxc-xminc+1)*(ymaxc-yminc+1);
	dr= buf;
	while(a--) {
		if(*dr) *dr= framebuffer_to_index(*dr);
		dr++;
	}

	/* put clipped result back, if needed */
	if(xminc==xmin && xmaxc==xmax && yminc==ymin && ymaxc==ymax) return buf;
	else {
		unsigned int *buf1= MEM_callocN( (xmax-xmin+1)*(ymax-ymin+1)*sizeof(int), "sample rect2");
		unsigned int *rd;
		short xs, ys;

		rd= buf;
		dr= buf1;

		for(ys= ymin; ys<=ymax; ys++) {
			for(xs= xmin; xs<=xmax; xs++, dr++) {
				if( xs>=xminc && xs<=xmaxc && ys>=yminc && ys<=ymaxc) {
					*dr= *rd;
					rd++;
				}
			}
		}
		MEM_freeN(buf);
		return buf1;
	}

	return buf;
}

/* smart function to sample a rect spiralling outside, nice for backbuf selection */
static unsigned int he_sample_backbuf_rect(unsigned int *buf, int size, int min, int max, short *dist)
{
	unsigned int *bufmin, *bufmax;
	int a, b, rc, nr, amount, dirvec[4][2];
	short distance=0;

	amount= (size-1)/2;
	rc= 0;

	dirvec[0][0]= 1; dirvec[0][1]= 0;
	dirvec[1][0]= 0; dirvec[1][1]= -size;
	dirvec[2][0]= -1; dirvec[2][1]= 0;
	dirvec[3][0]= 0; dirvec[3][1]= size;

	bufmin= buf;
	bufmax= buf+ size*size;
	buf+= amount*size+ amount;

	for(nr=1; nr<=size; nr++) {

		for(a=0; a<2; a++) {
			for(b=0; b<nr; b++, distance++) {

				if(*buf && *buf>=min && *buf<max ) {
					*dist= (short) sqrt( (float)distance );
					return *buf - min+1;	// messy yah, but indices start at 1
				}

				buf+= (dirvec[rc][0]+dirvec[rc][1]);

				if(buf<bufmin || buf>=bufmax) {
					return 0;
				}
			}
			rc++;
			rc &= 3;
		}
	}
	return 0;
}

/******************Let the findnearest stuff begin!**************/
/* samples a single pixel (copied from edmesh_mods, which copied from vpaint) */
static unsigned int he_sample_backbuf(int x, int y)
{
	unsigned int col=0;

	if(x>=curarea->winx || y>=curarea->winy) return 0;
	x+= curarea->winrct.xmin;
	y+= curarea->winrct.ymin;

#ifdef __APPLE__
	glReadBuffer(GL_AUX0);
#endif
	glReadPixels(x,  y, 1, 1, GL_RGB, GL_UNSIGNED_BYTE,  &col);
	glReadBuffer(GL_BACK);

	if(G.order==B_ENDIAN) SWITCH_INT(col);

	return framebuffer_to_index(col);
}

void HE_Border_Select(short min[2], short max[2], int selmode)
{
	if (G.scene->selectmode & SCE_SELECT_VERTEX) he_selectvert_border(min, max, selmode);
	if (G.scene->selectmode & SCE_SELECT_EDGE) he_selectedge_border(min, max, selmode);
	if (G.scene->selectmode & SCE_SELECT_FACE) he_selectface_border(min, max, selmode);
}


static i_arraypointer *he_read_bordersel(short min[2], short max[2])
{
	int *buf, *iden, *tmp, index, i, g;
	int *read = MEM_callocN(sizeof(int)*6000, "read_backbufsel");
	int cursor=0, totread=6000;
	static i_arraypointer retval_arr;

	printf("before [min: {%d, %d}, max: {%d, %d}]\n", (int)min[0], (int)min[1], (int)max[0], (int)max[1]);

	if (min[0]<0) min[0]=0;
	if (min[1]<0) min[1]=0;
	if (max[0]<0) max[0]=0;
	if (max[1]<0) max[1]=0;

	if (max[0] >= curarea->winx) max[0] = curarea->winx;
	if (max[1] >= curarea->winy) max[1] = curarea->winy;
	if (min[0] >= curarea->winx) min[0] = curarea->winx;
	if (min[1] >= curarea->winy) min[1] = curarea->winy;

	printf("min: {%d, %d}, max: {%d, %d}\n", (int)min[0], (int)min[1], (int)max[0], (int)max[1]);
	buf = iden = he_read_backbuf(min[0], min[1], max[0], max[1]);
	if (!iden || (max[0] == 0 || max[1]==0)) {
		retval_arr.array = read;
		retval_arr.len = 0;
	}

	i = (max[0]-min[0]+1)*(max[1]-min[1]+1);
	while (i--) { 
		index = *iden;
		for (g=0; g<cursor; g++) if (read[g]==index) index=0;
		/*simple auto-realloc'd array. . .*/
		if (index != 0) {
			if (cursor < totread) {
				read[cursor] = index;
			} else {
				printf("reallocating backbuf border list. . .\n");
				tmp = MEM_callocN(sizeof(int)*(totread+1000), "backbufselborderarray");
				for (g=0; g<totread; g++) tmp[g] = read[g];
				totread+= 1000;
				read = tmp;
				read[cursor] = index;
			}
			cursor++;
		}
		iden++;
	}
	MEM_freeN(buf);
	retval_arr.array = read;
	retval_arr.len = cursor;
	return &retval_arr;
}

static int _bint_cmp(int *a, int *b)
{
	if (*a < *b) return -1;
	else if (*a > *b) return 1;
	else return 0;
}

/*
 * Returns an uint array.
 * Array is null-terminated;
 * position 0 is low value,
 * position 1 is high value.
 */
unsigned int *he_read_bacbbufcircle(short *mval, short rads)
{
	unsigned int *buf, *dr, *retarr, *arr;
	short xmin, ymin, xmax, ymax, xc, yc, xs=mval[0], ys=mval[1];
	int radsq, i, lastdr, low=-1, high=0;
	
	if(G.obedit==NULL || G.vd->drawtype<OB_SOLID || (G.vd->flag & V3D_ZBUF_SELECT)==0) return 0;
	
	xmin= xs-rads; xmax= xs+rads;
	ymin= ys-rads; ymax= ys+rads;
	/*if (xmin<0) xmin=1;
	if (ymin<0) ymin=1;
	if (xmax<0) xmax=1;
	if (ymax<0) ymax=1;

	if (xmax >= curarea->winx) xmax = curarea->winx-1;
	if (ymax >= curarea->winy) ymax = curarea->winy-1;
	if (xmin >= curarea->winx) xmin = curarea->winx-1;
	if (ymin >= curarea->winy) ymin = curarea->winy-1;*/
	
	dr= buf= he_read_backbuf(xmin, ymin, xmax, ymax);
	if (buf==NULL) return 0;
	
	/* build selection lookup */
	retarr = arr = MEM_callocN(sizeof(int)*(2)*(xmax-xmin)*(ymax-ymin), "selbuf");
	arr += 2;
	radsq = rads*rads;
	i = 0;
	lastdr = 0;
	for(yc= -rads; yc<=rads; yc++) {
		for(xc= -rads; xc<=rads; xc++, dr++) {
			if(xc*xc + yc*yc < radsq) {
				if (*dr && (*dr != lastdr)) {
    				arr[i]= *dr;
    				if (*dr < low || low==-1) low = *dr;
    				if (*dr > high) high = *dr;
    				i++;
			    }
			    lastdr = *dr;
			}
		}
	}
	
	/*sort the array. . .*/
	qsort(arr, i, sizeof(unsigned int), &_bint_cmp);
	//printf("arr 0-2: %d, %d, %d\n", (int)arr[0], (int)arr[10], (int)arr[20]);
	arr[i+1] = 0; /* NULL-terminate */
	/*set low, high fields, to help with n^2 issues*/
	retarr[0] = (unsigned int) low;
	retarr[1] = (unsigned int) high;
	//printf("low: %d, high: %d\n", low, high);
	if (buf) MEM_freeN(buf);
	return retarr;
	
}

void he_selectvert_border_f(short min[2], short max[2], int selmode)
{
	/*well, select stuff. . .*/
	HE_EditMesh *em = G.he_editMesh;
	HE_EditVert *eve;

	calc_hemesh_vertdots_ext();

	for (eve=em->verts.first; eve; eve=eve->next) {
		if (eve->h) continue;
		if (eve->xs[0] >= min[0] && eve->xs[0] <= max[0]) {
			if (eve->xs[1] >= min[1] && eve->xs[1] <= max[1]) {
				if (selmode==1) eve->f |= SELECT;
				else eve->f &= ~SELECT;
			}
		}
	}
} 

void he_selectvert_border(short minr[2], short maxr[2], int selmode)
{
	HE_EditMesh *em = G.he_editMesh;
	if(G.vd->drawtype>OB_WIRE && (G.vd->flag & V3D_ZBUF_SELECT)) {
		i_arraypointer *array;
		HE_EditVert *eve;
		int i, j;

		if (!em) return;
		if (G.scene->selectmode != SCE_SELECT_VERTEX) {
			//force_backbufdraw_ext(0, NULL, &he_draw_bbsverts);
		}

		array = he_read_bordersel(minr, maxr);
		for (eve=em->verts.first, i=1; eve; i++, eve=eve->next) {
			for (j=0; j<array->len; j++) {
				if (array->array[j]==i) {
					if (selmode==1) eve->f |= SELECT;
					else eve->f &= ~SELECT;
					//eve->edge->enext->v1->f |= SELECT;
				}
			}
		}
		MEM_freeN(array->array);
	} else he_selectvert_border_f(minr, maxr, selmode);
	HE_update_edges(em);
	HE_update_faces(em);
}   

void he_selectedge_border(short min[2], short max[2], int selmode)
{
	HE_EditMesh *em = G.he_editMesh;
	HE_EditVert *eve;
	HE_EditEdge **elist;
	int i;

	he_selectvert_border(min, max, selmode);

	for (eve=em->verts.first; eve; eve=eve->next) {
		if (eve->f & SELECT) {
			elist = he_getHEdgesAroundVert(em, eve);
			i=0;
			if (elist) {
				while (elist[i]) {
					elist[i]->ieed->f |= SELECT;
					i++;
				}
			}
		}
	}
}   

void he_selectface_border_f(short min[2], short max[2], int selmode)
{
	/*well, select stuff. . .*/
	HE_EditMesh *em = G.he_editMesh;
	HE_EditFace *efa;

	calc_hemesh_facedots_ext();

	for (efa=em->faces.first; efa; efa=efa->next) {
		if (efa->h) continue;
		if (efa->xs[0] >= min[0] && efa->xs[0] <= max[0]) {
			if (efa->xs[1] >= min[1] && efa->xs[1] <= max[1]) {
				if (selmode==1) HE_select_face(efa);
				else HE_deselect_face(efa);
			}
		}
	} 
}

void he_selectface_border(short min[2], short max[2], int selmode)
{
	if(G.vd->drawtype>OB_WIRE && (G.vd->flag & V3D_ZBUF_SELECT)) {
		int i, j;
		HE_EditFace *efa;
		HE_EditMesh *em = G.he_editMesh;

		i_arraypointer *array;
		if (!em) return;

		if (G.scene->selectmode != SCE_SELECT_FACE) {
			//force_backbufdraw_ext(0, NULL, &he_draw_bbsfaces);
		}

		array = he_read_bordersel(min, max);
		for (efa=em->faces.first, i=1; efa; i++, efa=efa->next) {
			for (j=0; j<array->len; j++) {
				if (array->array[j]==i) {
					if (selmode==1) HE_select_face(efa);
					else HE_deselect_face(efa);
					//eve->edge->enext->v1->f |= SELECT;
				}
			}
		}
		MEM_freeN(array->array);
	} else he_selectface_border_f(min, max, selmode);
}   

/*masks rect with shape shape*/
unsigned int *he_sample_polyshape(short shape[][2], int len, int sizex, int sizey)
{
	short minr[2] = {0, 0}, maxr[2] = {0, 0};
	int i, j, a, readlen=1024;
	unsigned int *buf, *msk, *tmp;
	unsigned int *read = MEM_callocN(sizeof(unsigned int)*1024, "read buf for he_sample_polyshape");
	static unsigned int protect[2] = {0, 0};

	for (i=0; i<len; i++) {
		if (shape[i][0] < 0 || shape[i][0] > curarea->winx) continue; 
		if (shape[i][1] < 0 || shape[i][1] > curarea->winy) continue; 
		if (shape[i][0] < minr[0]) minr[0] = shape[i][0];
		if (shape[i][0] > maxr[0]) maxr[0] = shape[i][0];
		if (shape[i][1] < minr[1]) minr[1] = shape[i][1];
		if (shape[i][1] > maxr[1]) maxr[1] = shape[i][1];
	}
	//printf("minr: {%d, %d}, maxr: {%d, %d}\n", (int)minr[0], (int)minr[1], (int)maxr[0], (int)maxr[1]);
	persp(PERSP_WIN);
	buf = he_read_backbuf(minr[0], minr[1], maxr[0], maxr[1]);
	if (!buf) return protect;

	glColor3f(1, 1, 1);
	glRects(minr[0], minr[1], maxr[0], maxr[1]);
	glColor3ub(0, 0, 0);
	glBegin(GL_POLYGON);
	for (i=0; i<len; i++) {
		glVertex2s(shape[i][0], shape[i][1]);
	}
	glEnd();
	persp(PERSP_VIEW);
	glFinish();

	msk = he_read_backbuf(minr[0], minr[1], maxr[0], maxr[1]);

	i = (minr[0]-maxr[0]+1)*(minr[1]-maxr[1]+1);
	a = 0;
	while (i--) {
		if (*msk==0) {
			if (a >= readlen) {
				printf("a: %d\n", a);
				tmp = MEM_callocN(sizeof(int)*(readlen+(1024)), "realloc read for polyshape");
				for (j=1; j<readlen; j++) tmp[j] = read[j];
				MEM_freeN(read);
				read = tmp;
				readlen += 1024;
			}
			read[a] = *buf;
		}	 
		msk++;
		buf++;
		a++;
	}
	read[0] = a; 
	return read;
}

HE_EditVert *HE_findnearestvert_f(HE_EditMesh *em, short max_dis) 
{
	/*well, select stuff. . .*/
	HE_EditVert *eve, *vert=NULL;
	short mval[2];
	short dis;

	calc_hemesh_vertdots_ext();
	getmouseco_areawin(mval);

	for (eve=em->verts.first; eve; eve=eve->next) {
		if (eve->h) continue;
		dis = sqrt(pow(mval[0]-eve->xs[0],2) + pow(mval[1]-eve->xs[1],2));
		//printf("dis %d\n", (int)dis);
		if (dis<max_dis) {
			if (!vert) {
				vert = eve;
				vert->f1 = dis;
			} else if (dis<vert->f1) {
				vert = eve;
				vert->f1 = dis;
			}
		}
	}

	return vert;
}   

HE_EditVert *HE_findnearestvert(HE_EditMesh *em, short dis) 
{
	HE_EditVert *eve = NULL;
	if (!em) return NULL;
	/*well, select stuff. . .*/
	if(G.vd->drawtype>OB_WIRE && (G.vd->flag & V3D_ZBUF_SELECT)) {
		unsigned int *buf;
		int a=1, index;
		short mval[2], distance=255;

		getmouseco_areawin(mval);
		/*check to see if we're in multiple select mode, if so then
		  force backbuf sel redraw*/
		if ((G.vd->flag & V3D_NEEDBACKBUFDRAW) || G.scene->selectmode != SCE_SELECT_VERTEX) {
			//force_backbufdraw_ext(0, NULL, &he_draw_bbsverts);
		}
		// sample colorcode 
		buf= he_read_backbuf(mval[0]-25, mval[1]-25, mval[0]+24, mval[1]+24);
		if (buf) {
			//index = he_sample_backbuf(mval[0], mval[1]);
			index= he_sample_backbuf_rect(buf, 50, 0, 0xFFFFFF, &distance); // globals, set in drawobject.c
			index = framebuffer_to_index(index);
			MEM_freeN(buf);
			if(distance < dis) {
				if(index>0) for(eve=em->verts.first, a=2; eve; a++, eve=eve->next) if(index==a) break;
			}
			return eve;
		}
	} else return HE_findnearestvert_f(em, dis);
	return NULL; /*warning fix*/
} 

float he_dist_mval_edge(short *mval, HE_EditEdge *eed)
{
	float v1[2], v2[2], mval2[2];

	mval2[0] = (float)mval[0];
	mval2[1] = (float)mval[1];

	v1[0] = eed->v1->xs[0];
	v1[1] = eed->v1->xs[1];
	v2[0] = eed->v2->xs[0];
	v2[1] = eed->v2->xs[1];

	return PdistVL2Dfl(mval2, v1, v2);
}

HE_EditEdgeInfo *HE_findnearestedge_f(HE_EditMesh *em, short max_dis) 
{
	/*well, select stuff. . .*/
	HE_EditEdgeInfo *ieed, *edge=NULL;
	short mval[2];
	short dis;

	calc_hemesh_vertdots_ext();
	getmouseco_areawin(mval);

	for (ieed=em->einfo.first; ieed; ieed=ieed->next) {
		if (ieed->h) continue;
		if (!ieed->e1) continue;

		dis = he_dist_mval_edge(mval, ieed->e1);
		//printf("dis %d\n", (int)dis);
		if (dis<max_dis) {
			if (!edge) {
				edge = ieed;
				edge->f1 = dis;
			} else if (dis<edge->f1) {
				edge = ieed;
				edge->f1 = dis;
			}
		}
	}
	return edge;
}   

HE_EditEdgeInfo *HE_findnearestedge(HE_EditMesh *em, short dis) 
{
	/*well, select stuff. . .*/
	if(G.vd->drawtype>OB_WIRE && (G.vd->flag & V3D_ZBUF_SELECT)) {
		HE_EditEdgeInfo *ieed=NULL;
		short mval[2];
		unsigned int *buf=NULL;
		int a=1, index;
		short distance=255;

		getmouseco_areawin(mval);

		/* backbuffer has *not* been nicely prepared for us if we're
		   in multiple-select mode. */
		if ((G.vd->flag & V3D_NEEDBACKBUFDRAW) || G.scene->selectmode != SCE_SELECT_EDGE) {
			//force_backbufdraw_ext(0, NULL, he_draw_bbsedges);
		}

		buf= he_read_backbuf(mval[0]-25, mval[1]-25, mval[0]+24, mval[1]+24);
		if (buf) {
			index= he_sample_backbuf_rect(buf, 50, 0, 0xFFFFFF, &distance); // globals, set in drawobject.c
			index = framebuffer_to_index(index);
			MEM_freeN(buf);
			if(distance < dis) {
				if(index>0) for(ieed=em->einfo.first, a=2; ieed; a++, ieed=ieed->next) if(index==a) break;
			}
			return ieed;
		}
	} else return HE_findnearestedge_f(em, dis);
	return NULL; /*warning fix*/
} 

HE_EditFace *HE_findnearestface_f(HE_EditMesh *em, short max_dis) 
{
	/*well, select stuff. . .*/
	HE_EditFace *efa, *face=NULL;
	short mval[2];
	short dis;

	calc_hemesh_facedots_ext();
	getmouseco_areawin(mval);

	/*first search for nearest facedot. . .*/   
	for (efa=em->faces.first; efa; efa=efa->next) {
		if (efa->h) continue;
		dis = sqrt(pow(mval[0]-efa->xs[0],2) + pow(mval[1]-efa->xs[1],2));

		if (dis<max_dis) {
			if (!face) {
				face = efa;
				face->f1 = dis;
			} else if (dis<face->f1) {
				face = efa;
				face->f1 = dis;
			}
		}
	} 

	if (face) return face;
	return NULL;
	/*if not found, then do edge-face calculation--but disabled for now*/
	/*   ieed = HE_findnearestedge(em, max_dis);
		 if (!ieed) return NULL;
		 if (!ieed->e2) return ieed->e1->face;
		 
		 float edge_co[2];
		 float v1co[2] = {(float) ieed->e1->v1->xs[0], ieed->e1->v1->xs[1]};
		 float v2co[2] = {(float) ieed->e1->v2->xs[0], ieed->e1->v2->xs[1]};
		 float faceco[2] = {(float) ieed->e1->face->xs[0], ieed->e1->face->xs[1]};
		 float mco[2] = {(float) mval[0], (float) mval[1]};
		 
		 PptVL2Dfl(mco, v1co, v2co, edge_co);
		 float dis1 = sqrt(pow(edge_co[0]-faceco[0],2) + pow(edge_co[1]-faceco[1],2));
		 float dis2 = sqrt(pow(mval[0]-faceco[0],2) + pow(mval[1]-faceco[1],2));
		 
		 if (dis2 <= dis1) return ieed->e1->face; 
		 else return ieed->e2->face;*/

}  

HE_EditFace *HE_findnearestface(HE_EditMesh *em, short dis) 
{
	if(G.vd->drawtype>OB_WIRE && (G.vd->flag & V3D_ZBUF_SELECT)) {
		HE_EditFace *efa=NULL;
		unsigned int a=1, index;
		short mval[2];

		if (!em) return NULL;

		calc_hemesh_facedots_ext();	// shouldnt be needed each click
		getmouseco_areawin(mval);

		if ((G.vd->flag & V3D_NEEDBACKBUFDRAW) || G.scene->selectmode != SCE_SELECT_FACE) {
			//force_backbufdraw_ext(0, NULL, he_draw_bbsfaces);
		}
		// sample colorcode 
		index= he_sample_backbuf(mval[0], mval[1]);
		for (efa=em->faces.first, a=1; efa; a++, efa=efa->next) {
			if (a==index) return efa;
		}
		return NULL;
	} else return HE_findnearestface_f(em, dis);
}
/*------------Circle select functions-------------*/
static void he_circlesel_face_f(short selecting, short *mval, float rad)
{
	HE_EditMesh *em = G.he_editMesh;
	HE_EditFace *efa;
	float x,y,r;

	for (efa=em->faces.first; efa; efa=efa->next) {
		x = (float)(efa->xs[0]-mval[0]);
		y = (float)(efa->xs[1]-mval[1]);
		r = sqrt(x*x+y*y);
		if (r < rad) {
			if (selecting==1) {
				HE_select_face(efa);  
			} else {
				HE_deselect_face(efa);
			}
		}
	}
}
static void he_circlesel_vert(short selecting, short *mval, float rad);

static void he_circlesel_face(short selecting, short *mval, float rad)
{
	calc_hemesh_facedots_ext();
	if(G.vd->drawtype>OB_WIRE && (G.vd->flag & V3D_ZBUF_SELECT)) {
 		HE_EditMesh *em = G.he_editMesh;
		int *array, *start;
		HE_EditFace *efa, **farray;
		int i, flen;
		if (!em) return;	  

		i = 0;
		/*put all faces in pointer array. . .*/
		for (efa=em->faces.first; efa; efa=efa->next) i++;
		farray = MEM_callocN(sizeof(void*)*i, "farray");
		for (efa=em->faces.first, i=0; efa; i++, efa=efa->next) {
			farray[i] = efa;
		}
		flen = i;
		
		if (G.scene->selectmode != SCE_SELECT_FACE) {
			//force_backbufdraw_ext(0, NULL, &he_draw_bbsfaces);
		 }
			  
		array = start = he_read_bacbbufcircle(mval, rad);
		if (!array) return;
		for (array=start + 2; *array; array++) {
			if (*array > flen) continue;
			efa = farray[(*array)-1];
			if (selecting==1) HE_select_face(efa);
			else HE_deselect_face(efa);
		}
		if (start) MEM_freeN(start);
		if (farray) MEM_freeN(farray);
	} else he_circlesel_face_f(selecting, mval, rad);
}

static void he_circlesel_edge_f(short selecting, short *mval, float rad)
{
	HE_EditMesh *em = G.he_editMesh;
	HE_EditEdgeInfo *ieed;
	float r;

	calc_hemesh_vertdots_ext();
	for (ieed=em->einfo.first; ieed; ieed=ieed->next) {
		r = (float) he_dist_mval_edge(mval, ieed->e1);
		if (r < rad) {
			if (selecting==1) {
				ieed->f |= SELECT;
				ieed->e1->v1->f |= SELECT;
				ieed->e1->v2->f |= SELECT; 
			} else {
				ieed->f &= ~SELECT;
				ieed->e1->v1->f &= ~SELECT;
				ieed->e1->v2->f &= ~SELECT;	 
			}
		}
	}
}

static void he_circlesel_edge(short selecting, short *mval, float rad)
{
	calc_hemesh_facedots_ext();
	if(G.vd->drawtype>OB_WIRE && (G.vd->flag & V3D_ZBUF_SELECT)) {
 		HE_EditMesh *em = G.he_editMesh;
		int *array, *start;
		HE_EditEdgeInfo *ieed, **iarray;
		int i, ielen;
		if (!em) return;	  
		
		/*count info edge, for pointer array. . .*/
		i = 0;
		for (ieed=em->einfo.first; ieed; ieed=ieed->next) i++;
		iarray = MEM_callocN(sizeof(void*)*i, "iearray for circ sel heedge");
		for (ieed=em->einfo.first, i=0; ieed; i++, ieed=ieed->next) {
			iarray[i] = ieed;
		}
		ielen = i;
		
		if (G.scene->selectmode != SCE_SELECT_EDGE) {
			//force_backbufdraw_ext(0, NULL, &he_draw_bbsedges);
		 }
			  
		array = start = he_read_bacbbufcircle(mval, rad);
		for (array=start+2; *array; array++) {
			if (*array > ielen) continue;
			ieed = iarray[(*array)-1];
			if (selecting==1) {
				ieed->f |= SELECT;
				ieed->e1->v1->f |= SELECT;
				ieed->e1->v2->f |= SELECT; 
			} else {
				ieed->f &= ~SELECT;
				ieed->e1->v1->f &= ~SELECT;
				ieed->e1->v2->f &= ~SELECT;	 
			}
		}
			
		if (start) MEM_freeN(start);
		if (iarray) MEM_freeN(iarray);
	} else he_circlesel_edge_f(selecting, mval, rad);
	HE_update_faces(G.he_editMesh);
}

static void he_circlesel_vert_f(short selecting, short *mval, float rad)
{
	HE_EditMesh *em = G.he_editMesh;
	HE_EditVert *eve;
	float x,y,r;

	calc_hemesh_vertdots_ext();
	for (eve=em->verts.first; eve; eve=eve->next) eve->f1 = 0;
	for (eve=em->verts.first; eve; eve=eve->next) {
		x = (float)(eve->xs[0]-mval[0]);
		y = (float)(eve->xs[1]-mval[1]);
		r = sqrt(x*x+y*y);
		if (r < rad) {
			if (selecting==1) eve->f |= SELECT;
			else eve->f &= ~SELECT;
			eve->f1 = 1;
		}
	}
}

static void he_circlesel_vert(short selecting, short *mval, float rad)
{
	HE_EditMesh *em = G.he_editMesh;
	if(G.vd->drawtype>OB_WIRE && (G.vd->flag & V3D_ZBUF_SELECT)) {
		HE_EditVert *eve, **varray;
		int i, vlen, *buf, *start;
		if (!em) return;

		calc_hemesh_vertdots_ext();
		
		i = 0;
		for (eve=em->verts.first; eve; eve=eve->next) i++;
		varray = MEM_callocN(sizeof(void*)*i, "varray");
		for (eve=em->verts.first, i=0; eve; i++, eve=eve->next) {
			varray[i] = eve;
		}
		vlen = i;
		
		if (G.scene->selectmode != SCE_SELECT_VERTEX) {
			//force_backbufdraw_ext(0, NULL, &he_draw_bbsverts);
		}

		buf = start = he_read_bacbbufcircle(mval, rad);
		if (!buf) return;
		for (buf = start+2; *buf; buf++) {
			if (*buf > vlen) continue;
			eve = varray[(*buf)-1];
			if (selecting==1) eve->f |= SELECT;
			else eve->f &= ~SELECT;
			eve->f1 = 1;
		}
		if (start) MEM_freeN(start); //array->array);
		if (varray) MEM_freeN(varray);
	} else he_circlesel_vert_f(selecting, mval, rad);
	HE_update_edges(em);
	HE_update_faces(em);
}

/*border select functions*/
void HE_selectionCB(short selecting, Object *editobj, short *mval, float rad)
{
	/*we are out-of-bounds here. . .*/
	if (mval[0] >= curarea->winx || mval[1] >= curarea->winy) return;
	he_clear_all_mover();
	areawinset(curarea->win);
	if (G.scene->selectmode & SCE_SELECT_VERTEX) he_circlesel_vert(selecting, mval, rad);
	if (G.scene->selectmode & SCE_SELECT_EDGE) he_circlesel_edge(selecting, mval, rad);
	if (G.scene->selectmode & SCE_SELECT_FACE) he_circlesel_face(selecting, mval, rad);
	force_draw(0);
}

/***********Lasso Select****************/
