#include "BIF_meshHE.h"

#include "MEM_guardedalloc.h"

#include "BMF_Api.h"

#include "IMB_imbuf.h"

#include "MTC_matrixops.h"

#include "DNA_camera_types.h"
#include "DNA_curve_types.h"
#include "DNA_constraint_types.h" // for drawing constraint
#include "DNA_effect_types.h"
#include "DNA_ipo_types.h"
#include "DNA_lamp_types.h"
#include "DNA_lattice_types.h"
#include "DNA_material_types.h"
#include "DNA_mesh_types.h"
#include "DNA_meshdata_types.h"
#include "DNA_meta_types.h"
#include "DNA_object_types.h"
#include "DNA_space_types.h"
#include "DNA_scene_types.h"
#include "DNA_screen_types.h"
#include "DNA_view3d_types.h"
#include "DNA_world_types.h"

#include "BLI_blenlib.h"
#include "BLI_arithb.h"
#include "BLI_editVert.h"

#include "BKE_utildefines.h"
#include "BKE_curve.h"
#include "BKE_constraint.h" // for the get_constraint_target function
#include "BKE_deform.h"		// lattice_modifier()
#include "BKE_displist.h"
#include "BKE_effect.h"
#include "BKE_global.h"
#include "BKE_ipo.h"
#include "BKE_lattice.h"
#include "BKE_mesh.h"
#include "BKE_material.h"
#include "BKE_mball.h"
#include "BKE_object.h"

#include "BIF_gl.h"
#include "BIF_mywindow.h"
#include "BIF_screen.h"
#include "BIF_space.h"
#include "BIF_editarmature.h"
#include "BIF_editmesh.h"
#include "BIF_glutil.h"
#include "BIF_resources.h"

#include "BDR_drawmesh.h"
#include "BDR_drawobject.h"
#include "BDR_editobject.h"

#include "BSE_view.h"
#include "BSE_drawview.h"
#include "BSE_trans_types.h"

#include "blendef.h"
#include "mydevice.h"
#include "nla.h"

#include "BKE_deform.h"
#include "BKE_hemesh.h"

#include "he_editmesh.h"
#include "math.h"

#include "BLI_parray.h"

/*this function get's a list of all verts connected to a vert*/	 
static void he_recurse_vlink(HE_EditVert *seve, HE_EditVert *eve, PointerArray *arr, int *idx, float limit, int max_depth)
{
	HE_EditVert **vlist = he_getVertsAroundVert(G.he_editMesh, eve);
	int i = 0;
	float x, y, z, d;

#if 0
	eve->f3 = 1;
	PA_AddToArray(arr, eve, *idx);
	(*idx)++;
	
	if (!vlist) return;
	
	x = seve->co[0] - eve->co[0];
	y = seve->co[1] - eve->co[1];
	z = seve->co[2] - eve->co[2];
	d = sqrt(x*x+y*y+z*z);
	
	if ((limit > 0) && d >= limit) {
		MEM_freeN(vlist);
		return;
	}	
	else if (max_depth==0) {
		MEM_freeN(vlist);
		return;
	}	
	
	while (vlist[i]) {
		if (vlist[i]->f3==0) {
			he_recurse_vlink(seve, vlist[i], arr, idx, limit, max_depth-1);
		}
		i++;
	}
	MEM_freeN(vlist);
#endif
}

HE_EditVert **HE_GetLinkedVerts(HE_EditVert *seve, float limit, int skip_clearflags, int max_depth)
{
	HE_EditMesh *em = G.he_editMesh;
	HE_EditVert *eve;
	PointerArray array = {NULL, 0};
	int i=0;
#if 0
	
	if (!seve->edge) return NULL;
	
	if (!skip_clearflags) for (eve=em->verts.first; eve; eve=eve->next) eve->f3 = 0;
	
	i = 0;
	he_recurse_vlink(seve, seve, &array, &i, limit, max_depth);
	PA_AddToArray(&array, NULL, i);
#endif	
	return (HE_EditVert**) array.array;
}
/*don't for get to MEM_freeN the result after you're done!*/
HE_EditEdgeInfo **HE_GetEdgeLoop(HE_EditEdgeInfo *ieed)
{
	HE_EditMesh *em = G.he_editMesh;
	int i=0, s=0;
	HE_EditEdge *eed, *last_eed=NULL;
	PointerArray array = {NULL, 0};
#if 0
	
	if (!ieed) return NULL;
	if (!ieed->e1) return NULL;
	
	for (eed=em->edges.first; eed; eed=eed->next) {
		eed->f3 = 0;
		eed->f1 = 0;
	}	
	
	/*first rewind. . .*/
	eed = ieed->e1;
	s = 0;
	do {
		eed->f1 = 1;
		if (eed->eprev->pair) {
			eed=eed->eprev->pair->eprev;
		} else break;
		if (eed->enext->enext->enext==eed) break;
	} while (eed != ieed->e1 && eed->f1==0);
	
	i = 0;
	s = 0;
	last_eed = NULL;
	do {
		PA_AddToArray(&array, eed->ieed, i);
		eed->f1 = 2;
		//last_eed = eed;
		if (eed->enext->pair) {
			eed = eed->enext->pair->enext;
		} else break;
		i++;
		if (eed->enext->enext->enext==eed) {
			PA_AddToArray(&array, eed->ieed, i);
			break;
		}
	} while (eed->f1 != 2);
	
	/* NULL terminate*/
	PA_AddToArray(&array, NULL, i+1);
#endif
	return (HE_EditEdgeInfo **) array.array;
}

/*******************face loop function******************/

HE_EditFace **HE_GetFaceLoop(HE_EditEdgeInfo *ieed)
{
	HE_EditMesh *em = G.he_editMesh;
	int i=0, s=0;
	HE_EditEdge *eed=NULL;
	PointerArray array = {NULL, 0};

#if 0	
	if (!ieed) return NULL;
	if (!ieed->e1) return NULL;
	for (eed=em->edges.first; eed; eed=eed->next) eed->f1 = 0;
	eed = ieed->e1;
	do {
		/*accordion loop select if (G.qual & LR_CTRLKEY) {
		  if (s==1) {
		  s = 0;
		  eed->f1 = 1;
		  if (eed->pair) {
		  eed = eed->pair->enext->enext;
		  } else break;
		  continue;
		  } else s = 1;
		  }	 */
		PA_AddToArray(&array, eed->face, i);
		i++;
		eed->f1 = 1;
		if (eed->pair) {
			eed = eed->pair->enext->enext;
		} else break;
	} while (eed != ieed->e1 && eed->f1==0);
	s = 0;
	if (eed != ieed->e1) { /*it's not a closed loop*/
		eed = ieed->e1->eprev->eprev->pair;
		if (!eed) return NULL;
		do {
			if (G.qual & LR_CTRLKEY) {
				if (s==1) {
					s = 0;
					eed->f1 = 1;
					eed = eed->eprev->eprev;
					if (eed->pair) eed = eed->pair;
					continue;
				} else s = 1;
			} 
			PA_AddToArray(&array, eed->face, i);
			i++;
			eed->f1 = 1;
			eed = eed->eprev->eprev;
			if (eed->pair) eed = eed->pair;
			else break;
			if (eed->enext->enext->enext==eed) {
				break;
			}
		} while (eed != ieed->e1 && eed->f1==0);
	}
	
	/* NULL terminate*/
	PA_AddToArray(&array, NULL, i);
	PA_AddToArray(&array, NULL, i);
#endif
	return (HE_EditFace**) array.array;
}
