#include <stdio.h>
#include "MEM_guardedalloc.h"

#include "BIF_meshHE.h"
#include "BIF_meshHE_tools.h"
#include "BIF_space.h"
#include "BIF_interface.h"
#include "blendef.h"
#include "BLI_arithb.h"

#include "BKE_displist.h"
#include "BKE_hemesh.h"
#include "BKE_utildefines.h"

#include "he_editmesh.h"
#include "he_tools.h"
#include "mydevice.h"
#include <string.h>

/*
  Deledge (collapse edge) HE tool
*/
/*
  Delvert (collapse vert) HE tool
*/

char *HE_DelEdgeTool(char *arg_string, HE_EditMesh *em)
{
	
	HE_EditEdgeInfo *ieed, *next;
	if (!em) return NULL;
	
	for (ieed=em->einfo.first; ieed; ieed=next) {
		next = ieed->next;
		if (ieed->f & SELECT) {
			collapse_edge(em, ieed->e1);
			break;
		}
	}
	return NULL;
}

void HET_DelVert(void)
{
	short evt;
	
	evt = pupmenu("Delete%t|Collapse Vert%x1|Delete Face%x2");
	switch (evt) {
		case 1:
			HE_ExecToolFromString("Delete Vert", NULL);
			allqueue(REDRAWVIEW3D, 1);
			BIF_undo_push("Collapse Vert");
			break;
		case 2:
			HE_ExecToolFromString("Delete Face", NULL);
			allqueue(REDRAWVIEW3D, 1);
			HE_recalcnormals();
			BIF_undo_push("Delete Face");
			break;
	}
}

char *HE_DelFaceTool(char *arg_string, HE_EditMesh *em)
{
	
	HE_EditFace *efa, *next;
	HE_EditVert *eve, *vnext;
	HE_EditEdge *eed;
	if (!em) return NULL;
	
	for (efa=em->faces.first; efa; efa=next) {
		next = efa->next;
		if (efa->f & SELECT) {
			HE_delete_face(em, efa);
		}
	}
	
	for (eve=em->verts.first; eve; eve=eve->next) {
		eve->f1 = 0;
	}
	for (efa=em->faces.first; efa; efa=efa->next) {
		eed = efa->edge;
		do {
			eed->v1->f1++;
			eed=eed->enext;
		} while (eed != efa->edge);
	}
	
	for (eve=em->verts.first; eve; eve=vnext) {
		vnext = eve->next;
		if (eve->f1==0) {
			HE_delete_vert(em, eve);
		}
	}   
	return NULL;
}

HE_EditFace *HE_pop_face(HE_EditFace *efa, HE_EditVert *eve, HE_EditMesh *em)
{
	HE_EditEdge *eed;
	HE_EditVert *vlist[6000];
	HE_EditFace *newfa=NULL;
	int vlen=0, f = efa->f, mat = efa->mat;
	
	eed = efa->edge;
	do {
		if (eed->v1 != eve) {
			vlist[vlen] = eed->v1;
			vlen++;
		}
		eed=eed->enext;
	} while (eed != efa->edge);
	
	HE_delete_face(em, efa);
	if (vlen>2) {
		newfa = HE_addfacelist(em, vlist, vlen);
		newfa->f = f;
		newfa->mat = mat;
	}
	return newfa;
}

char *HE_DelVertTool(char *arg_string, HE_EditMesh *em)
{
	
	HE_EditVert *eve, **vlist, *vnext;
	HE_EditFace **flist=NULL, *newfa;
	HE_EditEdge **solid_list;
	HE_HashEdge *hash;
	int i, si, mat, f, go_on = 0;
	
	if (!em) return NULL;
	for (eve=em->verts.first; eve; eve=eve->next) {
		if (eve->f & SELECT) eve->f1 = SELECT;
		else eve->f1 = 0;
	}
	
	for (eve=em->verts.first; eve; eve=vnext) {
		vnext = eve->next;
		if (eve->f1 & SELECT) {
			solid_list = he_getSolidsAroundVert(em, eve);
			bw_assert(solid_list);
			if (!solid_list) continue;
			si = 0;
			while (solid_list[si]) {
				eve->edge = solid_list[si];
				mat = 0;
				f = 0;
				newfa = NULL;
				vlist = he_getVertsAroundVert(em, eve);
				flist = he_getFacesAroundVert(em, eve);
				if (!flist || !vlist) break;
				i = 0;
				while (flist[i]) {
				    HE_select_face(flist[i]);
					mat = flist[i]->mat;
					f = flist[i]->f;
					HE_pop_face(flist[i], eve, em);
					i++;
				}
				
				i = 0;
				while (vlist[i]) {
					i++;
				}
				
				if (i >= 3) newfa = HE_addfacelist(em, vlist, i);
				if (newfa) {
					HE_select_face(newfa);
					newfa->f = f | SELECT;
					newfa->mat = mat;
				}
				if (vlist) MEM_freeN(vlist);
				if (flist) MEM_freeN(flist);
				si++;
			};
			MEM_freeN(solid_list);
			HE_delete_vert(em, eve);
		}
	}
	HE_recalcnormals();
	return NULL;
}
