#include <stdio.h>
#include "MEM_guardedalloc.h"
#include "BIF_meshHE.h" 
#include "BIF_space.h"
#include "blendef.h"
#include "BLI_arithb.h"

#include "BKE_displist.h"
#include "BKE_hemesh.h"
#include "BKE_global.h"
#include "BKE_utildefines.h"

#include "he_editmesh.h"
#include "he_tools.h"
#include "he_error.h"
#include "mydevice.h"

#include "math.h"
#include <string.h>

/*
  Edge crease tool.
*/

void HET_EdgeCrease(void)
{
	HE_ExecToolFromString("Edge Crease", NULL);
	allqueue(REDRAWVIEW3D, 1);
	BIF_undo_push("Edge Crease");
}

/*
  arg_string format:
  [%f] where f is float specifying val
*/
char *HE_EdgeCreaseTool(char *arg_string, HE_EditMesh *em)
{
	
	HE_EditEdgeInfo *ieed;
	float *edge_creases;
	int i, len;
	float val, co[3] = {0, 0, 0};
	char *str;
	char header[1000] = "";
	
	if (!em) return NULL;
	
	if (arg_string) {
		sscanf(arg_string, "%f", &val);
		for (ieed=em->einfo.first; ieed; ieed=ieed->next) {
			if (ieed->f & SELECT) {
				ieed->crease += val;
			}
		}
		return arg_string;
	}
	
	/*ok, if no arg string then we enter Interactive Mode!!*/
	
	/*calc center and make crease array. . .*/
	len=0;
	for (ieed=em->einfo.first, i=0; ieed; i++, ieed=ieed->next) {
		if (ieed->f & SELECT) {
			VecAddf(co, co, ieed->e1->v1->co);
			VecAddf(co, co, ieed->e1->v2->co);
			len += 2;
		}
	}
	
	VecMulf( co, 1.0f/((float)len) );
	Mat4MulVecfl(G.obedit->obmat, co);
	
	edge_creases = MEM_mallocN(sizeof(float)*len, "edge_creases_he");
	
	i = 0;
	for (ieed=em->einfo.first; ieed; ieed=ieed->next) {
		if (ieed->f & SELECT) {
			edge_creases[i] = ieed->crease;
			i++;
		}
	}
	/*init the drag interface*/
	HED_Init_Drag(DIS_FROM_CENTER, co);
	
	/*enter loop*/
	while (!HED_Check_Cancel()) {
		val = HED_GetDragValuef();
		val -= 1;
		val *= -1;
		if (val < -1) val = -3;
		
		len = 0;
		
		for (i=0, ieed=em->einfo.first; ieed; i++, ieed=ieed->next) {
			if (ieed->f & SELECT) {
				ieed->crease = edge_creases[len] + val;
				if (ieed->crease < 0) ieed->crease = 0;
				len++;
			}
		}
		sprintf(header, "Edge crease offset: %f", val);
		HED_SetHeaderPrint(header);
		DAG_object_flush_update(G.scene, G.obedit, OB_RECALC_DATA);

		HED_DoDraw(NULL, NULL);
	}
	
	MEM_freeN(edge_creases);
	
	/*cancelled? or completed*/
	if (HED_Check_Cancel()==2) {
		/*luckily we do not have to restore anything, undo mesh
		  does it all for us!*/
		HE_RaiseError(HE_CANCELED);
		return NULL;
	} else {
		/*form arg string return*/
		char *str = MEM_mallocN(100, "edgesharpen_hetool");
		sprintf(str, "%f", val);
		return str;
	}
}
