#include <stdio.h>
#include "MEM_guardedalloc.h"
#include "BIF_meshHE.h" 
#include "BIF_space.h"
#include "BIF_transform.h"
#include "BIF_interface.h"
#include "blendef.h"

#include "BLI_arithb.h"
#include "BLI_parray.h"

#include "BKE_utildefines.h"
#include "BKE_displist.h"
#include "BKE_hemesh.h"
#include "BKE_global.h"

#include "DNA_scene_types.h"
#include "DNA_object_types.h"
#include "DNA_hemesh_types.h"
#include "DNA_mesh_types.h"
#include "DNA_material_types.h"

#include "he_editmesh.h"
#include "he_tools.h"
#include "he_error.h"
#include "mydevice.h"
#include "math.h"
#include <string.h>

/*
  Extrude HE tool
*/

void HET_Extrude(void) 
{
	//return;
	short type = pupmenu("Extrude%t|Region%x1|Faces%x2|Verts%x3|Extend Edges%x4");
	switch (type) {
		case 1:
			HE_ExecToolFromString("Extrude Region", NULL);
			allqueue(REDRAWVIEW3D, 1);  
			DAG_object_flush_update(G.scene, G.obedit, OB_RECALC_DATA);
			BIF_undo_push("Extrude Region");  
			break;
		case 2:
			/*we have to be in face select mode for this. . .*/
			G.scene->selectmode |= SCE_SELECT_FACE;
			HE_ExecToolFromString("Extrude Faces", NULL);
			allqueue(REDRAWVIEW3D, 1);  
			DAG_object_flush_update(G.scene, G.obedit, OB_RECALC_DATA);

			BIF_undo_push("Extrude Faces"); 
			break;
		case 3:
			/*we have to be in vertex select mode for this. . .*/
			G.scene->selectmode |= SCE_SELECT_VERTEX;
			HE_ExecToolFromString("Extrude Verts", NULL);
			allqueue(REDRAWVIEW3D, 1);  
			DAG_object_flush_update(G.scene, G.obedit, OB_RECALC_DATA);

			BIF_undo_push("Extrude Verts"); 
			break;
		case 4:
			HE_ExecToolFromString("Extend Edges", NULL);
			allqueue(REDRAWVIEW3D, 1);  

			DAG_object_flush_update(G.scene, G.obedit, OB_RECALC_DATA);

			BIF_undo_push("Extend Edges"); 
			break;
   	}
}

char *HE_ExtrudeFaceTool(char *arg_string, HE_EditMesh *em)
{
	
	HE_EditVert *vlist[5000], *vlist2[5000], *eve;
	HE_EditVert *quad[4];
	HE_EditEdge *eed;
	HE_EditEdgeInfo *ieed;
	HE_EditFace *efa, *fnext, *newfa;
	HE_HashEdge *hash;
	int f, i, len, mat;
	
	if (!em) {
		HE_RaiseError(HE_NO_MESHDATA);
		return NULL;
	}
	
	/*deselect all edges. . .*/
	for (ieed=em->einfo.first; ieed; ieed=ieed->next) ieed->f &= ~SELECT;
	
	/*deselect all verts. . .*/
	for (eve=em->verts.first; eve; eve=eve->next) eve->f &= ~SELECT;
	
	/*clear ->f1 face flags. . .*/
	for (efa=em->faces.first; efa; efa=efa->next) efa->f1 = 0;
	
	/*main loop*/
	for (efa=em->faces.first; efa; efa=fnext) {
		fnext = efa->next;
		/*we get list of verts first. . 
		  we grab iedges for this to copy edge flags correctly.*/
		if ((efa->f & SELECT) && efa->f1==0) {
			len = 0;
			eed = efa->edge;
			do {
				/*we have to dupalloc it because we are about to 
				  delete it*/
				//ielist[len] = MEM_dupallocN(eed->ieed);
				vlist[len] = eed->v1;
				len++;
				eed=eed->enext;
			} while (eed != efa->edge);
			
			/*delete the old face*/
			f = efa->f;
			mat = efa->mat;
			HE_delete_face(em, efa);
			
			/*form new face*/
			for (i=0; i<len; i++) {
				vlist2[i] = HE_addvertlist(em, vlist[i]->co);
			}
			newfa = HE_addfacelist(em, vlist2, len);
			newfa->f = f;
			newfa->f1 = 1;
			newfa->mat = mat;
			HE_select_face(newfa);
			
			/*set normals on verts*/
			for (i=0; i<len; i++) {
				VECCOPY(vlist2[i]->no, newfa->nor);
			}
			/*and form border edges*/
			for (i=0; i<len; i++) {
				quad[3] = vlist[i];
				quad[2] = vlist2[i];
				if (i != len-1) {
					quad[1] = vlist2[i+1];
					quad[0] = vlist[i+1];
				} else {
					quad[1] = vlist2[0];
					quad[0] = vlist[0];
				}
				newfa = HE_addfacelist(em, quad, 4);
				newfa->f = f & ~SELECT;
				newfa->f1 = 1;
				newfa->mat = mat;
			}
		}
	}
	
   	initTransform(TFM_SHRINKFATTEN, CTX_NONE);
   	Transform();
	return NULL;
}

char *HE_ExtrudeVertTool(char *arg_string, HE_EditMesh *em)
{
	HE_EditFace *newfa, **flist;
	HE_EditEdge *eed, **elist, **solid_list;
	HE_EditEdgeInfo *ieed;
	HE_EditVert *eve, *v1, *start;
	HE_EditVert *vlist[4], *vbuffer[6000];
	HE_HashEdge *hash;
	int i, si, vlen, f, f1, mat, tot, elen;
	float co[3], no[3], d, x, y, z;
	float *edge_dis;							  /*array of origional edge distances*/

	if (!em)
	{
		HE_RaiseError(HE_NO_MESHDATA);
		return NULL;
	}

	for (ieed=em->einfo.first; ieed; ieed=ieed->next) elen++;
	edge_dis = (float*) MEM_callocN(sizeof(float)*elen, "edge_dis");
	for (ieed=em->einfo.first, i=0; ieed; i++, ieed=ieed->next)
	{
		x = ieed->e1->v1->co[0] - ieed->e1->enext->v1->co[0];
		y = ieed->e1->v1->co[1] - ieed->e1->enext->v1->co[1];
		z = ieed->e1->v1->co[2] - ieed->e1->enext->v1->co[2];
		d = sqrt(x*x+y*y+z*z);
		edge_dis[i] = d;
		ieed->f1 = i;							  /*put indice in ->f1*/
	}
/*clear flags. . .*/
	for (eed=em->edges.first; eed; eed=eed->next)
	{
		eed->f1 = 0;
		eed->ieed->f &= ~SELECT;
		eed->face->f &= ~SELECT;
		eed->f2 = 0;
		eed->f3 = 0;							  /*used for incident-vert handler*/
		eed->ieed->f1 = 0;
		eed->face->f1 = 0;
		eed->v1->f1 = 0;
		eed->v1->f2 = 0;
	}

/*step 1: bevel vert*/
	for (eve=em->verts.first; eve; eve=eve->next)
	{
		if ((eve->f & SELECT) && eve->f1==0)
		{
			si = 0;
			solid_list = he_getSolidsAroundVert(em, eve);
			if (!solid_list) continue;
			while (solid_list[si])
			{
				eve->f1 = 1;					  /*to prevent recursive looping*/
												  /*we need this for later*/
				eve->edge = solid_list[si];
				flist = he_getFacesAroundVert(em, eve);
				elist = he_getHEdgesAroundVert(em, eve);
				bw_assert(elist);
				if (!elist[0] || !flist[0]) {
					if (elist) MEM_freeN(elist);
					if (flist) MEM_freeN(flist);
					continue;
				}
				if (elist[2]==NULL) {
					if (elist) MEM_freeN(elist);
					if (flist) MEM_freeN(flist);
					continue;
				}
				i = 0;

				/*split edges, putting new vert close to eve*/
				tot = 0;
				while (elist[i])
				{
					elist[i]->f1 = 1;
					if (elist[i]->v1==eve) start = elist[i]->v1;
					else start = elist[i]->enext->v1;
					f1 = elist[i]->ieed->f1;
					v1 = split_edge(em, elist[i]);
					elist[i]->f3 = 1;
					/*ieed->f1 is the indice in edge_dis,
					  which holds the origional edge lengths.
					  however, split_edge sets it, so we must restore it*/
					elist[i]->ieed->f1 = f1;
					elist[i]->enext->ieed->f1 = f1;
					//elist[i]->eprev->ieed->f1 = f1;

					vbuffer[i] = v1;
					v1->f1 = 2;
					v1->f &= ~SELECT;
					x = co[0]-v1->co[0];
					y = co[1]-v1->co[1];
					z = co[2]-v1->co[2];
					d = edge_dis[f1];			  //sqrt(x*x+y*y+z*z);
					VECCOPY(co, start->co);
					VecSubf(no, start->co, v1->co);
					Normalise(no);
					VecMulf(no, d*-0.1f);
					VecAddf(v1->co, start->co, no);
					i++;
					tot++;
				}
				/*pop vert from face. . .*/
				i = 0;
				f = flist[0]->f & ~SELECT;
				mat = flist[0]->mat;
				while (flist[i])
				{
					HE_pop_face(flist[i], eve, em);
					i++;
				}
				/*Form new faces*/
				for (i=0; i<tot; i++)
				{
					vlist[2] = eve;
					vlist[1] = vbuffer[i];
					if (i != tot-1) vlist[0] = vbuffer[i+1];
					else vlist[0] = vbuffer[0];
					newfa = HE_addfacelist(em, vlist, 3);
					newfa->f = f;
					newfa->mat = mat;
					/*if (HE_CatchError(HE_NON_MANIFOLD))
					{
						HE_delete_face(em, newfa);
						while (HE_CatchError(0));
					}*/
				}
				if (flist) MEM_freeN(flist);
				if (elist) MEM_freeN(elist);
				si++;
			}
			MEM_freeN(solid_list);
		}
	}

	if (edge_dis) MEM_freeN(edge_dis);
	initTransform(TFM_SHRINKFATTEN, CTX_NONE);
	Transform();
	return NULL;
}

char *disabledHE_ExtrudeVertTool(char *arg_string, HE_EditMesh *em)
{
	HE_EditFace *newfa, **flist;
	HE_EditVert *eve, *vlist[6000], *nlist[3];
	HE_EditEdge *eed, *next, *fedge, **elist;
	HE_EditEdgeInfo *ieed;
	HE_HashEdge *hash;
	int i, vlen, f, mat, go;
	float co[3], no[3], d;
	
	if (!em) {
		HE_RaiseError(HE_NO_MESHDATA);
		return NULL;
	}
	
	for (eve=em->verts.first; eve; eve=eve->next) {
		if (eve->f & SELECT) eve->f1 = 1;
		else eve->f1 = 0;
		eve->newvert = NULL;
		eve->f3 = 0;
	}
	
	
	/*clear all necassary selection flags. . .*/
	for (ieed=em->einfo.first; ieed; ieed=ieed->next) {
		ieed->f2 = 0;
		ieed->f1 = 0;
		ieed->f &= ~SELECT;
		ieed->e1->face->f &= ~SELECT;
		ieed->e1->f3=0;
		ieed->e1->f1 = 0;
		if (ieed->e2) ieed->e2->f1 = 0;
		if (ieed->e2) ieed->e2->f3 = 0;
	}
	
	/*begin main loop*/
	for (eve=em->verts.first; eve; eve=eve->next) {
		if (eve->f1==1) {
			go = 1;
			while (go) {
				f = 0;
				mat = 0;
				/*begin beveling*/
				flist = he_getFacesAroundVert(em, eve);
				elist = he_getHEdgesAroundVert(em, eve);
				i = 0;
				vlen = 0;
				bw_assert(flist);
				if (!flist) continue;
				/*count number of adjacent faces. . .*/
				while (elist[i]) {
					elist[i]->f1 = 1;
					i++;
				}
				if (i < 3) { /*not enough faces for a manifold vert. . .*/
					eve->f &= ~SELECT;
					continue;
				}
				i = 0;
				/*->f and ->mat is always from flist[0]*/
				if (flist[0]) {
					f = flist[0]->f;
					mat = flist[0]->mat;
				}
				/*first split edges around vert. . .*/
				while (flist[i]) {
					fedge = flist[i]->edge;
					eed = fedge;
					do {
						next = eed->eprev;
						if (eed->v1==eve) {
							eed->f1 = 1;
							if (eed->f3==0) { 
								/*slit edge, then position resultant vert*/
								VECCOPY(co, eed->enext->v1->co);
								VecMulf(co, -1);
								VecAddf(no, eed->v1->co, co);
								Normalise(no);
								VECCOPY(co, eve->co);
								d = sqrt(pow(eed->v1->co[0]-eed->v2->co[0],2) + pow(eed->v1->co[1]-eed->v2->co[1], 2) + pow(eed->v1->co[2]-eed->v2->co[2], 2));
								VecMulf(no, d * -0.2f);
								VecAddf(co, co, no);
								
								vlist[vlen] = split_edge(em, eed);
								eed->ieed->f &= ~SELECT;
								vlist[vlen]->f &= ~SELECT; /*split_edge copies flags 
															 from edge verts, so we must 
															 deselect it*/
								VECCOPY(vlist[vlen]->co, co);
								vlen++;
							}
							/*ok, handle non-solid mesh problem. . .*/
							if (!eed->eprev->pair) {
								VECCOPY(co, eed->v1->co);
								VecMulf(co, -1);
								VecAddf(no, eed->eprev->v1->co, co);
								Normalise(no);
								VECCOPY(co, eve->co);
								VecMulf(no, 0.2);
								VecAddf(co, co, no);
								vlist[vlen] = split_edge(em, eed->eprev);
								vlist[vlen]->f &= ~SELECT; /*split_edge copies flags 
															 from edge verts, so we must 
															 deselect it*/
								VECCOPY(vlist[vlen]->co, co);
								vlen++;
							}
						}
						eed=next;
					} while (eed != fedge);
					i++;
				}
				/*now pop vert out of faces surrounding it. . . */
				i = 0;
				while (flist[i]) {
					HE_pop_face(flist[i], eve, em);
					i++;
				}
				/*now form new faces*/
				for (i=0; i<vlen; i++) {
					nlist[0] = vlist[i];
					nlist[1] = eve;
					if (i != vlen-1) nlist[2] = vlist[i+1];
					else nlist[2] = vlist[0];
					newfa = HE_addfacelist(em, nlist, 3);
					newfa->f = f;
					newfa->mat = mat;
					eed = newfa->edge;
					do {
						eed->f1 = 1;
						eed=eed->enext;
					} while (eed != newfa->edge);
					if (HE_CatchError(HE_NON_MANIFOLD)) {
						while (HE_CatchError(HE_NON_MANIFOLD));
						HE_delete_face(em, newfa);
					}
				}
				if (flist) MEM_freeN(flist);
				if (elist) MEM_freeN(elist);  
				go = 0;
				for (hash=eve->hashedges.first; hash; hash=hash->next) {
					if (hash->eed->f1==0 && hash->eed->v1==eve) {
						printf("found new vert extrude! %d\n", hash->eed);
						go = 1;
						hash->eed->f1 = 1;
						eve->edge = hash->eed;
						break;
					}
				}
			} 
		}
	}
	
   	initTransform(TFM_SHRINKFATTEN, CTX_NONE);
   	Transform();
   	return NULL;
}

char *HE_ExtrudeRegionTool(char *arg_string, HE_EditMesh *em)
{
	HE_EditFace *efa, *newfa, *fnext, *newfa2;
	float nor[3] = {0, 0, 0};
	HE_EditVert *v1, *eve, *vlist[4];
	HE_EditEdge *eed, *eed2;
	PointerArray face_list = {NULL, 0};
	PointerArray eed_list = {NULL, 0};
	int i, len, nlen=0, f, mat;
#if 0
	
	for (efa=em->faces.first; efa; efa=efa->next) {
		/*this saves us one whole loop!  set all face ->f1 flags to 0!*/
		efa->f1 = 0;
	}
	for (eve=em->verts.first; eve; eve=eve->next) {
		/*eve->f1 is used to average new vert normals. . .*/
		eve->f1 = 0;
		eve->newvert = NULL;
		eve->f &= ~SELECT;
	}
	
	for (efa=em->faces.first; efa; efa=fnext) {
		fnext = efa->next;
		if ((efa->f & SELECT) && efa->f1==0) {
			f = efa->f;
			mat = efa->mat;
			/*add to normal to average normal. . .*/
			VecAddf(nor, efa->nor, nor);
			nlen++;
			
			/*if face is selected, duplicate and delete 
			  it (unweld, essentially). . .*/
			i = 0;
			PA_FreeArray(&face_list);
			eed = efa->edge;
			do {
				/*make the new vert. . .*/
				if (eed->v1->newvert) v1 = eed->v1->newvert;
				else v1 = HE_addvertlist(em, eed->v1->co);
				
				/*make vert have same normal as face. . .*/
				VECCOPY(v1->no, efa->nor);
				v1->f1 += 1;
				/*copy flags and select vert. . .*/
				v1->f |= SELECT;
				v1->f = eed->v1->f;
				PA_AddToArray(&face_list, v1, i);
				eed->v1->newvert = v1;
				/*we build a list of copies, since it's important to delete
				  the old face before building any new ones*/
				eed2 = MEM_dupallocN(eed);
				PA_AddToArray(&eed_list, eed2, i);
				eed=eed->enext;
				i++;
			} while (eed != efa->edge);
			/*set len to i*/
			len = i;
			/* delete old face. . . */
			HE_deselect_face(efa);
			HE_delete_face(em, efa);
			
			/*form new faces. . .*/
			for (i=0; i<len; i++) {
				eed = eed_list.array[i];
				if (i != len-1) eed2 = eed_list.array[i+1];
				else eed2 = eed_list.array[0];
				
				/*form a new face on the boundaries. . .*/
				if (eed->pair && !(eed->pair->face->f & SELECT)) {
					vlist[3] = eed->v1;
					vlist[2] = eed->v1->newvert;
					vlist[1] = eed2->v1->newvert;
					vlist[0] = eed2->v1;
					newfa2 = HE_addfacelist(em, vlist, 4);
					newfa2->f = f & ~SELECT;
					newfa2->mat = mat;
				}
				if (i > 0) MEM_freeN(eed);
			}
			MEM_freeN(eed_list.array[0]);
			
			newfa = HE_addfacelist(em, (HE_EditVert**)face_list.array, len);
			newfa->f = f;
			newfa->mat = mat;
			HE_select_face(newfa);
			/*set ignore flag, since as a rule we never add
			  to the head of an HE list.  But then, we could go into an
			  infinite loop :) .*/
			newfa->f1 = 1;
		}
	}
	
	VecMulf(nor, 1.0f/((float)nlen));
	for (eve=em->verts.first; eve; eve=eve->next) {
		if (eve->f1) {
			VECCOPY(eve->no, nor);
		}
	}
	
   	initTransform(TFM_SHRINKFATTEN, CTX_NONE);
   	Transform();
#endif
   	return NULL;
}

void HET_ExtendEdges(void) 
{
#if 0
	short ok;
	
	ok = okee("Extend?");
	if (ok) {
		HE_ExecToolFromString("Extend Edges", NULL);
		allqueue(REDRAWVIEW3D, 1);  
		BIF_undo_push("Extend Edges");  
	}
#endif
}

/*This tool extends edges that don't have a pair, e.g. boundary edges.*/
char *HE_ExtendEdgesTool(char *arg_string, HE_EditMesh *em)
{
	
	HE_EditVert *eve, *v1, *v2, *vlist[4];
	HE_EditEdge *eed, *newed;
	HE_EditEdgeInfo *ieed;
	HE_EditFace *newfa;
	char *str = NULL;
	float no[3];
	
	if (!em) return NULL;
	
	for (eve=em->verts.first; eve; eve=eve->next) {
		eve->newvert = NULL;
		eve->f &= ~SELECT;
		eve->f1 = 0;
		eve->no[0] = eve->no[1] = eve->no[2] = 0;
	}
	for (ieed=em->einfo.first; ieed; ieed=ieed->next) ieed->f1 = 0;
	
	for (ieed=em->einfo.first; ieed; ieed=ieed->next) {
		if ((ieed->f & SELECT) && (!ieed->f1)) {
			eed = ieed->e1;
			if (!eed->pair) {
				/*Make new verts, if necassary. . .*/
				if (!eed->v1->newvert) {
					v1 = HE_addvertlist(em, eed->v1->co);
					eed->v1->newvert = v1;
				} else v1 = eed->v1->newvert;
				if (!eed->enext->v1->newvert) {
					v2 = HE_addvertlist(em, eed->enext->v1->co);
					eed->enext->v1->newvert = v2;
				} else v2 = eed->enext->v1->newvert;
				
				/*Calculate the coplanier vert normals. . .*/
				VECCOPY(no, eed->eprev->v1->co);
				VecMulf(no, -1);
				VecAddf(no, eed->v1->co, no);
				Normalise(no);
				/*	if (v1->f1>0) { attempt to normalise vector properly--doesn't work
					dot = no[0] + v1->no[0] * no[1] + v1->no[1] * no[2] + v1->no[2];
					VecMulf(no, dot);
					dot = acosf(dot);
					dot = dot * 180 / 3.141592654;
					dot = dot / 360;
					VecMulf(no, 1-dot);
					}   	*/
				VecAddf(v1->no, v1->no, no);
				v1->f1++; /*for averaging the normals*/
				
		   		VECCOPY(no, eed->enext->enext->v1->co);
		   		VecMulf(no, -1);
		   		VecAddf(no, eed->enext->v1->co, no);
		   		Normalise(no);
				VecAddf(v2->no, v2->no, no);
				v2->f1++;/*for averaging the normals*/
				
				/*Now make the new face. . .*/
				vlist[0] = v1;
				vlist[1] = v2;
				vlist[2] = eed->enext->v1;
				vlist[3] = eed->v1;
				newfa = HE_addfacelist(em, vlist, 4);
				
				/*Find the correct new edge and select it.*/
				newed = newfa->edge;
				do {
					if (newed->v1==v1 && newed->enext->v1==v2) break;
					else if (newed->v1==v2 && newed->enext->v1==v1) break;
					newed=newed->enext;
				} while (newed != newfa->edge);
				newed->ieed->f |= SELECT;
				newed->v1->f |= SELECT;
				newed->enext->v1->f |= SELECT;
				newed->ieed->f1 = 1;
			}
			ieed->f &= ~SELECT;
		}
	}
	/* average the normals */ 
	for (eve=em->verts.first; eve; eve=eve->next) {
		if (eve->f1) {
			VecMulf(eve->no, 1.0f / ((float)eve->f1));
		}
	}
	
	initTransform(TFM_SHRINKFATTEN, CTX_NONE);
	Transform();
	return str;
}
