#include <stdio.h>
#include "MEM_guardedalloc.h"
#include "BIF_meshHE.h"
#include "BIF_space.h"
#include "BIF_interface.h"
#include "BIF_toolbox.h"
#include "blendef.h"
#include "BLI_arithb.h" 

#include "BKE_displist.h"
#include "BKE_global.h"
#include "BKE_hemesh.h"
#include "BKE_utildefines.h"

#include "he_editmesh.h"
#include "he_tools.h"
#include "he_error.h"

#include "math.h"
#include <string.h>

HE_EditTool he_tool_array[] = {
	{"Delete Vert", NULL, NULL, &HE_DelVertTool, 0, 0, 0},
	{"Subdivide", NULL, NULL, &HE_SubdivideTool, 0, 0, 0},
	{"Expand Sel", NULL, NULL, &HE_SelectMoreTool, 0, 0, 0},
	{"Contract Sel", NULL, NULL, &HE_SelectLessTool, 0, 0, 0},
	{"Extrude Region", NULL, NULL, &HE_ExtrudeRegionTool, 0, 0, 0},
	{"Extend Edges", NULL, NULL, &HE_ExtendEdgesTool, 0, 0, 0},
	{"Delete Face", NULL, NULL, &HE_DelFaceTool, 0, 0, 0},
	{"Merge", NULL, NULL, &HE_MergeTool, 0, 0, 0},
	{"SmoothShade", NULL, NULL, &HE_SmoothShadeTool, 0, 0, 0},
	{"SolidShade", NULL, NULL, &HE_SolidShadeTool, 0, 0, 0},
	{"Select Vert", NULL, NULL, &HE_SelVertTool, 0, 0, EXCLUDE_FROM_MACRO},
	{"Select Edge", NULL, NULL, &HE_SelEdgeTool, 0, 0, EXCLUDE_FROM_MACRO},
	{"Select Face", NULL, NULL, &HE_SelFaceTool, 0, 0, EXCLUDE_FROM_MACRO},
	{"JoinFaces", NULL, NULL, &HE_JoinFacesTool, 0, 0, 0},
	{"Delete Edge", NULL, NULL, &HE_DelEdgeTool, 0, 0, 0},
	{"Triangulate", NULL, NULL, &HE_TriangulateTool, 0, 0, 0},
	{"Extrude Verts", NULL, NULL, &HE_ExtrudeVertTool, 0, 0, 0},
	{"Extrude Faces", NULL, NULL, &HE_ExtrudeFaceTool, 0, 0, 0},
	{"Select Linked", NULL, NULL, &HE_SelLinkedTool, 0, 0, 0},
	{"Average", NULL, NULL, &HE_AverageTool, 0, 0, 0},
	{"Flip Normal", NULL, NULL, &HE_FlipAllNormalsTool, 0, 0, 0},
	{"Assign Mat", NULL, NULL, &HE_AssignMatTool, 0, 0, 0},
	{"Select Mat", NULL, NULL, &HE_SelectMatTool, 0, 0, 0},
	{"Hard Edge", NULL, NULL, &HE_HardEdgeTool, 0, 0, 0},
	{"Soft Edge", NULL, NULL, &HE_SoftEdgeTool, 0, 0, 0},
	{"Bridge", NULL, NULL, &HE_BridgeTool, 0, 0, 0},
	{"Subdivide Smooth", NULL, NULL, &HE_SubdivSmoothTool, 0, 0, 0},
	{"Sel-Desel All", NULL, NULL, &HE_DeselSelTool, 0, 0, 0},
	{"Select Swap", NULL, NULL, &HE_SelSwapTool, 0, 0, 0},
	{"Reveal Sel", NULL, NULL, &HE_RevealSelTool, 0, 0, 0},
	{"Hide Sel", NULL, NULL, &HE_HideSelTool, 0, 0, 0},
	{0, 0, 0, 0, 0, 0},
};

static int he_tottools = (sizeof(he_tool_array) / sizeof(HE_EditTool)) - 1;

void print_tool_history(ListBase *stack)
{
	HE_EditToolStackEntry *entry;
	printf("\n-----Printing Tool Macro History------\n\n");
	for (entry=stack->last; entry; entry=entry->prev) {
		printf("%s with argument %s\n", entry->tool->name, entry->arg_string);
	}
	printf("\n");
}

HE_EditTool *he_get_tool_from_name(char *name)
{
	HE_EditTool *tool = he_tool_array;
	while (tool->name) {
		if (strcmp(tool->name, name)==0) return tool;
	}
	return NULL;
}

/** in some cases, like FKEY bridge, we want full error 
	handling--just not error reporting
*/
char *he_exec_tool(HE_EditTool *tool, char *arg_string, int report)
{
	char *arg_string_pass; /*for passing to history push functions*/
	/*we make a undo mesh copy for if the tool raises a non-cancel error.
	  tools handle user cancel exceptions themselves--thats VERY important.*/
	heUndoMesh *undom = heeditmeshToUndoMesh();
	HE_EditMesh *em = G.he_editMesh;
	if (!em) return;
	
	/*clear mouseover highlight pointers. . .*/
	clear_all_mover();
	
	arg_string_pass = tool->tool(arg_string, em);
	if (!arg_string_pass) arg_string_pass = arg_string;
	
	if (( HE_CatchError(HE_CANCELED) || HE_CheckError() )) {
		int errorc = HE_CatchError(0);
		while (errorc) {
			if (errorc) {
				if (report) {
					char str[3000] = "";
					sprintf(str, "Tool \"%s\" raised error: %s", tool->name, HE_ErrorCodeToString(errorc));
					error(str);
				} else printf("error: %s\n",  HE_ErrorCodeToString(errorc));
			}
			errorc = HE_CatchError(0); 
		}
		/*restore mesh copy. . . */
		undomeshToHEEditMesh(undom);
	}
	freeundoHE(undom);
	return arg_string_pass;
}
char *HE_ExecToolFromString(char *name, char *arg_string)
{
	int i=0;
	char *retval = NULL;
	HE_Mesh *me = (HE_Mesh*) G.obedit->data;
	
	for (i=0; i<he_tottools; i++) {
		if (strcmp(name, he_tool_array[i].name)==0) {
			retval = he_exec_tool(&he_tool_array[i], arg_string, 1);

			if (me->f & HEM_SUBSURF) DAG_object_flush_update(G.scene, G.obedit, OB_RECALC_DATA);
			return retval;
		}
	}
	printf("Unknown tool %s.\n", name);
	return NULL;
}

/** HE_ExecToolFromString_NoReport
 * In some cases, such as FKEY bridge, it's
 * desirable to "try" to execute a tool, handle
 * any errors, but NOT tell the user that anything 
 * happened :)
 */
char *HE_ExecToolFromString_NoReport(char *name, char *arg_string)
{
	int i=0;
	char *retval = NULL;
	HE_Mesh *me = (HE_Mesh*) G.obedit->data;
	
	for (i=0; i<he_tottools; i++) {
		if (strcmp(name, he_tool_array[i].name)==0) {
			retval = he_exec_tool(&he_tool_array[i], arg_string, 0);

			if (me->f & HEM_SUBSURF) DAG_object_flush_update(G.scene, G.obedit, OB_RECALC_DATA);

			return retval;
		}
	}
	printf("Unknown tool %s.\n", name);
	return NULL;
}

void HE_ExecToolFromIndex(int index, char *arg_string) 
{
	int i=0;
	for (i=0; i<he_tottools; i++) {
		if (i==index) he_exec_tool(&he_tool_array[i], arg_string, 1);
		return;
	}
	
	printf("Bad tool index %d.\n", index);
}
