#include <stdio.h>
#include "MEM_guardedalloc.h"
#include "BIF_meshHE.h"
#include "BIF_space.h"
#include "blendef.h" 
#include "BLI_arithb.h"
#include "DNA_scene_types.h"

#include "BKE_displist.h"
#include "BKE_hemesh.h"
#include "BKE_global.h"
#include "BKE_main.h"
#include "BKE_utildefines.h"

#include "he_editmesh.h"
#include "he_tools.h"
#include "he_error.h"
#include "mydevice.h"
#include "math.h"
#include <string.h>


/*
  Various selection tools
*/

void HET_SelSwap(void)
{
	HE_ExecToolFromString("Select Swap", NULL);
	allqueue(REDRAWVIEW3D, 1);
	BIF_undo_push("Select Swap");
}

char *HE_SelSwapTool(char *arg_string, HE_EditMesh *em)
{
	HE_EditVert *eve;
	
	/*toggle sel flags in verts. . .*/
	for (eve=em->verts.first; eve; eve=eve->next) {
		if (eve->f & SELECT) eve->f &= ~SELECT;
		else eve->f |= SELECT;
	}
	
	/*refresh edge-face selections*/
	HE_update_edges(em);
	HE_update_faces(em);
	
	if (!em) return NULL;
	return NULL;
}
void HET_SelectLinkedVerts(void)
{
	HE_ExecToolFromString("Select Linked", NULL);
	allqueue(REDRAWVIEW3D, 1);
	BIF_undo_push("Select Linked");
}

char *HE_SelLinkedTool(char *arg_string, HE_EditMesh *em)
{
	
	HE_EditVert *eve, **vlist;
	int i;
	if (!em) return NULL;
	
	for (eve=em->verts.first; eve; eve=eve->next) {
		eve->f3 = 0;
	}
	
	for (eve=em->verts.first; eve; eve=eve->next) {
		if ((eve->f & SELECT) && eve->f3==0) {
			vlist = HE_GetLinkedVerts(eve, -1, 1, -1);
			if (vlist) {
				i = 0;
				while (vlist[i]) {
					vlist[i]->f |= SELECT;
					/*we NEVER EVER add to the head with BLI_addhead!
					  this would ruin construction history!*/
					vlist[i]->f1 = 1;
					i++;
				} 
				MEM_freeN(vlist);	
			}
		}
	}
	
	HE_update_edges(em);
	HE_update_faces(em);
	
	return NULL;   
}

char *HET_FindAndSelectVert(int dis)
{
	/*ok, these tools arg_strings are formatted as follows:
	  either F[max_dis] for Find nearest vert, or
	  S[some_number] for select vert at indice some_number*/
	/*SCRATCH THAT ^, let's make it 1/0 some_number, the 1/0 for shift
	  key status, the some_number the vert index*/
	//char str[100];
	//str[0] = 'F';
	//sprintf(str+1, "%d", dis); 
	char *pass;
	
	pass = HE_ExecToolFromString("Select Vert", NULL);
	allqueue(REDRAWVIEW3D, 1);
	BIF_undo_push("Select Vert");
	
	return pass;
}

char *HE_SelVertTool(char *arg_string, HE_EditMesh *em)
{
	HE_EditVert *eve, *loop;
	int i, shift, index;
	char *s;
	
	if (!em) return NULL;
	if (!arg_string) {
		if (G.scene->selectmode != SCE_SELECT_VERTEX)
			eve = HE_findnearestvert(em, 15);
		else eve = HE_findnearestvert(em, 75);
		
		if (!eve) {
			HE_RaiseError(HE_CANCELED);
			return NULL;
		} else {
			if (G.qual & LR_SHIFTKEY) {
				if (eve->f & SELECT) eve->f &= ~SELECT;
				else eve->f |= SELECT;
				for (loop=em->verts.first, i=0; loop; i++, loop=loop->next) {
					if (loop==eve) index = i;
				}
			} else {
				for (loop=em->verts.first, i=0; loop; i++, loop=loop->next) {
					if (loop==eve) index = i;
					loop->f &= ~SELECT;
				}
				eve->f |= SELECT;
			}	
		}
		HE_update_edges(em);
		HE_update_faces(em);
		s = MEM_callocN(6000, "selverttool's argstring");
		if (G.qual & LR_SHIFTKEY) shift = 1;
		else shift = 0;
		sprintf(s, "%d %d", shift, index);
		return s;
		
	} else { /*we have an argument string. . .*/
		sscanf(arg_string, "%d %d", &shift, &index);
		for (eve=em->verts.first, i=0; eve; i++, eve=eve->next) {
			if (i==index) {
				if (shift) {
					if (eve->f & SELECT) eve->f &= ~SELECT;
					else eve->f |= SELECT;
				} else {
					for (loop=em->verts.first, i=0; loop; i++, loop=loop->next) {
						if (loop==eve) index = i;
						loop->f &= ~SELECT;
					}
					eve->f |= SELECT;
				}
				HE_update_edges(em);
				HE_update_faces(em);
				break;
			}
		}
		if (index>i) { /*we found no vert. . .*/
			HE_RaiseError(HE_CANCELED);
			return NULL;
		}
		return arg_string;
	}
}

void HET_FindAndSelectEdge(void)
{
}

char *HE_SelEdgeTool(char *arg_string, HE_EditMesh *em)
{
	return NULL;
}

void HET_FindAndSelectFace(void)
{
}

char *HE_SelFaceTool(char *arg_string, HE_EditMesh *em)
{
	return NULL;
}

void HET_select_more(void)
{
	HE_ExecToolFromString("Expand Sel", NULL);
	allqueue(REDRAWVIEW3D, 1);
	BIF_undo_push("Expand Sel");
}

char *HE_SelectMoreTool(char *arg_string, HE_EditMesh *em)
{
	HE_EditVert *eve, **vert;
	HE_EditEdge **elist;
	int i = 0;
	if (!em) return NULL;
	//printf("exec\n");
	for (eve=em->verts.first; eve; i++, eve=eve->next) {
		if (eve->f & SELECT) {eve->f1 = 1;}
		else eve->f1 = 0;
	}
	
	//printf("---begin loop\n---");
	for (eve=em->verts.first, i=0; eve; i++, eve=eve->next) {
		if (eve->f1 == 1) {
			/*DISABLED-->current hackish way of solving incident verts:
			  use hash edges for detecting and "starting"
			  new loops of edges! not really good solution, 
			  but hay it works.*/
			// do {
			i = 0;
			vert = he_getVertsAroundVert(em, eve);
			//elist = he_getHEdgesAroundVert(em, eve);
			//testing purposes --> if (elist) {
			//  i = 0;
			// while (elist[i]) {
			//	 elist[i]->ieed->f |= SELECT;
			//	 i++;
			//  }
			//}	
			if (vert) {
				i = 0;
				while (vert[i]) {
					vert[i]->f |= SELECT;
					i++;
				}
				MEM_freeN(vert);
			}
			/*	bw_assert(eve);
				if (eve->hashedges.first) {
				for (hash=eve->hashedges.first; hash; hash=hash->next) {
				// bw_assert(hash);
				// bw_assert(hash->eed);
				// bw_assert(hash->eed->v1);
				// bw_assert(hash->eed->v2);
				// if (!hash->eed->v2) continue;
				// if (hash->eed->v1==eve && (!(hash->eed->v1->f & SELECT))) break;
				}
				if (hash->eed->v1==eve && (!(hash->eed->v2->f & SELECT))) {
				go_on = 1;
				eve->edge = hash->eed;
				} else go_on = 0;  
				} else go_on = 0;
				} while (go_on);	*/
		}
	}
	HE_update_edges(em);
	HE_update_faces(em);
	return arg_string;
} 

void HET_select_less(void)
{
	HE_ExecToolFromString("Contract Sel", NULL);
	allqueue(REDRAWVIEW3D, 1);
	BIF_undo_push("Contract Sel");
}

char *HE_SelectLessTool(char *arg_string, HE_EditMesh *em)
{
	HE_EditVert *eve, **vert;
	
	int i = 0, totusel;
	if (!em) return NULL;
	//printf("exec\n");
	for (eve=em->verts.first; eve; i++, eve=eve->next) {
		if (eve->f & SELECT) eve->f1 = 1;
		else eve->f1 = 0;
	}
	
	//printf("---begin loop\n---");
	for (eve=em->verts.first, i=0; eve; i++, eve=eve->next) {
		if (eve->f1 == 1) {
			i = 0;
			totusel = 0;
			vert = he_getVertsAroundVert(em, eve);
			while (vert[i]) {
				if (!(vert[i]->f & SELECT) && vert[i]->f1==0) eve->f &= ~SELECT; //totusel++;
				i++;
			}   
			MEM_freeN(vert);
			//if (totusel >= 2) eve->f &= ~SELECT; 
		}
	}
	HE_update_edges(em);
	HE_update_faces(em);
	
	return arg_string;
}

void HET_SelOrDeselAll(void)
{
	HE_ExecToolFromString("Sel-Desel All", NULL);
	allqueue(REDRAWVIEW3D, 1);
}

char *HE_DeselSelTool(char *arg_string, HE_EditMesh *em)
{
	HE_EditVert *eve;
	HE_EditEdgeInfo *ieed;
	HE_EditFace *efa;
	int desel = 0;
	
	if (!em) return NULL;
	
	for (eve=em->verts.first; eve; eve=eve->next) if (eve->f & SELECT) desel = 1;
	for (ieed=em->einfo.first; ieed; ieed=ieed->next) if (ieed->f & SELECT) desel = 1;
	for (efa=em->faces.first;  efa; efa=efa->next) if (efa->f & SELECT) desel = 1;
	
	if (desel == 1) {
		for (eve=em->verts.first; eve; eve=eve->next) eve->f &= ~SELECT;
		for (ieed=em->einfo.first; ieed; ieed=ieed->next) ieed->f &= ~SELECT;
		for (efa=em->faces.first; efa; efa=efa->next) efa->f &= ~SELECT;
		BIF_undo_push("Deselect all");
	} else {
		for (eve=em->verts.first; eve; eve=eve->next) eve->f |= SELECT;
		for (ieed=em->einfo.first; ieed; ieed=ieed->next) ieed->f |= SELECT;
		for (efa=em->faces.first; efa; efa=efa->next) efa->f |= SELECT;
		BIF_undo_push("Select all");
	}  
	
	return NULL;
}	  
