#include <stdio.h>
#include "MEM_guardedalloc.h"
#include "BIF_meshHE.h" 
#include "BIF_space.h"
#include "blendef.h"
#include "BLI_arithb.h"

#include "BKE_displist.h"
#include "BKE_hemesh.h"
#include "BKE_global.h"
#include "BKE_main.h"
#include "BKE_utildefines.h"

#include "he_editmesh.h"
#include "he_error.h"
#include "he_tools.h"
#include "mydevice.h"
#include "math.h"
#include <string.h>


void HET_Triangulate(void)
{
	HE_ExecToolFromString("Triangulate", NULL);
	HE_recalcnormals();
	
	allqueue(REDRAWVIEW3D, 1);
	BIF_undo_push("Triangulate");
}

char *HE_TriangulateTool(char *arg_string, HE_EditMesh *em)
{
	
	HE_EditFace *efa, *newfa, *next;
	HE_EditEdge *eed;
	HE_EditVert *vlist[6000], *nlist[3], *cvert;
	int mat, len, f, i, do_all=0;
	float cent[3];
	
	if (!em) return arg_string;
	
	if (arg_string) {
		sscanf(arg_string, "%d", &do_all);
	}
	
	for (efa=em->faces.first; efa; efa=efa->next) {
		efa->f1 = 0;
	}
	
	for (efa=em->faces.first; efa; efa=next) {
		next = efa->next;
		if (do_all || (efa->f1==0 && (efa->f & SELECT))) {
			eed = efa->edge;
			f = efa->f;
			mat = efa->mat;
			len = 0;
			cent[0] = cent[1] = cent[2] = 0;
			do {
				vlist[len] = eed->v1;
				VecAddf(cent, cent, eed->v1->co);
				
				eed=eed->enext;
				len++;
			} while (eed != efa->edge);
			
			VecMulf(cent, 1.0f/((float)len)); /*calc cent*/
			
			if (len <= 3) continue; /*already triangulated*/
			
			/*delete the face first, doing otherwise will
			  mess up half-edge integrity!!*/
			HE_delete_face(em, efa);
			
			/*do special cases first. . .*/ 
			if (len==4) {
				nlist[0] = vlist[0];
				nlist[1] = vlist[1];
				nlist[2] = vlist[2];
				newfa = HE_addfacelist(em, nlist, 3);
				newfa->f = f;
				newfa->mat = mat;
				if (newfa->f & SELECT) HE_select_face(newfa);
				
				nlist[0] = vlist[2];
				nlist[1] = vlist[3];
				nlist[2] = vlist[0];
				newfa = HE_addfacelist(em, nlist, 3);
				newfa->f = f;
				newfa->mat = mat;
				if (newfa->f & SELECT) HE_select_face(newfa);
			} else if (len==5) {
				nlist[0] = vlist[0];
				nlist[1] = vlist[1];
				nlist[2] = vlist[4];
				newfa = HE_addfacelist(em, nlist, 3);
				newfa->f = f;
				newfa->mat = mat;
				if (newfa->f & SELECT) HE_select_face(newfa);
				
				nlist[0] = vlist[1];
				nlist[1] = vlist[2];
				nlist[2] = vlist[4];
				newfa = HE_addfacelist(em, nlist, 3);
				newfa->f = f;
				newfa->mat = mat;
				if (newfa->f & SELECT) HE_select_face(newfa);
				
				nlist[0] = vlist[2];
				nlist[1] = vlist[3];
				nlist[2] = vlist[4];
				newfa = HE_addfacelist(em, nlist, 3);
				newfa->f = f;
				newfa->mat = mat;
				if (newfa->f & SELECT) HE_select_face(newfa);
			} else { /*triangulate around center. . .*/
				cvert = HE_addvertlist(em, cent);
				cvert->f |= SELECT;
				
				for (i=0; i<len; i++) {
					nlist[0] = vlist[i];
					nlist[1] = cvert;
					if (i != len-1) nlist[2] = vlist[i+1];
					else nlist[2] = vlist[0];
					
					newfa = HE_addfacelist(em, nlist, 3);
					newfa->f = f;
					newfa->mat = mat;
					if (newfa->f & SELECT) HE_select_face(newfa);
				}
			} 
		}
	}
	return arg_string;
}
