#ifndef _BIF_MESHHE
#define _BIF_MESHHE
/*note to hos: I've noticed that a lot of blender includes include
 * this file, seems like a standard can-do. */

#include "DNA_listBase.h"

/* *b*blender *w*arning assert macro*/
#define bw_assert(a) if (!(a)) printf("%s not true\n", #a);

/* A few notes:
 * Remeber that all half-edges should be treated
 * as if they don't have a ->v2 member.  The only
 * reason why they do is because it really is [not anymore] necassary
 * for the conversion and selection functions.
 */
struct Link;
struct HE_Mesh;
struct HE_EditMesh;
struct HE_EditVert;
struct HE_EditEdge;
struct HE_EditEdgeInfo;
struct HE_EditFace;
struct HE_HashEdge;
struct Object;
struct SubSurf;

#define MAKE_LINK(stct) struct stct *next, *prev

typedef struct HE_EditVert
{
	MAKE_LINK(HE_EditVert);
	float co[3], ssco[3];						  /*subsurf coordinates*/
	float no[3];
	signed short xs[2];
	unsigned int f, f1, f2, f3, h;
	unsigned int glname;
	struct HE_EditEdge *edge;
	double vcol[3];
	short uv_sticky[2];
	float weight;
	char fast, hl;								  /*mouseover highligt*/
	struct HE_EditVert *newvert;
/*this next one is ONLY for edge lookup!  Trust me,
  maintaining adjacency info like this is not necassary
  and very slow!!*/
	ListBase hashedges;							  /*linked list of hash_references*/
} HE_EditVert;

typedef struct HE_Ref
{
	MAKE_LINK(HE_Ref);
	void *ref;
} HE_Ref;

typedef struct HE_EditEdge
{
	MAKE_LINK(HE_EditEdge);
	char fast;
	struct HE_EditEdge *pair;
	struct HE_EditVert *v1, *v2;

	struct HE_EditFace *face;
	struct HE_EditEdge *enext, *eprev;

	struct HE_EditEdgeInfo *ieed;
/** half-edges are invisible to the user, as such they have
  no "offical' flags (e.g. a ->f member).*/
	unsigned int f1, f2, f3, freed, stag; /*stag is used by solid geometry
	                                       *  iterators, NEVER USE IT!*/
	short uv[2];
/** half-edges do have their own normal, though*/
	float no[3];
} HE_EditEdge;

typedef struct HE_HashEdge
{
	MAKE_LINK(HE_HashEdge);
	struct HE_EditEdge *eed;
	char fast;
//struct HE_EditVert *eve;
} HE_HashEdge;

/** no verts are stored in this struct deliberately,
  to avoid abusing it.*/
typedef struct HE_EditEdgeInfo
{
	MAKE_LINK(HE_EditEdgeInfo);
	unsigned int f, f1, f2, h;
	char seam, fast, hl;						  /*mouseover highligt*/
	float crease;

	HE_EditEdge *e1, *e2;
	unsigned int glname;
} HE_EditEdgeInfo;

/** this is for "drawing" edge-loops and then manually
	connecting them*/
/*typedef struct HE_EditDrawEdge {
	Link link;
	struct HE_EditVert *v1, *v2;
	unsigned int f, f1, f2, f3;
	char fast;
} HE_EditDrawEdge;*/

typedef struct HE_EditFace
{
	MAKE_LINK(HE_EditFace);
	struct HE_EditEdge *edge;
	float nor[3];
	float cent[3];
	signed short xs[2];
	unsigned int f, f1, f2, h;
	unsigned int glname;
	char switched, fast, hl;					  /*mouseover highligt*/
	int mat;
} HE_EditFace;

typedef struct HE_EditMesh
{
	ListBase faces;
	ListBase edges;
	ListBase verts;
	ListBase einfo;
	ListBase draw_edges;
/*block-allocated faces, verts, infoedges and half-edges
  this is a little messy, but the speed increase is worth it.*/
	HE_EditFace *blockf; HE_EditVert *blockv;
	HE_EditEdgeInfo *blockie; HE_EditEdge *blockhe;
	HE_HashEdge *blockhash;
	int totbf, totbv, totbie, totbhe;
	int curbf, curbv, curbie, curbhe;
	int totbhash, curbhash;

	struct HE_EditFace *f_mover;				  /*mouse-over face*/
	struct HE_EditEdgeInfo *e_mover;			  /*mouse-over edge*/
	struct HE_EditVert *v_mover;				  /*mouse_over vert*/
	struct SubSurf *ss;
} HE_EditMesh;

typedef struct HE_EditToolElemf
{
	float val;
	float fmin, fmax;
} HE_EditToolElemf;

typedef struct HE_EditToolElemi
{
	int val;
	int imin, imax;
} HE_EditToolElemi;

typedef struct HE_EditToolBox
{
	int type;
	HE_EditToolElemf *fboxes;
	int totfboxes;
	HE_EditToolElemi *iboxes;
	int totiboxes;
} HE_EditToolBox;

typedef struct HE_EditTool
{
	char *name;
	char *arg_validator;
	HE_EditToolBox *tool_box;
	char *(*tool)(char *arg_string, HE_EditMesh *em);
	char keycode;
	short qual;
	int flag;
} HE_EditTool;

extern HE_EditTool he_tool_array[];

typedef struct HE_EditToolStackEntry
{
	MAKE_LINK(HE_EditToolStackEntry);
	HE_EditTool *tool;
	char *arg_string;
	struct heUndoMesh *mesh;
} HE_EditToolStackEntry;

typedef struct HE_EditToolStack
{
	MAKE_LINK(HE_EditToolStack);
	ListBase stack;
} HE_EditToolStack;

struct HE_Mesh *HE_EmeshToHEMesh(struct HE_EditMesh *mesh);
struct HE_EditMesh *HE_HEmeshToEMesh(struct HE_Mesh *mesh);

void addPrimitiveHEMesh(int type);
void free_HEEditMesh(void);
void HE_Border_Select(short min[2], short max[2], int selmode);
void HE_selectionCB(short selecting, struct Object *editobj, short *mval, float rad);

/* tool stuff */
char *HE_ExecToolFromString(char *name, char *arg_string);
void HE_ExecTool(int index, char *arg_string);
char *he_exec_tool(HE_EditTool *tool, char *arg_string, int report);
char *HE_ExecToolFromString_NoReport(char *name, char *arg_string);

/*selection-set function. . .*/
void HE_selectmode_set(void);
void he_clear_all_mover(void);					  /*clear mouse-over highlights*/

/*and lastly, some defines!*/
#define HE_MOVER    64
#define FDIS    75								  /*findnearest dis*/

/*these are for multiple-select mode*/
#define FVDIS   20								  /*vert findnearest dis*/
#define FEDIS   15								  /*edge findnearest dis*/
#define FFDIS   40								  /*face findnearest dis*/
#endif											  /* _BIF_MESHHE */
