/**
 * $Id: buttons_script.c 1461 2005-09-24 19:43:27Z hos $ 
 *
 * ***** BEGIN GPL/BL DUAL LICENSE BLOCK *****
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version. The Blender
 * Foundation also sells licenses for use in proprietary software under
 * the Blender License.  See http://www.blender.org/BL/ for information
 * about this.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software Foundation,
 * Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 *
 * The Original Code is Copyright (C) 2001-2002 by NaN Holding BV.
 * All rights reserved.
 *
 * The Original Code is: all of this file.
 *
 * Contributor(s): none yet.
 *
 * ***** END GPL/BL DUAL LICENSE BLOCK *****
 */

#include <time.h>
#include <math.h>
#include <stdlib.h>
#include <string.h>

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include "MEM_guardedalloc.h"
#include "DNA_action_types.h"
#include "DNA_armature_types.h"
#include "DNA_camera_types.h"
#include "DNA_constraint_types.h"
#include "DNA_curve_types.h"
#include "DNA_effect_types.h"
#include "DNA_group_types.h"
#include "DNA_ID.h"
#include "DNA_image_types.h"
#include "DNA_key_types.h"
#include "DNA_lamp_types.h"
#include "DNA_lattice_types.h"
#include "DNA_material_types.h"
#include "DNA_meta_types.h"
#include "DNA_object_types.h"
#include "DNA_radio_types.h"
#include "DNA_screen_types.h"
#include "DNA_scene_types.h"
#include "DNA_sound_types.h"
#include "DNA_text_types.h"
#include "DNA_texture_types.h"
#include "DNA_userdef_types.h"
#include "DNA_vfont_types.h"
#include "DNA_view3d_types.h"
#include "DNA_world_types.h"
#include "DNA_space_types.h"

#include "BKE_anim.h"
#include "BKE_armature.h"
#include "BKE_constraint.h"
#include "BKE_curve.h"
#include "BKE_displist.h"
#include "BKE_effect.h"
#include "BKE_font.h"
#include "BKE_image.h"
#include "BKE_ipo.h"
#include "BKE_lattice.h"
#include "BKE_main.h"
#include "BKE_material.h"
#include "BKE_mball.h"
#include "BKE_mesh.h"
#include "BKE_object.h"
#include "BKE_sound.h"
#include "BKE_texture.h"
#include "BKE_global.h"
#include "BKE_utildefines.h"

#include "BLI_blenlib.h"
#include "BLI_dynstr.h"

#include "BSE_filesel.h"

#include "BIF_gl.h"
#include "BIF_graphics.h"
#include "BIF_keyval.h"
#include "BIF_mainqueue.h"
#include "BIF_resources.h"
#include "BIF_screen.h"
#include "BIF_mywindow.h"
#include "BIF_space.h"
#include "BIF_glutil.h"
#include "BIF_interface.h"
#include "BIF_toolbox.h"

#include "BIF_butspace.h"

#include "mydevice.h"
#include "blendef.h"
#include "butspace.h"

/* ************************ function prototypes ********************** */
int draw_scriptlink(uiBlock *, ScriptLink *, int , int , int ) ;

/* *************************** SCRIPT ******************************** */


static char *scriptlinks_menu_string(void)
{
	char *menu = NULL;
	DynStr *ds = BLI_dynstr_new();
	Text *text = G.main->text.first;
	int txtcounter = 0;
	
	if (text) {
		BLI_dynstr_append(ds, "Select Script Link%t");
		while (text) {
			BLI_dynstr_append(ds, "|");
			BLI_dynstr_append(ds, text->id.name+2);
			txtcounter += 1;
			text = text->id.next;
		}
		if (txtcounter) menu = BLI_dynstr_get_cstring(ds);
	}
	BLI_dynstr_free(ds);
	return menu;
}

static void scriptlinks_pupmenu(ScriptLink *slink)
{
	short menuitem;
	char *menustr = scriptlinks_menu_string();
	
	if (menustr) {
		menuitem = pupmenu_col(menustr, 20);
		MEM_freeN(menustr);
		if (menuitem > 0) {
			Text *text = G.main->text.first;
			while (--menuitem) text = text->id.next;
			if (text) slink->scripts[slink->totscript - 1]= (ID *)text;
		}
	}
}

static void extend_scriptlink(ScriptLink *slink)
{
	void *stmp, *ftmp;

	if (!slink) return;
		
	stmp= slink->scripts;		
	slink->scripts= MEM_mallocN(sizeof(ID*)*(slink->totscript+1), "scriptlistL");
	
	ftmp= slink->flag;		
	slink->flag= MEM_mallocN(sizeof(short*)*(slink->totscript+1), "scriptlistF");
	
	if (slink->totscript) {
		memcpy(slink->scripts, stmp, sizeof(ID*)*(slink->totscript));
		MEM_freeN(stmp);

		memcpy(slink->flag, ftmp, sizeof(short)*(slink->totscript));
		MEM_freeN(ftmp);
	}

	slink->scripts[slink->totscript]= NULL;
	slink->flag[slink->totscript]= SCRIPT_FRAMECHANGED;

	slink->totscript++;
	slink->actscript = slink->totscript;
}

static void add_scriptlink_func (void *arg1v, void *arg2v)
{
	ScriptLink *slink = arg1v;
	
	extend_scriptlink(slink);
	BIF_undo_push("Add Scriptlink");
	scriptlinks_pupmenu(slink);
	
	allqueue(REDRAWBUTSSCRIPT, 0);
}

static void del_scriptlink_func (void *arg1v, void *arg2v)
{
	ScriptLink *slink = arg1v;
	int delscript = (int) arg2v;

	int i;

	if (!slink) return;
	
	if (slink->totscript>0) {
		for (i=delscript-1; i<slink->totscript-1; i++) {
			slink->flag[i]= slink->flag[i+1];
			slink->scripts[i]= slink->scripts[i+1];
		}
		
		slink->totscript--;
	}
	
	CLAMP(slink->actscript, 1, slink->totscript);
	
	if (slink->totscript==0) {
		if (slink->scripts) MEM_freeN(slink->scripts);
		if (slink->flag) MEM_freeN(slink->flag);
		
		slink->scripts= NULL;
		slink->flag= NULL;
		slink->totscript= slink->actscript= 0;			
	}
	
	BIF_undo_push("Delete Scriptlink");
	allqueue(REDRAWBUTSSCRIPT, 0);
}

static void delete_scriptlink(ScriptLink *slink)
{
	int i;
	
	if (!slink) return;
	
	if (slink->totscript>0) {
		for (i=slink->actscript-1; i<slink->totscript-1; i++) {
			slink->flag[i]= slink->flag[i+1];
			slink->scripts[i]= slink->scripts[i+1];
		}
		
		slink->totscript--;
	}
		
	CLAMP(slink->actscript, 1, slink->totscript);
		
	if (slink->totscript==0) {
		if (slink->scripts) MEM_freeN(slink->scripts);
		if (slink->flag) MEM_freeN(slink->flag);

		slink->scripts= NULL;
		slink->flag= NULL;
		slink->totscript= slink->actscript= 0;			
	}
}

void do_scriptbuts(unsigned short event)
{
	Object *ob=NULL;
	ScriptLink *slink=NULL;
	Material *ma;
	
	switch (event) {
	case B_SSCRIPT_ADD:
		slink = &G.scene->scriptlink;
		extend_scriptlink(slink);
		BIF_undo_push("Add scriptlink");
		scriptlinks_pupmenu(slink);
		break;
	case B_SSCRIPT_DEL:
		BIF_undo_push("Delete scriptlink");
		delete_scriptlink(&G.scene->scriptlink);
		break;
		
	case B_SCRIPT_ADD:
	case B_SCRIPT_DEL:
		ob= OBACT;

		if (ob && G.buts->scriptblock==ID_OB) {
				slink= &ob->scriptlink;

		} else if (ob && G.buts->scriptblock==ID_MA) {
			ma= give_current_material(ob, ob->actcol);
			if (ma)	slink= &ma->scriptlink;

		} else if (ob && G.buts->scriptblock==ID_CA) {
			if (ob->type==OB_CAMERA)
				slink= &((Camera *)ob->data)->scriptlink;

		} else if (ob && G.buts->scriptblock==ID_LA) {
			if (ob->type==OB_LAMP)
				slink= &((Lamp *)ob->data)->scriptlink;

		} else if (G.buts->scriptblock==ID_WO) {
			if (G.scene->world) 
				slink= &(G.scene->world->scriptlink);
		}
		
		if (event==B_SCRIPT_ADD) {
			extend_scriptlink(slink);
			BIF_undo_push("Add scriptlink");
			scriptlinks_pupmenu(slink);
		}
		else {
			delete_scriptlink(slink);
			BIF_undo_push("Delete scriptlink");
		}
		break;
	default:
		break;
	}

	allqueue(REDRAWBUTSSCRIPT, 0);
}

int draw_scriptlink(uiBlock *block, ScriptLink *script, int sx, int sy, int scene) 
{
	char str[256];
	int width = 300;
	uiBut *but = NULL;
	int delscript;

	if (scene) {
		if (script->totscript<32767) 
			uiDefBut(block, BUT, B_SSCRIPT_ADD, "Add Scriptlink", (short)(sx), (short)sy+30, 110, 19, 0, 0, 0, 0, 0, "Add a new Script link");
	} else {
		if (script->totscript<32767) 
			but = uiDefBut(block, BUT, NULL, "Add Scriptlink", (short)(sx), (short)sy+30, 110, 19, 0, 0, 0, 0, 0, "Add a new Script link");
			uiButSetFunc(but, add_scriptlink_func, script, NULL);
	}

	if (script->totscript) {
		script->actscript = 1;
		
		while(script->actscript <= script->totscript) {
	
			strcpy(str, "FrameChanged%x 1|");
			strcat(str, "Redraw%x 4|");
			strcat(str, "Render%x 16|");
			if (scene) {
				strcat(str, "OnLoad%x 2|");
				strcat(str, "OnSave%x 8");
			}

			BIF_ThemeColorShade(TH_HEADER, -35);
			uiSetRoundBox(15);
			uiRoundBox((float)sx-5, (float)sy-6, (float)sx+width+5, (float)sy+17, 13.0);
			
			uiDefButS(block, MENU, 1, str, (short)sx+7, (short)sy, 140, 19, &script->flag[script->actscript-1], 0, 0, 0, 0, "Script links for this event");
			uiDefIDPoinBut(block, test_scriptpoin_but, 1, "", (short)(sx+160),(short)sy, 135, 19, &script->scripts[script->actscript-1], "Name of Script to link");
		 
			but = uiDefIconBut(block, BUT, NULL, ICON_X, (short)sx+300, (short)sy, 19, 19, NULL, 0.0, 0.0, 0.0, 0.0, "Delete scriptlink");

			delscript = (int) script->actscript;
			uiButSetFunc(but, del_scriptlink_func, script, 
						 (void *) delscript);
			
			sy -= 27;
		
			script->actscript += 1;
		}
	}
	return sy;
}

/* ************************************* */

static void  script_panel_scriptlink_scene(void)
{
	uiBlock *block;
	int xco = 10, yco = 204;
	
	block= uiNewBlock(&curarea->uiblocks, "script_panel_scriptlink_scene", UI_EMBOSS, UI_HELV, curarea->win);
	if(uiNewPanel(curarea, block, "Scene Scriptlinks", "Script", 0, 0, 318, 204)==0) return;
	
	/* this is a variable height panel, newpanel doesnt force new size on existing panels */
	/* so first we make it default height */
	uiNewPanelHeight(block, 204);
	
	uiDefBut(block, LABEL, 0, " ", xco,180,319,20, 0, 0, 0, 0, 0, "");
	
	uiPanelPush(block);
	
	yco = draw_scriptlink(block, &G.scene->scriptlink, 10, 150, 1);
	
	if(yco < 0) uiNewPanelHeight(block, 204-yco);
	
	uiPanelPop(block);
}

static void  script_panel_scriptlink_world(void)
{
	uiBlock *block;
	ScriptLink *script=NULL;
	int xco = 10, yco = 204;
	
	block= uiNewBlock(&curarea->uiblocks, "script_panel_scriptlink_world", UI_EMBOSS, UI_HELV, curarea->win);
	uiNewPanelTabbed("Scene Scriptlinks", "Script");
	if(uiNewPanel(curarea, block, "World Scriptlinks", "Script", 0, 0, 318, 204)==0) return;
	
	/* this is a variable height panel, newpanel doesnt force new size on existing panels */
	/* so first we make it default height */
	uiNewPanelHeight(block, 204);
	
	uiDefBut(block, LABEL, 0, " ", xco,180,319,20, 0, 0, 0, 0, 0, "");
	
	if (G.scene->world)
		script= &(G.scene->world->scriptlink);
		
	if (script) {
		uiPanelPush(block);
		yco = draw_scriptlink(block, script, 10, 150, 0);
		uiPanelPop(block);
	}
	
	if(yco < 0) uiNewPanelHeight(block, 204-yco);

}

static void  script_panel_scriptlink_object(void)
{
	uiBlock *block;
	Object *ob=NULL;
	ScriptLink *script=NULL;
	int xco = 10, yco = 204;
	
	block= uiNewBlock(&curarea->uiblocks, "script_panel_scriptlink_object", UI_EMBOSS, UI_HELV, curarea->win);
	if(uiNewPanel(curarea, block, "Object Scriptlinks", "Script", 0, 0, 318, 204)==0) return;
	
	/* this is a variable height panel, newpanel doesnt force new size on existing panels */
	/* so first we make it default height */
	uiNewPanelHeight(block, 204);
	
	ob= OBACT;
	
	uiDefBut(block, LABEL, 0, " ", xco,180,319,20, 0, 0, 0, 0, 0, "");
	
	if(ob) 
		script= &ob->scriptlink;
	
	if (script) {
		uiPanelPush(block);
		yco = draw_scriptlink(block, script, 10, 150, 0);
		uiPanelPop(block);
	}
	
	if(yco < 0) uiNewPanelHeight(block, 204-yco);
}

static void  script_panel_scriptlink_camera(void)
{
	uiBlock *block;
	Object *ob=NULL;
	ScriptLink *script=NULL;
	int xco = 10, yco = 204;
	
	block= uiNewBlock(&curarea->uiblocks, "script_panel_scriptlink_camera", UI_EMBOSS, UI_HELV, curarea->win);
	uiNewPanelTabbed("Object Scriptlinks", "Script");
	if(uiNewPanel(curarea, block, "Camera Scriptlinks", "Script", 0, 0, 318, 204)==0) return;
	
	/* this is a variable height panel, newpanel doesnt force new size on existing panels */
	/* so first we make it default height */
	uiNewPanelHeight(block, 204);
	
	ob= OBACT;
	
	uiDefBut(block, LABEL, 0, " ", xco,180,319,20, 0, 0, 0, 0, 0, "");
	
	if (ob->type==OB_CAMERA)
		script= &((Camera *)ob->data)->scriptlink;
	
	if (script) {
		uiPanelPush(block);
		yco = draw_scriptlink(block, script, 10, 150, 0);
		uiPanelPop(block);
	}
	
	if(yco < 0) uiNewPanelHeight(block, 204-yco);
}

static void  script_panel_scriptlink_lamp(void)
{
	uiBlock *block;
	Object *ob=NULL;
	ScriptLink *script=NULL;
	int xco = 10, yco = 204;
	
	block= uiNewBlock(&curarea->uiblocks, "script_panel_scriptlink_lamp", UI_EMBOSS, UI_HELV, curarea->win);
	uiNewPanelTabbed("Object Scriptlinks", "Script");
	if(uiNewPanel(curarea, block, "Lamp Scriptlinks", "Script", 0, 0, 318, 204)==0) return;
	
	/* this is a variable height panel, newpanel doesnt force new size on existing panels */
	/* so first we make it default height */
	uiNewPanelHeight(block, 204);
	
	ob= OBACT;
	
	uiDefBut(block, LABEL, 0, " ", xco,180,319,20, 0, 0, 0, 0, 0, "");
	
	if (ob->type==OB_LAMP)
		script= &((Lamp *)ob->data)->scriptlink;
	
	if (script) {
		uiPanelPush(block);
		yco = draw_scriptlink(block, script, 10, 150, 0);
		uiPanelPop(block);
	}
	
	if(yco < 0) uiNewPanelHeight(block, 204-yco);
}

static void  script_panel_scriptlink_material(void)
{
	uiBlock *block;
	Object *ob=NULL;
	Material *ma;
	ScriptLink *script=NULL;
	int xco = 10, yco = 204;
	
	block= uiNewBlock(&curarea->uiblocks, "script_panel_scriptlink_material", UI_EMBOSS, UI_HELV, curarea->win);
	uiNewPanelTabbed("Object Scriptlinks", "Script");
	if(uiNewPanel(curarea, block, "Material Scriptlinks", "Script", 0, 0, 318, 204)==0) return;
	
	/* this is a variable height panel, newpanel doesnt force new size on existing panels */
	/* so first we make it default height */
	uiNewPanelHeight(block, 204);
	
	ob= OBACT;
	
	uiDefBut(block, LABEL, 0, " ", xco,180,319,20, 0, 0, 0, 0, 0, "");
	
	ma= give_current_material(ob, ob->actcol);
	
	if (ma)
		script= &ma->scriptlink;
	
	if (script) {
		uiPanelPush(block);
		yco = draw_scriptlink(block, script, 10, 150, 0);
		uiPanelPop(block);
	}
	
	if(yco < 0) uiNewPanelHeight(block, 204-yco);
}

void script_panels()
{
	Object *ob=NULL;
	ob = OBACT;
	
	script_panel_scriptlink_scene();
	script_panel_scriptlink_world();
	
	if(ob) {
		script_panel_scriptlink_object();
	
		if(give_current_material(ob, ob->actcol))
			script_panel_scriptlink_material();
		else if(ob->type==OB_CAMERA)
			script_panel_scriptlink_camera();
		else if(ob->type==OB_LAMP)
			script_panel_scriptlink_lamp();
	}
}
