#include <string.h>
#include <math.h>

#ifdef WIN32
#include "BLI_winstuff.h"
#endif

#include "MEM_guardedalloc.h"

#include "BLI_blenlib.h"
#include "BLI_arithb.h"
#include "MTC_matrixops.h"

#include "IMB_imbuf_types.h"

#include "DNA_image_types.h"
#include "DNA_lamp_types.h"
#include "DNA_material_types.h"
#include "DNA_hemesh_types.h"
#include "DNA_object_types.h"
#include "DNA_property_types.h"
#include "DNA_scene_types.h"
#include "DNA_screen_types.h"
#include "DNA_view3d_types.h"
#include "DNA_meshdata_types.h"

#include "BKE_bmfont.h"
#include "BKE_displist.h"
#include "BKE_effect.h"
#include "BKE_global.h"
#include "BKE_image.h"
#include "BKE_main.h"
#include "BKE_material.h"
#include "BKE_mesh.h"
#include "BKE_object.h"
#include "BKE_property.h"
#include "BKE_utildefines.h"
#include "BKE_dmesh.h"

#include "BIF_resources.h"
#include "BIF_gl.h"
#include "BIF_glutil.h"
#include "BIF_mywindow.h"
#include "BIF_resources.h"
#include "BIF_meshHE.h"
#include "BIF_screen.h"

#include "BDR_editface.h"
#include "BDR_vpaint.h"
#include "BDR_drawmesh.h"

#include "BSE_drawview.h"
#include "BSE_view.h"

#include "blendef.h"

extern ScrArea *curarea;

void HEE_drawHEEditMesh(unsigned int dt);
static void hee_draw_faces_for_picking(void);
void he_draw_edges(unsigned char alpha);

/*1 = backbuffered faces, 2 = black, 3=normal draw*/
void HED_drawSubSurfFaces(char mode)
{
	HE_EditMesh *em = G.he_editMesh;
	HE_EditFace *efa;
	HE_Mesh *hme = G.obedit->data;
	DMesh *dm = hme->dmesh;
	DFace *df;
	DVert *dv;
	int *verts;
	int i, a;

	for (efa=em->faces.first, a=1; efa; efa=efa->next)
	{
		efa->f1 = a;
		a++;
	}
	if (mode==1 || mode==2) glDisable(GL_LIGHTING);
	if (mode==2) glColor4ub(0, 0, 0, 0);

	if (mode>0)
	{
		a = 0;
		df = dm->faces;
		for (i=0; i<dm->totface; i++, df++)
		{
			if (df->ref)
			{
				efa = (HE_EditFace*) df->ref;
				if (mode==1) {set_framebuffer_index_color(efa->f1);}
			}

			glBegin(GL_POLYGON);
			verts = &dm->array->array[df->verts];
			for (a=0; a<df->vlen; a++)
			{
				dv = &dm->verts[verts[a]];
				glVertex3fv(dv->co);
			}
			glEnd();
		}
	}
}


void HED_drawSubSurfEdges(char mode, char alpha)
{
	HE_EditMesh *em = G.he_editMesh;
	HE_EditEdgeInfo *ieed;
	HE_Mesh *hme = G.obedit->data;
	DMesh *dm = hme->dmesh;
	DEdge *de;
	DVert *dv1, *dv2;
	unsigned char selcol[3];
	unsigned char hlight[3];
	unsigned char wirecol[3];
	int i, a;

/*draw edges. . .*/
	bglPolygonOffset(1.0);
//glGetFloatv(GL_LINE_WIDTH, widbuffer);
//wid = widbuffer[0];
	glLineWidth(1);
	BIF_GetThemeColor3ubv(TH_EDGE_SELECT, selcol);
	BIF_GetThemeColor3ubv(TH_EDGEMOVER, hlight);
	BIF_GetThemeColor3ubv(TH_WIRE, wirecol);

	glBegin(GL_LINES);
	a = 1;
	for (ieed=em->einfo.first; ieed; ieed=ieed->next)
	{
		ieed->f1 = a;
		a++;
	}

	de = dm->edges;
	for (i=0; i<dm->totedge; i++, de++)
	{
		if (de->ref)
		{
			ieed = (HE_EditEdgeInfo*) de->ref;
			if (mode != 1)
			{
				if (ieed->f & SELECT)
				{
					if (ieed->hl==1)			  /*draw blended highlight*/
					{
						glColor4ub((hlight[0]-selcol[0])/3, (hlight[1]+selcol[1])/3, (hlight[2]+selcol[2])/3, alpha);
					}							  /*draw normal blue highlight*/
					else if (ieed->hl==2)
					{
						glColor4ub(hlight[0], hlight[1], hlight[2], alpha);
					}
					else
					{
						glColor4ub(selcol[0], selcol[1], selcol[2], alpha);
					}
				} else if (ieed->hl) glColor4ub(hlight[0], hlight[1], hlight[2], alpha);
				else glColor4ub(wirecol[0], wirecol[1], wirecol[2], alpha);
			}
			else
			{
				set_framebuffer_index_color(ieed->f1);
			}

			dv1 = &dm->verts[de->v1];
			dv2 = &dm->verts[de->v2];

			glVertex3fv(dv1->co);
			glVertex3fv(dv2->co);
		}
	}
	glEnd();
}


/** Draw a solid HE EditMesh
	Mode 0 just draws it normally.
	Mode 1 draws it for face backbuffer sel, with indices encoded
	 as three-byte integers inside the colors.
	Mode 2 draws it for backbuffer sel, but with just black (no
	 face indices), for vert and edge backbuffered sel.
*/
void drawHEEMeshSolid(char mode)
{
	HE_EditMesh *em = G.he_editMesh;
	HE_Mesh *me = G.obedit->data;
	Material *mat;
	HE_EditFace *efa;
	HE_EditEdge *eed;
	float col[4];
	unsigned int i;
	int subsurf=0;

	if (!em) return;

	if ((me->f & HEM_SUBSURF) && me->subdivLevels > 0) subsurf = 1;
	if (G.f & G_EDITDEFORM) subsurf = 1;
	if (subsurf)
	{
		HED_drawSubSurfFaces(mode);
		return;
	}

	glEnable(GL_LIGHTING);
/*these make occluded wireframe drawing, but. . .why?*/
//glEnable(GL_BLEND);
//glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
	glDisable(GL_CULL_FACE);
	if (mode != 2)
		glColor4f(0.5, 0.5, 0.5, 0.5);
	else
	{
		glColor3f(0, 0, 0);
		glDisable(GL_LIGHTING);
	}

	for (efa=em->faces.first, i=0; efa; i++, efa=efa->next)
	{
		if (efa->h) continue;
		eed = efa->edge;
		if (mode==1)
		{
			glDisable(GL_LIGHTING);
			set_framebuffer_index_color(i+1);
		}
		if (mode != 2) glNormal3fv(efa->nor);
		if (mode == 0)
		{
			if (efa->mat < me->totcol)
			{
				mat = me->mat[efa->mat];
				col[0] = mat->r; col[1] = mat->g;
				col[2] = mat->b; col[3] = 1;
				glMaterialfv(GL_FRONT_AND_BACK, GL_DIFFUSE, col);
			}
			if (efa->f & HEM_SMOOTH) glShadeModel(GL_SMOOTH);
			else glShadeModel(GL_FLAT);
		}

		glBegin(GL_POLYGON);					  //GL_TRIANGLE_FAN);
//if (mode==0 && (efa->f & HEM_SMOOTH)) glNormal3fv(efa->nor);
//glVertex3fv(efa->cent);
		do
		{
			if (mode == 0 && (efa->f & HEM_SMOOTH))
			{
//if (eed->ieed->f & HEM_HARD) glNormal3fv(efa->nor);
				glNormal3fv(eed->v1->no);
//glNormal3fv(eed->no);
			}
			glVertex3fv(eed->v1->co);
			eed=eed->enext;
		} while (eed != efa->edge);
		glVertex3fv(eed->v1->co);
		glEnd();
	}

	glDisable(GL_LIGHTING);
	glDisable(GL_CULL_FACE);
	glDisable(GL_BLEND);
}


void drawMEWire(HE_Mesh *me)
{
	HE_Face *mef;
	HE_Edge *med;
	HE_Vert *v1, *v2;
	int i;
	glBegin(GL_LINES);
	for (i=0; i<me->totfaces; i++)
	{
		mef = &me->faces[i];
		med = &me->half_edges[me->faces[i].edge];
		do
		{
			v1 = &me->verts[med->v1];
			v2 = &me->verts[med->v2];
			glVertex3fv(v1->co);
			glVertex3fv(v2->co);
			med = &me->half_edges[med->next];
		} while (med != &me->half_edges[me->faces[i].edge]);
	}
	glEnd();
}


void drawMESolid(HE_Mesh *me)
{
	HE_Vert *v1;
	HE_Edge *med;
	HE_Face *mfa;
	Material *mat;
	float col[4];
	int i=0;

	glEnable(GL_LIGHTING);
	glDisable(GL_CULL_FACE);
	glColor3f(0.5, 0.5, 0.5);

	for (i=0; i<me->totfaces; i++)
	{
		mfa = &me->faces[i];
		glNormal3fv(mfa->nor);

/*choose shade model*/
		if (mfa->f & HEM_SMOOTH) glShadeModel (GL_SMOOTH);
		else glShadeModel (GL_FLAT);
		if (mfa->mat_index < me->totcol)
		{
			mat = me->mat[mfa->mat_index];
			col[0] = mat->r; col[1] = mat->g;
			col[2] = mat->b; col[3] = 1;
			glMaterialfv(GL_FRONT_AND_BACK, GL_DIFFUSE, col);
		}

		glBegin(GL_POLYGON);
		med = &(me->half_edges[me->faces[i].edge]);
		do
		{
			v1 = &(me->verts[med->v1]);
			if (mfa->f & HEM_SMOOTH)
			{
// mied = &me->iedges[med->iedge];
//if (mied->f & HEM_HARD) glNormal3fv(me->faces[i].nor);
				glNormal3fv(v1->no);
/*we store shade-normals in half-edge now*/
//glNormal3fv(med->no);
			}
			glVertex3fv(v1->co);
			med = &(me->half_edges[med->next]);
		} while (med != &(me->half_edges[me->faces[i].edge]));
		glEnd();
	}
	glDisable(GL_LIGHTING);
	glDisable(GL_CULL_FACE);
}


void he_draw_displist(Object *ob, unsigned int dt)
{
	HE_Mesh *me = ob->data;
	DMesh *dm;

	if (!me->dmesh) makeDispListHEMesh(ob); 

	dm = me->dmesh;
	bw_assert(dm);
	
	if (!dm) return;
	
	if ((dt<=OB_WIRE) && (ob==G.obedit)) return;
	draw_DMesh(dm, dt);
}


void drawHEMesh(Object *ob, unsigned int dt)
{
	int subsurf = 0;
	Material *ma;
	HE_Mesh *me=ob->data;
	float spec[4], diff[4];

	me->f |= HEM_TWOSIDE;						  /*quick hack. . .*/

	glDisable(GL_CULL_FACE);
	if (dt >= OB_SOLID) glEnable(GL_LIGHTING);
	glLightModeli(GL_LIGHT_MODEL_TWO_SIDE, me->f & HEM_TWOSIDE);
	if ((me->f & HEM_SUBSURF) && me->subdivLevels > 0) subsurf = 1;
	if (G.f & G_EDITDEFORM) subsurf = 1;

	if (dt > OB_WIRE)
	{
		if (me->mat) ma = me->mat[0];
		else ma = G.main->mat.first;
		if (ma)
		{
			diff[0] = (ma->ref+ma->emit)*ma->r;
			diff[1] = (ma->ref+ma->emit)*ma->g;
			diff[2] = (ma->ref+ma->emit)*ma->b;
			diff[3] = 1.0;

			spec[0] = ma->spec*ma->specr;
			spec[1] = ma->spec*ma->specr;
			spec[2] = ma->spec*ma->specr;
			spec[3] = 1.0;

			glMaterialfv(GL_FRONT_AND_BACK, GL_DIFFUSE, diff);
			glMaterialfv(GL_FRONT_AND_BACK, GL_SPECULAR, spec);
		}
	}

	if (ob==G.obedit)
	{
/*eventually this will be how occluded
  wireframe will work.  something to do with drawing
  with alpha 0--not really sure what, but it occludes edges
  while maintaining background transparency*/
//if (dt < OB_SOLID) {
/*occluded wireframe*/
//glEnable(GL_BLEND);
//glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
//}
		if (!subsurf)
		{
			if (dt >= OB_SOLID) drawHEEMeshSolid(0);
		}
		else
		{
			he_draw_displist(ob, dt);
		}
		HEE_drawHEEditMesh(dt);
	}
	else if (G.obedit && ob->data == G.obedit->data)
	{
		if (!subsurf)
		{
			if (dt >= OB_SOLID) drawHEEMeshSolid(0);
			else if (dt < OB_SOLID) he_draw_edges(255);
		} else he_draw_displist(ob, dt);
	} else he_draw_displist(ob, dt);
	glDisable(GL_LIGHTING);
}


void calc_hemesh_vertdots_ext(void)
{
	HE_EditMesh *em = G.he_editMesh;
	HE_Mesh *me = G.obedit->data;
	HE_EditVert *eve;
	float mat[4][4];
	float co[3];
	int subsurf = 0;

	if ((me->f & HEM_SUBSURF) && me->subdivLevels > 0) subsurf = 1;

	if(em->verts.first==NULL) return;
	eve= em->verts.first;

	areawinset(curarea->win);
	persp(PERSP_VIEW);

	mymultmatrix(G.obedit->obmat);

	MTC_Mat4SwapMat4(G.vd->persmat, mat);
	mygetsingmatrix(G.vd->persmat);

	eve= em->verts.first;
	while(eve)
	{
		if (subsurf) project_float(eve->ssco, co);
		else project_float(eve->co, co);		  /*to fix off-screen exclusion problems. . .*/
		eve->xs[0] = (signed short) co[0];
		eve->xs[1] = (signed short) co[1];
		eve= eve->next;
	}
	MTC_Mat4SwapMat4(G.vd->persmat, mat);

	myloadmatrix(G.vd->viewmat);
}


void calc_hemesh_facedots_ext(void)
{
	HE_EditMesh *em = G.he_editMesh;
	HE_EditFace *efa;
	HE_EditVert *eve;
	HE_EditEdge *eed;
	float mat[4][4];
	float co[3];

	if(em->faces.first==NULL) return;
	efa= em->faces.first;

	areawinset(curarea->win);
	persp(PERSP_VIEW);

	mymultmatrix(G.obedit->obmat);

	MTC_Mat4SwapMat4(G.vd->persmat, mat);
	mygetsingmatrix(G.vd->persmat);

	efa= em->faces.first;
	for (eve=em->verts.first; eve; eve=eve->next)
	{
		eve->f2 = 0;
	}

	while(efa)
	{
		float vec[3];							  // = G.vd->viewinv[2];
		VECCOPY(vec, G.vd->viewinv[2]);
		eed = efa->edge;

		do
		{
			eed=eed->enext;
		} while (eed != efa->edge);

		if( efa->h==0)
		{
			project_float(efa->cent, co);
			efa->xs[0] = (short) co[0];
			efa->xs[1] = (short) co[1];
		}
		efa= efa->next;
	}
	MTC_Mat4SwapMat4(G.vd->persmat, mat);

	myloadmatrix(G.vd->viewmat);
}


static void hee_draw_faces_for_picking(void)
{
	HE_EditMesh *em = G.he_editMesh;
	HE_EditFace *efa;
	HE_EditEdge *eed;
	int name = 1;
//glEnable(GL_CULL_FACE);
	glPointSize(50);
	for (efa=em->faces.first; efa; efa=efa->next)
	{
		efa->glname = name;
		glLoadName(name++);
		bglBegin(GL_POINTS);
		bglVertex3fv(efa->cent);
		bglEnd();
//glColor3f(0, 0, 0);
		glBegin(GL_LINE_LOOP);
		eed = efa->edge;
		do
		{
			glVertex3fv(eed->v1->co);
//glVertex3fv(eed->v2->co);
			eed=eed->enext;
		} while (eed != efa->edge);
		glEnd();
	}
}


void he_draw_verts(char alpha)
{
	HE_EditMesh *em = G.he_editMesh;
	HE_Mesh *me = G.obedit->data;
	HE_EditVert *eve;
	float vsize;
	unsigned char col[3], selcol[3];
	unsigned char hlight[3];
	int i=0, subsurf=0;

	if (!em) return;

	BIF_GetThemeColor3ubv(TH_VERTEX, col);
	BIF_GetThemeColor3ubv(TH_VERTEX_SELECT, selcol);
	BIF_GetThemeColor3ubv(TH_VERTMOVER, hlight);

	vsize= BIF_GetThemeValuef(TH_VERTEX_SIZE);
	if ((me->f & HEM_SUBSURF) && me->subdivLevels > 0) subsurf = 1;
	if (G.f & G_EDITDEFORM) subsurf = 1;
	bglPolygonOffset(1.0);
	glPointSize(vsize);

	for (eve=em->verts.first, i=0; eve; i++, eve=eve->next)
	{
/*we blend/average selection and highlight colors. . .*/
		if (eve->h) continue;
		if (eve->f & SELECT)
		{
			if (eve == em->v_mover && (G.f & G.f & G_HEMOVER)) glColor4ub(((char)hlight[0]-selcol[0])/2, ((char)hlight[1]+selcol[1])/2, ((char)hlight[2]+selcol[2])/2, alpha);
			else glColor4ub(selcol[0], selcol[1], selcol[2], 100);
		} else if (eve==em->v_mover) glColor4ub((char)hlight[0], (char)hlight[1], (char)hlight[2], alpha);
		else glColor4ub(col[0], col[1], col[2], 100);

		bglBegin(GL_POINTS);
		if (subsurf) bglVertex3fv(eve->ssco);
		else bglVertex3fv(eve->co);
		bglEnd();
	}

}


void he_draw_edges(unsigned char alpha)
{
	HE_EditMesh *em = G.he_editMesh;
	HE_Mesh *me = G.obedit->data;
	HE_EditEdgeInfo *ieed;
	float vco1[3], vco2[3];
	unsigned char hlight[4];
	unsigned char wirecol[3];
	unsigned char selcol[3];
	int subsurf=0;

	if (!em) return;

	if ((me->f & HEM_SUBSURF) && me->subdivLevels > 0) subsurf = 1;
	if (G.f & G_EDITDEFORM) subsurf = 1;
	if (subsurf)
	{
		HED_drawSubSurfEdges(0, alpha);
		return;
	}
/*draw edges. . .*/
	bglPolygonOffset(1.0);
	BIF_GetThemeColor3ubv(TH_EDGE_SELECT, selcol);
	BIF_GetThemeColor3ubv(TH_EDGEMOVER, hlight);
	BIF_GetThemeColor3ubv(TH_WIRE, wirecol);

	for (ieed=em->einfo.first; ieed; ieed=ieed->next)
	{
		if (ieed->h) continue;
		if (ieed->f & SELECT)
		{
												  /*draw blended highlight*/
			if (ieed->hl==1 && (G.f & G.f & G_HEMOVER))
			{
				glColor4ub((hlight[0]-selcol[0])/3, (hlight[1]+selcol[1])/3, (hlight[2]+selcol[2])/3, alpha);
			}									  /*draw normal blue highlight*/
			else if (ieed->hl==2 && (G.f & G.f & G_HEMOVER))
			{
				glColor4ub(hlight[0], hlight[1], hlight[2], alpha);
			}
			else
			{
				glColor4ub(selcol[0], selcol[1], selcol[2], alpha);
			}
		} else if (ieed->hl) glColor4ub(hlight[0], hlight[1], hlight[2], alpha);
		else glColor4ub(wirecol[0], wirecol[1], wirecol[2], alpha);

		if (!ieed->e1) continue;
		glBegin(GL_LINES);
		vco1[0] = ieed->e1->v1->co[0];
		vco1[1] = ieed->e1->v1->co[1];
		vco1[2] = ieed->e1->v1->co[2];

		vco2[0] = ieed->e1->enext->v1->co[0];
		vco2[1] = ieed->e1->enext->v1->co[1];
		vco2[2] = ieed->e1->enext->v1->co[2];

		bglVertex3fv(vco1);
		bglVertex3fv(vco2);
		glEnd();
	}
}


/*draws the nice little transparent faces*/
static void he_draw_subsurf_faces(unsigned char alpha)
{
	HE_Mesh *me = G.obedit->data;
	HE_EditMesh *em = G.he_editMesh;
	HE_EditFace *efa;
	DMesh *dm = me->dmesh;
	DFace *df;
	DVert *dv;
	int *verts;
	int i;
	unsigned char mfcol[4];
	unsigned char mfcol_sel[4];
	unsigned char fselcol[4];
	unsigned char fcol[4];
	int j;

	if (!G.obedit || !G.he_editMesh || !dm) return;

	BIF_GetThemeColor4ubv(TH_FACE_SELECT, fselcol);
	BIF_GetThemeColor4ubv(TH_FACEMOVER, mfcol);
	BIF_GetThemeColor4ubv(TH_FACEMOVER_SEL, mfcol_sel);
	BIF_GetThemeColor4ubv(TH_FACE, fcol);

	fselcol[3] -= 255 - alpha;
	fcol[3] -= 255 - alpha;

	glEnable(GL_BLEND);
	glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
	glDisable(GL_LIGHTING);
	glDepthMask(0);
/*draw faces*/
	df = dm->faces;
	for (i=0; i<dm->totface; i++, df++)
	{
		if (df->ref)
		{
			efa = (HE_EditFace*) df->ref;
			if (efa == em->f_mover && (G.f & G.f & G_HEMOVER))
			{
				if (efa->f & SELECT) glColor4ub(mfcol_sel[0], mfcol_sel[1], mfcol_sel[2], mfcol_sel[3]);
				else glColor4ub(mfcol[0], mfcol[1], mfcol[2], mfcol[3]);
			} else if (efa->f & SELECT) glColor4ub(fselcol[0], fselcol[1], fselcol[2], fselcol[3]);
			else glColor4ub(fcol[0], fcol[1], fcol[2], fcol[3]);
			verts = &dm->array->array[df->verts];
			glBegin(GL_POLYGON);
			for (j=0; j<df->vlen; j++)
			{
				dv = &dm->verts[verts[j]];
				glVertex3fv(dv->co);
			}
			glEnd();
		}
	}
	glDisable(GL_BLEND);
	glDepthMask(1);
}


static void he_draw_trans_faces(unsigned char alpha)
{
	HE_EditMesh *em = G.he_editMesh;
	HE_EditFace *efa;
	HE_EditEdge *eed;
	HE_Mesh *me = G.obedit->data;
	unsigned char fselcol[4];
	unsigned char fcol[4];
	unsigned char mfcol[4];
	unsigned char mfcol_sel[4];

/*do subsurf. . .*/
	if ((me->f & HEM_SUBSURF) && me->subdivLevels>0)
	{
		he_draw_subsurf_faces(alpha);
		return;
	}

	if (!em) return;

	BIF_GetThemeColor4ubv(TH_FACE_SELECT, fselcol);
	BIF_GetThemeColor4ubv(TH_FACEMOVER, mfcol);
	BIF_GetThemeColor4ubv(TH_FACEMOVER_SEL, mfcol_sel);
	BIF_GetThemeColor4ubv(TH_FACE, fcol);
	fselcol[3] -= 255 - alpha;
	fcol[3] -= 255 - alpha;

	glEnable(GL_BLEND);
	glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
	glDisable(GL_LIGHTING);
	glDepthMask(0);

	for (efa=em->faces.first; efa; efa=efa->next)
	{
		if (efa->h) continue;
		if (efa == em->f_mover && (G.f & G.f & G_HEMOVER))
		{
			if (efa->f & SELECT) glColor4ub(mfcol_sel[0], mfcol_sel[1], mfcol_sel[2], mfcol_sel[3]);
			else glColor4ub(mfcol[0], mfcol[1], mfcol[2], mfcol[3]);
		} else if (efa->f & SELECT) glColor4ub(fselcol[0], fselcol[1], fselcol[2], fselcol[3]);
		else glColor4ub(fcol[0], fcol[1], fcol[2], fcol[3]);
		glBegin(GL_POLYGON);
		eed = efa->edge;
		do
		{
			glVertex3fv(eed->v1->co);
			eed=eed->enext;
		} while (eed != efa->edge);
		glEnd();
	}
	glDisable(GL_BLEND);
	glDepthMask(1);
}


void he_draw_facedots(unsigned char alpha)
{
	HE_EditMesh *em = G.he_editMesh;
	HE_EditFace *efa;

	float fsize;
	int i=0;
	char fselcol[3];

	BIF_GetThemeColor3ubv(TH_FACE_DOT, fselcol);

	fsize= BIF_GetThemeValuef(TH_FACEDOT_SIZE);
	glPointSize(fsize);
	for (efa=em->faces.first, i=0; efa; i++, efa=efa->next)
	{
		if (efa->h) continue;
		if (efa->f & SELECT)
		{
			glColor4ub(fselcol[0], fselcol[1], fselcol[2], alpha);
		} else
		glColor4ub(0, 0, 0, alpha);

		bglBegin(GL_POINTS);
		bglVertex3fv(efa->cent);
		bglEnd();
	}
}


void HEE_drawHEEditMesh(unsigned int dt)
{
	HE_Mesh *hem = G.obedit->data;
	HE_EditMesh *em = G.he_editMesh;
	HE_EditVert *eve;
	HE_EditFace *efa;
	int subsurf = 0;

	glDisable(GL_LIGHTING);
	if (hem->f & HEM_SUBSURF && hem->subdivLevels > 0) subsurf = 1;
/*draw edges first. . .*/
	he_draw_edges(255);

/* draw verts. . .*/
	if(G.scene->selectmode & SCE_SELECT_VERTEX)
	{
		he_draw_verts(255);
	}

/*draw face dots. . .*/
/*draw tranparency*/
	if(G.scene->selectmode & SCE_SELECT_FACE)
	{
		glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
		glEnable(GL_BLEND);
		he_draw_facedots(100);
		glDisable(GL_BLEND);
	}

/*draw face normal lines. . .*/
	if (G.f & G_DRAWNORMALS)
	{
		float nor[3];

		glBegin(GL_LINES);
		for (efa=em->faces.first; efa; efa=efa->next)
		{
			glColor3f(0.7, 0.1, 0.7);
			VecCopyf (nor, efa->nor);
			VecMulf(nor, 0.16);
			glVertex3fv(efa->cent);
			glVertex3f(efa->cent[0]+nor[0], efa->cent[1]+nor[1], efa->cent[2]+nor[2]);
		}
		glEnd();

/*draw vert normal lines. . .*/
		glBegin(GL_LINES);
		for (eve=em->verts.first; eve; eve=eve->next)
		{
			glColor3f(0.7, 0.1, 0.7);
			VecCopyf (nor, eve->no);
			VecMulf(nor, 0.333);
			if (subsurf)
			{
				glVertex3fv(eve->ssco);
				glVertex3f(eve->ssco[0]+nor[0], eve->ssco[1]+nor[1], eve->ssco[2]+nor[2]);
			}
			else
			{
				glVertex3fv(eve->co);
				glVertex3f(eve->co[0]+nor[0], eve->co[1]+nor[1], eve->co[2]+nor[2]);
			}
		}
		glEnd();
	}

/*draw transparent verts/wires/facedots, if not in backbuffered select mode. . .*/
	if ((G.vd->flag & V3D_ZBUF_SELECT) == 0)
	{
		glDisable(GL_DEPTH_TEST);
		glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
		glEnable(GL_BLEND);

		he_draw_edges(100);
		if (G.scene->selectmode & SCE_SELECT_VERTEX) he_draw_verts(100);
		if (G.scene->selectmode & SCE_SELECT_FACE) he_draw_facedots(100);
		if( (G.f & (G_FACESELECT+G_DRAWFACES))) he_draw_trans_faces(255);

		glEnable(GL_DEPTH_TEST);
	}

	if( (G.f & (G_FACESELECT+G_DRAWFACES))) he_draw_trans_faces(255);
	glEnable(GL_LIGHTING);
}
