/**
 * $Id: unwrapper.c 1425 2005-08-25 02:11:41Z hos $
 *
 * ***** BEGIN GPL/BL DUAL LICENSE BLOCK *****
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version. The Blender
 * Foundation also sells licenses for use in proprietary software under
 * the Blender License.  See http://www.blender.org/BL/ for information
 * about this.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software Foundation,
 * Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 *
 * The Original Code is Copyright (C) 2001-2002 by NaN Holding BV.
 * All rights reserved.
 *
 * The Original Code is: all of this file.
 *
 * Contributor(s): none yet.
 *
 * ***** END GPL/BL DUAL LICENSE BLOCK *****
 */

#include <string.h>
#include <stdlib.h>
#include <math.h>

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include "blendef.h"
#include "mydevice.h" 	 
 
#include "MEM_guardedalloc.h"

#include "DNA_mesh_types.h"
#include "DNA_meshdata_types.h"
#include "DNA_scene_types.h"
#include "DNA_screen_types.h"
#include "DNA_space_types.h"

#include "BKE_utildefines.h"
#include "BKE_global.h"
#include "BKE_mesh.h"

#include "BLI_arithb.h"
#include "BLI_rand.h"
#include "BLI_linklist.h"
#include "BLI_memarena.h"
#include "BLI_edgehash.h"

#include "BIF_editsima.h"
#include "BIF_space.h"
#include "BIF_interface.h"
#include "BIF_editsima.h"
#include "BIF_toolbox.h"
#include "BIF_screen.h"

#include "BDR_editface.h"

#include "PIL_time.h"

#include "MEM_guardedalloc.h"

#include "ONL_opennl.h"

/* ************************** STRUCTS **************************** */

typedef struct _UvVert UvVert;
typedef struct _UvEdge UvEdge;
typedef struct _UvFace UvFace;
typedef struct _UvIsland UvIsland;
typedef struct _UvAtlas UvAtlas;

typedef enum {
	ATLAS_SEAMS,
	ATLAS_IMPLICIT_SELECTED,
	ATLAS_IMPLICIT_ALL
} AtlasType;

struct _UvAtlas {
	UvIsland *islands;

	MemArena *arena;
};

#define UVISLAND_NOPACK 1
#define UVISLAND_LIVE_UNWRAP 2
#define UVISLAND_DO_OPTIMIZE 4

struct _UvIsland {
	UvIsland *next;

	int nverts, nedges, nfaces;

	UvVert *verts;
	UvEdge *edges;
	UvFace *faces;

	float min[3], max[3], center[3];
	float radius, scale;

	short flag;

	MemArena *arena;
	Mesh *me;
};

#define UVVERT_PINNED 1
#define UVVERT_TFACE_PINNED 2
#define UVVERT_SPLIT 4
#define UVVERT_SELECTED 8

struct _UvVert {
	UvVert *next;

	LinkNode *edges;
	LinkNode *faces;

	float uv[2], old_uv[2], *co;

	short flag;
	union {
		float distortion;
		float stretch;
		float angle;
		int index;
		int totpinned;
	} u;

	struct {
		int index;
	} orig;
};

#define UVEDGE_SEAM 1
#define UVEDGE_TFACE_SEAM 2
#define UVEDGE_VISITED 4
#define UVEDGE_SPLIT 8

struct _UvEdge {
	UvEdge *next;

	UvVert *v[2];
	LinkNode *faces;

	short flag;
	union {
		float distortion;
		int totedge;
	} u;
};

#define UVFACE_SPLIT 1
#define UVFACE_VERTEX_SPLIT 2
#define UVFACE_FAKE 4

struct _UvFace {
	UvFace *next;

	UvVert *v[3];
	UvEdge *e[3];

	short flag;
	union {
		float stretch;
		float *tface_uv;
	} u;

	struct {
		char findex[3];
		int index;
	} orig;
};

/* ************************ CONSTRUCTION ************************* */

static UvVert *uv_island_add_vert(UvIsland *island, float *co)
{
	UvVert *v = BLI_memarena_alloc(island->arena, sizeof(UvVert));

	v->edges = NULL;
	v->faces = NULL;

	v->uv[0] = v->uv[1] = 0.0f;
	v->co = co;
	v->flag = 0;

	v->next = island->verts;
	island->verts = v;
	island->nverts++;

	return v;
}

static UvEdge *uv_island_find_edge(UvVert *v1, UvVert *v2)
{
    LinkNode *link;

	for (link= v1->edges; link; link= link->next) {
		UvEdge *e= (UvEdge*)link->link;

		if (e->v[0] == v2 || e->v[1] == v2)
			return e;
	}

	return NULL;
}

static UvEdge *uv_island_add_edge(UvIsland *island, UvVert *v1, UvVert *v2)
{
	UvEdge *e = uv_island_find_edge(v1, v2);

	if (e)
		return e;

	e = BLI_memarena_alloc(island->arena, sizeof(UvEdge));

	BLI_linklist_prepend_arena(&v1->edges, e, island->arena);
	BLI_linklist_prepend_arena(&v2->edges, e, island->arena);

	e->v[0] = v1;
	e->v[1] = v2;

	e->faces = NULL;
	e->flag = 0;

	e->next = island->edges;
	island->edges = e;
	island->nedges++;

	return e;
}

static UvFace *uv_island_add_face(UvIsland *island, UvVert *v1, UvVert *v2, UvVert *v3)
{
	UvFace *f;
	UvEdge *e1, *e2, *e3;

	f = BLI_memarena_alloc(island->arena, sizeof(UvFace));

	BLI_linklist_prepend_arena(&v1->faces, f, island->arena);
	BLI_linklist_prepend_arena(&v2->faces, f, island->arena);
	BLI_linklist_prepend_arena(&v3->faces, f, island->arena);

	f->v[0] = v1;
	f->v[1] = v2;
	f->v[2] = v3;

	e1 = uv_island_add_edge(island, v1, v2);
	e2 = uv_island_add_edge(island, v2, v3);
	e3 = uv_island_add_edge(island, v3, v1);

	BLI_linklist_prepend_arena(&e1->faces, f, island->arena);
	BLI_linklist_prepend_arena(&e2->faces, f, island->arena);
	BLI_linklist_prepend_arena(&e3->faces, f, island->arena);

	f->e[0] = e1;
	f->e[1] = e2;
	f->e[2] = e3;

	f->flag = 0;

	f->next = island->faces;
	island->faces = f;
	island->nfaces++;
	
	return f;
}

/* ************************** BOUNDARY *************************** */

static UvEdge *boundary_next(UvEdge *e, int *vindex)
{
	LinkNode *link;
	UvEdge *next_e = NULL;
	UvVert *v = e->v[(*vindex)];

	for (link=v->edges; link; link=link->next) {
		next_e = (UvEdge*)link->link;
		if ((next_e->flag & UVEDGE_SEAM) && (e != next_e))
			break;
	}

	(*vindex) = (next_e->v[0] == v)? 1: 0;

	return next_e;
}

static LinkNode *get_boundaries(UvIsland *island)
{
	UvEdge *e, *loop_e, *outer_e;
	LinkNode *list;
	int vindex;
	float len, vlen, maxlen, maxvlen;

	for (e=island->edges; e; e=e->next)
		e->flag &= ~UVEDGE_VISITED;
	
	list = NULL;
	outer_e = NULL;
	maxlen = -1.0f;

	for (e=island->edges; e; e=e->next) {
		if (!(e->flag & UVEDGE_SEAM) || e->flag & UVEDGE_VISITED)
			continue;

		len = 0.0f;
		maxvlen = -1.0f;
		vindex = 0;
		loop_e = e;
		e->u.totedge = 0;

		while (!(loop_e->flag & UVEDGE_VISITED)) {
			e->u.totedge++;

			vlen = VecLenf(loop_e->v[vindex]->co, island->center);
			maxvlen = MAX2(vlen, maxvlen);
			len += vlen;

			loop_e->flag |= UVEDGE_VISITED;
			loop_e = boundary_next(loop_e, &vindex);
		}

		len /= e->u.totedge; /* FIXME: unused */

		if (maxvlen > maxlen) {
			maxlen = maxvlen;
			if (outer_e)
				BLI_linklist_prepend(&list, (void*)outer_e);
			outer_e = e;
		}
		else
			BLI_linklist_prepend(&list, (void*)e);
	}

	if (outer_e)
		BLI_linklist_prepend(&list, (void*)outer_e);
	
	return list;
}

/* *************************** COMPUTE *************************** */

static void compute_uv_bbox(UvIsland *island)
{
	UvVert *v;

	INIT_MINMAX2(island->min, island->max);

	for (v=island->verts; v; v=v->next) {
		DO_MINMAX2(v->uv, island->min, island->max);
	}

	Vec2Addf(island->center, island->min, island->max);
	Vec2Mulf(island->center, 0.5f);
}

static void compute_co_bbox_center_radius(UvIsland *island)
{
	UvVert *v;
	float sub[3];

	island->center[0]= island->center[1]= island->center[2]= 0.0f;
	INIT_MINMAX(island->min, island->max);

	for (v=island->verts; v; v=v->next) {
		DO_MINMAX(v->co, island->min, island->max);
		VecAddf(island->center, island->center, v->co);
	}

	if (island->nverts > 1)
		VecMulf(island->center, 1.0f/(float)(island->nverts));

	VecSubf(sub, island->center, island->max);
	island->radius = Normalise(sub);

	if(island->radius < 1e-20)
		island->radius = 1.0;
}

static float compute_angle(float e1, float e2, float e3)
{
	float n, angle;

	n = 2.0f*e2*e3;	

	if (n > 0.0f) {
		angle = (e2*e2 + e3*e3 - e1*e1)/n;

		if (angle > 0.9999f)
			return 0.0f;
		else if (angle < -0.9999f)
			return M_PI;
		else
			return acos(angle);
	}
	else
		return 0.0f;
}

/* ********************** LOAD/FLUSH TFACE *********************** */

static UvVert *load_add_vert(UvVert **vtable, UvIsland *island, unsigned int v, float *co)
{
	if (vtable[v])
		return vtable[v];
	else {
		UvVert *newv;
		
		newv = uv_island_add_vert(island, co);
		newv->orig.index = v;
		vtable[v] = newv;
		return newv;
	}
}

static int load_quad_split(MVert *mv, MFace *mf)
{
	float e1, e2, e3, e4, d1, d2, a1, a2, a3, a4, max;

	e1 = VecLenf(mv[mf->v1].co, mv[mf->v2].co);
	e2 = VecLenf(mv[mf->v2].co, mv[mf->v3].co);
	e3 = VecLenf(mv[mf->v3].co, mv[mf->v4].co);
	e4 = VecLenf(mv[mf->v4].co, mv[mf->v1].co);

	d1 = VecLenf(mv[mf->v1].co, mv[mf->v3].co);
	d2 = VecLenf(mv[mf->v2].co, mv[mf->v4].co);

	a1 = compute_angle(d2, e4, e1);
	a2 = compute_angle(d1, e1, e2);
	a3 = compute_angle(d2, e2, e3);
	a4 = compute_angle(d1, e3, e4);

	max = MAX4(a1, a2, a3, a4);

	return (max == a2 || max == a4);
}

static void load_mesh(UvIsland *island, Mesh *me, AtlasType type)
{
	MVert *mv;
	MFace *mf;
	TFace *tf;
	int a, splitdir, mask;
	UvVert *v1, *v2, *v3, **vtable;
	UvFace *face;

	vtable = MEM_callocN(sizeof(UvVert*)*me->totvert, "UvIslandVTable");

	mv= me->mvert;
	mf= me->mface;
	tf= me->tface;

	for (a=me->totface; a>0; a--, tf++, mf++) {
		if (mf->v3 && !(tf->flag & TF_HIDE) && (tf->flag & TF_SELECT)) { 
			if (type == ATLAS_IMPLICIT_SELECTED) {
				if (mf->v4)
					mask = (TF_SEL1|TF_SEL2|TF_SEL3|TF_SEL4);
				else
					mask = (TF_SEL1|TF_SEL2|TF_SEL3);

				if ((tf->flag & mask)!=mask)
					continue;
			}

			splitdir = (mf->v4)? load_quad_split(mv, mf): 0;

			if (splitdir) {
				v1 = load_add_vert(vtable, island, mf->v4, mv[mf->v4].co);
				v2 = load_add_vert(vtable, island, mf->v1, mv[mf->v1].co);
				v3 = load_add_vert(vtable, island, mf->v2, mv[mf->v2].co);

				face = uv_island_add_face(island, v1, v2, v3);

				face->orig.index = me->totface - a;
				face->orig.findex[0] = 3;
				face->orig.findex[1] = 0;
				face->orig.findex[2] = 1;

				if (tf->unwrap & TF_SEAM4)
					face->e[0]->flag |= UVEDGE_TFACE_SEAM;

				if (tf->unwrap & TF_SEAM1)
					face->e[1]->flag |= UVEDGE_TFACE_SEAM;

				v1 = load_add_vert(vtable, island, mf->v2, mv[mf->v2].co);
				v2 = load_add_vert(vtable, island, mf->v3, mv[mf->v3].co);
				v3 = load_add_vert(vtable, island, mf->v4, mv[mf->v4].co);

				face = uv_island_add_face(island, v1, v2, v3);

				face->orig.index = me->totface - a;
				face->orig.findex[0] = 1;
				face->orig.findex[1] = 2;
				face->orig.findex[2] = 3;

				if (tf->unwrap & TF_SEAM2)
					face->e[0]->flag |= UVEDGE_TFACE_SEAM;

				if (tf->unwrap & TF_SEAM3)
					face->e[1]->flag |= UVEDGE_TFACE_SEAM;
			}
			else {
				v1 = load_add_vert(vtable, island, mf->v1, mv[mf->v1].co);
				v2 = load_add_vert(vtable, island, mf->v2, mv[mf->v2].co);
				v3 = load_add_vert(vtable, island, mf->v3, mv[mf->v3].co);

				face = uv_island_add_face(island, v1, v2, v3);

				face->orig.index = me->totface - a;
				face->orig.findex[0] = 0;
				face->orig.findex[1] = 1;
				face->orig.findex[2] = 2;

				if (tf->unwrap & TF_SEAM1)
					face->e[0]->flag |= UVEDGE_TFACE_SEAM;

				if (tf->unwrap & TF_SEAM2)
					face->e[1]->flag |= UVEDGE_TFACE_SEAM;

				if ((!mf->v4) && (tf->unwrap & TF_SEAM3))
					face->e[2]->flag |= UVEDGE_TFACE_SEAM;

				if (mf->v4) {
					v1 = load_add_vert(vtable, island, mf->v1, mv[mf->v1].co);
					v2 = load_add_vert(vtable, island, mf->v3, mv[mf->v3].co);
					v3 = load_add_vert(vtable, island, mf->v4, mv[mf->v4].co);

					face = uv_island_add_face(island, v1, v2, v3);

					face->orig.index = me->totface - a;
					face->orig.findex[0] = 0;
					face->orig.findex[1] = 2;
					face->orig.findex[2] = 3;

					if (tf->unwrap & TF_SEAM3)
						face->e[1]->flag |= UVEDGE_TFACE_SEAM;

					if (tf->unwrap & TF_SEAM4)
						face->e[2]->flag |= UVEDGE_TFACE_SEAM;
				}
			}
		}
	}

	MEM_freeN(vtable);
}

static void load_pins_selection(UvIsland *island, int *pinned, int *selected, int *pinselected)
{
	UvFace *f;
	UvVert *v;
	TFace *tface, *tf;
	int totpin, totsel, totpinsel, j, findex;

	tface = (TFace*)island->me->tface;

	for (f=island->faces; f; f=f->next) {
		if (f->flag & UVFACE_FAKE)
			continue;

		for (j = 0; j < 3; j++) {
			tf = tface + f->orig.index;
			findex = f->orig.findex[j];

			if (tf->unwrap & TF_PIN_MASK(findex))
				f->v[j]->flag |= (UVVERT_PINNED|UVVERT_TFACE_PINNED);

			if (tf->flag & TF_SEL_MASK(findex))
				f->v[j]->flag |= UVVERT_SELECTED;
		}
	}
	
	if (pinned || selected || pinselected) {
		totpin = 0;
		totsel = 0;
		totpinsel = 0;
		for (v=island->verts; v; v=v->next) {
			if (v->flag & UVVERT_PINNED) {
				totpin++;
				if (v->flag & UVVERT_SELECTED) {
					totsel++;
					totpinsel++;
				}
			}
			else if (v->flag & UVVERT_SELECTED)
				totsel++;
		}
		
		if (pinned)
			*pinned = totpin;
		if (selected)
			*selected = totsel;
		if (pinselected)
			*pinselected = totpinsel;
	}
}

static void load_uvs(UvIsland *island)
{
	LinkNode *link;
	UvFace *f;
	UvVert *v;
	int tot, index;
	char findex;
	TFace *tf, *tface = (TFace*)island->me->tface;

	for (v=island->verts; v; v=v->next) {
		/* assign average of (pinned) uv's */
		
		v->uv[0] = v->uv[1] = 0.0f;
		tot = 0;

		for (link=v->faces; link; link=link->next) {
			f = (UvFace*)link->link;

			if (f->flag & UVFACE_FAKE)
				continue;

			/* if any of the uv's is pinned, add only the pinned ones,
			   otherwise add all of them */

			if (f->v[0] == v) index = 0;
			else if (f->v[1] == v) index = 1;
			else index = 2;

			tf = tface + f->orig.index;
			findex = f->orig.findex[index];

			if (v->flag & UVVERT_TFACE_PINNED)
				if (!(tf->unwrap & TF_PIN_MASK(findex)))
					continue;
				
			Vec2Addf(v->uv, v->uv, tf->uv[findex]);
			tot++;
		}

		if (tot > 1)
			Vec2Mulf(v->uv, 1.0/(float)tot);
		Vec2Copyf(v->old_uv, v->uv);
	}
}

static float *load_get_vertex_uv(UvIsland *island, UvVert *v, UvFace *f)
{
	TFace *tface, *tf;
	char id;

	tface = (TFace*)island->me->tface;

	if (f->v[0] == v) id = 0;
	else if (f->v[1] == v) id = 1;
	else id = 2;

	tf = tface + f->orig.index;
	return tf->uv[f->orig.findex[id]];
}

static void flush_old_uvs(UvIsland *island)
{
	UvFace *f;
	TFace *tf, *tface = (TFace*)island->me->tface;

	for (f=island->faces; f; f=f->next) {
		if (f->flag & UVFACE_FAKE)
			continue;

		tf = tface + f->orig.index;

		Vec2Copyf(tf->uv[f->orig.findex[0]], f->v[0]->old_uv);
		Vec2Copyf(tf->uv[f->orig.findex[1]], f->v[1]->old_uv);
		Vec2Copyf(tf->uv[f->orig.findex[2]], f->v[2]->old_uv);
	}
}

static void flush_uvs(UvIsland *island)
{
	UvFace *f;
	TFace *tf, *tface = (TFace*)island->me->tface;

	for (f=island->faces; f; f=f->next) {
		if (f->flag & UVFACE_FAKE)
			continue;

		tf = tface + f->orig.index;

		Vec2Copyf(tf->uv[f->orig.findex[0]], f->v[0]->uv);
		Vec2Copyf(tf->uv[f->orig.findex[1]], f->v[1]->uv);
		Vec2Copyf(tf->uv[f->orig.findex[2]], f->v[2]->uv);
	}
}

/* *************************** WHEEL ***************************** */

static void wheel_spokes_id(UvFace *f, UvVert *v, char *id1, char *id2, char *id3)
{
	char findex;

	if (f->v[0] == v) findex = 0;
	else if (f->v[1] == v) findex = 1;
	else findex = 2;

	*id1 = findex;
	*id2 = (findex+2) % 3;
	*id3 = (findex+1) % 3;
}

static void wheel_spokes(UvFace *f, UvVert *v, UvEdge **e1, UvEdge **e2, UvEdge **e3)
{
	char id1, id2, id3;

	wheel_spokes_id(f, v, &id1, &id2, &id3);

	*e1 = f->e[id1];
	*e2 = f->e[id2];
	*e3 = f->e[id3];
}

static UvEdge *wheel_next_edge(UvFace *f, UvEdge *e, UvVert *v)
{
	UvEdge *other_e;
	char index;

	if (f->e[0] == e) index = 0;
	else if (f->e[1] == e) index = 1;
	else if (f->e[2] == e) index = 2;
	else return NULL;
	
	other_e = f->e[(index+1) % 3];

	if (other_e->v[0] == v || other_e->v[1] == v)
		return other_e;
	else
		return f->e[(index+2) % 3];
}

/* ************************* SPLITTING *************************** */

static void split_edge(UvIsland *island, UvVert *v, UvVert *new_v, UvFace *f, char id)
{
	UvEdge *e;
	UvFace *face;
	LinkNode *link, *edge_link, *last_link;

	e = f->e[id];

	/* update e->v */
	for (link=e->faces; link; link=link->next)
		if (!(((UvFace*)link->link)->flag & UVFACE_VERTEX_SPLIT))
			break;
	
	if (link == NULL) {
		/* don't split, only move to new vertex */

		BLI_linklist_extract(&(v->edges), (UvEdge*)e, &edge_link);

		if (edge_link) {
			edge_link->next = new_v->edges;
			new_v->edges = edge_link;

			if (e->v[0] == v)
				e->v[0] = new_v;
			else
				e->v[1] = new_v;
		}
	}
	else {
		/* split off edge */

		UvEdge *new_e;
		UvVert *v2;
		LinkNode *next;

		v2 = (e->v[0] == v)? e->v[1]: e->v[0];
		new_e = uv_island_add_edge(island, new_v, v2);

		/* replace edge in face */
		f->e[id] = new_e;

		/* add link from new_v to new_e */
		BLI_linklist_prepend_arena(&new_v->edges, new_e, island->arena);

		/* move face links from e to new_e */
		last_link = NULL;
		for (link=e->faces; link; link=next) {
			next = link->next;
			face = (UvFace*)link->link;

			if (face->flag & UVFACE_VERTEX_SPLIT) {
				if (last_link)
					last_link->next = link->next;
				else
					e->faces = link->next;
				link->next = new_e->faces;
				new_e->faces = link;

			}
			else
				last_link = link;
		}

		/* set seam flag */
		if (new_e->faces->next == NULL)
			new_e->flag |= UVEDGE_SEAM;
	}
}

static void split_vertex(UvIsland *island, UvVert *v, LinkNode *faces)
{
	UvVert *new_v;
	UvFace *f;
	LinkNode *link;
	char id1, id2, id3;

	new_v = uv_island_add_vert(island, v->co);
	new_v->faces = faces;

	/* update f->v in splitted faces */
	for (link=new_v->faces; link; link=link->next) {
		f = (UvFace*)link->link;

		if (f->v[0] == v) f->v[0] = new_v;
		else if (f->v[1] == v) f->v[1] = new_v;
		else f->v[2] = new_v;
		
		f->flag |= UVFACE_VERTEX_SPLIT;
	}

	/* split edges */
	for (link=new_v->faces; link; link=link->next) {
		f = (UvFace*)link->link;

		wheel_spokes_id(f, new_v, &id1, &id2, &id3);
		split_edge(island, v, new_v, f, id1);
		split_edge(island, v, new_v, f, id2);
	}

	/* clear flags again */
	for (link=new_v->faces; link; link=link->next)
		((UvFace*)link->link)->flag &= ~UVFACE_VERTEX_SPLIT;
}

/* ********************** WHEEL SPLITTING ************************
 * - mark boundaries and non-manifold edges as seams.
 * - split based on UVEDGE_TFACE_SEAM flag.
 * - order faces by wheel.
 */

static LinkNode *wheel_get_adjacent(UvIsland *island, UvVert *v, UvEdge *e)
{
	LinkNode *facelist, *link, *last_link, *next;
	UvFace *f;
	UvEdge *wheel_e;
	int found = 0;

	/* make a list of adjacent faces, until a seam is reached */

	facelist = NULL;
	found = 1;

	while (found) {
		found = 0;
		last_link = NULL;

		for (link=v->faces; link; link=next) {
			next = link->next;
			f = (UvFace*)link->link;

			wheel_e = wheel_next_edge(f, e, v);
			if (wheel_e) {
				/* add face to facelist, remove from v->faces */
				link->next = facelist;
				facelist = link;

				if (last_link)
					last_link->next = next;
				else
					v->faces = next;

				/* stop */
				if (wheel_e->flag & UVEDGE_SEAM)
					break;

				/* continue with new edge */
				e = wheel_e;
				found = 1;

				break;
			}
			else
				last_link = link;
		}
	}

	return facelist;
}

static void wheel_split_seams(UvIsland *island)
{
	UvVert *v;
	UvEdge *e, *e1, *e2, *e3;
	LinkNode *left_faces, *right_faces, *facelist, *first;

	/* FIXME: this might not give good results with isolated seams */

	for (e=island->edges; e; e=e->next) {
		if (e->flag & UVEDGE_TFACE_SEAM)
			e->flag |= UVEDGE_SEAM;
		else if (e->faces->next==NULL || e->faces->next->next)
			e->flag |= UVEDGE_SEAM; /* edge has != 2 faces */
	}

	for (v=island->verts; v; v=v->next) {
		do {
			facelist = NULL;

			/* remove first from v->faces */
			first = v->faces;
			v->faces = first->next;

			wheel_spokes((UvFace*)first->link, v, &e1, &e2, &e3);

			/* add adjacent faces on the left */
			if (!(e1->flag & UVEDGE_SEAM)) {
				left_faces = wheel_get_adjacent(island, v, e1);
				BLI_linklist_reverse(&left_faces);
				facelist = left_faces;
			}

			/* add face in the middle */
			first->next = facelist;
			facelist = first;

			/* add adjacent faces on the right */
			if (!(e2->flag & UVEDGE_SEAM)) {
				right_faces = wheel_get_adjacent(island, v, e2);
				BLI_linklist_join(&right_faces, facelist);
				facelist = right_faces;
			}

			if (v->faces)
				split_vertex(island, v, facelist);
			
		} while (v->faces);

		v->faces = facelist;
	}
}

/* ********************** IMPLICIT SPLITTING ************************
 * - mark boundaries as seams.
 * - split based on connectivity of uv coords.
 */

void implicit_split_seams(UvIsland *island)
{
	UvEdge *e;
	UvVert *v;
	UvFace *f, *firstf;
	LinkNode *first, *link, *last_link, *next, *facelist;
	float limit[2];

	get_connected_limit_tface_uv(limit);

	for (v=island->verts; v; v=v->next) {
		if (v->faces->next == NULL)
			continue;

		for (link=v->faces; link; link=link->next) {
			f = (UvFace*)link->link;
			f->u.tface_uv = load_get_vertex_uv(island, v, f);
		}

		do {
			/* remove first from v->faces and add to facelist*/
			first = v->faces;
			v->faces = first->next;

			first->next = NULL;
			facelist = first;

			firstf = (UvFace*)first->link;

			/* split off faces within dist and add them to facelist */
			last_link = NULL;
			for (link=v->faces; link; link=next) {
				next = link->next;
				f = (UvFace*)link->link;

				if (fabs(f->u.tface_uv[0]-firstf->u.tface_uv[0]) < limit[0]) {
					if(fabs(f->u.tface_uv[1]-firstf->u.tface_uv[1]) < limit[1]) {
						if (last_link)
							last_link->next = link->next;
						else
							v->faces = link->next;
						link->next = facelist;
						facelist = link;
					}
					else
						last_link = link;
				}
				else
					last_link = link;
			}

			if (v->faces)
				split_vertex(island, v, facelist);

		} while (v->faces);

		v->faces = facelist;
	}

	for (e=island->edges; e; e=e->next)
		if (e->faces->next==NULL)
			e->flag |= UVEDGE_SEAM; /* edge has exactly 1 face */
}


/* ********************* ISLAND SPLITTING *********************** */

static void split_off_island(UvIsland *continent, UvIsland *island)
{
	int doit, i;
	UvFace *last_f, *f, *next;
	UvEdge *e, *last_e, *next_e;
	UvVert *v, *last_v, *next_v;
	LinkNode *link;

	/* pick island containing first face */
	continent->faces->flag |= UVFACE_SPLIT; 
	doit = 1;

	while (doit) {
		doit = 0;

		last_f = NULL;
		for (f=continent->faces; f; f=next) {
			next = f->next;

			if (f->flag & UVFACE_SPLIT) {
				doit = 1;

				for (i=0; i<3; i++) {
					f->v[i]->flag |= UVVERT_SPLIT;
					f->e[i]->flag |= UVEDGE_SPLIT;

					for (link=f->e[i]->faces; link; link=link->next)
						((UvFace*)link->link)->flag |= UVFACE_SPLIT;
				}

				/* move face from continent to island */
				if (last_f == NULL)
					continent->faces = f->next;
				else
					last_f->next = f->next;
				continent->nfaces--;

				f->next = island->faces;
				island->faces = f;
				island->nfaces++;
			}
			else
				last_f = f;
		}
	}

	/* move verts from continent to island */
	last_v = NULL;
	for (v=continent->verts; v; v=next_v) {
		next_v = v->next;

		if (v->flag & UVVERT_SPLIT) {
			if (last_v == NULL)
				continent->verts = v->next;
			else
				last_v->next = v->next;
			continent->nverts--;

			v->next = island->verts;
			island->verts = v;
			island->nverts++;
		}
		else
			last_v = v;
	}

	/* move edges from continent to island */
	last_e = NULL;
	for (e=continent->edges; e; e=next_e) {
		next_e = e->next;

		if (e->flag & UVEDGE_SPLIT) {
			if (last_e == NULL)
				continent->edges = e->next;
			else
				last_e->next = e->next;
			continent->nedges--;

			e->next = island->edges;
			island->edges = e;
			island->nedges++;
		}
		else
			last_e = e;
	}
}

/* *************************** ATLAS ***************************** */

static UvAtlas *new_atlas(Mesh *me, AtlasType type)
{
	UvAtlas *atlas;
	UvIsland *continent, *island;
	
	atlas = (UvAtlas*)MEM_callocN(sizeof(UvAtlas), "UvAtlas");
	atlas->arena = BLI_memarena_new(BLI_MEMARENA_STD_BUFSIZE);
	atlas->islands = NULL;

	continent = (UvIsland*)MEM_callocN(sizeof(UvIsland), "UvIsland");
	continent->arena = atlas->arena;
	continent->me = me;

	load_mesh(continent, me, type);
	if (type == ATLAS_SEAMS)
		wheel_split_seams(continent);
	else
		implicit_split_seams(continent);

	while (continent->faces) {
		island = (UvIsland*)MEM_callocN(sizeof(UvIsland), "UvIsland");
		island->arena = atlas->arena;
		island->me = me;

		island->next = atlas->islands;
		atlas->islands = island;

		split_off_island(continent, island);
	}

	MEM_freeN(continent);

	return atlas;
}

static void free_atlas(UvAtlas *atlas)
{
	UvIsland *next;

	while (atlas->islands) {
		next = atlas->islands->next;
		MEM_freeN(atlas->islands);
		atlas->islands = next;
	}

	BLI_memarena_free(atlas->arena);
	MEM_freeN(atlas);
}

/* ************************ FILLING HOLES ************************ */

static void fill_spokes(UvVert *v, UvEdge **e1, UvEdge **e2, UvVert **v1, UvVert **v2)
{
	LinkNode *link;
	UvEdge *e;

	*e1 = NULL;

	for (link=v->edges; link; link=link->next) {
		e = (UvEdge*)link->link;
		if (e->flag & UVEDGE_SEAM) {
			if (*e1) {
				*e2 = e;
				break;
			}
			else
				*e1 = e;
		}
	}

	*v1 = ((*e1)->v[0] == v) ? (*e1)->v[1]: (*e1)->v[0];
	*v2 = ((*e2)->v[0] == v) ? (*e2)->v[1]: (*e2)->v[0];
}

static void fill_hole(UvIsland *island, UvEdge *e)
{
	UvVert *v, **boundary, *v1, *v2;
	UvEdge *loop_e, *e1, *e2, *e3;
	UvFace *f, *spokef;
	LinkNode *link;
	int totvert, totface, vindex, i, mini, dir;
	float l1, l2, l3, minangle;

	totvert = e->u.totedge;
	boundary = MEM_mallocN(totvert*sizeof(UvVert*), "UvBoundary");

	vindex = 0;
	i = 0;
	loop_e = e;

	/* compute vertex angles in hole */
	while (!(loop_e->flag & UVEDGE_VISITED)) {
		v = loop_e->v[vindex];
		boundary[i] = v;
		v->u.angle = 2*M_PI;

		for (link=v->faces; link; link=link->next) {
			f = (UvFace*)link->link;
			wheel_spokes(f, v, &e1, &e2, &e3);

			l1 = VecLenf(e1->v[0]->co, e1->v[1]->co);
			l2 = VecLenf(e2->v[0]->co, e2->v[1]->co);
			l3 = VecLenf(e3->v[0]->co, e3->v[1]->co);

			v->u.angle -= compute_angle(l3, l1, l2);
		}
		
		loop_e->flag |= UVEDGE_VISITED;
		loop_e = boundary_next(loop_e, &vindex);
		i++;
	}

	totface = totvert - 2;
	while (totface > 0) {
		/* find smallest angle */
		minangle = 1e10f;
		mini = 0;
		for (i=0; i<totvert; i++) {
			v = boundary[i];
			if (v && v->u.angle < minangle) {
				minangle = v->u.angle;
				mini = i;
			}
		}
		
		/* add face */
		v = boundary[mini];
		boundary[mini] = NULL;

		fill_spokes(v, &e1, &e2, &v1, &v2);
		
		/* choose normal based on one of spoke faces */
		spokef = (UvFace*)e1->faces->link;

		if (spokef->e[0] == e1)
			dir = (spokef->v[0] == v);
		else if (spokef->e[1] == e1)
			dir = (spokef->v[1] == v);
		else
			dir = (spokef->v[2] == v);

		if (dir) {
			f = uv_island_add_face(island, v1, v, v2);
			e3 = f->e[2];
		}
		else {
			f = uv_island_add_face(island, v, v1, v2);
			e3 = f->e[1];
		}

		f->flag |= UVFACE_FAKE;
		totface--;

		/* update seams */

		e1->flag &= ~UVEDGE_SEAM;
		e2->flag &= ~UVEDGE_SEAM;

		if (totface == 0)
			e3->flag &= ~UVEDGE_SEAM;
		else {
			e3->flag |= UVEDGE_SEAM;

			/* update angles */

			l1 = VecLenf(v->co, v1->co);
			l2 = VecLenf(v->co, v2->co);
			l3 = VecLenf(v1->co, v2->co);

			v1->u.angle -= compute_angle(l2, l1, l3);
			v2->u.angle -= compute_angle(l1, l2, l3);
		}
	}

	MEM_freeN(boundary);
}

static void fill_holes(UvIsland *island)
{
	LinkNode *list, *link;
	UvEdge *e;

	/* this breaks the faces wheel order */
	
	list = get_boundaries(island);

	if (list) {
		/* fill all boundaries, except the first (the outer boundary) */

		for (e=island->edges; e; e=e->next)
			e->flag &= ~UVEDGE_VISITED;

		for (link=list->next; link; link=link->next)
			fill_hole(island, (UvEdge*)link->link);
	}

	BLI_linklist_free(list, NULL);
}

/* ***************************** LSCM ***************************************
 * Least Squares Conformal Maps parameterization, based on:
 * Bruno Levy, Sylvain Petitjean, Nicolas Ray, Jerome Maillot. Least Squares
 * Conformal Maps for Automatic Texture Atlas Generation. In Siggraph 2002,
 * July 2002.
 */

static void lscm_projection_axis(UvIsland *island, float *p1, float *p2)
{
	float dx, dy, dz;

	dx= island->max[0] - island->min[0];
	dy= island->max[1] - island->min[1];
	dz= island->max[2] - island->min[2];

	p1[0]= p1[1]= p1[2]= 0.0;
	p2[0]= p2[1]= p2[2]= 0.0;

	if(dx < dy && dx < dz) {
		if(dy > dz) p1[1]= p2[2]= 1.0;   /* y, z */
		else p1[2]= p2[1]= 1.0;          /* z, y */
	}
	else if(dy < dx && dy < dz) {
		if(dx > dz) p1[0]= p2[2]= 1.0;   /* x, z */
		else p1[2]= p2[0]= 1.0;          /* z, x */
	}
	else {
		if(dx > dy) p1[0]= p2[1]= 1.0;   /* x, y */
		else p1[1]= p2[0]= 1.0;          /* y, x */
	}
}

static void lscm_index_vertices(UvIsland *island)
{
	UvVert *v;
	int index;

	for (v=island->verts, index=0; v; v=v->next, index++)
		v->u.index = index;
}

static int lscm_set_pinned_solution(UvIsland *island)
{
	UvVert *v;
	float min[2], max[2];
	int pinned;
	
	INIT_MINMAX2(min, max);
	pinned = 0;

	for (v=island->verts; v; v=v->next) {
		nlSetVariable(2*v->u.index, v->uv[0]);
		nlSetVariable(2*v->u.index + 1, v->uv[1]);

		if (v->flag & UVVERT_PINNED) {
			DO_MINMAX2(v->uv, min, max);

			nlLockVariable(2*v->u.index);
			nlLockVariable(2*v->u.index + 1);

			pinned++;
		}
	}

	if (pinned >= 2) {
		/* abuse umax vmax for caculating euclidian distance */
		max[0] -= min[0];
		max[1] -= min[1];

		/* check for degenerated pinning box */
		if (((max[0]*max[0])+(max[1]*max[1])) < 1e-10)
			return -1;
	}

	return pinned;
}

static void lscm_set_projected_solution(UvIsland *island, float *p1, float *p2)
{
	UvVert *v;
	float umin, umax;
	int index_min, index_max;

	umin = 1.0e30f;
	umax = -1.0e30f;

	index_min = 0;
	index_max = 2;

	for (v=island->verts; v; v=v->next) {
		v->uv[0] = Inpf(v->co, p1);
		v->uv[1] = Inpf(v->co, p2);
		
		if(v->uv[0] < umin) {
			index_min = v->u.index;
			umin = v->uv[0];
		}

		if(v->uv[0] > umax) {
			index_max = v->u.index;
			umax = v->uv[0];
		}

		nlSetVariable(2*v->u.index, v->uv[0]);
		nlSetVariable(2*v->u.index + 1, v->uv[1]);
	}

	nlLockVariable(2*index_min);
	nlLockVariable(2*index_min + 1);
	nlLockVariable(2*index_max);
	nlLockVariable(2*index_max + 1);
}

static void lscm_normalize(float *co, float *center, float radius)
{
	/* normalize relative to complete surface */
	VecSubf(co, co, center);
	VecMulf(co, (float)1.0/radius);
}

static void lscm_add_triangle(UvIsland *island, UvVert *v1, UvVert *v2, UvVert *v3)
{
	float x[3], y[3], z[3], sub[3], z1[2], z2[2], co1[3], co2[3], co3[3];
	int id0[2], id1[2], id2[2];

	VecCopyf(co1, v1->co);
	VecCopyf(co2, v2->co);
	VecCopyf(co3, v3->co);

	/* project 3d triangle
	 * edge length is lost, as this algorithm is angle based */
	lscm_normalize(co1, island->center, island->radius);
	lscm_normalize(co2, island->center, island->radius);
	lscm_normalize(co3, island->center, island->radius);

	VecSubf(x, co2, co1);
	Normalise(x);

	VecSubf(sub, co3, co1);
	Crossf(z, x, sub);
	Normalise(z);

	Crossf(y, z, x);

	/* reduce to two 2d vectors */
	VecSubf(sub, co2, co1);
	z1[0]= Normalise(sub);
	z1[1]= 0;

	VecSubf(sub, co3, co1);
	z2[0]= Inpf(sub, x);
	z2[1]= Inpf(sub, y);

	/* split id's up for u and v
	   id = u, id + 1 = v */
	id0[0]= 2*v1->u.index;
	id0[1]= 2*v1->u.index + 1;
	id1[0]= 2*v2->u.index;
	id1[1]= 2*v2->u.index + 1;
	id2[0]= 2*v3->u.index;
	id2[1]= 2*v3->u.index + 1;

	/* The LSCM Equation:
	 * ------------------
	 * (u,v) are the uv coords we are looking for -> complex number u + i*v
	 * (x,y) are the above calculated local coords -> complex number x + i*y
	 * Uk = uk + i*vk
	 * Zk = xk + i*yk (= zk[0] + i*zk[1] in the code)
	 * 
	 * That makes the equation:
	 * (Z1 - Z0)(U2 - U0) = (Z2 - Z0)(U1 - U0)
	 *
	 * x0, y0 and y1 were made zero by projecting the triangle:
	 * (x1 + i*y1)(u2 + i*v2 - u0 - i*v0) = (x2 + i*y2)(u1 + i*v1 - u0 - i*v0)
	 *
	 * this gives the following coefficients:
	 * u0 * ((-x1 + x2) + i*(y2))
	 * v0 * ((-y2) + i*(-x1 + x2))
	 * u1 * ((-x2) + i*(-y2))
	 * v1 * ((y2) + i*(-x2))
	 * u2 * (x1)
	 * v2 * (i*(x1))
	 */

	/* real part */
	nlBegin(NL_ROW);
	nlCoefficient(id0[0], -z1[0] + z2[0]);
	nlCoefficient(id0[1], -z2[1]        );
	nlCoefficient(id1[0], -z2[0]        );
	nlCoefficient(id1[1],  z2[1]        );
	nlCoefficient(id2[0],  z1[0]        );
	nlEnd(NL_ROW);

	/* imaginary  part */
	nlBegin(NL_ROW);
	nlCoefficient(id0[0],  z2[1]        );
	nlCoefficient(id0[1], -z1[0] + z2[0]);
	nlCoefficient(id1[0], -z2[1]        );
	nlCoefficient(id1[1], -z2[0]        );
	nlCoefficient(id2[1],  z1[0]        );
	nlEnd(NL_ROW);
}

static void lscm_build_matrix(UvIsland *island)
{
	UvFace *f;

	for (f=island->faces; f; f=f->next)
		lscm_add_triangle(island, f->v[0], f->v[1], f->v[2]);
}

static void lscm_load_solution(UvIsland *island)
{
	UvVert *v;

	for (v=island->verts; v; v=v->next) {
		v->uv[0] = nlGetVariable(2*v->u.index);
		v->uv[1] = nlGetVariable(2*v->u.index + 1);
	}
}

static void lscm_begin(UvIsland *island)
{
	compute_co_bbox_center_radius(island);

	fill_holes(island);

	lscm_index_vertices(island); /* set indices for variables in solver */
}

static void lscm_solve(UvIsland *island, int autopin)
{
	int solve = 1;

	nlNewContext();
	/* n variables = 2 x n verts, for u and v */
	nlSolverParameteri(NL_NB_VARIABLES, 2*island->nverts);
	nlSolverParameteri(NL_LEAST_SQUARES, NL_TRUE);

	nlBegin(NL_SYSTEM);

	if (autopin) {
		/* auto pin 2 uv's and use projected uv's as intial solution */

		float p1[3], p2[3];

		lscm_projection_axis(island, p1, p2);
		lscm_set_projected_solution(island, p1, p2);
	}
	else {
		/* pin user defined verts and use tface uv's as intial solution */

		load_uvs(island);
		if (!lscm_set_pinned_solution(island))
			solve = 0;
	}

	if (solve) {
		/* add triangles to the solver */
		nlBegin(NL_MATRIX);
		lscm_build_matrix(island);
		nlEnd(NL_MATRIX);
			
		nlEnd(NL_SYSTEM);

		/* LSCM solver magic! */
		nlSolve();

		lscm_load_solution(island); /* get uv's from matrix */
	}
	else
		nlEnd(NL_SYSTEM);

	nlDeleteContext(nlGetCurrent());
}

static void lscm_end(UvIsland *island)
{
	/* empty - for now */
}

/* *************************** PACKING *************************** */

static void uv_island_scale(UvIsland *island, float scale)
{
	UvVert *v;

	for (v=island->verts; v; v=v->next)
		Vec2Mulf(v->uv, scale);
}

static void uv_island_move(UvIsland *island, float *add)
{
	UvVert *v;

	for (v=island->verts; v; v=v->next)
		Vec2Addf(v->uv, v->uv, add);
}

/* put within (0,0)->(1,1) boundbox */
static void uv_island_normalize(UvIsland *island)
{
	float dx, dy, scale, add[2];

	compute_uv_bbox(island);

	dx= island->max[0] - island->min[0];
	dy= island->max[1] - island->min[1];

	scale= MAX2(dx, dy);
	scale= (1.0/scale);

	add[0]= -island->min[0];
	add[1]= -island->min[1];

	uv_island_move(island, add);
	uv_island_scale(island, scale);
}

static void normalize_atlas(UvAtlas *atlas)
{
	float dx, dy, scale, add[2], min[2], max[2];
	UvIsland *island;

	INIT_MINMAX2(min, max);

	for (island=atlas->islands; island; island=island->next) {
		if (island->flag & UVISLAND_NOPACK)
			continue;

		compute_uv_bbox(island);

		DO_MINMAX2(island->min, min, max);
		DO_MINMAX2(island->max, min, max);
	}

	dx= max[0]-min[0];
	dy= max[1]-min[1];

	scale= MAX2(dx, dy);
	scale= (1.0/scale);
	scale*= 0.98;

	add[0]= -min[0] + 0.01;
	add[1]= -min[1] + 0.01;

	for (island=atlas->islands; island; island=island->next) {
		if (island->flag & UVISLAND_NOPACK)
			continue;

		uv_island_move(island, add);
		uv_island_scale(island, scale);
	}
}

/* get scale relative to mesh */
static float uv_island_relative_scale(UvIsland *island)
{
	UvEdge *e;
	float len_xyz, len_uv;

	len_xyz = len_uv = 0.0f;

	for (e=island->edges; e; e=e->next) {
		len_uv += Vec2Lenf(e->v[0]->uv, e->v[1]->uv);
		len_xyz += VecLenf(e->v[0]->co, e->v[1]->co);
	}

	if (len_xyz == 0.0f)
		return 0.0f;
	else
		return (len_uv/len_xyz);
}

/* very primitive packing */
static void pack_atlas(UvAtlas *atlas)
{
	float minscale, scale, add[2], groupw, dx, dy, packx, packy, rowh;
	int totisland;
	UvIsland *island;

	minscale= 1e30;

	totisland = 0;
	for (island=atlas->islands; island; island=island->next) {
		if (island->flag & UVISLAND_NOPACK)
			continue;

		uv_island_normalize(island);
		island->scale= uv_island_relative_scale(island);
		minscale= MIN2(minscale, island->scale);
		totisland++;
	}

	packx= packy= 0.0;
	rowh= 0.0;
	groupw= 1.0/sqrt(totisland);

	for (island=atlas->islands; island; island=island->next) {
		if (island->flag & UVISLAND_NOPACK)
			continue;

		/* FIXME, uses bbox for scale now, should be area? */

		/* scale so all groups have the same size relative to the mesh */
		scale= minscale/island->scale;
		scale *= groupw;

		compute_uv_bbox(island);
		dx= (island->max[0]-island->min[0])*scale;
		dy= (island->max[1]-island->min[1])*scale;

		/* for padding */
		dx += 0.01;
		dy += 0.01;

		add[0]= add[1]= 0.0;

		if(dx > 1.0) {
			add[0]= 0.0;
			add[1]= packy;

			packy += dy;
			packx= 0.0;
			rowh= 0.0;
		}
		else if(dx <= (1.0-packx)) {
			add[0]= packx;
			add[1]= packy;

			packx += dx;
			rowh= MAX2(rowh, dy);
		}
		else {
			packy += rowh;
			packx= dx;
			rowh= dy;

			add[0]= 0.0;
			add[1]= packy;
		}

		/* for padding */
		add[0] += 0.005;
		add[1] += 0.005;

		uv_island_scale(island, scale);
		uv_island_move(island, add);
	}

	normalize_atlas(atlas);
}

/* *************************** STRETCH **************************************
 * Optimize Stretch, based on algorithm from:
 * "Texture Mapping Progressive Meshes" (2001) by Sander et.al.
 * The code was inspired by the implementation of algorithm in Graphite,
 * Copyright (C) 2000 Bruno Levy.
 */

void stretch_pin_boundary(UvIsland *island, int *pinselected)
{
	UvVert *v;
	UvEdge *e;
	int totpinsel;

	for (e=island->edges; e; e=e->next) {
		/* edge on boundary? */
		if (e->flag & UVEDGE_SEAM) {
			e->v[0]->flag |= UVVERT_PINNED;
			e->v[1]->flag |= UVVERT_PINNED;
		}
	}

	if (pinselected) {
		totpinsel = 0;
		for (v=island->verts; v; v=v->next)
			if ((v->flag & UVVERT_PINNED) && (v->flag & UVVERT_PINNED))
				totpinsel++;

		*pinselected = totpinsel;
	}
}

float stretch_compute_face(UvFace *f)
{
	float A, L2, w, tmp[3];
	float Ss[3], St[3];
	UvVert *v1, *v2, *v3;

	v1 = f->v[0];
	v2 = f->v[1];
	v3 = f->v[2];

	/* calculate uv area */
	A = 0.5 * (((v2->uv[0]-v1->uv[0]) * (v3->uv[1]-v1->uv[1])) - 
	           ((v3->uv[0]-v1->uv[0]) * (v2->uv[1]-v1->uv[1])));
	
	if (A<0)
		return 1e20;

	w= 1.0/(2.0*A);

	/* compute derivatives */
	Ss[0]= Ss[1]= Ss[2]= 0.0;
	VecCopyf(tmp, v1->co);
	VecMulf(tmp, (v2->uv[1] - v3->uv[1]));
	VecAddf(Ss, Ss, tmp);
	VecCopyf(tmp, v2->co);
	VecMulf(tmp, (v3->uv[1] - v1->uv[1]));
	VecAddf(Ss, Ss, tmp);
	VecCopyf(tmp, v3->co);
	VecMulf(tmp, (v1->uv[1] - v2->uv[1]));
	VecAddf(Ss, Ss, tmp);
	VecMulf(Ss, w);

	St[0]= St[1]= St[2]= 0.0;
	VecCopyf(tmp, v1->co);
	VecMulf(tmp, (v3->uv[0] - v2->uv[0]));
	VecAddf(St, St, tmp);
	VecCopyf(tmp, v2->co);
	VecMulf(tmp, (v1->uv[0] - v3->uv[0]));
	VecAddf(St, St, tmp);
	VecCopyf(tmp, v3->co);
	VecMulf(tmp, (v2->uv[0] - v1->uv[0]));
	VecAddf(St, St, tmp);
	VecMulf(St, w);

	L2 = sqrt(0.5 * (Inpf(Ss, Ss) + Inpf(St, St)));
	if (L2 < 1e20)
		L2 = L2*L2*AreaT3Dfl(v1->co, v2->co, v3->co);

	f->u.stretch = L2;

	return L2;
}

float stretch_compute_vertex(UvVert *v)
{
	LinkNode *link;
	float sum;

	sum = 0.0f;
	for (link=v->faces; link; link=link->next)
		sum += stretch_compute_face((UvFace*)link->link);
	
	v->u.stretch= sum;
	
	return sum;
}

float stretch_sum_vertex(UvVert *v)
{
	LinkNode *link;
	float sum;

	sum = 0.0f;
	for (link=v->faces; link; link=link->next)
		sum += ((UvFace*)link->link)->u.stretch;
	
	v->u.stretch= sum;
	
	return sum;
}

#define STRETCH_ITER 20

void stretch_optimize(UvIsland *island)
{
	UvVert *v, *wheel_v;
	UvEdge *e;
	LinkNode *link;
	int j, totedge, ok;
	float orig_stretch, low, stretch_low, high, stretch_high, m, stretch_m;
	float orig_pos[2], dir[2], rand, trusted_radius;

	for(v=island->verts; v; v=v->next) {
		if((v->flag & UVVERT_PINNED) || !(v->flag & UVVERT_SELECTED))
			continue;

		orig_stretch = v->u.stretch;
		orig_pos[0] = v->uv[0];
		orig_pos[1] = v->uv[1];

		/* move vertex in a random direction */

		trusted_radius = 0.0f;
		totedge = 0;
		for (link=v->edges; link; link=link->next) {
			e = (UvEdge*)link->link;
			wheel_v = (e->v[0] == v) ? e->v[1]: e->v[0]; 
			trusted_radius += Vec2Lenf(wheel_v->uv, v->uv);
			totedge++;
		}

		trusted_radius /= 2 * totedge;

		rand = BLI_frand() * 2.0 * M_PI;
		dir[0] = trusted_radius * cos(rand);
		dir[1] = trusted_radius * sin(rand);

		/* calculate old and new stretch */
		low = 0;
		stretch_low = v->u.stretch;

		Vec2Addf(v->uv, orig_pos, dir);
		high = 1;
		stretch_high = stretch_compute_vertex(v);

		/* iterate to get to the most optimal position for this vertex */
		j = 0;
		while ((stretch_high > 1e10 || j < STRETCH_ITER) && j < STRETCH_ITER) {

			j++;

			m = 0.5 * (low + high);
			v->uv[0]= orig_pos[0] + m*dir[0];
			v->uv[1]= orig_pos[1] + m*dir[1];
			stretch_m = stretch_compute_vertex(v);

			if (stretch_low < stretch_high) {
				high = m;
				stretch_high = stretch_m;
			}
			else {
				low = m;
				stretch_low = stretch_m;
			}
		}

		/* no luck, stretch has increased, reset to old values */
		if(v->u.stretch > orig_stretch) {
			Vec2Copyf(v->uv, orig_pos);
			stretch_compute_vertex(v);
		}

		/* check if stretch is at an acceptable level in wheel verts */
		ok = 1;
		for (link=v->edges; link && ok; link=link->next) {
			e = (UvEdge*)link->link;
			wheel_v = (e->v[0] == v) ? e->v[1]: e->v[0]; 
			ok = (stretch_sum_vertex(wheel_v) < 1e10);
		}

		/* stretch not acceptable in other vertices, resetting */
		if(!ok) {
			Vec2Copyf(v->uv, orig_pos);
			stretch_compute_vertex(v);

			for (link=v->edges; link && ok; link=link->next) {
				e = (UvEdge*)link->link;
				wheel_v = (e->v[0] == v) ? e->v[1]: e->v[0]; 
				stretch_sum_vertex(wheel_v);
			}
		}
	}
}

/* ***************************** UI ****************************** */

void unwrap_lscm()
{
	UvAtlas *atlas;
	UvIsland *island;
	Mesh *me;
	int pinned, autopin;

	if (is_uv_tface_editing_allowed()==0) return;
	me = get_mesh(OBACT);

	atlas = new_atlas(me, ATLAS_SEAMS);

	for (island=atlas->islands; island; island=island->next) {
		load_pins_selection(island, &pinned, NULL, NULL);
		autopin = (pinned < 2);

		lscm_begin(island);
		lscm_solve(island, autopin);
		lscm_end(island);

		if (!autopin)
			island->flag |= UVISLAND_NOPACK;
	}

	pack_atlas(atlas);

	for (island=atlas->islands; island; island=island->next)
		flush_uvs(island);

	free_atlas(atlas);

	BIF_undo_push("UV lscm unwrap");

	allqueue(REDRAWVIEW3D, 0);
	allqueue(REDRAWIMAGE, 0);
}

void stretch_optimize_tface_uv()
{
	UvAtlas *atlas;
	UvIsland *island;
	UvFace *f;
	UvVert *v;
	Mesh *me;
	int pinselected, selected, doit;
	short val;
	unsigned short event= 0;
	double lasttime;

	if (is_uv_tface_editing_allowed()==0)
		return;
	me= get_mesh(OBACT);

	atlas = new_atlas(me, ATLAS_IMPLICIT_SELECTED);

	doit = 0;
	for (island=atlas->islands; island; island=island->next) {
		load_pins_selection(island, NULL, &selected, NULL);
		stretch_pin_boundary(island, &pinselected);

		if (pinselected > 3 && pinselected < selected) {
			island->flag |= UVISLAND_DO_OPTIMIZE;
			load_uvs(island);

			/* initial stretch computation */
			for (f=island->faces; f; f=f->next)
				stretch_compute_face(f);
			
			for (v=island->verts; v; v=v->next)
				stretch_sum_vertex(v);

			doit = 1;
		}
	}

	if (!doit) {
		free_atlas(atlas);
		return;
	}

	lasttime = PIL_check_seconds_timer();

	while (doit) {
		for (island=atlas->islands; island; island=island->next) {
			if (island->flag & UVISLAND_DO_OPTIMIZE) {
				stretch_optimize(island);
				flush_uvs(island);
			}

			while (qtest()) {
				event= extern_qread(&val);
				if (val)
					if (event==ESCKEY || event==RETKEY || event==PADENTER)
						doit = 0;
			}
			
			if (!doit)
				break;
		}

		headerprint("Enter to finish. Esc to escape.");

		if (PIL_check_seconds_timer() - lasttime > 0.5) {
			lasttime = PIL_check_seconds_timer();
			/* FIXME: bug in force_draw also draws the 3d view */
			if(G.sima->lock) force_draw_plus(SPACE_VIEW3D, 0);
			else force_draw(0);
		}
	}

	if(event==ESCKEY)
		for (island=atlas->islands; island; island=island->next)
			if (island->flag & UVISLAND_DO_OPTIMIZE)
				flush_old_uvs(island);

	free_atlas(atlas);

	if(event!=ESCKEY)
		BIF_undo_push("UV optimize stretch");

	allqueue(REDRAWVIEW3D, 0);
	allqueue(REDRAWIMAGE, 0);
}

/* ********************* FACESELECT TOOLS ************************ */

#define EDHASH(a, b) ((a)*256 + (b))
#define EDHASHSIZE 65536
#define EDHMAX 256

typedef struct HashEdge {
	unsigned int v1, v2;
	struct HashEdge *next;
} HashEdge;

static HashEdge *get_hashedge(HashEdge *table, unsigned int v1, unsigned int v2)
{
	unsigned int hv1, hv2;

	hv1= v1 % EDHMAX;
	hv2= v2 % EDHMAX;
	if(hv1 > hv2)
		SWAP(unsigned int, hv1, hv2);

	return (table + EDHASH(hv1, hv2));
}

static int has_hashedge(HashEdge *he, unsigned int v1, unsigned int v2)
{
	while(he) {
		if(he->v1 || he->v2) {
			if(he->v1==v1 && he->v2==v2) return 1;
			else if(he->v1==v2 && he->v2==v1) return 1;
		}
		he= he->next;
	}

	return 0;
}

static void add_hashedge(HashEdge *first, unsigned int v1, unsigned int v2)
{
	if(first->v1 == 0 && first->v2 == 0) {
		first->v1 = v1;
		first->v2 = v2;
	}
	else {
		HashEdge *he= (HashEdge*)MEM_mallocN(sizeof(HashEdge), "mini");
		he->v1= v1;
		he->v2= v2;
		he->next= first->next;
		first->next= he;
	}
}

static int edge_in_hash(HashEdge *htable, unsigned int v1, unsigned int v2)
{
	return has_hashedge(get_hashedge(htable, v1, v2), v1, v2);
}

static void hash_add_edge(HashEdge *htable, unsigned int v1, unsigned int v2)
{
	HashEdge *he = get_hashedge(htable, v1, v2);

	if (!has_hashedge(he, v1, v2))
		add_hashedge(he, v1, v2);
}

static void hash_add_face(HashEdge *htable, MFace *mface)
{
	hash_add_edge(htable, mface->v1, mface->v2);
	hash_add_edge(htable, mface->v2, mface->v3);
	if(mface->v4) {
		hash_add_edge(htable, mface->v3, mface->v4);
		hash_add_edge(htable, mface->v4, mface->v1);
	}
	else
		hash_add_edge(htable, mface->v3, mface->v1);
}

static HashEdge *make_hash_edge_table(Mesh *me, short fill)
{
	HashEdge *htable, *he;
	MEdge *medge;
	unsigned int a;

	if(fill && (me->medge==NULL))
		return NULL;

	htable= MEM_callocN(EDHASHSIZE*sizeof(HashEdge), "lscmedgehashtable");

	if(fill) {
		medge= me->medge;
		for(a=me->totedge; a>0; a--, medge++) {
			if(medge->flag & ME_SEAM) {
				he= get_hashedge(htable, medge->v1, medge->v2);
				add_hashedge(he, medge->v1, medge->v2);
			}
		}
	}

	return htable;
}

static void free_hash_edge_table(HashEdge *htable)
{
	HashEdge *first, *he, *hen;
	int a;

	if(htable) {
		first= htable;
		for(a=EDHASHSIZE; a>0; a--, first++) {
			he= first->next;
			while(he) {
				hen= he->next;
				MEM_freeN(he);
				he= hen;
			}
		}
		MEM_freeN(htable);
	}
}

/* Set tface seams based on edge data, uses hash table to find seam edges. */

void set_seamtface()
{
	Mesh *me;
	EdgeHash *ehash;
	int a;
	MFace *mf;
	TFace *tf;
	MEdge *medge;

	me= get_mesh(OBACT);
	if(!me || !me->tface || !(G.f & G_FACESELECT)) return;
	
	ehash= BLI_edgehash_new();

	for(medge=me->medge, a=me->totedge; a>0; a--, medge++)
		if(medge->flag & ME_SEAM)
			BLI_edgehash_insert(ehash, medge->v1, medge->v2, NULL);

	mf= me->mface;
	tf= me->tface;
	for(a=me->totface; a>0; a--, mf++, tf++) {
		tf->unwrap &= ~(TF_SEAM1|TF_SEAM2|TF_SEAM3|TF_SEAM4);

		if(!ehash) continue;

		if(BLI_edgehash_haskey(ehash, mf->v1, mf->v2)) tf->unwrap |= TF_SEAM1;
		if(BLI_edgehash_haskey(ehash, mf->v2, mf->v3)) tf->unwrap |= TF_SEAM2;

		if(mf->v4) {
			if(BLI_edgehash_haskey(ehash, mf->v3, mf->v4)) tf->unwrap |= TF_SEAM3;
			if(BLI_edgehash_haskey(ehash, mf->v4, mf->v1)) tf->unwrap |= TF_SEAM4;
		}
		else if(BLI_edgehash_haskey(ehash, mf->v3, mf->v1)) tf->unwrap |= TF_SEAM3;
	}

	BLI_edgehash_free(ehash, NULL);
}

void select_linked_tfaces_with_seams(int mode, Mesh *me, unsigned int index)
{
	TFace *tf;
	MFace *mf;
	int a, doit=1, mark=0;
	char *linkflag;
	EdgeHash *ehash;

	ehash= BLI_edgehash_new();
	linkflag= MEM_callocN(sizeof(char)*me->totface, "linkflaguv");

	if (mode==0 || mode==1) {
		/* only put face under cursor in array */
		mf= ((MFace*)me->mface) + index;
		hash_add_face(ehash, mf);
		linkflag[index]= 1;
	}
	else {
		/* fill array by selection */
		tf= me->tface;
		mf= me->mface;
		for(a=0; a<me->totface; a++, tf++, mf++) {
			if(tf->flag & TF_HIDE);
			else if(tf->flag & TF_SELECT) {
				hash_add_face(ehash, mf);
				linkflag[a]= 1;
			}
		}
	}
	
	while(doit) {
		doit= 0;
		
		/* expand selection */
		tf= me->tface;
		mf= me->mface;
		for(a=0; a<me->totface; a++, tf++, mf++) {
			if(tf->flag & TF_HIDE);
			else if(!linkflag[a]) {
				mark= 0;

				if(!(tf->unwrap & TF_SEAM1))
					if(BLI_edgehash_haskey(ehash, mf->v1, mf->v2))
						mark= 1;
				if(!(tf->unwrap & TF_SEAM2))
					if(BLI_edgehash_haskey(ehash, mf->v2, mf->v3))
						mark= 1;
				if(!(tf->unwrap & TF_SEAM3)) {
					if(mf->v4) {
						if(BLI_edgehash_haskey(ehash, mf->v3, mf->v4))
							mark= 1;
					}
					else if(BLI_edgehash_haskey(ehash, mf->v3, mf->v1))
						mark= 1;
				}
				if(mf->v4 && !(tf->unwrap & TF_SEAM4))
					if(BLI_edgehash_haskey(ehash, mf->v4, mf->v1))
						mark= 1;

				if(mark) {
					linkflag[a]= 1;
					hash_add_face(ehash, mf);
					doit= 1;
				}
			}
		}
		
	}

	if(mode==0 || mode==2) {
		for(a=0, tf=me->tface; a<me->totface; a++, tf++)
			if(linkflag[a])
				tf->flag |= TF_SELECT;
			else
				tf->flag &= ~TF_SELECT;
	}
	else if(mode==1) {
		for(a=0, tf=me->tface; a<me->totface; a++, tf++)
			if(linkflag[a] && (tf->flag & TF_SELECT))
				break;

		if (a<me->totface) {
			for(a=0, tf=me->tface; a<me->totface; a++, tf++)
				if(linkflag[a])
					tf->flag &= ~TF_SELECT;
		}
		else {
			for(a=0, tf=me->tface; a<me->totface; a++, tf++)
				if(linkflag[a])
					tf->flag |= TF_SELECT;
		}
	}
	
	BLI_edgehash_free(ehash, NULL);
	MEM_freeN(linkflag);
	
	BIF_undo_push("Select linked UV face");
	object_tface_flags_changed(OBACT, 0);
}

