/**
 * $Id: vpaint.c 1462 2005-09-26 23:53:58Z broken $
 *
 * ***** BEGIN GPL/BL DUAL LICENSE BLOCK *****
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version. The Blender
 * Foundation also sells licenses for use in proprietary software under
 * the Blender License.  See http://www.blender.org/BL/ for information
 * about this.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software Foundation,
 * Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 *
 * The Original Code is Copyright (C) 2001-2002 by NaN Holding BV.
 * All rights reserved.
 *
 * The Original Code is: all of this file.
 *
 * Contributor(s): none yet.
 *
 * ***** END GPL/BL DUAL LICENSE BLOCK *****
 */

#include <math.h>
#include <string.h>

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#ifdef WIN32
#include <io.h>
#else
#include <unistd.h>
#endif   

#include "MEM_guardedalloc.h"

#include "IMB_imbuf.h"
#include "IMB_imbuf_types.h"

#include "BLI_blenlib.h"
#include "BLI_linklist.h"
#include "BLI_arithb.h"
#include "MTC_matrixops.h"

#include "DNA_image_types.h"
#include "DNA_action_types.h"
#include "DNA_armature_types.h"
#include "DNA_mesh_types.h"
#include "DNA_meshdata_types.h"
#include "DNA_modifier_types.h"
#include "DNA_object_types.h"
#include "DNA_object_force.h"
#include "DNA_screen_types.h"
#include "DNA_scene_types.h"
#include "DNA_view3d_types.h"
#include "DNA_userdef_types.h"

#include "BKE_DerivedMesh.h"
#include "BKE_depsgraph.h"
#include "BKE_deform.h"
#include "BKE_displist.h"
#include "BKE_global.h"
#include "BKE_image.h"
#include "BKE_mesh.h"
#include "BKE_modifier.h"
#include "BKE_object.h"
#include "BKE_utildefines.h"

#include "BIF_editview.h"
#include "BIF_gl.h"
#include "BIF_glutil.h"
#include "BIF_graphics.h"
#include "BIF_interface.h"
#include "BIF_mywindow.h"
#include "BIF_screen.h"
#include "BIF_space.h"
#include "BIF_toolbox.h"
#include "BIF_writeimage.h"

#include "BDR_vpaint.h"
#include "BDR_drawmesh.h"

#include "BSE_drawview.h"
#include "BSE_trans_types.h"
#include "BSE_view.h"

#include "mydevice.h"
#include "blendef.h"

#include "BIF_editdeform.h"

	/* Gvp.mode */
#define VP_MIX	0
#define VP_ADD	1
#define VP_SUB	2
#define VP_MUL	3
#define VP_FILT	4

#define MAXINDEX	512000

VPaint Gvp= {1.0, 1.0, 1.0, 0.2, 25.0, 1.0, 1.0, 0, VP_AREA+VP_SOFT+VP_SPRAY};
VPaint Gwp= {1.0, 1.0, 1.0, 0.2, 25.0, 1.0, 1.0, 0, VP_AREA+VP_SOFT};
float vpimat[3][3];
unsigned int *vpaintundobuf= NULL;
int totvpaintundo;
int *indexar= NULL;

int totwpaintundo;
MDeformVert *wpaintundobuf=NULL;

/* in contradiction to cpack drawing colors, the MCOL colors (vpaint colors) are per byte! 
   so not endian sensitive. Mcol = ABGR!!! so be cautious with cpack calls */

unsigned int rgba_to_mcol(float r, float g, float b, float a)
{
	int ir, ig, ib, ia;
	unsigned int col;
	char *cp;
	
	ir= floor(255.0*r);
	if(ir<0) ir= 0; else if(ir>255) ir= 255;
	ig= floor(255.0*g);
	if(ig<0) ig= 0; else if(ig>255) ig= 255;
	ib= floor(255.0*b);
	if(ib<0) ib= 0; else if(ib>255) ib= 255;
	ia= floor(255.0*a);
	if(ia<0) ia= 0; else if(ia>255) ia= 255;
	
	cp= (char *)&col;
	cp[0]= ia;
	cp[1]= ib;
	cp[2]= ig;
	cp[3]= ir;
	
	return col;
	
}

unsigned int vpaint_get_current_col(void)
{
	return rgba_to_mcol(Gvp.r, Gvp.g, Gvp.b, 1.0);
}

void do_shared_vertexcol(Mesh *me)
{
	/* if no mcol: do not do */
	/* if tface: only the involved faces, otherwise all */
	MFace *mface;
	TFace *tface;
	int a;
	short *scolmain, *scol;
	char *mcol;
	
	if(me->mcol==0 || me->totvert==0 || me->totface==0) return;
	
	scolmain= MEM_callocN(4*sizeof(short)*me->totvert, "colmain");
	
	tface= me->tface;
	mface= me->mface;
	mcol= (char *)me->mcol;
	for(a=me->totface; a>0; a--, mface++, mcol+=16) {
		if(tface==0 || (tface->mode & TF_SHAREDCOL) || (G.f & G_FACESELECT)==0) {
			scol= scolmain+4*mface->v1;
			scol[0]++; scol[1]+= mcol[1]; scol[2]+= mcol[2]; scol[3]+= mcol[3];
			scol= scolmain+4*mface->v2;
			scol[0]++; scol[1]+= mcol[5]; scol[2]+= mcol[6]; scol[3]+= mcol[7];
			scol= scolmain+4*mface->v3;
			scol[0]++; scol[1]+= mcol[9]; scol[2]+= mcol[10]; scol[3]+= mcol[11];
			if(mface->v4) {
				scol= scolmain+4*mface->v4;
				scol[0]++; scol[1]+= mcol[13]; scol[2]+= mcol[14]; scol[3]+= mcol[15];
			}
		}
		if(tface) tface++;
	}
	
	a= me->totvert;
	scol= scolmain;
	while(a--) {
		if(scol[0]>1) {
			scol[1]/= scol[0];
			scol[2]/= scol[0];
			scol[3]/= scol[0];
		}
		scol+= 4;
	}
	
	tface= me->tface;
	mface= me->mface;
	mcol= (char *)me->mcol;
	for(a=me->totface; a>0; a--, mface++, mcol+=16) {
		if(tface==0 || (tface->mode & TF_SHAREDCOL) || (G.f & G_FACESELECT)==0) {
			scol= scolmain+4*mface->v1;
			mcol[1]= scol[1]; mcol[2]= scol[2]; mcol[3]= scol[3];
			scol= scolmain+4*mface->v2;
			mcol[5]= scol[1]; mcol[6]= scol[2]; mcol[7]= scol[3];
			scol= scolmain+4*mface->v3;
			mcol[9]= scol[1]; mcol[10]= scol[2]; mcol[11]= scol[3];
			if(mface->v4) {
				scol= scolmain+4*mface->v4;
				mcol[13]= scol[1]; mcol[14]= scol[2]; mcol[15]= scol[3];
			}
		}
		if(tface) tface++;
	}

	MEM_freeN(scolmain);
}

void make_vertexcol()	/* single ob */
{
	Object *ob;
	Mesh *me;
	int i;

	/*
	 * Always copies from shadedisplist to mcol.
	 * When there are tfaces, it copies the colors and frees mcol
	 */
	
	if(G.obedit) {
		error("Unable to perform function in Edit Mode");
		return;
	}
	
	ob= OBACT;
	me= get_mesh(ob);
	if(me==0) return;

	if(me->mcol) MEM_freeN(me->mcol);
	mesh_create_shadedColors(ob, 1, (unsigned int**) &me->mcol, NULL);

	for (i=0; i<me->totface*4; i++) {
		me->mcol[i].a = 255;
	}
		
	if(me->tface) mcol_to_tface(me, 1);
	
	DAG_object_flush_update(G.scene, ob, OB_RECALC_DATA);
	
	allqueue(REDRAWBUTSEDIT, 0);
	allqueue(REDRAWVIEW3D, 0);
}



void copy_vpaint_undo(unsigned int *mcol, int tot)
{
	if(vpaintundobuf) MEM_freeN(vpaintundobuf);
	vpaintundobuf= 0;
	totvpaintundo= tot;	// because of return, it is used by weightpaint
	
	if(mcol==0 || tot==0) return;
	
	vpaintundobuf= MEM_mallocN(4*sizeof(int)*tot, "vpaintundobuf");
	memcpy(vpaintundobuf, mcol, 4*sizeof(int)*tot);
	
}

void vpaint_undo()
{
	Mesh *me;
	Object *ob;
	unsigned int temp, *from, *to;
	int a;
	
	if((G.f & G_VERTEXPAINT)==0) return;
	if(vpaintundobuf==0) return;

	ob= OBACT;
	me= get_mesh(ob);
	if(me==0 || me->totface==0) return;

	if(me->tface) tface_to_mcol(me);
	else if(me->mcol==0) return;
	
	a= MIN2(me->totface, totvpaintundo);
	from= vpaintundobuf;
	to= (unsigned int *)me->mcol;
	a*= 4;
	while(a--) {
		temp= *to;
		*to= *from;
		*from= temp;
		to++; from++;
	}
	DAG_object_flush_update(G.scene, ob, OB_RECALC_DATA);
	
	allqueue(REDRAWVIEW3D, 0);
	if(me->tface) mcol_to_tface(me, 1);
}

void clear_vpaint()
{
	Mesh *me;
	Object *ob;
	unsigned int *to, paintcol;
	int a;
	
	if((G.f & G_VERTEXPAINT)==0) return;

	ob= OBACT;
	me= get_mesh(ob);
	if(ob->id.lib) return;

	if(me==0 || me->totface==0) return;

	if(me->tface) tface_to_mcol(me);
	if(me->mcol==0) return;

	paintcol= vpaint_get_current_col();

	to= (unsigned int *)me->mcol;
	copy_vpaint_undo(to, me->totface);
	a= 4*me->totface;
	while(a--) {
		*to= paintcol;
		to++; 
	}
	BIF_undo_push("Clear vertex colors");
	DAG_object_flush_update(G.scene, ob, OB_RECALC_DATA);
	
	allqueue(REDRAWVIEW3D, 0);
	if(me->tface) mcol_to_tface(me, 1);
}

void clear_vpaint_selectedfaces()
{
	Mesh *me;
	TFace *tf;
	Object *ob;
	unsigned int paintcol;
	int i;

	ob= OBACT;

	me= get_mesh(ob);
	tf = me->tface;
	if (!tf) return; /* should not happen, but you never know */

	if(me==0 || me->totface==0) return;

	paintcol= vpaint_get_current_col();

	for (i = 0; i < me->totface; i++) {
		if (tf[i].flag & TF_SELECT) {
			tf[i].col[0] = paintcol;
			tf[i].col[1] = paintcol;
			tf[i].col[2] = paintcol;
			tf[i].col[3] = paintcol;
		}
	}
	BIF_undo_push("Clear vertex colors");
	allqueue(REDRAWVIEW3D, 0);
}

void vpaint_dogamma()
{
	Mesh *me;
	Object *ob;
	float igam, fac;
	int a, temp;
	char *cp, gamtab[256];

	if((G.f & G_VERTEXPAINT)==0) return;

	ob= OBACT;
	me= get_mesh(ob);
	if(me==0 || me->totface==0) return;

	if(me->tface) tface_to_mcol(me);
	else if(me->mcol==0) return;

	copy_vpaint_undo((unsigned int *)me->mcol, me->totface);

	igam= 1.0/Gvp.gamma;
	for(a=0; a<256; a++) {
		
		fac= ((float)a)/255.0;
		fac= Gvp.mul*pow( fac, igam);
		
		temp= 255.9*fac;
		
		if(temp<=0) gamtab[a]= 0;
		else if(temp>=255) gamtab[a]= 255;
		else gamtab[a]= temp;
	}

	a= 4*me->totface;
	cp= (char *)me->mcol;
	while(a--) {
		
		cp[1]= gamtab[ cp[1] ];
		cp[2]= gamtab[ cp[2] ];
		cp[3]= gamtab[ cp[3] ];
		
		cp+= 4;
	}
	allqueue(REDRAWVIEW3D, 0);
	
	if(me->tface) mcol_to_tface(me, 1);
}

/* used for both 3d view and image window */
void sample_vpaint()	/* frontbuf */
{
	unsigned int col;
	int x, y;
	short mval[2];
	char *cp;
	
	getmouseco_areawin(mval);
	x= mval[0]; y= mval[1];
	
	if(x<0 || y<0) return;
	if(x>=curarea->winx || y>=curarea->winy) return;
	
	x+= curarea->winrct.xmin;
	y+= curarea->winrct.ymin;
	
	glReadBuffer(GL_FRONT);
	glReadPixels(x, y, 1, 1, GL_RGBA, GL_UNSIGNED_BYTE, &col);
	glReadBuffer(GL_BACK);

	cp = (char *)&col;
	
	Gvp.r= cp[0];
	Gvp.r /= 255.0;

	Gvp.g= cp[1];
	Gvp.g /= 255.0;

	Gvp.b= cp[2];
	Gvp.b /= 255.0;

	allqueue(REDRAWBUTSEDIT, 0);
	addqueue(curarea->win, REDRAW, 1); // needed for when panel is open...
}

/* only used in drawobject.c now... */
void weight_to_rgb(float input, float *fr, float *fg, float *fb)
{
	float blend;
	
	blend= ((input/2.0f)+0.5f);
	
	if (input<=0.25f){	// blue->cyan
		*fr= 0.0f;
		*fg= blend*input*4.0f;
		*fb= blend;
	}
	else if (input<=0.50f){	// cyan->green
		*fr= 0.0f;
		*fg= blend;
		*fb= blend*(1.0f-((input-0.25f)*4.0f)); 
	}
	else if (input<=0.75){	// green->yellow
		*fr= blend * ((input-0.50f)*4.0f);
		*fg= blend;
		*fb= 0.0f;
	}
	else if (input<=1.0){ // yellow->red
		*fr= blend;
		*fg= blend * (1.0f-((input-0.75f)*4.0f)); 
		*fb= 0.0f;
	}
}

void init_vertexpaint()
{
	
	indexar= MEM_mallocN(sizeof(int)*MAXINDEX + 2, "vertexpaint");
}


void free_vertexpaint()
{
	
	if(indexar) MEM_freeN(indexar);
	indexar= NULL;
	if(vpaintundobuf) MEM_freeN(vpaintundobuf);
	vpaintundobuf= NULL;
	if(wpaintundobuf) 
		free_dverts(wpaintundobuf, totwpaintundo);
	wpaintundobuf= NULL;
}


static unsigned int mcol_blend(unsigned int col1, unsigned int col2, int fac)
{
	char *cp1, *cp2, *cp;
	int mfac;
	unsigned int col=0;
	
	if(fac==0) return col1;
	if(fac>=255) return col2;

	mfac= 255-fac;
	
	cp1= (char *)&col1;
	cp2= (char *)&col2;
	cp=  (char *)&col;
	
	cp[0]= 255;
	cp[1]= (mfac*cp1[1]+fac*cp2[1])>>8;
	cp[2]= (mfac*cp1[2]+fac*cp2[2])>>8;
	cp[3]= (mfac*cp1[3]+fac*cp2[3])>>8;
	
	return col;
}

static unsigned int mcol_add(unsigned int col1, unsigned int col2, int fac)
{
	char *cp1, *cp2, *cp;
	int temp;
	unsigned int col=0;
	
	if(fac==0) return col1;
	
	cp1= (char *)&col1;
	cp2= (char *)&col2;
	cp=  (char *)&col;
	
	cp[0]= 255;
	temp= cp1[1] + ((fac*cp2[1])>>8);
	if(temp>254) cp[1]= 255; else cp[1]= temp;
	temp= cp1[2] + ((fac*cp2[2])>>8);
	if(temp>254) cp[2]= 255; else cp[2]= temp;
	temp= cp1[3] + ((fac*cp2[3])>>8);
	if(temp>254) cp[3]= 255; else cp[3]= temp;
	
	return col;
}

static unsigned int mcol_sub(unsigned int col1, unsigned int col2, int fac)
{
	char *cp1, *cp2, *cp;
	int temp;
	unsigned int col=0;
	
	if(fac==0) return col1;
	
	cp1= (char *)&col1;
	cp2= (char *)&col2;
	cp=  (char *)&col;
	
	cp[0]= 255;
	temp= cp1[1] - ((fac*cp2[1])>>8);
	if(temp<0) cp[1]= 0; else cp[1]= temp;
	temp= cp1[2] - ((fac*cp2[2])>>8);
	if(temp<0) cp[2]= 0; else cp[2]= temp;
	temp= cp1[3] - ((fac*cp2[3])>>8);
	if(temp<0) cp[3]= 0; else cp[3]= temp;
	
	return col;
}

static unsigned int mcol_mul(unsigned int col1, unsigned int col2, int fac)
{
	char *cp1, *cp2, *cp;
	int mfac;
	unsigned int col=0;
	
	if(fac==0) return col1;

	mfac= 255-fac;
	
	cp1= (char *)&col1;
	cp2= (char *)&col2;
	cp=  (char *)&col;
	
	/* first mul, then blend the fac */
	cp[0]= 255;
	cp[1]= (mfac*cp1[1] + fac*((cp2[1]*cp1[1])>>8)  )>>8;
	cp[2]= (mfac*cp1[2] + fac*((cp2[2]*cp1[2])>>8)  )>>8;
	cp[3]= (mfac*cp1[3] + fac*((cp2[3]*cp1[3])>>8)  )>>8;

	
	return col;
}

static void vpaint_blend( unsigned int *col, unsigned int *colorig, unsigned int paintcol, int alpha)
{

	if(Gvp.mode==VP_MIX || Gvp.mode==VP_FILT) *col= mcol_blend( *col, paintcol, alpha);
	else if(Gvp.mode==VP_ADD) *col= mcol_add( *col, paintcol, alpha);
	else if(Gvp.mode==VP_SUB) *col= mcol_sub( *col, paintcol, alpha);
	else if(Gvp.mode==VP_MUL) *col= mcol_mul( *col, paintcol, alpha);
	
	/* if no spray, clip color adding with colorig & orig alpha */
	if((Gvp.flag & VP_SPRAY)==0) {
		unsigned int testcol=0, a;
		char *cp, *ct, *co;
		
		alpha= (int)(255.0*Gvp.a);
		
		if(Gvp.mode==VP_MIX || Gvp.mode==VP_FILT) testcol= mcol_blend( *colorig, paintcol, alpha);
		else if(Gvp.mode==VP_ADD) testcol= mcol_add( *colorig, paintcol, alpha);
		else if(Gvp.mode==VP_SUB) testcol= mcol_sub( *colorig, paintcol, alpha);
		else if(Gvp.mode==VP_MUL) testcol= mcol_mul( *colorig, paintcol, alpha);
		
		cp= (char *)col;
		ct= (char *)&testcol;
		co= (char *)colorig;
		
		for(a=0; a<4; a++) {
			if( ct[a]<co[a] ) {
				if( cp[a]<ct[a] ) cp[a]= ct[a];
				else if( cp[a]>co[a] ) cp[a]= co[a];
			}
			else {
				if( cp[a]<co[a] ) cp[a]= co[a];
				else if( cp[a]>ct[a] ) cp[a]= ct[a];
			}
		}
	}
}


static int sample_backbuf_area(int x, int y, float size)
{
	unsigned int rect[129*129], *rt;
	int x1, y1, x2, y2, a, tot=0, index;
	
	if(totvpaintundo>=MAXINDEX) return 0;
	
	if(size>64.0) size= 64.0;
	
	x1= x-size;
	x2= x+size;
	CLAMP(x1, 0, curarea->winx);
	CLAMP(x2, 0, curarea->winx);
	y1= y-size;
	y2= y+size;
	CLAMP(y1, 0, curarea->winy);
	CLAMP(y2, 0, curarea->winy);
#ifdef __APPLE__
	glReadBuffer(GL_AUX0);
#endif
	glReadPixels(x1+curarea->winrct.xmin, y1+curarea->winrct.ymin, x2-x1+1, y2-y1+1, GL_RGBA, GL_UNSIGNED_BYTE,  rect);
	glReadBuffer(GL_BACK);	

	if(G.order==B_ENDIAN) IMB_convert_rgba_to_abgr( (int)(4*size*size), rect);

	rt= rect;
	size= (y2-y1)*(x2-x1);
	if(size<=0) return 0;

	memset(indexar, 0, sizeof(int)*totvpaintundo+2);	/* plus 2! first element is total */
	
	while(size--) {
			
		if(*rt) {
			index= framebuffer_to_index(*rt);
			if(index>0 && index<=totvpaintundo)
				indexar[index] = 1;
		}
	
		rt++;
	}
	
	for(a=1; a<=totvpaintundo; a++) {
		if(indexar[a]) indexar[tot++]= a;
	}
	
	return tot;
}

static unsigned int sample_backbuf(int x, int y)
{
	unsigned int col;
	
	if(x>=curarea->winx || y>=curarea->winy) return 0;
	
	x+= curarea->winrct.xmin;
	y+= curarea->winrct.ymin;

#ifdef __APPLE__
	glReadBuffer(GL_AUX0);
#endif
	glReadPixels(x,  y, 1, 1, GL_RGBA, GL_UNSIGNED_BYTE,  &col);
	glReadBuffer(GL_BACK);	

	if(G.order==B_ENDIAN) SWITCH_INT(col);
		
	return framebuffer_to_index(col);
}

static int calc_vp_alpha_dl(VPaint *vp, DerivedMesh *dm, int vert, short *mval)
{
	float co[3], no[3];
	float fac, dx, dy;
	int alpha;
	short vertco[2];
	
	if(vp->flag & VP_SOFT) {
		dm->getVertCo(dm, vert, co);

	 	project_short_noclip(co, vertco);
		dx= mval[0]-vertco[0];
		dy= mval[1]-vertco[1];
		
		fac= sqrt(dx*dx + dy*dy);
		if(fac > vp->size) return 0;

		alpha= 255.0*vp->a*(1.0-fac/vp->size);
	}
	else {
		alpha= 255.0*vp->a;
	}

	if(vp->flag & VP_NORMALS) {
		dm->getVertNo(dm, vert, no);

			/* transpose ! */
		fac= vpimat[2][0]*no[0]+vpimat[2][1]*no[1]+vpimat[2][2]*no[2];
		if(fac>0.0) {
			dx= vpimat[0][0]*no[0]+vpimat[0][1]*no[1]+vpimat[0][2]*no[2];
			dy= vpimat[1][0]*no[0]+vpimat[1][1]*no[1]+vpimat[1][2]*no[2];
			
			alpha*= fac/sqrt(dx*dx + dy*dy + fac*fac);
		}
		else return 0;
	}
	
	return alpha;
}


void wpaint_undo (void)
{
	Object *ob= OBACT;
	Mesh	*me;
	MDeformVert *swapbuf;

	me = get_mesh(ob);
	if (!me)
		return;

	if (!wpaintundobuf)
		return;

	if (!me->dvert)
		return;

	if (totwpaintundo != me->totvert)
		return;

	swapbuf= me->dvert;

	/* copy undobuf to mesh */
	me->dvert= MEM_mallocN(sizeof(MDeformVert)*me->totvert, "deformVert");
	copy_dverts(me->dvert, wpaintundobuf, totwpaintundo);
	
	/* copy previous mesh to undo */
	free_dverts(wpaintundobuf, me->totvert);
	wpaintundobuf= MEM_mallocN(sizeof(MDeformVert)*me->totvert, "wpaintundo");
	copy_dverts(wpaintundobuf, swapbuf, totwpaintundo);
	
	/* now free previous mesh dverts */
	free_dverts(swapbuf, me->totvert);

	DAG_object_flush_update(G.scene, ob, OB_RECALC_DATA);
	DAG_object_flush_update(G.scene, modifiers_isDeformedByArmature(ob), OB_RECALC_DATA);
	scrarea_do_windraw(curarea);
	
}

void copy_wpaint_undo (MDeformVert *dverts, int dcount)
{
	if (wpaintundobuf)
		free_dverts(wpaintundobuf, totwpaintundo);

	wpaintundobuf = MEM_mallocN (sizeof(MDeformVert)*dcount, "wpaintundo");
	totwpaintundo = dcount;
	copy_dverts (wpaintundobuf, dverts, dcount);
}

static void wpaint_blend(MDeformWeight *dw, MDeformWeight *uw, float alpha, float paintval)
{
	
	if(dw==NULL || uw==NULL) return;
	
	if(Gwp.mode==VP_MIX || Gwp.mode==VP_FILT)
		dw->weight = paintval*alpha + dw->weight*(1.0-alpha);
	else if(Gwp.mode==VP_ADD)
		dw->weight += paintval*alpha;
	else if(Gwp.mode==VP_SUB) 
		dw->weight -= paintval*alpha;
	else if(Gwp.mode==VP_MUL) 
		/* first mul, then blend the fac */
		dw->weight = ((1.0-alpha) + alpha*paintval)*dw->weight;
	
	CLAMP(dw->weight, 0.0f, 1.0f);
	
	/* if no spray, clip result with orig weight & orig alpha */
	if((Gwp.flag & VP_SPRAY)==0) {
		float testw=0.0f;
		
		alpha= Gwp.a;
		
		if(Gwp.mode==VP_MIX || Gwp.mode==VP_FILT)
			testw = paintval*alpha + uw->weight*(1.0-alpha);
		else if(Gwp.mode==VP_ADD)
			testw = uw->weight + paintval*alpha;
		else if(Gwp.mode==VP_SUB) 
			testw = uw->weight - paintval*alpha;
		else if(Gwp.mode==VP_MUL) 
			/* first mul, then blend the fac */
			testw = ((1.0-alpha) + alpha*paintval)*uw->weight;
		
		CLAMP(testw, 0.0f, 1.0f);
		
		if( testw<uw->weight ) {
			if(dw->weight < testw) dw->weight= testw;
			else if(dw->weight > uw->weight) dw->weight= uw->weight;
		}
		else {
			if(dw->weight > testw) dw->weight= testw;
			else if(dw->weight < uw->weight) dw->weight= uw->weight;
		}
	}
	
}

static MDeformWeight *get_defweight(MDeformVert *dv, int defgroup)
{
	int i;
	for (i=0; i<dv->totweight; i++){
		if (dv->dw[i].def_nr == defgroup)
			return dv->dw+i;
	}
	return NULL;
}

/* used for 3d view */
/* cant sample frontbuf, weight colors are interpolated too unpredictable */
/* so we return the closest value to vertex, wich is actually correct anyway */
void sample_wpaint()
{
	extern float editbutvweight;
	Object *ob= OBACT;
	Mesh *me= get_mesh(ob);
	int index;
	short mval[2], sco[2];
	
	getmouseco_areawin(mval);
	index= sample_backbuf(mval[0], mval[1]);
	
	if(index && index<=me->totface) {
		MFace *mface;
		DerivedMesh *dm;
		MDeformWeight *dw;
		float w1, w2, w3, w4, co[3], fac;
		int needsFree;
		
		dm = mesh_get_derived_deform(ob, &needsFree);
		
		mface= ((MFace *)me->mface) + index-1;
		
		/* calc 3 or 4 corner weights */
		dm->getVertCo(dm, mface->v1, co);
	 	project_short_noclip(co, sco);
		w1= ((mval[0]-sco[0])*(mval[0]-sco[0]) + (mval[1]-sco[1])*(mval[1]-sco[1]));
		
		dm->getVertCo(dm, mface->v2, co);
	 	project_short_noclip(co, sco);
		w2= ((mval[0]-sco[0])*(mval[0]-sco[0]) + (mval[1]-sco[1])*(mval[1]-sco[1]));
		
		dm->getVertCo(dm, mface->v3, co);
	 	project_short_noclip(co, sco);
		w3= ((mval[0]-sco[0])*(mval[0]-sco[0]) + (mval[1]-sco[1])*(mval[1]-sco[1]));
		
		if(mface->v4) {
			dm->getVertCo(dm, mface->v4, co);
			project_short_noclip(co, sco);
			w4= ((mval[0]-sco[0])*(mval[0]-sco[0]) + (mval[1]-sco[1])*(mval[1]-sco[1]));
		}
		else w4= 1.0e10;
		
		fac= MIN4(w1, w2, w3, w4);
		if(w1==fac) {
			dw= get_defweight(me->dvert+mface->v1, ob->actdef-1);
			if(dw) editbutvweight= dw->weight; else editbutvweight= 0.0f;
		}
		else if(w2==fac) {
			dw= get_defweight(me->dvert+mface->v2, ob->actdef-1);
			if(dw) editbutvweight= dw->weight; else editbutvweight= 0.0f;
		}
		else if(w3==fac) {
			dw= get_defweight(me->dvert+mface->v3, ob->actdef-1);
			if(dw) editbutvweight= dw->weight; else editbutvweight= 0.0f;
		}
		else if(w4==fac) {
			if(mface->v4) {
				dw= get_defweight(me->dvert+mface->v4, ob->actdef-1);
				if(dw) editbutvweight= dw->weight; else editbutvweight= 0.0f;
			}
		}
		
		if (needsFree)
			dm->release(dm);
		
	}
	allqueue(REDRAWBUTSEDIT, 0);
	
}


void weight_paint(void)
{
	extern float editbutvweight;
	MDeformWeight	*dw, *uw;
	Object *ob; 
	Mesh *me;
	MFace *mface;
	TFace *tface;
	float mat[4][4], imat[4][4], paintweight;
	int index, totindex, alpha, totw;
	short mval[2], mvalo[2], firsttime=1, mousebut;

	if((G.f & G_WEIGHTPAINT)==0) return;
	if(G.obedit) return;
	
	if(G.qual & LR_CTRLKEY) {
		sample_wpaint();
		return;
	}
	
	if(indexar==NULL) init_vertexpaint();
	
	ob= OBACT;
	if(ob->id.lib) return;

	me= get_mesh(ob);
	if(me==NULL || me->totface==0) return;
	
	/* if nothing was added yet, we make dverts and a vertex deform group */
	if (!me->dvert)
		create_dverts(me);
	
	/* this happens on a Bone select, when no vgroup existed yet */
	if(ob->actdef==0) {
		Object *modob;
		if((modob = modifiers_isDeformedByArmature(ob))) {
			bPoseChannel *pchan;
			for(pchan= modob->pose->chanbase.first; pchan; pchan= pchan->next)
				if(pchan->bone->flag & SELECT)
					break;
			if(pchan) {
				bDeformGroup *dg= get_named_vertexgroup(ob, pchan->name);
				if(dg==NULL)
					dg= add_defgroup_name(ob, pchan->name);	// sets actdef
				else
					ob->actdef= get_defgroup_num(ob, dg);
				allqueue(REDRAWBUTSEDIT, 0);
			}
		}
	}
	if(ob->defbase.first==NULL) {
		add_defgroup(ob);
		allqueue(REDRAWBUTSEDIT, 0);
	}	
	
	if(ob->lay & G.vd->lay); else error("Active object is not in this layer");
	
	persp(PERSP_VIEW);
	/* imat for normals */
	Mat4MulMat4(mat, ob->obmat, G.vd->viewmat);
	Mat4Invert(imat, mat);
	Mat3CpyMat4(vpimat, imat);
	
	/* load projection matrix */
	mymultmatrix(ob->obmat);
	mygetsingmatrix(mat);
	myloadmatrix(G.vd->viewmat);
	
	getmouseco_areawin(mvalo);
	
	if(me->tface) tface_to_mcol(me);
	copy_vpaint_undo( (unsigned int *)me->mcol, me->totface);
	copy_wpaint_undo(me->dvert, me->totvert);
	
	getmouseco_areawin(mval);
	mvalo[0]= mval[0];
	mvalo[1]= mval[1];
	
	if (U.flag & USER_LMOUSESELECT) mousebut = R_MOUSE;
	else mousebut = L_MOUSE;
	
	while (get_mbut() & mousebut) {
		getmouseco_areawin(mval);
		
		if(firsttime || mval[0]!=mvalo[0] || mval[1]!=mvalo[1]) {
			DerivedMesh *dm;
			int needsFree;

			firsttime= 0;
			
			/* which faces are involved */
			if(Gwp.flag & VP_AREA) {
				totindex= sample_backbuf_area(mval[0], mval[1], Gwp.size);
			}
			else {
				indexar[0]= sample_backbuf(mval[0], mval[1]);
				if(indexar[0]) totindex= 1;
				else totindex= 0;
			}
			
			MTC_Mat4SwapMat4(G.vd->persmat, mat);
			
			if(Gwp.flag & VP_COLINDEX) {
				for(index=0; index<totindex; index++) {
					if(indexar[index] && indexar[index]<=me->totface) {
					
						mface= ((MFace *)me->mface) + (indexar[index]-1);
					
						if(mface->mat_nr!=ob->actcol-1) {
							indexar[index]= 0;
						}
					}					
				}
			}

			if((G.f & G_FACESELECT) && me->tface) {
				for(index=0; index<totindex; index++) {
					if(indexar[index] && indexar[index]<=me->totface) {
					
						tface= ((TFace *)me->tface) + (indexar[index]-1);
					
						if((tface->flag & TF_SELECT)==0) {
							indexar[index]= 0;
						}
					}					
				}
			}
			
			/* make sure each vertex gets treated only once */
			/* and calculate filter weight */
			totw= 0;
			if(Gwp.mode==VP_FILT) 
				paintweight= 0.0f;
			else
				paintweight= editbutvweight;
			
			for(index=0; index<totindex; index++) {
				if(indexar[index] && indexar[index]<=me->totface) {
					mface= me->mface + (indexar[index]-1);
					
					(me->dvert+mface->v1)->flag= 1;
					(me->dvert+mface->v2)->flag= 1;
					(me->dvert+mface->v3)->flag= 1;
					if(mface->v4) (me->dvert+mface->v4)->flag= 1;
					
					if(Gwp.mode==VP_FILT) {
						dw= verify_defweight(me->dvert+mface->v1, ob->actdef-1);
						if(dw) {paintweight+= dw->weight; totw++;}
						dw= verify_defweight(me->dvert+mface->v2, ob->actdef-1);
						if(dw) {paintweight+= dw->weight; totw++;}
						dw= verify_defweight(me->dvert+mface->v3, ob->actdef-1);
						if(dw) {paintweight+= dw->weight; totw++;}
						if(mface->v4) {
							dw= verify_defweight(me->dvert+mface->v4, ob->actdef-1);
							if(dw) {paintweight+= dw->weight; totw++;}
						}
					}
				}
			}
			
			if(Gwp.mode==VP_FILT) 
				paintweight/= (float)totw;
			
			dm = mesh_get_derived_deform(ob, &needsFree);

			for(index=0; index<totindex; index++) {
				
				if(indexar[index] && indexar[index]<=me->totface) {
					mface= me->mface + (indexar[index]-1);
					
					if((me->dvert+mface->v1)->flag) {
						alpha= calc_vp_alpha_dl(&Gwp, dm, mface->v1, mval);
						if(alpha) {
							dw= verify_defweight(me->dvert+mface->v1, ob->actdef-1);
							uw= verify_defweight(wpaintundobuf+mface->v1, ob->actdef-1);
							wpaint_blend(dw, uw, (float)alpha/255.0, paintweight);
						}
						(me->dvert+mface->v1)->flag= 0;
					}
					
					if((me->dvert+mface->v2)->flag) {
						alpha= calc_vp_alpha_dl(&Gwp, dm, mface->v2, mval);
						if(alpha) {
							dw= verify_defweight(me->dvert+mface->v2, ob->actdef-1);
							uw= verify_defweight(wpaintundobuf+mface->v2, ob->actdef-1);
							wpaint_blend(dw, uw, (float)alpha/255.0, paintweight);
						}
						(me->dvert+mface->v2)->flag= 0;
					}
					
					if((me->dvert+mface->v3)->flag) {
						alpha= calc_vp_alpha_dl(&Gwp, dm, mface->v3, mval);
						if(alpha) {
							dw= verify_defweight(me->dvert+mface->v3, ob->actdef-1);
							uw= verify_defweight(wpaintundobuf+mface->v3, ob->actdef-1);
							wpaint_blend(dw, uw, (float)alpha/255.0, paintweight);
						}
						(me->dvert+mface->v3)->flag= 0;
					}
					
					if((me->dvert+mface->v4)->flag) {
						if(mface->v4) {
							alpha= calc_vp_alpha_dl(&Gwp, dm, mface->v4, mval);
							if(alpha) {
								dw= verify_defweight(me->dvert+mface->v4, ob->actdef-1);
								uw= verify_defweight(wpaintundobuf+mface->v4, ob->actdef-1);
								wpaint_blend(dw, uw, (float)alpha/255.0, paintweight);
							}
							(me->dvert+mface->v4)->flag= 0;
						}
					}
				}
			}
			if (needsFree)
				dm->release(dm);
			
			MTC_Mat4SwapMat4(G.vd->persmat, mat);
			
		}
		else BIF_wait_for_statechange();
		
		if(mval[0]!=mvalo[0] || mval[1]!=mvalo[1]) {

			DAG_object_flush_update(G.scene, ob, OB_RECALC_DATA);
			scrarea_do_windraw(curarea);
			
			if(Gwp.flag & (VP_AREA|VP_SOFT)) {
				/* draw circle in backbuf! */
				persp(PERSP_WIN);
				fdrawXORcirc((float)mval[0], (float)mval[1], Gwp.size);
				persp(PERSP_VIEW);
			}

			screen_swapbuffers();
			backdrawview3d(0);
	
			mvalo[0]= mval[0];
			mvalo[1]= mval[1];
		}
	}
	
	if(me->tface) {
		MEM_freeN(me->mcol);
		me->mcol= 0;
	}
	
	DAG_object_flush_update(G.scene, ob, OB_RECALC_DATA);
	// this flag is event for softbody to refresh weightpaint values
	if(ob->soft) ob->softflag |= OB_SB_REDO;
	
	allqueue(REDRAWVIEW3D, 0);

}

void vertex_paint()
{
	Object *ob;
	Mesh *me;
	MFace *mface;
	TFace *tface;
	float mat[4][4], imat[4][4];
	unsigned int paintcol=0, *mcol, *mcolorig, fcol1, fcol2;
	int index, alpha, totindex;
	short mval[2], mvalo[2], firsttime=1, mousebut;
	
	if((G.f & G_VERTEXPAINT)==0) return;
	if(G.obedit) return;
	
	if(indexar==NULL) init_vertexpaint();
	
	ob= OBACT;
	if(ob->id.lib) return;

	me= get_mesh(ob);
	if(me==NULL || me->totface==0) return;
	if(ob->lay & G.vd->lay); else error("Active object is not in this layer");
	
	if(me->tface==NULL && me->mcol==NULL) make_vertexcol();

	if(me->tface==NULL && me->mcol==NULL) return;
	
	persp(PERSP_VIEW);
	/* imat for normals */
	Mat4MulMat4(mat, ob->obmat, G.vd->viewmat);
	Mat4Invert(imat, mat);
	Mat3CpyMat4(vpimat, imat);
	
	/* load projection matrix */
	mymultmatrix(ob->obmat);
	mygetsingmatrix(mat);
	myloadmatrix(G.vd->viewmat);
	
	paintcol= vpaint_get_current_col();
	
	getmouseco_areawin(mvalo);
	
	if(me->tface) tface_to_mcol(me);
	copy_vpaint_undo( (unsigned int *)me->mcol, me->totface);
	
	getmouseco_areawin(mval);
	mvalo[0]= mval[0];
	mvalo[1]= mval[1];
	
	if (U.flag & USER_LMOUSESELECT) mousebut = R_MOUSE;
	else mousebut = L_MOUSE;
	
	while (get_mbut() & mousebut) {
		getmouseco_areawin(mval);
		
		if(firsttime || mval[0]!=mvalo[0] || mval[1]!=mvalo[1]) {
			DerivedMesh *dm;
			int needsFree;

			firsttime= 0;

			/* which faces are involved */
			if(Gvp.flag & VP_AREA) {
				totindex= sample_backbuf_area(mval[0], mval[1], Gvp.size);
			}
			else {
				indexar[0]= sample_backbuf(mval[0], mval[1]);
				if(indexar[0]) totindex= 1;
				else totindex= 0;
			}
			
			MTC_Mat4SwapMat4(G.vd->persmat, mat);
			
			if(Gvp.flag & VP_COLINDEX) {
				for(index=0; index<totindex; index++) {
					if(indexar[index] && indexar[index]<=me->totface) {
					
						mface= ((MFace *)me->mface) + (indexar[index]-1);
					
						if(mface->mat_nr!=ob->actcol-1) {
							indexar[index]= 0;
						}
					}					
				}
			}
			if((G.f & G_FACESELECT) && me->tface) {
				for(index=0; index<totindex; index++) {
					if(indexar[index] && indexar[index]<=me->totface) {
					
						tface= ((TFace *)me->tface) + (indexar[index]-1);
					
						if((tface->flag & TF_SELECT)==0) {
							indexar[index]= 0;
						}
					}					
				}
			}

			dm= mesh_get_derived_deform(ob, &needsFree);
			for(index=0; index<totindex; index++) {

				if(indexar[index] && indexar[index]<=me->totface) {
				
					mface= ((MFace *)me->mface) + (indexar[index]-1);
					mcol=	  ( (unsigned int *)me->mcol) + 4*(indexar[index]-1);
					mcolorig= ( (unsigned int *)vpaintundobuf) + 4*(indexar[index]-1);

					if(Gvp.mode==VP_FILT) {
						fcol1= mcol_blend( mcol[0], mcol[1], 128);
						if(mface->v4) {
							fcol2= mcol_blend( mcol[2], mcol[3], 128);
							paintcol= mcol_blend( fcol1, fcol2, 128);
						}
						else {
							paintcol= mcol_blend( mcol[2], fcol1, 170);
						}
						
					}
					
					alpha= calc_vp_alpha_dl(&Gvp, dm, mface->v1, mval);
					if(alpha) vpaint_blend( mcol, mcolorig, paintcol, alpha);
					
					alpha= calc_vp_alpha_dl(&Gvp, dm, mface->v2, mval);
					if(alpha) vpaint_blend( mcol+1, mcolorig+1, paintcol, alpha);
	
					alpha= calc_vp_alpha_dl(&Gvp, dm, mface->v3, mval);
					if(alpha) vpaint_blend( mcol+2, mcolorig+2, paintcol, alpha);

					if(mface->v4) {
						alpha= calc_vp_alpha_dl(&Gvp, dm, mface->v4, mval);
						if(alpha) vpaint_blend( mcol+3, mcolorig+3, paintcol, alpha);
					}
				}
			}
			if (needsFree)
				dm->release(dm);
				
			MTC_Mat4SwapMat4(G.vd->persmat, mat);
			
			do_shared_vertexcol(me);
			if(me->tface) {
				mcol_to_tface(me, 0);
			}
	
			DAG_object_flush_update(G.scene, ob, OB_RECALC_DATA);
			scrarea_do_windraw(curarea);

			if(Gvp.flag & (VP_AREA|VP_SOFT)) {
				/* draw circle in backbuf! */
				persp(PERSP_WIN);
				fdrawXORcirc((float)mval[0], (float)mval[1], Gvp.size);
				persp(PERSP_VIEW);
			}
			screen_swapbuffers();
			backdrawview3d(0);
			
			mvalo[0]= mval[0];
			mvalo[1]= mval[1];
		}
		else BIF_wait_for_statechange();
	}
	
	if(me->tface) {
		MEM_freeN(me->mcol);
		me->mcol= 0;
	}
	
	allqueue(REDRAWVIEW3D, 0);
}

void set_wpaint(void)		/* toggle */
{		
	Object *ob;
	Mesh *me;
	
	scrarea_queue_headredraw(curarea);
	ob= OBACT;
	if(ob->id.lib) return;
	me= get_mesh(ob);
		
	if(me && me->totface>=MAXINDEX) {
		error("Maximum number of faces: %d", MAXINDEX-1);
		G.f &= ~G_WEIGHTPAINT;
		return;
	}
	
	if(G.f & G_WEIGHTPAINT) G.f &= ~G_WEIGHTPAINT;
	else G.f |= G_WEIGHTPAINT;
	
	allqueue(REDRAWVIEW3D, 0);
	allqueue(REDRAWBUTSEDIT, 0);
	
		/* Weightpaint works by overriding colors in mesh,
		 * so need to make sure we recalc on enter and
		 * exit (exit needs doing regardless because we
		 * should redeform).
		 */
	if (me) {
		DAG_object_flush_update(G.scene, OBACT, OB_RECALC_DATA);
	}

	if(G.f & G_WEIGHTPAINT) {
		setcursor_space(SPACE_VIEW3D, CURSOR_VPAINT);
	}
	else {
		freefastshade();	/* to be sure */
		if(!(G.f & G_FACESELECT))
			setcursor_space(SPACE_VIEW3D, CURSOR_STD);
	}
}


void set_vpaint(void)		/* toggle */
{		
	Object *ob;
	Mesh *me;
	
	scrarea_queue_headredraw(curarea);
	ob= OBACT;
	if(ob->id.lib) {
		G.f &= ~G_VERTEXPAINT;
		return;
	}
	
	me= get_mesh(ob);
	
	if(me && me->totface>=MAXINDEX) {
		error("Maximum number of faces: %d", MAXINDEX-1);
		G.f &= ~G_VERTEXPAINT;
		return;
	}
	
	if(me && me->tface==NULL && me->mcol==NULL) make_vertexcol();
	
	if(G.f & G_VERTEXPAINT){
		G.f &= ~G_VERTEXPAINT;
	}
	else {
		G.f |= G_VERTEXPAINT;
		/* Turn off weight painting */
		if (G.f & G_WEIGHTPAINT)
			set_wpaint();
	}
	
	allqueue(REDRAWVIEW3D, 0);
	allqueue(REDRAWBUTSEDIT, 0);
	
	if(G.f & G_VERTEXPAINT) {
		setcursor_space(SPACE_VIEW3D, CURSOR_VPAINT);
	}
	else {
		freefastshade();	/* to be sure */
		if (me) {
			DAG_object_flush_update(G.scene, ob, OB_RECALC_DATA);
		}
		if((G.f & G_FACESELECT)==0) setcursor_space(SPACE_VIEW3D, CURSOR_STD);
	}
}

/* ************* vertex paint baking ************** */

/* accumalte color info in a pixel */
void accumulateColor(unsigned int *pixel, float fcol[4])
{
	float icol[4];

	/* get color values from pixel */
	image_to_rgba(*pixel, icol, icol+1, icol+2, icol+3);

	VecMulf(icol, icol[3]);		/* multiply accumulated color by last total	*/
	VecAddf(icol, icol, fcol);	/* add the colors							*/
	icol[3] += 1.0f;			/* increment the total						*/
	VecMulf(icol, 1.0f/icol[3]);/* renormalize with new total				*/

	/* put color values back in pixel */
	*pixel = rgba_to_image(icol[0], icol[1], icol[2], icol[3]);
}

/* stupid modulo function, because modulo in C is not IEEE standard	*/
/*		meaning that -1 % x != x - 1 (like in Python for example)	*/
static int Modulo(int x, int y) {
	int val = x % y;
	if (val < 0)
		val += y;
	return val;
}

/* fill a single horizontal line					*/
/*		starting at x1, col1 and ending on x2, col2	*/
void fillLine(ImBuf *imbuf, int x1, int x2, int y, float col1[4], float col2[4])
{
	int x;
	unsigned int *pixel;
	float dcol[4];	/* color variation	*/
	float fcol[4];	/* current color	*/

	y %= imbuf->y;	/* wrap on y axis */

	if (y<0)
		y = Modulo(y, imbuf->y);

	/* point to starting pixel */
	pixel = imbuf->rect + imbuf->x * y + Modulo(x1, imbuf->x);

	if (x2 == x1) {	/* drawing a single pixel */
		accumulateColor(pixel, col1);
	}
	else {			/* drawing many pixels */
		int dx = x2 - x1;	/* variation in x */

		/* calculate color variation */
		dcol[0] = (col2[0] - col1[0]) / dx;
		dcol[1] = (col2[1] - col1[1]) / dx;
		dcol[2] = (col2[2] - col1[2]) / dx;
		dcol[3] = (col2[3] - col1[3]) / dx;

		/* initialize current color */
		memcpy(fcol, col1, sizeof(float) * 4);

		/* paint each pixel */
		for (x=x1 ; x<=x2 ; x++, pixel++) {
			/* wrap pixel on x axis */
			if (x!=0 && x % imbuf->x == 0) pixel -= imbuf->x;

			accumulateColor(pixel, fcol);

			/* add color variation to current color */
			VecAddf(fcol, fcol, dcol);
		}
	}
}

/* UV pixel structure							*/
/*		contains the coordinates and the color	*/
typedef struct PUV {
	float x, y;
	float col[4];
} PUV;

/* sort triangle from the top down	*/
/* using a reduce bubble sort		*/
void sortPUV(PUV p[3]) {
	PUV temp;
	/* swap 1st and 2nd if needed	*/
	if (p[0].y < p[1].y) {
		temp = p[0];
		p[0] = p[1];
		p[1] = temp;
	}
	/* swap 2nd and 3rd if needed	*/
	/* lowest is now 3rd			*/
	if (p[1].y < p[2].y) {
		temp = p[1];
		p[1] = p[2];
		p[2] = temp;
	}
	/* swap 1st and 2nd if needed	*/
	/* highest is now 1st			*/
	if (p[0].y < p[1].y) {
		temp = p[0];
		p[0] = p[1];
		p[1] = temp;
	}
}

/* paint a single triangle into the accumulation buffer		*/
/*		uv1, uv2, uv3 are the uv coordinates corresponding	*/
/*		to the colors col1, col2, col3						*/
/*		imbuf is the accumulation buffer					*/
/* ceil and floor are used to make sure the maximum number	*/
/* of pixel gets painted on.								*/
void paintTriangle(ImBuf *imbuf, float *uv1, float *uv2, float *uv3, unsigned int col1, unsigned int col2, unsigned int col3)
{
	PUV p[3];	/* the triangle to fill */
	float width, height;	/* buffer dimensions */
	int y;	/* y coordinates, when painting lines */
	float dx1, dx2, dx3;		/* variances on X axis	*/
	float dcol1[4], dcol2[4];	/* color variances, two per subtriangle */

	width = (float)(imbuf->x - 1);
	height = (float)(imbuf->y - 1);

	/* filling the triangle */
	p[0].x = (width * uv1[0]);
	p[0].y = (height * uv1[1]);
	mcol_to_rgba(col1, p[0].col, p[0].col+1, p[0].col+2, p[0].col+3);

	p[1].x = (width * uv2[0]);
	p[1].y = (height * uv2[1]);
	mcol_to_rgba(col2, p[1].col, p[1].col+1, p[1].col+2, p[1].col+3);

	p[2].x = (width * uv3[0]);
	p[2].y = (height * uv3[1]);
	mcol_to_rgba(col3, p[2].col, p[2].col+1, p[2].col+2, p[2].col+3);

	/* sorting the coordinates/col pair from the top down */
	sortPUV(p);

	/* variance from the first to the third point */
	dx3 = (p[0].x - p[2].x) / (p[0].y - p[2].y);

	/* if lower half of the triangle is more than 1 pixel of height */
	if (p[1].y - p[2].y > 0.5f) {
		float fcol1[4], fcol2[4];	/* current colors on each side of the triangle	*/
		float ix1, ix2;				/* x coordinates on each side of the triangle	*/
		float dy;			/* variation on y */
		int top, bottom;	/* top and bottom of the triangle */

		/* first side of the triangle */
		dy = (float)(ceil(p[0].y) - floor(p[2].y));

		/* setting the color variance */
		dcol1[0] = (p[0].col[0] - p[2].col[0]) / dy;
		dcol1[1] = (p[0].col[1] - p[2].col[1]) / dy;
		dcol1[2] = (p[0].col[2] - p[2].col[2]) / dy;
		dcol1[3] = (p[0].col[3] - p[2].col[3]) / dy;

		/* second side of the triangle */
		dy = (float)(ceil(p[1].y) - floor(p[2].y));

		/* setting the color variance */
		dcol2[0] = (p[1].col[0] - p[2].col[0]) / dy;
		dcol2[1] = (p[1].col[1] - p[2].col[1]) / dy;
		dcol2[2] = (p[1].col[2] - p[2].col[2]) / dy;
		dcol2[3] = (p[1].col[3] - p[2].col[3]) / dy;

		/* variance from the second to the third point */
		dx2 = (p[1].x - p[2].x) / dy;

		/* initial color on both sides */
		memcpy(fcol1, p[2].col, sizeof(float) * 4);
		memcpy(fcol2, p[2].col, sizeof(float) * 4);

		/* initial x coordinates on both sides */
		ix1 = ix2 = p[2].x;

		/* floor and ceil to get the maximum number of scanlines */
		bottom = (int)floor(p[2].y);
		top = (int)ceil(p[1].y);

		/* paint each horizontal scanline */
		for(y=bottom ; y <= top ; y++) {
			/* fill the line with the correct order of x parameters */
			if (ix1 < ix2) {
				fillLine(imbuf, (int)floor(ix1), (int)ceil(ix2), y, fcol1, fcol2);
			}
			else {
				fillLine(imbuf, (int)floor(ix2), (int)ceil(ix1), y, fcol2, fcol1);
			}

			/* increase x coordinates and colors on both sides */
			ix1 += dx3;
			ix2 += dx2;

			fcol1[0] += dcol1[0];
			fcol1[1] += dcol1[1];
			fcol1[2] += dcol1[2];
			fcol1[3] += dcol1[3];

			fcol2[0] += dcol2[0];
			fcol2[1] += dcol2[1];
			fcol2[2] += dcol2[2];
			fcol1[3] += dcol1[3];
		}
	}

	/* if upper half of the triangle is more than 1 pixel of height			*/
	/* SAME LOGIC AS UPPER HALF EXCEPT THAT THE TRIANGLE IS DRAWN TOP DOWN	*/
	/* INSTEAD OF BOTTOM UP. LESS COMMENTS BECAUSE OF THAT					*/
	if (p[0].y - p[1].y > 0.5f) {
		float fcol1[4], fcol2[4];	/* current colors on each side of the triangle	*/
		float ix1, ix2;				/* x coordinates on each side of the triangle	*/
		float dy;			/* variation on y */
		int top, bottom;	/* top and bottom of the triangle */

		dy = (float)(ceil(p[0].y) - floor(p[2].y));

		dcol1[0] = (p[0].col[0] - p[2].col[0]) / dy;
		dcol1[1] = (p[0].col[1] - p[2].col[1]) / dy;
		dcol1[2] = (p[0].col[2] - p[2].col[2]) / dy;
		dcol1[3] = (p[0].col[3] - p[2].col[3]) / dy;

		dy = (float)(ceil(p[0].y) - floor(p[1].y));

		dcol2[0] = (p[0].col[0] - p[1].col[0]) / dy;
		dcol2[1] = (p[0].col[1] - p[1].col[1]) / dy;
		dcol2[2] = (p[0].col[2] - p[1].col[2]) / dy;
		dcol2[3] = (p[0].col[3] - p[1].col[3]) / dy;

		/* variance from the first to the second point */
		dx1 = (p[0].x - p[1].x) / dy;

		memcpy(fcol1, p[0].col, sizeof(float) * 4);
		memcpy(fcol2, p[0].col, sizeof(float) * 4);

		ix1 = ix2 = (float)p[0].x;

		bottom = (int)floor(p[1].y);
		top = (int)ceil(p[0].y);

		for(y=top ; y >= bottom ; y--) {
			if (ix1 < ix2) {
				fillLine(imbuf, (int)floor(ix1), (int)ceil(ix2), y, fcol1, fcol2);
			}
			else {
				fillLine(imbuf, (int)floor(ix2), (int)ceil(ix1), y, fcol2, fcol1);
			}

			ix1 -= dx3;
			ix2 -= dx1;

			fcol1[0] -= dcol1[0];
			fcol1[1] -= dcol1[1];
			fcol1[2] -= dcol1[2];
			fcol1[3] -= dcol1[3];

			fcol2[0] -= dcol2[0];
			fcol2[1] -= dcol2[1];
			fcol2[2] -= dcol2[2];
			fcol2[3] -= dcol2[3];
		}
	}
	/* if upper half of the triangle is only 1 pixel of height, paint it directly */
	else if (p[1].y != p[2].y) {
		if (p[0].x < p[1].x) {
			fillLine(imbuf, (int)floor(p[0].x), (int)ceil(p[1].x), (int)p[0].y, p[0].col, p[1].col);
		}
		else {
			fillLine(imbuf, (int)floor(p[1].x), (int)ceil(p[0].x), (int)p[0].y, p[1].col, p[0].col);
		}
	}
}


#define CLAMPINSIDE(limit, value) (value<0?value+limit:(value>=limit?value-limit:value))

/* blur the mask before mulitplying with the image	*/
/*		modes are:									*/
/*			BAKE_BLURALL	1						*/
/*			BAKE_BLURBORDER	2						*/
/*		mulbuf is the multiplication buffer			*/
void blurMask(ImBuf *mulbuf, short mode) {
	float col[4], bcol[4];		/* buffer color and blended color	*/
	unsigned int *rout, *rin;	/* rect output and rect input		*/
	unsigned int *pixel;		/* current pixel that is written	*/
	unsigned int p1, p2, p3, p4, p5, p6, p7, p8, p9;	/* The 9 pixel values for blurring */
	int x, y, width, height;	/* x and y coordinates, width and height of buffer */

	/* allocate new buffer as the result of the blur */
	pixel = rout = MEM_callocN(mulbuf->x * mulbuf->y * sizeof(unsigned int), "temp rect");

/*
	Stupid blur matrix
	0.0625	0.125	0.0625
	0.125	0.25	0.125
	0.0625	0.125	0.0625

	Positions
	1		2		3
	4		5		6
	7		8		9

*/
	/* input rect */
	rin = mulbuf->rect;

	width = mulbuf->x;
	height = mulbuf->y;

	/* each line */
	for (y=0 ; y<height ; y++)
	{
		int yU, yD;	/* y up and down coordinates */

		/**/
		yU = CLAMPINSIDE(height, y-1);
		yD = CLAMPINSIDE(height, y+1);

		for (x=0 ; x<width ; x++, pixel++)
		{
			int xL, xR;	/* x left and right coordinates */

			/* center pixel */
			p5 = *(rin + y*width + x);

			/* */
			if (mode==BAKE_BLURBORDER && p5!=0) {
				*pixel = p5;
				continue;
			}

			xL = CLAMPINSIDE(width, x-1);
			xR = CLAMPINSIDE(width, x+1);

			/* get the rest of the pixels */
			p1 = *(rin + yU*width + xL);
			p2 = *(rin + yU*width + x);
			p3 = *(rin + yU*width + xR);
			p4 = *(rin + y*width + xL);
			p6 = *(rin + y*width + xR);
			p7 = *(rin + yD*width + xL);
			p8 = *(rin + yD*width + x);
			p9 = *(rin + yD*width + xR);

			/* if no neighbour has been painted, no need to process */
			if (p1 + p2 + p3 + p4 + p6 + p7 + p8 + p9 == 0) {
				continue;
			}

			/* empty blend color */
			bcol[0] = bcol[1] = bcol[2] = bcol[3] = 0.0f;

			/* For each pixel, check if the pixel holds color			*/
			/* if is does, multiply by blur factor corresponding		*/
			/* add to the blend color									*/
			/* add the factor to the running sum for normalizing later	*/
			if (p1) {
				image_to_rgba(p1, col, col+1, col+2, col+3);
				VecMulf(col, 0.0625f);
				VecAddf(bcol, bcol, col);
				bcol[3] += 0.0625f;
			}
			if (p2) {
				image_to_rgba(p2, col, col+1, col+2, col+3);
				VecMulf(col, 0.125f);
				VecAddf(bcol, bcol, col);
				bcol[3] += 0.125f;
			}
			if (p3) {
				image_to_rgba(p3, col, col+1, col+2, col+3);
				VecMulf(col, 0.0625f);
				VecAddf(bcol, bcol, col);
				bcol[3] += 0.0625f;
			}
			if (p4) {
				image_to_rgba(p4, col, col+1, col+2, col+3);
				VecMulf(col, 0.125f);
				VecAddf(bcol, bcol, col);
				bcol[3] += 0.125f;
			}
			if (p5) {
				image_to_rgba(p5, col, col+1, col+2, col+3);
				VecMulf(col, 0.25f);
				VecAddf(bcol, bcol, col);
				bcol[3] += 0.25f;
			}
			if (p6) {
				image_to_rgba(p6, col, col+1, col+2, col+3);
				VecMulf(col, 0.125f);
				VecAddf(bcol, bcol, col);
				bcol[3] += 0.125f;
			}
			if (p7) {
				image_to_rgba(p7, col, col+1, col+2, col+3);
				VecMulf(col, 0.0625f);
				VecAddf(bcol, bcol, col);
				bcol[3] += 0.0625f;
			}
			if (p8) {
				image_to_rgba(p8, col, col+1, col+2, col+3);
				VecMulf(col, 0.125f);
				VecAddf(bcol, bcol, col);
				bcol[3] += 0.125f;
			}
			if (p9) {
				image_to_rgba(p9, col, col+1, col+2, col+3);
				VecMulf(col, 0.0625f);
				VecAddf(bcol, bcol, col);
				bcol[3] += 0.0625f;
			}

			/* Normalize the color with the running sum */
			VecMulf(bcol, 1.0f/bcol[3]);

			*pixel = rgba_to_image(bcol[0], bcol[1], bcol[2], 1.0f);
		}
	}

	/* free the old rect	*/
	MEM_freeN(mulbuf->rect);
	/* point to the new one	*/
	mulbuf->rect = rout;
}

/* multiple a source buffer with a multiplication buffer */
/*		BOTH BUFFERS ARE OF THE SAME SIZE				*/
void multBuffers(ImBuf *srcbuf, ImBuf *mulbuf)
{
	float icol[4], mcol[4];	/* initial color and multiplication color */
	int step = 0;	/* step on the buffer */
	int size;		/* size of the buffer */
	unsigned int *pixel, *mul;	/* source pixel and multiplication pixel */

	/* initialize both pixels */
	pixel = srcbuf->rect;
	mul = mulbuf->rect;

	/* calculate size */
	size = srcbuf->x * srcbuf->y;

	/* step on all the buffer */
	for(step=0 ; step<size ; step++, pixel++, mul++)
	{
		/* if multiplication pixel holds data */
		if (*mul) {
			/* convert both pixel to RGBA*/
			image_to_rgba(*pixel, icol, icol+1, icol+2, icol+3);
			image_to_rgba(*mul, mcol, mcol+1, mcol+2, mcol+3);

			/* multiply each color */
			icol[0] *= mcol[0];
			icol[1] *= mcol[1];
			icol[2] *= mcol[2];

			/* assign back to the source pixel */
			*pixel = rgba_to_image(icol[0], icol[1], icol[2], icol[3]);
		}
	}
}

/* create destination buffer and set its path */
ImBuf *createDestBuffer(Image *ima, char *name, char *path)
{
	char dir[FILE_MAXDIR], file[FILE_MAXFILE];
	ImBuf *ibuf;

	/* split of image name from the image */
	BLI_split_dirfile(ima->ibuf->name, dir, file);

	/* new name is meshname_oldfile.extension */
	sprintf(path, "%s%s_%s", dir, name, file);
	/* add corresponding image type extension */
	addImageExtension(path);

	/* duplicate the image buffer */
	ibuf = IMB_dupImBuf(ima->ibuf);
	/* copy name in */
	BLI_strncpy(ibuf->name, path, sizeof(ibuf->name));

	return ibuf;
}

/* A baking item				*/
/*		Contains:				*/
/*			source image		*/
/*			destination buffer	*/
/*			save path			*/
/*			mult buffer			*/
/*		AND CAN BE CHAINNED		*/
typedef struct BakeItems {
	void *next, *prev;
	Image *ima;
	ImBuf *ibuf;
	char path[FILE_MAXDIR+FILE_MAXFILE];
	ImBuf mulmask;
} BakeItems;

/* get the mult buffer corresponding to an image */
ImBuf *getBakeBuffer(ListBase *masklist, Image *ima, char *name)
{
	BakeItems *item = masklist->first;

	/* look if it's in the list already, return it if it is */
	while (item) {
		if (item->ima == ima)
			return &item->mulmask;
		item = item->next;
	}

	/* not in list, need to create it */

	/* load image buffer if not there */
	if (ima->ibuf==0)
		load_image(ima, IB_rect, "", 0);

	/* create a new Bake Item */
	item = MEM_callocN(sizeof(BakeItems), "BakeItem");

	/* append to the list */
	BLI_addtail(masklist, item);

	/* assign the image to the item */
	item->ima = ima;

	/* create a new destination buffer */
	item->ibuf = createDestBuffer(ima, name, item->path);

	/* copy in the mulmask buffer */
	item->mulmask = *ima->ibuf;
	/* create a new rect area for the mulmask buffer */
	item->mulmask.rect = MEM_callocN(item->mulmask.x * item->mulmask.y * sizeof(unsigned int), "multiply mask");

	/* return a pointer to the multiplication buffer */
	return &item->mulmask;
}

/* For each item in the list, apply it's multiplication mask to the image */
void applyBakeBuffer(ListBase *masklist, Mesh *me, short blurmode, short blurpass)
{
	BakeItems *item = masklist->first;
	int i;		/* counter on the faces	*/
	short bp;	/* blur passes counter	*/

	/* act on every item in the list */
	while (item) {
		TFace *tf;
		Image *ima;

		/* blur the mask */
		bp = blurpass;
		while(bp--)
			blurMask(&item->mulmask, blurmode);

		/* aply it */
		multBuffers(item->ibuf, &item->mulmask);
		/* free the mask's rect area */
		MEM_freeN(item->mulmask.rect);

		/* write the resulting image to disk */
		BIF_write_ibuf(item->ibuf, item->path);
		/* free the image buffer */
		IMB_freeImBuf(item->ibuf);

		/* load the resulting image from disk */
		ima = add_image(item->path);

		/* make the faces point to the new image if they pointed to the old one */
		tf = me->tface;
		for(i=0;i<me->totface;i++,tf++)
		{
			if (tf->tpage == item->ima)
				tf->tpage=ima;
		}
		item = item->next;
	}
}

/* Bake the vertex color of a mesh into its textures	*/
/*	Creates copies, leaves the originals untouched		*/
void bakeMeshPaint(Mesh *me, short blurmode, short blurpass)
{
	ListBase masklist = {0,0};	/* bake item list */
	TFace *tf;			/* current TFace		*/
	MFace *mf;			/* current MFace		*/
	ImBuf *mulmask=0;	/* current mult buffer	*/
	unsigned int *col;	/* current face colors	*/
	int i;	/* counter on the faces */

	/* for safety, but shouldn't get called if no tfaces */
	if (me->tface==0)
		return;

	/* loop on each face */
	tf = me->tface;
	mf = me->mface;
	for(i=0;i<me->totface;i++,tf++,mf++)
	{
		/* skip to next if no image assigned */
		if (tf->tpage==0)
			continue;

		/* colors for that face */
		col = tf->col;

		/* Get the multiplication buffer corresponding to that image	*/
		/* need mesh name (minus ME) to create new buffers if needed	*/
		mulmask = getBakeBuffer(&masklist, tf->tpage, me->id.name+2);

		/* paint the first triangle */
		paintTriangle(mulmask, tf->uv[0], tf->uv[1], tf->uv[2], col[0], col[1], col[2]);

		if (mf->v4) {	/* paint the other triangle if the face is a quad */
			paintTriangle(mulmask, tf->uv[2], tf->uv[3], tf->uv[0], col[2], col[3], col[0]);
		}

		/* clear vertex colors */
		col[0] = col[1] = col[2] = col[3] = 0xFFFFFFFF;
	}

	/* apply the baking to all the buffers */
	applyBakeBuffer(&masklist, me, blurmode, blurpass);

	/* free the buffers list */
	BLI_freelistN(&masklist);
}
