/**
 * blenlib/DNA_effect_types.h (mar-2001 nzc)
 *	
 * $Id: DNA_effect_types.h 774 2004-08-02 21:49:45Z hos $ 
 *
 * ***** BEGIN GPL/BL DUAL LICENSE BLOCK *****
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version. The Blender
 * Foundation also sells licenses for use in proprietary software under
 * the Blender License.  See http://www.blender.org/BL/ for information
 * about this.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software Foundation,
 * Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 *
 * The Original Code is Copyright (C) 2001-2002 by NaN Holding BV.
 * All rights reserved.
 *
 * The Original Code is: all of this file.
 *
 * Contributor(s): none yet.
 *
 * ***** END GPL/BL DUAL LICENSE BLOCK *****
 */
#ifndef DNA_EFFECT_TYPES_H
#define DNA_EFFECT_TYPES_H

#include "DNA_object_types.h"

/* don't forget, new effects also in writefile.c for dna!!! */

#define PAF_MAXMULT		4

	/* paf->flag (keep bit 0 free for compatibility) */
#define PAF_BSPLINE		2
#define PAF_STATIC		4
#define PAF_FACE		8
#define PAF_ANIMATED	16

	/* eff->type */
#define EFF_BUILD		0
#define EFF_PARTICLE	1
#define EFF_WAVE		2
#define EFF_RVERT		3
#define EFF_SOFT		7	/* Softbody effect*/

	/* eff->flag */
#define EFF_SELECT		1
#define EFF_CYCLIC		2

	/* paf->stype */
#define PAF_NORMAL		0
#define PAF_VECT		1

	/* paf->texmap */
#define PAF_TEXINT		0
#define PAF_TEXRGB		1
#define PAF_TEXGRAD		2

	/* wav->flag */
#define WAV_X			2
#define WAV_Y			4
#define WAV_CYCL		8

	/* rvert->flag */
#define RVERT_X			2
#define RVERT_Y			4
#define RVERT_Z			8
#define RVERT_N			16

	/* rvert->opts */
#define RVERT_TIME		2
#define RVERT_SEED		4

	/* sft->mode */
#define SFT_FIELD		0
#define SFT_COLL		1
#define SFT_CLOTH		2

	/* sft->fomode */
#define SFT_NOFO		0
#define SFT_SMOOTH		1
#define SFT_SHARP		2

	/* sft->flag */
#define SFT_OFFSET		2
#define SFT_TARGETNORS	4
#define SFT_DISPLIST	8	/* not implemented */
#define SFT_QUAD		16
#define SFT_ORIGMESH	32	/* sft->cloth contains the original mesh */
#define SFT_MASS		64	/* SFT_CLOTH */
#define SFT_RESTITUTION	128	/* SFT_CLOTH */
#define SFT_BAKED		256 /* SFT_CLOTH */
#define SFT_EXTCOLL		512 /* SFT_CLOTH */

	/* Effector Fields types */
#define PFIELD_FORCE	1
#define PFIELD_VORTEX	2
#define PFIELD_MAGNET	3

typedef struct Effect {
	struct Effect *next, *prev;
	short type, flag, buttype, rt;
	
} Effect;

typedef struct BuildEff {
	struct BuildEff *next, *prev;
	short type, flag, buttype, rt;
	
	float len, sfra;
	
} BuildEff;

#
#
typedef struct Particle {
	float co[3], no[3];
	float time, lifetime;
	short mat_nr, rt;
} Particle;


typedef struct PartEff {
	struct PartEff *next, *prev;
	short type, flag, buttype, stype;
	
	float sta, end, lifetime;
	int totpart, totkey, seed;
	
	float normfac, obfac, randfac, texfac, randlife;
	float force[3];
	float damp;
	
	float nabla, vectsize, defvec[3];
	
	float mult[4], life[4];
	short child[4], mat[4];
	short texmap, curmult;
	short texmap_space;
	short staticstep;

	Particle *keys;

} PartEff;


typedef struct WaveEff {
	struct WaveEff *next, *prev;
	short type, flag, buttype, stype;
	
	float startx, starty, height, width;
	float narrow, speed, minfac, damp;
	
	float timeoffs, lifetime;
	
} WaveEff;

typedef struct RVertEff {
	struct RVertsEff *next, *prev;
	short type, flag, buttype, stype;

	float xammount, yammount, zammount, nammount;

	int seed, opts;

} RVertEff;

typedef struct SoftEff {
	struct SoftEff *next, *prev;
	short type, flag, buttype, stype;
	
	Object *ob;
	struct MVert *cloth;	/* SFT_CLOTH */
	float *vs;				/* SFT_CLOTH */
	float *baked;			/* SFT_CLOTH */
	float *oldco;			/* SFT_CLOTH */
	int totvert;			/* SFT_CLOTH */
	
	float falloff, range, shift;
	float sta, end;
	
	float stiffness;		/* SFT_CLOTH */
	float gravity;			/* SFT_CLOTH */
	float damp;			/* SFT_CLOTH */
	float movedamp;		/* SFT_CLOTH */
	float restitution;		/* SFT_CLOTH */
	float colldist;		/* SFT_CLOTH */
	short lastcalc;		/* SFT_CLOTH */

	short pad3;
	
	short mode, fomode;
} SoftEff;

#endif

